/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                ANALYSIS OF CLAUSE SETS                 * */
/* *                                                        * */
/* *  $Module:   ANALYZE                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.57.1.1 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/11/30 19:20:48 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: analyze.c,v $ */

#include "analyze.h"

static LIST ana_CalculatePredicatePrecedence(LIST, LIST);
static LIST ana_CalculateFunctionPrecedence(LIST, LIST);

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

BOOL ana_EQUATIONS;                /* Problem contains any equations at all */
static BOOL ana_PEQUATIONS;        /* Problem contains positive equations */
static BOOL ana_NEQUATIONS;        /* Problem contains negative equations */
static BOOL ana_FUNCTIONS;         /* Problem contains non-constant function symbols */
static BOOL ana_PROP;              /* Problem contains propositional variables */
static BOOL ana_GROUND;            /* Problem contains non-propositional, non-equational ground atoms */
static BOOL ana_NONUNIT;           /* Problem contains non-unit clauses */
static BOOL ana_MONADIC;           /* Problem contains non-ground monadic predicates */
static BOOL ana_NONMONADIC;        /* Problem contains non-ground n-place predicates, n>1, not equality */
static BOOL ana_DYADIC;            /* Problem contains non-ground dyadic predicates */
static BOOL ana_NONDYADIC;         /* Problem contains non-ground n-place predicates, n>2, not equality */
BOOL ana_SORTRES;                  /* Problem contains literal not(S(x)) for some S */
static BOOL ana_USORTRES;          /* Problem contains literal not(S(t)) for some S */
static BOOL ana_FINDOMAIN;         /* Problem contains clause implying a finite domain */
static BOOL ana_CONGROUND;         /* Conjecture is ground */

static BOOL ana_PUREEQUATIONAL;    /* Problem is purely equational */
static BOOL ana_PUREPROPOSITIONAL; /* Problem is purely propositional */

BOOL ana_SORTDECEQUATIONS;         /* True if all positive equations are sort decreasing with respect to
                                      the static sort theory contained in the problem */

static NAT  ana_AXIOMCLAUSES;      /* Number of axiom clauses */
static NAT  ana_CONCLAUSES;        /* Number of conjecture clauses */
static NAT  ana_NONHORNCLAUSES;    /* Number of Non-Horn clauses */


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void ana_AnalyzeProblem(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: Void.
  EFFECT:  Analyzes the clauses and sets the anlayze variables.
           Recomputes the weight for the clauses.
***************************************************************/
{
  LIST   Scan;
  CLAUSE Clause;

  ana_EQUATIONS   = FALSE;  
  ana_PEQUATIONS  = FALSE;             /* Defaults for properties */
  ana_NEQUATIONS  = FALSE; 
  ana_FUNCTIONS   = FALSE;
  ana_FINDOMAIN   = FALSE;
  ana_MONADIC     = FALSE;
  ana_NONMONADIC  = FALSE;
  ana_DYADIC      = FALSE;
  ana_NONDYADIC   = FALSE;
  ana_PROP        = FALSE;
  ana_GROUND      = FALSE;
  ana_SORTRES     = FALSE;
  ana_USORTRES    = FALSE;
  ana_NONUNIT     = FALSE;
  ana_CONGROUND   = TRUE;

  ana_AXIOMCLAUSES    = 0; 
  ana_CONCLAUSES      = 0;
  ana_NONHORNCLAUSES  = 0;

  if (list_Empty(Clauses))
    return;
  Scan = Clauses;
  while(!list_Empty(Clauses)) {
    Clause = (CLAUSE)list_Car(Clauses);
    clause_SetWeight(Clause, clause_ComputeWeight(Clause));

    if (clause_GetFlag(Clause,CONCLAUSE))
      ana_CONCLAUSES++;
    else
      ana_AXIOMCLAUSES++;

    if (clause_NumOfSuccLits(Clause) > 1)
      ana_NONHORNCLAUSES++;

    if (ana_CONGROUND && clause_GetFlag(Clause,CONCLAUSE) &&
	clause_MaxVar(Clause) != symbol_GetInitialStandardVarCounter())
      ana_CONGROUND = FALSE;
    if (!ana_PEQUATIONS && clause_ContainsPositiveEquations(Clause)) {
      ana_PEQUATIONS = TRUE;
    }
    if (!ana_NEQUATIONS && clause_ContainsNegativeEquations(Clause)) {
      ana_NEQUATIONS = TRUE;
    }
    if (!ana_MONADIC || !ana_NONMONADIC || !ana_DYADIC || !ana_NONDYADIC
		    || !ana_PROP || !ana_GROUND)
      clause_ContainsFolAtom(Clause, &ana_PROP, &ana_GROUND,
		&ana_MONADIC, &ana_NONMONADIC, &ana_DYADIC, &ana_NONDYADIC);

    if (!ana_FUNCTIONS && clause_ContainsFunctions(Clause)) {
      ana_FUNCTIONS = TRUE;
    }
    if (!ana_FINDOMAIN && clause_ImpliesFiniteDomain(Clause)) {
      ana_FINDOMAIN = TRUE;
    }
    if (!ana_NONUNIT && clause_Length(Clause) > 1) {
      ana_NONUNIT = TRUE;
    }
    if (!ana_SORTRES || !ana_USORTRES) 
      clause_ContainsSortRestriction(Clause,&ana_SORTRES,&ana_USORTRES);
    
    Clauses = list_Cdr(Clauses);
  }

  ana_PUREEQUATIONAL    = ((ana_PEQUATIONS || ana_NEQUATIONS) && !ana_MONADIC && !ana_NONMONADIC 
			   && !ana_PROP && !ana_GROUND);
  ana_EQUATIONS         = (ana_PEQUATIONS || ana_NEQUATIONS);
  ana_PUREPROPOSITIONAL = (!ana_PEQUATIONS && !ana_NEQUATIONS &&!ana_MONADIC && !ana_NONMONADIC && ana_PROP);

  if (ana_PEQUATIONS && ana_SORTRES)
    ana_SORTDECEQUATIONS = sort_EquationsAreSortDecreasing(Scan);

  return;
}

void ana_AnalyzeClauses(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: Void.
  EFFECT:  Analyzes and prints each clause, more expensive than
           ana_AnalyzeProblem.
***************************************************************/
{
  LIST   Scan;
  CLAUSE Clause;
  BOOL   conclause, horn, findomain, unit, groundcl, 
         equations, pequations, nequations, constants, functions, 
         monadic, nonmonadic, dyadic, nondyadic, prop, ground, 
	 sortres, usortres, regular;

  if (list_Empty(Clauses))
    return;

  printf("\nDiagnostics\n");
  Scan = Clauses;
  while(!list_Empty(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    clause_Print(Clause);

    /* clause_SetWeight(Clause, clause_ComputeWeight(Clause)); */

    /* Clause properties */
    conclause  = (clause_GetFlag(Clause,CONCLAUSE) ? TRUE : FALSE);
    horn       = ((clause_NumOfSuccLits(Clause) > 1) ? FALSE : TRUE);
    findomain  = (clause_ImpliesFiniteDomain(Clause) ? TRUE : FALSE);
    unit       = ((clause_Length(Clause) > 1) ? FALSE : TRUE);
    groundcl   = ((clause_MaxVar(Clause) == symbol_GetInitialStandardVarCounter()) ? TRUE : FALSE);
    regular    = (reg_IsKRegularClause(Clause,1) ? TRUE : FALSE);

    /* Existence of symbols/literals */
    pequations = (clause_ContainsPositiveEquations(Clause) ? TRUE : FALSE);
    nequations = (clause_ContainsNegativeEquations(Clause) ? TRUE : FALSE);
    equations  = (pequations || nequations);
    constants  = (clause_ContainsConstants(Clause) ? TRUE : FALSE);
    functions  = (clause_ContainsFunctions(Clause) ? TRUE : FALSE);

    monadic     = FALSE;
    nonmonadic  = FALSE;
    dyadic      = FALSE;
    nondyadic   = FALSE;
    prop        = FALSE;
    ground      = FALSE;
    clause_ContainsFolAtom(Clause, &prop, &ground, &monadic,
		    &nonmonadic, &dyadic, &nondyadic);

    sortres     = FALSE;
    usortres    = FALSE;
    clause_ContainsSortRestriction(Clause,&sortres,&usortres);
    
    /* Print diagnostics */
    printf("\n    Properties:\t");
    if (conclause) 
      printf("ConCl, "); 
    else 
      printf("AxCl, ");
    if (horn) 
      printf("Horn, ");
    if (findomain) 
      printf("FinDom, ");
    if (unit) 
      printf("Unit, ");
    if (groundcl) 
      printf("Gr, ");
    if (equations && !monadic && !nonmonadic && !prop && !ground)
      printf("Eq, "); 
    if (!equations && !monadic && !nonmonadic && prop) 
      printf("Prop, ");
    if (regular) 
      printf("Reg, ");
    else 
      printf("NonReg, ");
    printf("Depth=%d, ", clause_ComputeTermDepth(Clause));

    printf("\n    Literals:\t");
    if (pequations) 
      printf("PEq, ");
    if (nequations) 
      printf("NEq, ");
    if (constants) 
      printf("Const, ");
    if (functions) 
      printf("Func, ");
    if (monadic) 
      printf("Mon, ");
    if (nonmonadic) 
      printf("NonMon, ");
    if (dyadic) 
      printf("Dy, ");
    if (nondyadic) 
      printf("NonDy, ");
    if (prop) 
      printf("Prop, ");
    if (ground) 
      printf("Gr, ");
    if (sortres) 
      printf("SortRes, ");
    if (usortres) 
      printf("USortRes, ");

    printf("\n");
    Scan = list_Cdr(Scan);

  }
  return;
}


void ana_Print(void)
{
  char* Horn;

  if (ana_NONHORNCLAUSES == 0)
    Horn = "Horn";
  else
    Horn = "Non-Horn";
  
  if (ana_MONADIC && !ana_NONMONADIC) {
    printf("\n This is a monadic %s problem",Horn);
    if (ana_PEQUATIONS || ana_NEQUATIONS)
      printf(" with equality.");
    else
      printf(" without equality.");
  }
  else
    if (ana_PEQUATIONS || ana_NEQUATIONS) {
      if (ana_NONMONADIC || ana_MONADIC || ana_PROP)
	printf("\n This is a first-order %s problem containing equality.",Horn);
      else
	if (ana_NONUNIT)
	  printf("\n This is a pure equality %s problem.",Horn);
	else
	  printf("\n This is a unit equality problem.");
    }
    else
      if (ana_NONMONADIC || ana_MONADIC)
	printf("\n This is a first-order %s problem without equality.",Horn);

  if (ana_PUREPROPOSITIONAL)
    printf("\n This is a propositional %s problem.",Horn);
  else
    if (ana_FINDOMAIN || !ana_FUNCTIONS) {
      printf("\n This is a problem that has, if any, a finite domain model.");
      if (ana_FINDOMAIN)
	printf("\n There is a finite domain clause.");
      if (!ana_FUNCTIONS)
	printf("\n There are no function symbols.");
    }
      
  
  if (ana_SORTRES) {
    printf("\n This is a problem that contains sort information.");
    if (ana_PEQUATIONS && ana_SORTDECEQUATIONS)
      printf("\n All equations are sort-decreasing.");
  }

  if (ana_CONGROUND && !ana_PUREPROPOSITIONAL)
    printf("\n The conjecture is ground.");

  printf("\n Axiom clauses: %d Conjecture clauses: %d",ana_AXIOMCLAUSES,ana_CONCLAUSES);

  flag_PrintInferenceRules();
  flag_PrintReductionRules();
  printf("\n Extras    : ");
  if (flag_GetFlagValue(flag_SATINPUT))
    printf("Input Saturation, ");
  else
    printf("No Input Saturation, ");
  if (flag_GetFlagValue(flag_SELECT) == 0)
    printf("No Selection, ");
  else
    if (flag_GetFlagValue(flag_SELECT) == 1)
      printf("Dynamic Selection, ");
    else
      printf("Always Selection, ");
  if (flag_GetFlagValue(flag_SPLITS) == -1)
    printf("Full Splitting, ");
  else
    if (flag_GetFlagValue(flag_SPLITS) == 0)
      printf("No Splitting, ");
    else
      printf("Maximum of %d Splits, ",flag_GetFlagValue(flag_SPLITS));
  if (flag_GetFlagValue(flag_FULLRED))
    printf("Full Reduction, ");
  else
    printf("Lazy Reduction, ");
  printf(" Ratio: %d, FuncWeight: %d, VarWeight: %d",flag_GetFlagValue(flag_WDRATIO),
	 flag_GetFlagValue(flag_FUNCWEIGHT), flag_GetFlagValue(flag_VARWEIGHT));
  printf("\n Precedence: ");
  fol_PrintPrecedence();
  printf("\n Ordering  : %s",(flag_GetFlagValue(flag_ORD) == flag_ORDKBO ? "KBO" : "RPOS"));
}

void ana_AutoConfiguration(LIST Clauses)
{
  LIST       Functions, Predicates;
  SYMBOL     Symbol;
  SIGNATURE  Entry;
  int        i;

  Functions  = list_Nil();            /* Set precedence */
  Predicates = list_Nil();

  for (i=1;i<symbol_ACTINDEX;i++) {
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i)) {
      Entry  = symbol_Signature(i);
      Symbol = Entry->info;
      if (symbol_IsPredicate(Symbol))
	Predicates = list_Cons((POINTER)Symbol,Predicates);
      else
	if (symbol_IsFunction(Symbol))
	  Functions = list_Cons((POINTER)Symbol,Functions);
    }
  }
  
  Predicates = ana_CalculatePredicatePrecedence(Predicates, Clauses);
  Functions  = ana_CalculateFunctionPrecedence(Functions, Clauses);

  for ( ; !list_Empty(Functions); Functions = list_Pop(Functions))
    symbol_SetOrdering((SYMBOL)list_Car(Functions),
		       symbol_GetIncreasedOrderingCounter());
  /* Predicates < Functions */
  for ( ; !list_Empty(Predicates); Predicates = list_Pop(Predicates))
    symbol_SetOrdering((SYMBOL)list_Car(Predicates),
		       symbol_GetIncreasedOrderingCounter());

  flag_ClearInferenceRules();
  flag_ClearReductionRules();

  flag_SetFlagValue(flag_ROBV,1);
  flag_SetFlagValue(flag_RTAUT,1);
  flag_SetFlagValue(flag_RFSUB,1);
  flag_SetFlagValue(flag_RBSUB,1);
  flag_SetFlagValue(flag_RFCLR,1);
  flag_SetFlagValue(flag_RBCLR,1);
  flag_SetFlagValue(flag_RUNC,1);
  flag_SetFlagValue(flag_FULLRED,1);
  /*flag_SetFlagValue(flag_FUNCWEIGHT,1);
  flag_SetFlagValue(flag_VARWEIGHT,1);*/
  flag_SetFlagValue(flag_WDRATIO,5);

  if (ana_NEQUATIONS)
    flag_SetFlagValue(flag_IEQR,1);

  if (ana_PEQUATIONS) {
    flag_SetFlagValue(flag_ISPR,1);
    flag_SetFlagValue(flag_ISPL,1);
    if (ana_NONHORNCLAUSES >0)
      flag_SetFlagValue(flag_IEQF,1);
    if (ana_NONUNIT) 
      flag_SetFlagValue(flag_RCON,1);
    flag_SetFlagValue(flag_RFREW,1);
    flag_SetFlagValue(flag_RBREW,1);
  }
  
  if (ana_SORTRES) {
    flag_SetFlagValue(flag_IEMS,1);
    flag_SetFlagValue(flag_ISOR,1);
    flag_SetFlagValue(flag_RSSI,1);
    if (!ana_PEQUATIONS || ana_SORTDECEQUATIONS)
      flag_SetFlagValue(flag_RSST,1);
  }
  else
    flag_SetFlagValue(flag_SORTS,0);

  if (ana_MONADIC || ana_NONMONADIC) { /* Problem contains real predicates */
    flag_SetFlagValue(flag_IGER,1);
    if (ana_NONHORNCLAUSES > 0)
      flag_SetFlagValue(flag_IGEF,1);
    if (ana_NONUNIT) 
      flag_SetFlagValue(flag_RCON,1);
  }


  if (!ana_FUNCTIONS)
    flag_SetFlagValue(flag_SELECT,2);
  else
    if (ana_NONUNIT)
      flag_SetFlagValue(flag_SELECT,1);
    else
      flag_SetFlagValue(flag_SELECT,0);

  if (ana_CONCLAUSES < ana_AXIOMCLAUSES || (ana_CONGROUND && !ana_PUREPROPOSITIONAL))
    flag_SetFlagValue(flag_SATINPUT,1);
  else
    flag_SetFlagValue(flag_SATINPUT,0);

  if (ana_NONHORNCLAUSES > 0)
    flag_SetFlagValue(flag_SPLITS,-1);
  else
    flag_SetFlagValue(flag_SPLITS,0);
}


static LIST ana_CalculatePredicatePrecedence(LIST Predicates, LIST Clauses)
/**************************************************************
  INPUT:   A list of predicates and a list of clauses.
  RETURNS: A list of predicate symbols, which should be used
           for setting the symbol precedence. The list is sorted
           in descending order, that means predicates with highest
           precedence come first.
  EFFECT:  Analyze the clause list to build a directed graph G where
           the predicates are vertices. There's an edge (P,Q) in
           G iff a clause exists where P is a negative literal
           and Q is a positive literal and P != Q. Apply DFS to
           find the strongly connected components of this graph.
	   The <Predicates> list is deleted.
  WARNING: The predicate list must contain ALL predicates
           ocurring in the clause list!
***************************************************************/
{
  GRAPH  graph;
  LIST   result, scan;
  int    i, j;
  NAT    count;
  SYMBOL s;

  /* clause_ListPrint(Clauses); DBG */

  if (list_Empty(Predicates)) {
    return Predicates;
  }

  graph = graph_Create();

  /* First create the nodes: one node for every predicate symbol. */
  for ( ; !list_Empty(Predicates); Predicates = list_Pop(Predicates))
    graph_AddNode(graph, symbol_Index((SYMBOL)list_Car(Predicates)));

  /* Now scan the clause clause list to create the edges */
  /* An edge (P,Q) means P is smaller than Q */
  for (scan = Clauses; !list_Empty(scan); scan = list_Cdr(scan)) {
    CLAUSE c = list_Car(scan);

    for (i = clause_FirstLitIndex(); i < clause_FirstSuccedentLitIndex(c); i++) {
      SYMBOL negPred = term_TopSymbol(clause_GetLiteralAtom(c, i));
      if (!symbol_Equal(negPred, fol_Equality())) { /* negative predicate */
	for (j = clause_FirstSuccedentLitIndex(c); j < clause_Length(c); j++) {
	  SYMBOL posPred = term_TopSymbol(clause_GetLiteralAtom(c, j));
	  if (!symbol_Equal(posPred, fol_Equality()) && /* positive predicate */
	      negPred != posPred) {  /* No self loops! */
	    graph_AddEdge(graph_GetNode(graph, symbol_Index(negPred)),
			  graph_GetNode(graph, symbol_Index(posPred)));
	  }
	}
      }
    }
  }

  /* graph_Print(graph); fflush(stdout); DBG */

  /* Calculate the strongly connected components of the graph */
  count = graph_StronglyConnectedComponents(graph);

  /* Now create the precedence list by scanning the nodes.        */
  /* If there's a link between two strongly connected components  */
  /* c1 and c2 then component_num(c1) > component_num(c2), so the */
  /* following code creates a valid precedence list in descending */
  /* order.                                                       */
  result = list_Nil();
  for (i = count - 1; i >= 0; i--) {
    for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
      GRAPHNODE n = list_Car(scan);
      if (graph_NodeCompNum(n) == i) {
	/* The symbol represented by the node <<n> belongs to component <i> */
	s = symbol_IndexPredSymbol(graph_NodeNumber(n));
	result = list_Cons((POINTER)s, result);
      }
    }
  }

  /* putchar('\n');
     for (scan = result; !list_Empty(scan); scan = list_Cdr(scan)) {
     s = (SYMBOL) list_Car(scan);
     printf("%s ", symbol_Name(s));
     }
     putchar('\n'); fflush(stdout); DBG */

  graph_Delete(graph);

  return result;
}

/* We use the node info to store the degree of the node */
static __inline__ NAT ana_NodeDegree(GRAPHNODE Node)
{
  return (NAT)graph_NodeInfo(Node);
}

static __inline__ void ana_IncNodeDegree(GRAPHNODE Node)
{
  graph_NodeSetInfo(Node, (POINTER)(ana_NodeDegree(Node)+1));
}

static BOOL ana_NodeGreater(GRAPHNODE N1, GRAPHNODE N2)
/**************************************************************
  INPUT:   Two graph nodes.
  RETURNS: TRUE, if N1 is greater than N2.
  EFFECT:  This function is used to sort the node list
           of the graph in ana_CalculateFunctionPrecedence.
***************************************************************/
{
  return (symbol_Arity(symbol_IndexPredSymbol(graph_NodeNumber(N1))) >
	  symbol_Arity(symbol_IndexPredSymbol(graph_NodeNumber(N2))));
}

static BOOL ana_BidirectionalDistributivity(LIST SymbolPairs)
/**************************************************************
  INPUT:   A list of symbol pairs defining distributivity.
  RETURNS: TRUE, if the list contains two pairs (s1, s2) and (s2, s1)
           FALSE otherwise.
  EFFECT:  This function is used to detect symbols that are distributive
           in both directions, logical OR and AND for example.
***************************************************************/
{
  LIST scan, actPair, nextPair;

  for ( ; !list_Empty(SymbolPairs); SymbolPairs = list_Cdr(SymbolPairs)) {
    actPair = list_Car(SymbolPairs);
    /* If actPair = (s1, s2), check wether there's a pair (s2, s1) in list */
    for (scan = list_Cdr(SymbolPairs); !list_Empty(scan); scan = list_Cdr(scan)) {
      nextPair = list_Car(scan);
      if (symbol_Equal((SYMBOL)list_PairFirst(actPair),(SYMBOL)list_PairSecond(nextPair)) &&
	  symbol_Equal((SYMBOL)list_PairSecond(actPair),(SYMBOL)list_PairFirst(nextPair)))
	return TRUE;
    }
  }
  return FALSE;
}

static LIST ana_CalculateFunctionPrecedence(LIST Functions, LIST Clauses)
/**************************************************************
  INPUT:   A list of functions and a list of clauses.
  RETURNS: A list of function symbols, which should be used
           for setting the symbol precedence. The list is sorted
           in descending order, that means function with highest
           precedence come first.
  EFFECT:  Analyzes the clauses to build a directed graph G with
           function symbol as nodes. An edge (f,g) or in G means
           f should have lower precedence than g.
           An edge (f,g) or (g,f) is created if there's an equation
           equal(f(...), g(...)) in the clause list.
	   The direction of the edge depends on the degree of the
           nodes and the symbol arity.
	   Then find the strongly connected components of this
           graph.
***************************************************************/
{
  GRAPH     graph;
  GRAPHNODE n1, n2;
  LIST      result, scan, scan2, distrPairs;
  NAT       i, j;
  SYMBOL    s, Add, Mult;

  if (list_Empty(Functions))
    return Functions;   /* Problem contains no functions */
  else if (!ana_PEQUATIONS) {
    Functions = list_NumberSort(Functions, symbol_PositiveArity);;
    return Functions;
  }

  graph = graph_Create();
  /* First create the nodes: one node for every function symbol. */
  for (; !list_Empty(Functions); Functions = list_Pop(Functions))
    graph_AddNode(graph, symbol_Index((SYMBOL)list_Car(Functions)));

  /* Now sort the node list wrt descending symbol arity. */
  graph_SortNodes(graph, ana_NodeGreater);

  /* A list of pairs (add, multiply) of distributive symbols */
  distrPairs = list_Nil();

  /* Now add undirected edges: there's an undirected edge between  */
  /* two nodes if the symbols occur as top symbols in a positive   */
  /* equation. */
  for (scan = Clauses; !list_Empty(scan); scan = list_Cdr(scan)) {
    CLAUSE c = list_Car(scan);
    for (i = clause_FirstSuccedentLitIndex(c);
	 i <= clause_LastSuccedentLitIndex(c); i++) {
      if (clause_LiteralIsEquality(clause_GetLiteral(c, i))) {
	/* Consider only positive equations */
	TERM t1, t2;

	if (fol_DistributiveEquation(clause_GetLiteralAtom(c,i), &Add, &Mult)) {
	  /* Add a pair (Add, Mult) to <distrTerms> */
	  distrPairs = list_Cons(list_PairCreate((POINTER)Add, (POINTER)Mult),
				 distrPairs);
	  /*printf("\nDISTRIBUTIVITY: ");
	    term_PrintPrefix(clause_GetLiteralAtom(c,i));
	    printf(" Add="); symbol_Print(Add);
	    printf(" Mult="); symbol_Print(Mult); fflush(stdout); DBG */
	}

	t1 = term_FirstArgument(clause_GetLiteralAtom(c, i));
	t2 = term_SecondArgument(clause_GetLiteralAtom(c, i));

	if  (!term_IsVariable(t1) && !term_IsVariable(t2) &&
	     !term_EqualTopSymbols(t1, t2) &&  /* No self loops! */
	     !term_HasSubterm(t1, t2) &&       /* No subterm property */
	     !term_HasSubterm(t2, t1)) {
	  n1 = graph_GetNode(graph, symbol_Index(term_TopSymbol(t1)));
	  n2 = graph_GetNode(graph, symbol_Index(term_TopSymbol(t2)));
	  /* Create an undirected edge by adding two directed edges */
	  graph_AddEdge(n1, n2);
	  graph_AddEdge(n2, n1);
	  /* Use the node info for the degree of the node */
	  ana_IncNodeDegree(n1);
	  ana_IncNodeDegree(n2);
	}
      }
    }
  }
  
  /* putchar('\n');
     for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
     n1 = list_Car(scan);
     printf("(%s,%d,%u), ",
     symbol_Name(symbol_IndexPredSymbol(graph_NodeNumber(n1))),
     graph_NodeNumber(n1), ana_NodeDegree(n1));
     }
     graph_Print(graph); fflush(stdout); DBG */

  graph_DeleteDuplicateEdges(graph);
  
  /* Transform the undirected graph into a directed graph. */
  for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
    n1 = list_Car(scan);
    result = list_Nil(); /* Collect edges from n1 that shall be deleted */ 
    for (scan2 = graph_NodeNeighbors(n1); !list_Empty(scan2);
	 scan2 = list_Cdr(scan2)) {
      int a1, a2;
      n2 = list_Car(scan2);
      /* Get the node degrees in the undirected graph with multiple edges */
      i  = ana_NodeDegree(n1);
      j  = ana_NodeDegree(n2);
      a1 = symbol_Arity(symbol_IndexPredSymbol(graph_NodeNumber(n1)));
      a2 = symbol_Arity(symbol_IndexPredSymbol(graph_NodeNumber(n2)));

      if (i > j || (i==j && a1 >= a2)) {
	/* symbol2 <= symbol1, so remove edge n1 -> n2 */
	result = list_Cons(n2, result);
      }
      if (i < j || (i==j && a1 <= a2)) {
	/* symbol1 <= symbol2, so remove edge n2 -> n1 */
	graph_DeleteEdge(n2, n1);
      }
      /* NOTE: If (i==j && a1==a2) both edges are deleted! */
    }
    /* Now delete edges from n1 */
    for ( ; !list_Empty(result); result = list_Pop(result))
      graph_DeleteEdge(n1, list_Car(result));
  }

  if (!list_Empty(distrPairs) && !ana_BidirectionalDistributivity(distrPairs)) {
    /* Enable RPO ordering, otherwise the default KBO will be used. */
    flag_SetFlagValue(flag_ORD, flag_ORDRPOS);
  }

  /* Now examine the list of distribute symbols */
  /* since they've highest priority.                  */
  for ( ; !list_Empty(distrPairs); distrPairs = list_Pop(distrPairs)) {
    scan = list_Car(distrPairs); /* A pair (Add, Mult) */
    /* Addition */
    n1 = graph_GetNode(graph,
		       symbol_Index((SYMBOL)list_PairFirst(scan)));
    /* Multiplication */
    n2 = graph_GetNode(graph, 
		       symbol_Index((SYMBOL)list_PairSecond(scan)));
    /* Remove any edges between n1 and n2 */
    graph_DeleteEdge(n1, n2);
    graph_DeleteEdge(n2, n1);
    /* Add one edge Addition -> Multiplication */
    graph_AddEdge(n1, n2);
    list_PairFree(scan);
  }

  /* fputs("\n------------------------",stdout);
     graph_Print(graph); fflush(stdout); DBG */

  /* Calculate the strongly connected components of the graph. */
  /* <i> is the number of SCCs. */
  i = graph_StronglyConnectedComponents(graph);

  /* Now create the precedence list by scanning the nodes.        */
  /* If there's a link between two strongly connected components  */
  /* c1 and c2 then component_num(c1) > component_num(c2), so the */
  /* following code creates a valid precedence list in descending */
  /* order.                                                       */
  result = list_Nil();
  while (i-- > 0) {   /* for i = numberOfSCCs -1 dowto 0 */
    for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
      n1 = list_Car(scan);
      if (graph_NodeCompNum(n1) == i) {
	/* The symbol represented by the node <n> belongs to component <i> */
	s = symbol_IndexPredSymbol(graph_NodeNumber(n1));
	result = list_Cons((POINTER)s, result);
      }
    }
  }

  /* putchar('\n');
     for (scan = result; !list_Empty(scan); scan = list_Cdr(scan)) {
     s = (SYMBOL) list_Car(scan);
     printf("%s > ", symbol_Name(s));
     }
     putchar('\n'); fflush(stdout); DBG */

  graph_Delete(graph);

  return result;
}
