/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                  CLAUSE APPROXIMATION                  * */
/* *                                                        * */
/* *  $Module:   APPROX                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.42 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 18:05:42 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: approx.c,v $ */

#include <math.h>
#include "approx.h"
#include "clause.h"


static char   approx_BUFFER[symbol__SYMBOLMAXLEN];
static NAT    approx_PREDICATECOUNTER; /* Used by approx_MonadicFlattenHeads */

static SYMBOL approx_TRUESYMBOL; /* Used by approx_MonadicByTermEncoding */
const  char   approx_TRUENAME[] = "True";

typedef struct {
  SYMBOL original;
  LIST   newSymbols;
} APPROX_NODE, *APPROX_TRANSLATION;

void approx_Init(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Initialize the module's variables.
***************************************************************/
{
  approx_PREDICATECOUNTER = 0;
  approx_TRUESYMBOL = symbol_Null();
}

static void approx_DeleteTranslation(LIST Translation)
/**************************************************************
  INPUT:   A list of APPROX_TRANSLATIONs.
  RETURNS: Nothing.
  EFFECT:  All memory used by the translation is freed.
***************************************************************/
{
  for ( ; !list_Empty(Translation); Translation = list_Pop(Translation)) {
    APPROX_TRANSLATION t = list_Car(Translation);
    list_Delete(t->newSymbols);
    memory_Free(t, sizeof(APPROX_NODE));
  }
}

static APPROX_TRANSLATION approx_LookupTranslationSymbols(LIST Translation,
							  SYMBOL Symbol)
/**************************************************************
  INPUT:   A list of APPROX_TRANSLATIONs and a symbol.
  RETURNS: If the list contains a translation for the <Symbol>,
           the corresponding APPROX_TRANSLATION is returned,
           otherwise NULL is returned.
***************************************************************/
{
  APPROX_TRANSLATION trans;
  for (trans = NULL; trans == NULL && !list_Empty(Translation);
       Translation = list_Cdr(Translation)) {
    SYMBOL s = ((APPROX_TRANSLATION)list_Car(Translation))->original;
    if (symbol_Equal(Symbol, s))
      trans = list_Car(Translation);
  }
  return trans;
}


/**************************************************************/
/* Help functions for approx_MonadicByProjection              */
/**************************************************************/

static APPROX_TRANSLATION approx_CreateProjectionSymbols(SYMBOL Symbol)
/**************************************************************
  INPUT:   A n-ary predicate symbol.
  RETURNS: An APPROX_TRANSLATION containing n new,
           unary predicate symbols.
***************************************************************/
{
  APPROX_TRANSLATION result;
  SYMBOL predicate;
  NAT    i, suffix, pos, arity;

  result = (APPROX_TRANSLATION) memory_Malloc(sizeof(APPROX_NODE));
  result->original   = Symbol;
  result->newSymbols = list_Nil();

  arity = symbol_Arity(Symbol);
  i     = symbol__SYMBOLMAXLEN - (NAT)log10(arity) - 2;/* maximum characters */
  pos   = strlen(symbol_Name(Symbol));
  if (pos >= i)
    pos = i;
  strcpy(approx_BUFFER, symbol_Name(Symbol));
  /* If <Symbol> has arity <n> create <n> new symbols */
  for (i = 1; i <= arity; ++i) {
    suffix = i;
    /* Ensure that new symbol wasn't previously defined */
    do {
      sprintf(approx_BUFFER+pos, "%u", suffix++);
    } while (symbol_Lookup(approx_BUFFER) != 0);
    predicate = symbol_CreatePredicate(approx_BUFFER,1,symbol_STATLEX);
    result->newSymbols = list_Cons((POINTER)predicate, result->newSymbols);
  }
  result->newSymbols = list_NReverse(result->newSymbols);
  return result;
}

static LIST approx_ApplyProjection(APPROX_TRANSLATION Translation,
				   CLAUSE Clause, int LitIndex)
/**************************************************************
  INPUT:   An APPROX_TRANSLATION, a clause and a literal index
           within the clause.
  RETURNS: A list of new clauses generated from <Clause>.
  EFFECT:  The non-monadic literal with arity n at index <LitIndex>
           is replaced by monadic literals, resulting in n new
           clauses.
***************************************************************/
{
  SYMBOL symbol;
  LIST   result, scan, argscan;
  TERM   atom;
  CLAUSE newClause;
  
  result = list_Nil();
  atom   = clause_GetLiteralAtom(Clause, LitIndex);
  /* Now create the new clauses */
  for (scan = Translation->newSymbols, argscan = term_ArgumentList(atom);
       !list_Empty(scan); scan = list_Cdr(scan), argscan = list_Cdr(argscan)) {
    symbol = (SYMBOL) list_Car(scan);  /* The new Predicate */
    /* Create the new monadic atom */
    atom = term_Create(symbol, list_List(term_Copy(list_Car(argscan))));
    newClause = clause_Copy(Clause);
    /* Delete the old literal */
    clause_LiteralDelete(clause_GetLiteral(newClause,LitIndex));
    /* and replace it with the new one */
    if (LitIndex < clause_FirstSuccedentLitIndex(newClause)) {
      clause_SetLiteral(newClause, LitIndex,
			clause_LiteralCreateNegative(atom, newClause));
    } else {
      clause_SetLiteral(newClause, LitIndex,
			clause_LiteralCreate(atom, newClause));
    }
    clause_Normalize(newClause);
    clause_SetWeight(newClause, clause_ComputeWeight(newClause));
    clause_UpdateMaxVar(newClause);
    result = list_Nconc(result, list_List(newClause));
  }
  return result;
}

static LIST approx_ApplyProjectionToList(LIST* Translation, LIST Clauses)
/**************************************************************
  INPUT:   A list of APPROX_TRANSLATIONs and a list of clauses
           that were approximated from the same "father" clause.
  RETURNS: A list of clauses.
  EFFECT:  This recursive function is initially called with
           a one-element list containing a clause.
           It searches for a non-monadic literal in the first
           clause of the list. If there's no such literal
           the list is just returned. If a non-monadic literal
           was found, all clauses in the list are transformed
           in the same way, adding new clauses to the list.
           Finally the function is called recursively with the
           new list to check for other non-monadic literals.
***************************************************************/
{
  APPROX_TRANSLATION trans;
  CLAUSE clause;
  SYMBOL symbol;
  LIST   result;
  BOOL   hit;
  NAT    i;

  /* Check wether projection is possible for the first clause */
  clause = list_Car(Clauses);
  for (hit = FALSE, i = clause_FirstLitIndex();
       !hit && i <= clause_LastLitIndex(clause); i++) {
    symbol = term_TopSymbol(clause_GetLiteralAtom(clause, i));
    if (!symbol_Equal(fol_Equality(),symbol) && symbol_Arity(symbol) > 1)
      hit = TRUE;
  }
  --i;   /* Important! */
  /* If projection wasn't possible for the first clause, */
  /* it isn't possible for the other clauses.            */
  /* If projection was possible for the first clause,    */
  /* it's also applicable to the other clauses at the    */
  /* same literal index.                                 */
  if (!hit) {
    return Clauses;     /* Stop recursion */
  } else {
    symbol = term_TopSymbol(clause_GetLiteralAtom(clause, i));
    /* Examine wether the symbol was translated earlier */
    trans = approx_LookupTranslationSymbols(*Translation, symbol);
    if (trans == NULL) {  /* Symbol not found in translation list */
      trans = approx_CreateProjectionSymbols(symbol);
      *Translation = list_Cons(trans, *Translation);
    }
    /* Apply projection at literal index <i> */
    result = list_Nil();
    for ( ; !list_Empty(Clauses); Clauses = list_Pop(Clauses)) {
      result = list_Nconc(result,
			  approx_ApplyProjection(trans,list_Car(Clauses),i));
      /* Delete the old clause */
      clause_Delete(list_Car(Clauses));
    }
    return approx_ApplyProjectionToList(Translation, result);  /* recursion */
  }
}
			  
LIST approx_MonadicByProjection(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: A list of clauses with monadic literals. If a clause
           contains a n-ary atom P(t_1,...t_n), n clause copies
	   are created where the atom in the i-th new clause is
	   replaced by Pi(t_i).
           Equality is left unchanged.
  EFFECT:  The original clauses are not changed.
           Copies are created even if no projection was applied.
***************************************************************/
{
  LIST result, translation;

  result = translation = list_Nil();
  for( ; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) {
    result = list_Nconc(result, approx_ApplyProjectionToList(&translation,
                list_List(clause_Copy(list_Car(Clauses)))));
  }
  approx_DeleteTranslation(translation);
  list_Apply(clause_NewNumber, result);
  return result;
}


/**************************************************************/
/* Help functions for approx_MonadicByTermEncoding            */
/**************************************************************/

static APPROX_TRANSLATION approx_CreateFunctionForPredicate(SYMBOL Symbol)
/**************************************************************
  INPUT:   A predicate.
  RETURNS: A one-element list with an APPROX_TRANSLATION for
           the predicate.
  EFFECT:  If the predicate has the name "P", a function "P_f" is
           created for the predicate.
***************************************************************/
{
  APPROX_TRANSLATION result;
  SYMBOL predicate;
  NAT    i, pos;

  result = (APPROX_TRANSLATION) memory_Malloc(sizeof(APPROX_NODE));
  result->original   = Symbol;
  result->newSymbols = list_Nil();

  i     = symbol__SYMBOLMAXLEN - 3;  /* maximum characters */
  pos   = strlen(symbol_Name(Symbol));
  if (pos >= i)
    pos = i;
  strcpy(approx_BUFFER, symbol_Name(Symbol));
  strcpy(approx_BUFFER+pos, "_f");
  /* If the predicate symbol name + "_f" is already defined, try additional numbers */
  i = 0;
  while (symbol_Lookup(approx_BUFFER) != 0) {
    sprintf(approx_BUFFER+pos+2, "%u", ++i);
  }
  predicate = symbol_CreateFunction(approx_BUFFER,1,symbol_STATLEX);
  result->newSymbols = list_Cons((POINTER)predicate, result->newSymbols);
  return result;
}

static void approx_EncodeClause(LIST* Translation, CLAUSE Clause)
/**************************************************************
  INPUT:   A clause and a list of APPROX_TRANSLATIONs.
  RETURNS: Nothing.
  EFFECT:  All non-monadic literals of the clause are
           transformed into monadic ones.
           The clause is destructively changed!
***************************************************************/
{
  APPROX_TRANSLATION trans;
  SYMBOL symbol;
  TERM   atom;
  NAT    i;

  for (i = clause_FirstLitIndex(); i <= clause_LastLitIndex(Clause); ++i) {
    symbol = term_TopSymbol(clause_GetLiteralAtom(Clause, i));
    if (!symbol_Equal(fol_Equality(),symbol) && symbol_Arity(symbol) > 1) {
      trans = approx_LookupTranslationSymbols(*Translation, symbol);
      if (trans == NULL) {
	trans = approx_CreateFunctionForPredicate(symbol);
	*Translation = list_Cons(trans, *Translation);
      }
      atom = clause_GetLiteralAtom(Clause, i);
      term_RplacTop(atom, (SYMBOL)list_Car(trans->newSymbols));
      atom = term_Create(approx_TRUESYMBOL, list_List(atom));
      clause_LiteralSetAtom(clause_GetLiteral(Clause, i), atom);
    }      
  }
}

LIST approx_MonadicByTermEncoding(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: A list of clauses with monadic literals. A non-monadic
           literal P(t1,t2,...) is translated to True(p(t1,t2,...))
           where p is a function with the same arity as P
           and True is a special predicate with boolean domain.
	   Equality is left unchanged.
  EFFECT:  The clauses are not changed.
***************************************************************/
{
  LIST result, translation;
  NAT  i, length;

  /* Ensure that "True" is defined */
  if (symbol_Equal(approx_TRUESYMBOL, symbol_Null())) {
    i = 0;
    length = strlen(approx_TRUENAME);
    strcpy(approx_BUFFER, approx_TRUENAME);
    while (symbol_Lookup(approx_BUFFER) != 0) {
      sprintf(approx_BUFFER+length, "%u", ++i);
    }
    approx_TRUESYMBOL = symbol_CreatePredicate(approx_BUFFER, 1, symbol_STATLEX);
  }

  translation = list_Nil();
  result = list_CopyWithElement(Clauses, (POINTER(*)())clause_Copy);
  for(Clauses = result; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) {
    approx_EncodeClause(&translation, list_Car(Clauses));
  }
  approx_DeleteTranslation(translation);
  list_Apply(clause_NewNumber, result);
  return result;
}

/**************************************************************/
/* Help functions for approx_MonadicMakeLinear                */
/**************************************************************/

static LIST approx_MakeTermLinear(TERM Term, NAT Mark)
/**************************************************************
  INPUT:   A TERM and the current term binding mark.
  RETURNS: A list of pairs (oldvar, newvar) of variable renamings.
  EFFECT:  The term is destructively changed.
           The variables within Term are made unique with
           respect to previous variable occurrences since
           the last term_StartBinding call.
***************************************************************/
{
  LIST   scan, result;
  SYMBOL s;

  result = list_Nil();
  if (term_IsComplex(Term)) {
    for(scan=term_ArgumentList(Term);!list_Empty(scan);scan=list_Cdr(scan))
      result = list_Nconc(approx_MakeTermLinear(list_Car(scan),Mark), result);
  } else {
    s = term_TopSymbol(Term);
    if (symbol_IsVariable(s)) {
      if (term_VarIsMarked(s,(POINTER)Mark)) {
	term_RplacTop(Term, symbol_CreateStandardVariable());
	result = list_Cons(list_PairCreate((POINTER)s,(POINTER)term_TopSymbol(Term)),
			   result);
      } else {
        term_CreateValueBinding(s, (POINTER)Mark, NULL);
      }
    }
  }
  return result;
}

static CLAUSE approx_MakeClauseLinear(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause with monadic positive literals.
  RETURNS: The linear approximation of the clause.
  EFFECT:  The original clause is not changed.
  CAUTION: The MaxVar info of the clause must be up to date!
***************************************************************/
{
  LIST   renamings, scan, constraint, antecedent, succedent;
  SYMBOL newvar, oldvar;
  CLAUSE clause;
  TERM   atom, newAtom;
  NAT    mark;
  int    i;

#ifdef DEBUG
  if (term_InBindingPhase()) {
    fputs("\n\nError in approx_makeClauseLinear\n", stderr);
    misc_Error();
  } else {
    term_StartBinding();
  }
#endif

  for (i = clause_FirstLitIndex(); i <= clause_LastLitIndex(Clause); ++i) {
    atom = clause_GetLiteralAtom(Clause, i);
    if (fol_IsEquality(atom) || !term_IsDeclaration(atom))
      return clause_Copy(Clause);
  }

  /* Ensure that all new variables don't occur in the clause */
  symbol_SetStandardVarCounter(clause_MaxVar(Clause));
  
  term_NewMark();
  mark = term_ActMark();

  constraint = antecedent = succedent = list_Nil();
  renamings = list_Nil();
  /* Make the positive literals linear */
  for (i = clause_LastSuccedentLitIndex(Clause);
       i >= clause_FirstSuccedentLitIndex(Clause); --i) {
    atom = term_Copy(clause_GetLiteralAtom(Clause, i));
    succedent = list_Cons(atom, succedent);
    renamings = list_Nconc(approx_MakeTermLinear(atom, mark), renamings);
  }

#ifdef DEBUG
  term_StopBinding();
#endif

  /* Now examine the negative literals */
  for (i = clause_LastAntecedentLitIndex(Clause);
       i >= clause_FirstConstraintLitIndex(Clause); --i) {
    /* Keep a copy of the original literal */
    atom = term_Copy(clause_GetLiteralAtom(Clause, i));
    if (i <= clause_LastConstraintLitIndex(Clause))
      constraint = list_Cons(atom, constraint);
    else
      antecedent = list_Cons(atom, antecedent);
    for (scan = renamings; !list_Empty(scan); scan = list_Cdr(scan)) {
      oldvar = (SYMBOL) list_PairFirst(list_Car(scan));
      newvar = (SYMBOL) list_PairSecond(list_Car(scan));
      if (term_ContainsSymbol(atom, oldvar)) {
	newAtom = term_Copy(atom);
	term_ExchangeVariable(newAtom, oldvar, newvar);
	/* Add additional literals for renamed variables */
	if (i <= clause_LastConstraintLitIndex(Clause))
	  constraint = list_Cons(newAtom, constraint);
	else
	  antecedent = list_Cons(newAtom, antecedent);
      }
    }
  }
  /*putchar('\n');
    term_TermListPrintPrefix(constraint); putchar('\n');
    term_TermListPrintPrefix(antecedent); putchar('\n');
    term_TermListPrintPrefix(succedent); putchar('\n'); DBG */
  clause = clause_Create(constraint, antecedent, succedent);
  list_Delete(constraint);
  list_Delete(antecedent);
  list_Delete(succedent);
  list_DeleteWithElement(renamings, list_PairFree);
  return clause;
}


LIST approx_MonadicMakeLinear(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: A list of copies of the clauses.
  EFFECT:  The original clauses are not changed.
           The copy of a clause is made linear only if all literals
           are monadic and the clause contains no equality.
	   All positive literals are made linear as a whole.
           For every variable renaming X -> Y, that is necessary
           to make the positive literals linear, we collect
           all negative literals P(t) containing X and add new
           negative literals P(t) where X is replaced by Y.
  CAUTION: The MaxVar info of the clause must be up to date!
***************************************************************/
{
  LIST result;
  
  result = list_Copy(Clauses);
  list_NMapCar(result, (POINTER(*)())approx_MakeClauseLinear);
  return result;
}


/**************************************************************/
/* Help functions for approx_MonadicFlattenHeads              */
/**************************************************************/

static LIST approx_ApplyFlattening(CLAUSE Clause, int Index, LIST Item,
				   LIST varsInTerm)
/**************************************************************
  INPUT:   A clause, a literal index in <Clause>, where the
           transformation shall be applied, a list item
           from the argument list of a term in the literal
           and the list of all variable symbols occuring in
           the subterm.
  RETURNS: A list of two clauses.
  EFFECT:  The <Clause> is not changed.
           Creates a new predicate <NewPredicate>.
           1st clause: succedent: copy except <Index>,
               at <Index> replacce Car(Item) by a new variable X
             antecedent and constraint: keep only literals,
               that contain variables from  <varsInTerm>
               + a new literal NewPredicate(X)
           2nd clause: succedent = NewPredicate(term)
             antecedent + constraint: all negative literals from
               <Clause>, that weren't copied to the 1st clause.
***************************************************************/
{
  LIST   constraint1, antecedent1, succedent1;  /* literals for 1st clause */
  LIST   constraint2, antecedent2, succedent2;  /* literals for 2nd clause */
  TERM   term, newVarTerm;
  SYMBOL newPredicate;
  CLAUSE clause1, clause2;
  int    i;

  constraint1 = antecedent1 = succedent1 = list_Nil();
  constraint2 = antecedent2 = succedent2 = list_Nil();
  for (i = clause_LastAntecedentLitIndex(Clause);
       i >= clause_FirstConstraintLitIndex(Clause); --i) {
    SYMBOL var;

    term = term_Copy(clause_GetLiteralAtom(Clause, i));  /* an atom */
    var = term_TopSymbol(term_FirstArgument(term));
    if (i <= clause_LastConstraintLitIndex(Clause)) {
      if (list_PointerMember(varsInTerm, (POINTER)var))
	constraint2 = list_Cons(term, constraint2);
      else
	constraint1 = list_Cons(term, constraint1);
    } else {
      if (list_PointerMember(varsInTerm, (POINTER)var))
	antecedent2 = list_Cons(term, antecedent2);
      else
	antecedent1 = list_Cons(term, antecedent1);
    }
  }
  /* Create a new predicate */
  strcpy(approx_BUFFER, "Fl");
  /* Ensure that new symbol wasn't previously defined */
  do {
    sprintf(approx_BUFFER+2, "%u", ++approx_PREDICATECOUNTER);
  } while (symbol_Lookup(approx_BUFFER) != 0);
  newPredicate = symbol_CreatePredicate(approx_BUFFER, 1, symbol_STATLEX);
  /* Create a new variable not occurring in the clause */
  symbol_SetStandardVarCounter(clause_MaxVar(Clause));
  newVarTerm = term_CreateStandardVariable();
  antecedent1 = list_Cons(term_Create(newPredicate, list_List(newVarTerm)),
			  antecedent1);

  /* First manipulate the succedent literal */
  term = list_Car(Item);
  list_Rplaca(Item, newVarTerm);  /* This is safe since the term is restored later */
  for (i = clause_LastSuccedentLitIndex(Clause);
       i >= clause_FirstSuccedentLitIndex(Clause); --i) {
    TERM atom;
    atom = term_Copy(clause_GetLiteralAtom(Clause, i));
    succedent1 = list_Cons(atom, succedent1);
  }
  /* Now restore the succedent literal */
  list_Rplaca(Item, term);
  
  succedent2 = list_List(term_Create(newPredicate, list_List(term_Copy(term))));
  
  clause1 = clause_Create(constraint1, antecedent1, succedent1);
  clause2 = clause_Create(constraint2, antecedent2, succedent2);
  /* cleanup */
  list_Delete(constraint1);
  list_Delete(antecedent1);
  list_Delete(succedent1);
  list_Delete(constraint2);
  list_Delete(antecedent2);
  list_Delete(succedent2);
  
  return list_Cons(clause1, list_List(clause2)); 
}

static LIST approx_OneFlatteningStep(CLAUSE Clause, BOOL CheckVarCondition)
/**************************************************************
  INPUT:   A clause and a boolean flag.
  RETURNS: An empty list, if the clause wasn't flattened,
           else the list of clauses resulting from the
           flattening of the clause.
  EFFECT:  The <Clause> is not changed.
***************************************************************/
{
  TERM term;
  LIST scan;
  int  i;

  for (i = clause_FirstSuccedentLitIndex(Clause);
       i <= clause_LastSuccedentLitIndex(Clause); ++i) {
    term = clause_GetLiteralAtom(Clause, i);  /* an atom */
    if (fol_IsEquality(term) || !term_IsDeclaration(term)) {
      /* positive literal is equality or isn't monadic */
      return list_Nil();
    }
    term = term_FirstArgument(term); /* the term */
    for (scan = term_ArgumentList(term); !list_Empty(scan);
	 scan = list_Cdr(scan)) {
      term = list_Car(scan);
      if (term_IsComplex(term)) {
	/* found a candidate */
	LIST varsInTerm, varsInAtom, result;
	
	varsInTerm = term_VariableSymbols(term);
	if (CheckVarCondition) {
	  /* Now check the variable condition */
	  /* Replace the actual term by a simple constant */
	  list_Rplaca(scan, term_Create(fol_True(), list_Nil()));
	  varsInAtom = term_VariableSymbols(clause_GetLiteralAtom(Clause, i));
	  term_Delete(list_Car(scan)); /* Delete the constant term */
	  list_Rplaca(scan, term);     /* Restore the old atom */
	  
	  if (list_HasIntersection(varsInAtom, varsInTerm)) {
	    /* term and rest of the literal have common variables */
	    list_Delete(varsInTerm);
	    list_Delete(varsInAtom);
	    return list_Nil();
	  }
	  list_Delete(varsInAtom);
	}
	result = approx_ApplyFlattening(Clause, i, scan, varsInTerm);
	list_Delete(varsInTerm);
	return result;
      }
    }
  }
  return list_Nil();  /* Flattening wasn't necessary */
}


LIST approx_MonadicFlattenHeads(CLAUSE Clause, BOOL CheckVarCondition)
/**************************************************************
  INPUT:   A clause and a boolean flag.
  RETURNS: An empty list, if the clause wasn't flattened,
           else the list of clauses resulting from the
           flattening of the clause.
           The flattening is applied recursively to the clause
           as long as possible.
  EFFECT:  The <Clause> itself is not changed.
	   If the clause doesn't meet the following requirements,
	   it's not flattened and list_Nil() is returned:
           1) all literals must be monadic
           2) all negative literals must have a variable as
              argument
           3) a positive literal P(t[s]) where s isn't a variable
              or constant
           4) additionally, if <CheckVarCondition> is TRUE,
              s mustn't have any variables in common with the rest
	      of the literal
  CAUTION: The clause's maxvar information must be up to date.
***************************************************************/
{
  LIST   result, queue, newClauses;
  CLAUSE actClause;
  TERM   term;
  BOOL   firstTime;
  int    i;

  /* Check the negative literals */
  for (i = clause_FirstConstraintLitIndex(Clause);
       i <= clause_LastAntecedentLitIndex(Clause); ++i) {
    term = clause_GetLiteralAtom(Clause, i);   /* an atom */
    if (fol_IsEquality(term) || !term_IsDeclaration(term)) {
      /* negative literal isn't monadic */
      return list_Nil();
    }
    if (!term_IsVariable(term_FirstArgument(term))) {
      /* negative literal hasn't simple variable as argument */
      return list_Nil();
    }
  }

  result = list_Nil();
  firstTime = TRUE;
  queue = list_List(Clause);  /* Initialize the queue with the first clause */
  do {
    actClause = list_NCar(&queue);
    newClauses = approx_OneFlatteningStep(actClause, CheckVarCondition);
    if (list_Empty(newClauses)) {
      /* actClause is finished */
      result = list_Cons(actClause, result);
    } else {
      /* actClause was transformed */
      queue = list_Nconc(newClauses, queue);
      /* Delete temporary clauses */
      if (!firstTime)  /* Avoid deletion of the initial clause */
	clause_Delete(actClause);
    }
    firstTime = FALSE;
  } while (!list_Empty(queue));

  if (list_Empty(list_Cdr(result))) {
    /* list contains only the original clause */
    list_Delete(result);
    return list_Nil();
  } else
    return list_NReverse(result);
}


/**************************************************************/
/* approx_MakeHorn                                            */
/**************************************************************/

LIST approx_MakeHorn(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: An empty list if <Clause> is already a horn clause,
           otherwise a list of horn clauses resulting from <Clause>.
  EFFECT:  A non-horn clause with n positive literals
           is replaced by n horn clauses. The i-th horn clause
           has the i-th positive literal from <Clause> as
           only positive literal.
***************************************************************/
{
  LIST result, constraint, antecedent, succedent;
  int  i;
  
  result = list_Nil();
  if (clause_NumOfSuccLits(Clause) > 1) {
    constraint = antecedent = list_Nil();
    /* Collect constraint atoms */
    for (i = clause_LastConstraintLitIndex(Clause);
	 i >= clause_FirstConstraintLitIndex(Clause); --i)
      constraint = list_Cons(clause_GetLiteralAtom(Clause, i), constraint);
    /* Collect antecedent atoms */
    for (i = clause_LastAntecedentLitIndex(Clause);
	 i >= clause_FirstAntecedentLitIndex(Clause); --i)
      antecedent = list_Cons(clause_GetLiteralAtom(Clause, i), antecedent);
    for (i = clause_FirstSuccedentLitIndex(Clause);
	 i <= clause_LastSuccedentLitIndex(Clause); ++i) {
      /* Copy constraint and antecedent atoms */
      list_NMapCar(constraint, (POINTER (*)())term_Copy);
      list_NMapCar(antecedent, (POINTER (*)())term_Copy);
      succedent = list_List(term_Copy(clause_GetLiteralAtom(Clause, i)));
      result = list_Cons(clause_Create(constraint, antecedent, succedent),
			 result);
      list_Delete(succedent);
    }
    list_Delete(constraint);
    list_Delete(antecedent);
  }
  return list_NReverse(result);
}
