/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   CLAUSES                              * */
/* *                                                        * */
/* *  $Module:   CLAUSE                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.100 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:46:55 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: clause.c,v $ */

#include "clause.h"

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

NAT  clause_CLAUSECOUNTER;
NAT  clause_STAMPID;
LIST clause_SORT[clause_MAXWEIGHT+1];

/**************************************************************/
/* Primitive literal functions                                */
/**************************************************************/

BOOL clause_LiteralIsLiteral(LITERAL Literal)
/*********************************************************
  INPUT:   A literal.
  RETURNS: TRUE, if 'Literal' is not NULL and has a predicate
           symbol as its atoms top symbol.
  MEMORY:  No Effects.
**********************************************************/
{
  return ((Literal != NULL) &&
	  symbol_IsPredicate(clause_LiteralPredicate(Literal)));
}

static NAT clause_LiteralComputeWeight(LITERAL Literal)
/*********************************************************
  INPUT:   A literal.
  RETURNS: The weight of the literal.
  MEMORY:  No Effects.
**********************************************************/
{
  TERM Term;
  int  Bottom;
  NAT  Number;

#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralComputeWeight\n", stderr);
    misc_DumpCore();
  }
#endif 

  Term = clause_LiteralSignedAtom(Literal);
  Bottom = stack_Bottom();
  Number = 0;

  do {
    if (term_IsComplex(Term)) {
      /* Aenderung uh:  symbol_Weight geht hier mit ein*/
      Number += (symbol_Weight(term_TopSymbol(Term))+flag_GetFlagValue(flag_FUNCWEIGHT));
      stack_Push(term_ArgumentList(Term));
    }
    else
      if (term_IsVariable(Term))
	Number += flag_GetFlagValue(flag_VARWEIGHT);
      else
	/* Aenderung uh:  symbol_Weight geht hier mit ein*/
	Number += (symbol_Weight(term_TopSymbol(Term))+flag_GetFlagValue(flag_FUNCWEIGHT));
      
    while (!stack_Empty(Bottom) && list_Empty(stack_Top()))
      stack_Pop();
    if (!stack_Empty(Bottom)) {
      Term = (TERM) list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(Bottom));

  return(Number);

}

LITERAL clause_LiteralCreate(TERM Atom, CLAUSE Clause)
/**********************************************************
  INPUT:   A Pointer to a signed Predicate-Term and one to a
           clause it shall belong to.
  RETURNS: An appropriate literal where the other values 
           are defaulted.
  MEMORY:  A LITERAL_NODE is allocated.
***********************************************************/
{
  LITERAL Result;

#ifdef DEBUG
  if (!term_IsTerm(Atom)) {
    fputs("\n\nIllegal input in clause_LiteralCreate\n", stderr);
    misc_DumpCore();
  }
  if (!symbol_IsPredicate(term_TopSymbol(Atom)) &&
      (term_TopSymbol(Atom) != fol_Not())) {
    fputs("\n\nIllegal input (no Predicate- or negated Term!) "
	   "in clause_LiteralCreate\n", stderr);
    misc_DumpCore();
  }
#endif 

  Result                = (LITERAL)memory_Malloc(sizeof(LITERAL_NODE));
  
  Result->atomWithSign  = Atom;
  Result->oriented      = FALSE;
  Result->weight        = 0;
  Result->maxLit        = 0;
  Result->owningClause  = Clause;

  return Result;
}

LITERAL clause_LiteralCreateNegative(TERM Atom, CLAUSE Clause)
/**********************************************************
  INPUT:   A Pointer to a unsigned Predicate-Term and one to a
           clause it shall belong to.
  RETURNS: An appropriate literal where the other values 
           are defaulted and the term gets a sign.
  MEMORY:  A LITERAL_NODE is allocated.
***********************************************************/
{
  LITERAL Result;

#ifdef DEBUG
  if (!term_IsTerm(Atom)) {
    fputs("\n\nIllegal input in clause_LiteralCreateNegative\n", stderr);
    misc_DumpCore();
  }
  if (!symbol_IsPredicate(term_TopSymbol(Atom)) &&
      (term_TopSymbol(Atom) != fol_Not())) {
    fputs("\n\nIllegal input (no Predicate- or negated Term!) "
	  "in clause_LiteralCreateNegative\n", stderr);
    misc_DumpCore();
  }
#endif 

  Result                = (LITERAL)memory_Malloc(sizeof(LITERAL_NODE));
  
  term_RplacSupertermList(Atom, list_Nil());

  Result->atomWithSign  = term_Create(fol_Not(), list_List(Atom));
  Result->oriented      = FALSE;
  Result->maxLit        = 0;
  Result->weight        = 0;
  Result->owningClause  = Clause;

  return Result;
}


void clause_LiteralDelete(LITERAL Literal)
/*********************************************************
  INPUT:   A literal, which has an unshared Atom. For Shared
           literals clause_LiteralDeleteFromSharing(Lit) is 
	   available.
  RETURNS: Nothing.
  MEMORY:  The Atom of 'Literal' is deleted and its memory
           is freed as well as the LITERAL_NODE.
**********************************************************/
{
#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralDelete\n", stderr);
    misc_DumpCore();
  }
#endif 

  term_Delete(clause_LiteralSignedAtom(Literal));

  clause_LiteralFree(Literal);
}


void clause_LiteralInsertIntoSharing(LITERAL Lit, SHARED_INDEX ShIndex)
/**********************************************************
  INPUT:   A Literal with an unshared Atom and an Index.
  RETURNS: The literal with a shared Atom inserted into the
           'Index'.
  MEMORY:  Allocates TERM_NODE memory needed to insert 'Lit' 
           into the sharing and frees the memory of the 
           unshared Atom.
***********************************************************/
{

  TERM Atom;
  
#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Lit)){
    fputs("\n\nIllegal literal input in clause_LiteralInsertIntoSharing\n",
	  stderr);
    misc_DumpCore();
  }
#endif
  
  Atom = clause_LiteralAtom(Lit);
  
  clause_LiteralSetAtom(Lit, sharing_Insert(Lit, Atom, ShIndex));
  
  term_Delete(Atom);  
}


void clause_LiteralDeleteFromSharing(LITERAL Lit, SHARED_INDEX ShIndex)
/**********************************************************
  INPUT:   A Literal and an 'Index'.
  RETURNS: Nothing.
  MEMORY:  Deletes 'Lit' from Sharing, frees no more used
           TERM memory and frees the LITERAL_NODE.
********************************************************/
{

  TERM Atom;

#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Lit)){
    fputs("\n\nIllegal literal input in clause_LiteralDeleteFromSharing\n",
	  stderr);
    misc_DumpCore();
  }
#endif

  Atom = clause_LiteralAtom(Lit);

  /*puts("\nAufruf von clause_LiteralDeleteFromSharing 
    mit Literal bzw Atom");
    clause_LiteralPrintSigned(Lit); term_Print(Atom);*/

  if (clause_LiteralIsNegative(Lit)) {
    list_Free(term_ArgumentList(clause_LiteralSignedAtom(Lit)));
    term_Free(clause_LiteralSignedAtom(Lit));
  }
 
  sharing_Delete(Lit, Atom, ShIndex);

  clause_LiteralFree(Lit);

}


static LIST clause_CopyLitInterval(CLAUSE Clause, int Start, int End)
/**************************************************************
 INPUT:   A clause and two integers representing
          literal indeces.
 RETURNS: Copies of all literals in <Clause> with index i and
          in the interval [Start:End] are prepended to <List>.
 MEMORY:  All atoms are copied.
***************************************************************/
{
  TERM Atom;
  LIST List;

  List = list_Nil();

  for ( ; Start <= End; Start++) {
    Atom = term_Copy(clause_GetLiteralAtom(Clause, Start));
    List = list_Cons(Atom, List);
  }
  
  return(List);
}



LIST clause_CopyConstraint(CLAUSE Clause)
/**************************************************************
 INPUT:   A clause.
 RETURNS: The list of copied constraint literals from <Clause>.
 EFFECTS: None
***************************************************************/
{
  return(clause_CopyLitInterval(Clause,  
				clause_FirstConstraintLitIndex(Clause), 
				clause_LastConstraintLitIndex(Clause)));
}

LIST clause_CopySuccedent(CLAUSE Clause)
/**************************************************************
 INPUT:   A clause.
 RETURNS: The list of copied succedent literals from <Clause>.
 EFFECTS: None
***************************************************************/
{
  return(clause_CopyLitInterval(Clause,
				clause_FirstSuccedentLitIndex(Clause), 
				clause_LastSuccedentLitIndex(Clause)));
}


/**************************************************************/
/* Specials                                                   */
/**************************************************************/

BOOL clause_IsUnorderedClause(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the invariants concerning splitting etc. hold
           Invariants concerning maximality restrictions are not tested.
**********************************************************/
{
  return((Clause != NULL) &&
	 clause_CheckSplitLevel(Clause) &&
	 (clause_IsEmptyClause(Clause) ||
	  clause_LiteralIsLiteral(clause_GetLiteral(Clause,clause_FirstLitIndex()))) &&
	 (clause_SplitLevel(Clause) == 0 || Clause->splitfield_length>0) &&
	 clause_DependsOnSplitLevel(Clause,clause_SplitLevel(Clause)));
}


BOOL clause_IsClause(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if literals are correctly ordered and the
           invariants concerning splitting etc. hold
  MEMORY:  No Effects.
**********************************************************/
{
  int     i;
  LITERAL ActLit;

  for (i=clause_FirstAntecedentLitIndex(Clause);i<=clause_LastSuccedentLitIndex(Clause);i++) {
    ActLit = clause_GetLiteral(Clause,i);
    if (fol_IsEquality(clause_LiteralSignedAtom(ActLit))) {
      ord_RESULT HelpRes;
	
      HelpRes =
	ord_Compare(term_FirstArgument(clause_LiteralSignedAtom(ActLit)), 
		    term_SecondArgument(clause_LiteralSignedAtom(ActLit)));
	
      if (ord_IsSmallerThan(HelpRes)) { 
	printf("\nERROR, first argument smaller than second in: ");
	term_Print(clause_LiteralSignedAtom(ActLit));
	/* Can't print clause as clause_Print calls clause_IsClause !!! */
	/* printf("\nin clause: "); clause_Print(Clause); */
	return FALSE;
      }
    }
  }

  return clause_IsUnorderedClause(Clause);
	  
}


BOOL clause_ContainsPositiveEquations(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains a positive equality literal.
  MEMORY:  No Effects.
**********************************************************/
{
  int i;
  
  for (i = clause_FirstSuccedentLitIndex(Clause);
       i < clause_Length(Clause);
       i++)
    if (clause_LiteralIsEquality(clause_GetLiteral(Clause, i)))
      return TRUE;

  return FALSE;
}

BOOL clause_ContainsNegativeEquations(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains a positive equality literal.
  MEMORY:  No Effects.
**********************************************************/
{
  int i;
  
  for (i = clause_FirstAntecedentLitIndex(Clause);
       i < clause_FirstSuccedentLitIndex(Clause);
       i++)
    if (clause_LiteralIsEquality(clause_GetLiteral(Clause, i)))
      return TRUE;

  return FALSE;
}

int clause_ContainsFolAtom(CLAUSE Clause, BOOL *Prop, BOOL *Grd, 
		BOOL *Monadic, BOOL *NonMonadic,
		BOOL *Dyadic, BOOL *NonDyadic)
/*********************************************************
  INPUT:   A clause.
  RETURNS: The number of boolean variables changed.
           If <*Prop> is FALSE and the clause contains a propositional variable,
	      it is changed to TRUE
	   If <*Grd> is FALSE and the clause contains a non-propositional ground atom,
	      it is changed to TRUE
           If <*Monadic> is FALSE and the clause contains a monadic atom,
	      it is changed to TRUE
	   If <*NonMonadic> is FALSE and the clause contains a at least 2-place non-equality atom,
	      it is changed to TRUE
           If <*Dyadic> is FALSE and the clause contains a dyadic atom,
	      it is changed to TRUE
	   If <*NonDyadic> is FALSE and the clause contains a at least 3-place non-equality atom,
	      it is changed to TRUE
**********************************************************/
{
  int  i,Result,Arity;
  BOOL Ground;

  Result = 0;
  i      = clause_FirstLitIndex();

  while (Result < 6 && 
	 i < clause_Length(Clause) &&
	 (!(*Prop) || !(*Monadic) || !(*NonMonadic))) {
    Arity  = symbol_Arity(term_TopSymbol(clause_GetLiteralAtom(Clause,i)));
    Ground = term_IsGround(clause_GetLiteralAtom(Clause,i));
    if (!(*Prop) && Arity == 0) {
      Result++;
      *Prop = TRUE;
    }
    if (!(*Grd) && Arity > 0 && Ground &&
	!fol_IsEquality(clause_GetLiteralAtom(Clause,i))) {
      Result++;
      *Grd = TRUE;
    }
    if (!(*Monadic) && Arity == 1 && !Ground) {
      Result++;
      *Monadic = TRUE;
    }
    if (!(*Dyadic) && Arity == 2 && !Ground &&
	!fol_IsEquality(clause_GetLiteralAtom(Clause,i))) {
      Result++;
      *Dyadic = TRUE;
      *NonMonadic = TRUE;
    }
    if (!(*NonMonadic) && Arity > 1 && !Ground &&
	!fol_IsEquality(clause_GetLiteralAtom(Clause,i))) {
      Result++;
      *NonMonadic = TRUE;
    }
    if (!(*NonDyadic) && Arity > 2 && !Ground) {
      Result++;
      *NonDyadic = TRUE;
    }
    i++;
  }
  return Result;
}

BOOL clause_ContainsVariables(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains at least one variable
  MEMORY:  No Effects.
**********************************************************/
{
  int  i;
  TERM Term;

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
    Term = clause_GetLiteralAtom(Clause,i);
    if (term_NumberOfVarOccs(Term)>0)
      return TRUE;
  }

  return FALSE;
}

void clause_ContainsSortRestriction(CLAUSE Clause, BOOL *Sortres, BOOL *USortres)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains a negative monadic atom with
           a variable argument
  MEMORY:  No Effects.
**********************************************************/
{
  int  i;
  TERM Term;

  for (i = clause_FirstLitIndex(); 
       i <= clause_LastAntecedentLitIndex(Clause) && (!*Sortres || !*USortres); 
       i++) {
    Term = clause_GetLiteralAtom(Clause,i);
    if (symbol_IsBaseSort(term_TopSymbol(Term))) { 
      *USortres = TRUE;
      if (symbol_IsVariable(term_TopSymbol(term_FirstArgument(Term))))
	*Sortres = TRUE;
    }
  }
}

BOOL clause_ContainsConstants(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains at least one constant symbol
  MEMORY:  No Effects.
**********************************************************/
{
  int  i;
  TERM Term;

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
    Term = clause_GetLiteralAtom(Clause,i);
    if (term_ContainsConstants(Term))
      return TRUE;
  }

  return FALSE;
}

BOOL clause_ContainsFunctions(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains at least one (non-nullary)
           function symbol
  MEMORY:  No Effects.
**********************************************************/
{
  int  i;
  TERM Term;

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
    Term = clause_GetLiteralAtom(Clause,i);
    if (term_ContainsFunctions(Term))
      return TRUE;
  }

  return FALSE;
}

BOOL clause_ContainsSymbol(CLAUSE Clause, SYMBOL Symbol)
/*********************************************************
  INPUT:   A clause and a symbol.
  RETURNS: TRUE, if the clause contains the symbol
**********************************************************/
{
  int  i;

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++)
    if (term_ContainsSymbol(clause_GetLiteralAtom(Clause,i), Symbol))
      return TRUE;
  return FALSE;
}

BOOL clause_ImpliesFiniteDomain(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause consists of a positive disjunction
           of equations, where each equation is of the form "x=t" for
	   some variable "x" and ground term "t"
  MEMORY:  No Effects.
**********************************************************/
{
  int  i;
  TERM Term;

  if (clause_FirstLitIndex() != clause_FirstSuccedentLitIndex(Clause))
    return FALSE;

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
    Term = clause_GetLiteralTerm(Clause,i);
    if (!symbol_Equal(term_TopSymbol(Term),fol_Equality()) ||
	(!symbol_IsVariable(term_TopSymbol(term_FirstArgument(Term))) &&
	 !symbol_IsVariable(term_TopSymbol(term_SecondArgument(Term)))) ||
	(symbol_IsVariable(term_TopSymbol(term_FirstArgument(Term))) &&
	 !term_IsGround(term_SecondArgument(Term))) ||
	(symbol_IsVariable(term_TopSymbol(term_SecondArgument(Term))) &&
	 !term_IsGround(term_FirstArgument(Term))))
      return FALSE;
  }

  return TRUE;
}

NAT clause_NumberOfVarOccs(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The number of variable occurrences in the clause.
  MEMORY:  No effects.
***************************************************************/
{
  int i,n;
  NAT Result;

  Result = 0;
  n      = clause_Length(Clause);

  for (i=clause_FirstLitIndex();i<n;i++)
    Result += term_NumberOfVarOccs(clause_GetLiteralTerm(Clause,i));
  
  return Result;				   
}

NAT clause_ComputeWeight(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The Weight of the literals in the clause, 
           up to now the number of variable symbols plus
	   twice the number of signature symbols.
  MEMORY:  No effects.
***************************************************************/
{
  int     i,n;
  NAT     Weight,Help;
  LITERAL Lit;

  Weight = 0;
  n      = clause_Length(Clause);

  for (i = clause_FirstLitIndex();
       i < n;
       i++) {
    Lit    = clause_GetLiteral(Clause, i);
    Help   = clause_LiteralComputeWeight(Lit);
    clause_SetLiteralWeight(Lit,Help);
    Weight += Help;

  }
  return Weight;
}

NAT clause_ComputeTermDepth(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: Maximal depth of any literal in <Clause>.
  MEMORY:  No effects.
***************************************************************/
{
  int     i,n;
  NAT     Depth,Help;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_ComputeTermDepth\n", stderr);
    misc_DumpCore();
  }
#endif 


  Depth = 0;
  n     = clause_Length(Clause);

  for (i = clause_FirstLitIndex();i < n;i++) {
    Help   = term_Depth(clause_GetLiteralAtom(Clause,i));
    if (Help > Depth)
      Depth = Help;
  }
  return Depth;
}

NAT clause_MaxTermDepthClauseList(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: Maximal depth of a clause in <Clauses>.
  MEMORY:  No effects.
***************************************************************/
{
  NAT  Depth,Help;

  Depth = 0;

  while (!list_Empty(Clauses)) {
    Help = clause_ComputeTermDepth(list_Car(Clauses));
    if (Help > Depth)
      Depth = Help;
    Clauses = list_Cdr(Clauses);
  }

  return Depth;
}
  

NAT clause_ComputeSize(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The Size of the literals in the clause, 
           up to now the number of symbols.
  MEMORY:  No effects.
***************************************************************/
{
  int     i,n;
  NAT     Size;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_ComputeSize\n", stderr);
    misc_DumpCore();
  }
#endif 

  Size = 0;
  n    = clause_Length(Clause);

  for (i = clause_FirstLitIndex();i < n;i++)
    Size += term_ComputeSize(clause_GetLiteralTerm(Clause,i));
  
  return Size;
}


BOOL clause_WeightCorrect(CLAUSE Clause)
/*********************************************************
  INPUT:   A clause .
  RETURNS: TRUE iff the weight fields of the clause and its
           literals are correct.
  MEMORY:  None.
**********************************************************/
{
  int     i,n;
  NAT     Weight,Help;
  LITERAL Lit;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_WeightCorrect\n", stderr);
    misc_DumpCore();
  }
#endif 

  Weight = 0;
  n      = clause_Length(Clause);

  for (i = clause_FirstLitIndex(); i < n; i++) {
    Lit    = clause_GetLiteral(Clause, i);
    Help   = clause_LiteralComputeWeight(Lit);
    if (Help != clause_LiteralWeight(Lit))
      return FALSE;
    Weight += Help;
  }

  if (clause_Weight(Clause) != Weight)
    return FALSE;

  return TRUE;
}


LIST clause_InsertWeighed(CLAUSE Clause, LIST UsList)
/*********************************************************
  INPUT:   A clause and a list to insert the clause into.
  RETURNS: The list where the clause is inserted wrt its
           weight (Weight(Car(list)) <= Weight(Car(Cdr(list)))).
  MEMORY:  A new listnode is allocated.
**********************************************************/
{
  LIST Scan;
  NAT  Weight;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_InsertWeighed\n", stderr);
    misc_DumpCore();
  }
#endif 

  Weight = clause_Weight(Clause);

  Scan = UsList;

  if (list_Empty(Scan) ||
      (clause_Weight(list_Car(Scan)) > Weight)){
    return(list_Cons(Clause, Scan));

  } else {
    while (!list_Empty(list_Cdr(Scan)) &&
	   (clause_Weight(list_Car(list_Cdr(Scan))) <= Weight)){
      
      Scan = list_Cdr(Scan);
    }
    list_Rplacd(Scan, list_Cons(Clause, list_Cdr(Scan)));
    return UsList;
  }
}

LIST clause_ListSortWeighed(LIST Clauses)
/*********************************************************
  INPUT:   A list of clauses.
  RETURNS: The clause list sorted with respect to the weight
           of clauses, minimal weight first.
  MEMORY:  The original List is destructively changed!
**********************************************************/
{
  int  weight;
  LIST Scan;

  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    weight = clause_Weight(list_Car(Scan));
    if (weight < clause_MAXWEIGHT)
      clause_SORT[weight]  = list_Cons(list_Car(Scan),clause_SORT[weight]);
    else
      clause_SORT[clause_MAXWEIGHT]  = list_Cons(list_Car(Scan),clause_SORT[clause_MAXWEIGHT]);
  }
  Scan = list_NumberSort(clause_SORT[clause_MAXWEIGHT],clause_Weight);
  clause_SORT[clause_MAXWEIGHT] = list_Nil();
  for(weight = clause_MAXWEIGHT-1;weight>=0;weight--) {
    Scan                = list_Nconc(clause_SORT[weight],Scan);
    clause_SORT[weight] = list_Nil();
  }
  list_Delete(Clauses);
  return Scan;  
}


LITERAL clause_LiteralCopy(LITERAL Literal)
/*********************************************************
  INPUT:   A literal.
  RETURNS: An unshared copy of the literal, where the owning
           clause-slot is defaulted with NULL.
  MEMORY:  Memory for a new LITERAL_NODE and all its TERMs
           subterms is allocated.
**********************************************************/
{

  LITERAL Result;

#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralCopy\n", stderr);
    misc_DumpCore();
  }
#endif 

  Result                = (LITERAL)memory_Malloc(sizeof(LITERAL_NODE));
  
  Result->atomWithSign  = term_Copy(clause_LiteralSignedAtom(Literal));
  Result->oriented      = clause_LiteralIsOrientedEquality(Literal);

  Result->maxLit        = Literal->maxLit;
  Result->weight        = Literal->weight;
  Result->owningClause  = (POINTER)NULL;

  return(Result);
}


CLAUSE clause_Copy(CLAUSE Clause)
/*********************************************************
  INPUT:   A Clause.
  RETURNS: An unshared copy of the Clause.
  MEMORY:  Memory for a new CLAUSE_NODE, LITERAL_NODE and 
           all its TERMs subterms is allocated.
**********************************************************/
{

  CLAUSE Result;
  int i,c,a,s,l;

  Result                = (CLAUSE)memory_Malloc(sizeof(CLAUSE_NODE));

  Result->clausenumber  = clause_Number(Clause);
  Result->maxVar        = clause_MaxVar(Clause);
  Result->flags         = Clause->flags;
  clause_InitSplitData(Result);
  Result->validlevel    = clause_SplitLevel(Clause);
  clause_SetSplitField(Result, Clause->splitfield, Clause->splitfield_length);
  Result->depth         = clause_Depth(Clause);
  Result->weight        = clause_Weight(Clause);
  Result->parentCls     = list_Copy(clause_ParentClauses(Clause));
  Result->parentLits    = list_Copy(clause_ParentLiterals(Clause));
  Result->origin        = clause_Origin(Clause);

  Result->c             = (c = clause_NumOfConsLits(Clause));
  Result->a             = (a = clause_NumOfAnteLits(Clause));
  Result->s             = (s = clause_NumOfSuccLits(Clause));

  l = c + a + s;
  if (l != 0)
    Result->literals      = (LITERAL *)memory_Malloc(l * sizeof(LITERAL));

  for(i = 0; i < l; i++) {
    clause_SetLiteral(Result, i,
		      clause_LiteralCopy(clause_GetLiteral(Clause, i)));
    clause_LiteralSetOwningClause((Result->literals[i]), Result);
  }


  return(Result);
}


SYMBOL clause_LiteralMaxVar(LITERAL Literal)
/*********************************************************
  INPUT:   A literal.
  RETURNS: The maximal symbol of the literals variables,
           if the literal is ground, symbol_GetInitialStandardVarCounter().
  MEMORY:  No effects.
**********************************************************/
{

  TERM Term;
  int Bottom;
  SYMBOL MaxSym,Help;

#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralMaxVar\n", stderr);
    misc_DumpCore();
  }
#endif 

  Bottom = stack_Bottom();
  MaxSym = symbol_GetInitialStandardVarCounter();
  Term   = clause_LiteralAtom(Literal);

  do {
    if (term_IsComplex(Term))
      stack_Push(term_ArgumentList(Term));
    else
      if (term_IsVariable(Term))
	MaxSym = ((MaxSym < (Help = term_TopSymbol(Term))) ?
		  Help : MaxSym);
    while (!stack_Empty(Bottom) && list_Empty(stack_Top()))
      stack_Pop();
    if (!stack_Empty(Bottom)) {
      Term = (TERM) list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(Bottom));

  return(MaxSym);
}

SYMBOL clause_AtomMaxVar(TERM Term)
/*********************************************************
  INPUT:   A term.
  RETURNS: The maximal symbol of the lcontained variables,
           if <Term> is ground, symbol_GetInitialStandardVarCounter().
  MEMORY:  No effects.
**********************************************************/
{
  int Bottom;
  SYMBOL VarSym,Help;

  Bottom = stack_Bottom();
  VarSym = symbol_GetInitialStandardVarCounter();

  do {
    if (term_IsComplex(Term))
      stack_Push(term_ArgumentList(Term));
    else
      if (term_IsVariable(Term))
	VarSym = ((VarSym < (Help = term_TopSymbol(Term))) ?
		  Help : VarSym);
    while (!stack_Empty(Bottom) && list_Empty(stack_Top()))
      stack_Pop();
    if (!stack_Empty(Bottom)) {
      Term = (TERM)list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(Bottom));

  return VarSym;
}


void clause_SetMaxLitFlags(CLAUSE Clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: None, sets the maxLit-flag for maximal literals.
  MEMORY:  No effects.
***********************************************************/
{
  int        i,j,n,fa;
  LITERAL    ActLit,CompareLit;
  BOOL       Result, Twin;
  ord_RESULT HelpRes;
  
  n   = clause_Length(Clause);
  fa  = clause_FirstAntecedentLitIndex(Clause);
  clause_RemoveFlag(Clause,CLAUSESELECT);
  for (i = clause_FirstLitIndex(); i < n; i++)
    clause_LiteralFlagReset(clause_GetLiteral(Clause, i));
  if (term_StampOverflow(clause_STAMPID))
    for (i = clause_FirstLitIndex(); i < n; i++)
      term_ResetTermStamp(clause_LiteralSignedAtom(clause_GetLiteral(Clause, i)));
  term_StartStamp();

  /*printf("\n Start: "); clause_Print(Clause);*/

  for (i = fa; i < n; i++) {
    ActLit = clause_GetLiteral(Clause, i);
    if (!term_HasTermStamp(clause_LiteralSignedAtom(ActLit))) {
      Result  = TRUE;
      Twin    = FALSE;
      for (j = fa; j < n && Result; j++)
	if (i != j) {
	  CompareLit = clause_GetLiteral(Clause, j);
	  HelpRes    = ord_LiteralCompare(clause_LiteralSignedAtom(ActLit),
					  clause_LiteralIsOrientedEquality(ActLit),
					  clause_LiteralSignedAtom(CompareLit),
					  clause_LiteralIsOrientedEquality(CompareLit),FALSE);
	    
	  /* printf("\n\tWe compare: ");
	  fol_PrintDFG(clause_LiteralAtom(ActLit));
	  printf(" ");
	  fol_PrintDFG(clause_LiteralAtom(CompareLit));
	  printf(" Result: "); ord_Print(HelpRes); */

	  if (ord_IsEqual(HelpRes))
	    Twin = TRUE;
	  if (ord_IsSmallerThan(HelpRes))
	    Result = FALSE;
	  if (ord_IsGreaterThan(HelpRes))
	    term_SetTermStamp(clause_LiteralSignedAtom(CompareLit));
	}	
      if (Result) {
	clause_LiteralSetFlag(ActLit, MAXIMAL);
	if (!Twin)
	  clause_LiteralSetFlag(ActLit, STRICTMAXIMAL);
      }
    }
  }
  term_StopStamp();
  /*printf("\n End: "); clause_Print(Clause);*/
}


SYMBOL clause_SearchMaxVar(CLAUSE Clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: The maximal symbol of the clauses variables.
           If the clause is ground, symbol_GetInitialStandardVarCounter().
  MEMORY:  No effects.
***********************************************************/
{
  int i, n;
  SYMBOL Help, MaxSym;
  
  n = clause_Length(Clause);

  MaxSym = symbol_GetInitialStandardVarCounter();

  for (i = 0; i < n; i++) {
    Help = clause_LiteralMaxVar(clause_GetLiteral(Clause, i));
    if (Help > MaxSym)
      MaxSym = Help;
  }
  return MaxSym;
}


void clause_RenameVarsBiggerThan(CLAUSE Clause, SYMBOL MinVar)
/**********************************************************
  INPUT:   A clause and a variable symbol.
  RETURNS: The clause with variables renamed in a way, that
           all vars are (excl.) bigger than MinVar.
  MEMORY:  No effects.
***********************************************************/
{
  int i,n;
  
#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_RenameVarsBiggerThan\n", stderr);
    misc_DumpCore();
  }
#endif 

  if (MinVar != symbol_GetInitialStandardVarCounter()) {
    n = clause_Length(Clause);

    term_StartMaxRenaming(MinVar);

    for(i = clause_FirstLitIndex(); i < n; i++)
      term_Rename(clause_GetLiteralTerm(Clause, i));
  }
}

void clause_Normalize(CLAUSE Clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: The term with normalized Variables, DESTRUCTIVE
           on the variable subterms' topsymbols.
  MEMORY:  No effects.
***********************************************************/
{
  int i,n;
  
  n = clause_Length(Clause);

  term_StartMinRenaming();

  for(i = clause_FirstLitIndex(); i < n; i++)
    term_Rename(clause_GetLiteralTerm(Clause, i));
}


void clause_SubstApply(SUBST Subst, CLAUSE Clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: None. Applys the substitution to the clause.
  MEMORY:  No effects.
***********************************************************/
{
  int i,n;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_SubstApply\n", stderr);
    misc_DumpCore();
  }
#endif 

  n = clause_Length(Clause);

  for (i=clause_FirstLitIndex(); i<n; i++)
    clause_LiteralSetAtom(clause_GetLiteral(Clause, i), 
			  subst_Apply(Subst,clause_GetLiteralAtom(Clause,i)));
}


void clause_UpdateMaxVar(CLAUSE Clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: None. Actualizes the MaxVar slot wrt the actual
           literals.
  MEMORY:  No effects.
***********************************************************/
{
  clause_SetMaxVar(Clause, clause_SearchMaxVar(Clause));
}



void clause_OrientEqualities(CLAUSE Clause)
/**********************************************************
  INPUT:   An unshared clause.
  RETURNS: None. Reorders the arguments of equality literals
           wrt. the ordering. Thus first arguments aren't smaller
	   after the application.
  MEMORY:  No effects.
***********************************************************/
{
  int        i,length;
  LITERAL    EqLit;
  ord_RESULT HelpRes;

  /*printf("\n Clause: ");clause_Print(Clause);*/

  length = clause_Length(Clause);

  for (i = clause_FirstLitIndex(); i < length; i++) {
    EqLit = clause_GetLiteral(Clause, i);

    if (clause_LiteralIsEquality(EqLit)) {
      HelpRes = ord_Compare(term_FirstArgument(clause_LiteralAtom(EqLit)),
			    term_SecondArgument(clause_LiteralAtom(EqLit)));
      
      /*printf("\n\tWe compare: ");
      fol_PrintDFG(term_FirstArgument(clause_LiteralAtom(EqLit)));
      printf(" ");
      fol_PrintDFG(term_SecondArgument(clause_LiteralAtom(EqLit)));
      printf("\nResult: "); ord_Print(HelpRes);*/

      switch(HelpRes){

      case ord_SMALLER_THAN:
	/* printf("\nSwapping: ");
	   term_Print(clause_LiteralAtom(EqLit)); DBG */
	term_EqualitySwap(clause_LiteralAtom(EqLit)); 
	clause_LiteralSetOrientedEquality(EqLit);
	/*	Swapped = TRUE; */
	break;
      case ord_GREATER_THAN:
	clause_LiteralSetOrientedEquality(EqLit); 
	break;
      default:
	clause_LiteralSetNoOrientedEquality(EqLit);
	break;
      }
    }
    else
      clause_LiteralSetNoOrientedEquality(EqLit);
  }
}


void  clause_InsertFlatIntoIndex(CLAUSE Clause, st_INDEX Index)
/**********************************************************
  INPUT:   An unshared clause and an index.
  EFFECT:  The atoms of <Clause> are inserted into the index.
           A link from the atom to its literal via the supertermlist
	   is established.
***********************************************************/
{
  int     i,n;
  LITERAL Lit;
  TERM    Atom    ;

  n = clause_Length(Clause);

  for(i=clause_FirstLitIndex();i<n;i++) {
    Lit  = clause_GetLiteral(Clause,i);
    Atom = clause_LiteralAtom(Lit);
    term_RplacSupertermList(Atom, list_List(Lit));
    st_EntryCreate(Index, Atom, Atom, cont_LeftContext());
  }  
}

void  clause_DeleteFlatFromIndex(CLAUSE Clause, st_INDEX Index)
/**********************************************************
  INPUT:   An unshared clause and an index.
  EFFECT:  The clause is deleted from the index and deleted itself.
***********************************************************/
{
  int     i,n;
  LITERAL Lit;
  TERM    Atom    ;

  n = clause_Length(Clause);

  for(i=clause_FirstLitIndex();i<n;i++) {
    Lit  = clause_GetLiteral(Clause,i);
    Atom = clause_LiteralAtom(Lit);
    list_Delete(term_SupertermList(Atom));
    term_RplacSupertermList(Atom, list_Nil());
    st_EntryDelete(Index, Atom, Atom, cont_LeftContext());
  } 
  clause_Delete(Clause);
}


void  clause_DeleteClauseListFlatFromIndex(LIST Clauses, st_INDEX Index)
/**********************************************************
  INPUT:   An list of unshared clause and an index.
  EFFECT:  The clauses are deleted from the index and deleted itself.
           The list is deleted.
***********************************************************/
{
  LIST Scan;

  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    clause_DeleteFlatFromIndex(list_Car(Scan), Index);
  
  list_Delete(Clauses);
}

/**************************************************************/
/* Literal Output Functions                                   */
/**************************************************************/

void clause_LiteralPrint(LITERAL Literal)
/**************************************************************
  INPUT:	A Literal.
  RETURNS: none.
  MEMORY:  No effects.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralPrint\n", stderr);
    misc_DumpCore();
  }
#endif 

  term_PrintPrefix(clause_LiteralSignedAtom(Literal));
}


void clause_LiteralListPrint(LIST LitList)
/**************************************************************
  INPUT:   A list of literals.
  RETURNS: none.
  SUMMARY: Prints the literals  to stdout. 
  CAUTION: Uses the clause_LiteralPrint function.
  MEMORY:  No effects.
***************************************************************/
{
  while(!(list_Empty(LitList))) {
    clause_LiteralPrint(list_First(LitList));
    LitList = list_Cdr(LitList);

    if (!list_Empty(LitList))
      putchar(' ');
  }
}


void clause_LiteralPrintUnsigned(LITERAL Literal)
/**************************************************************
  INPUT:   A Literal.
  RETURNS: none.
  MEMORY:  No effects.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralPrintUnsigned\n", stderr);
    misc_DumpCore();
  }
#endif 

  term_PrintPrefix(clause_LiteralAtom(Literal));
  fflush(stdout);
}


void clause_LiteralPrintSigned(LITERAL Literal)
/**************************************************************
  INPUT:   A Literal.
  RETURNS: none.
  MEMORY:  No effects.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralPrintSigned\n", stderr);
    misc_DumpCore();
  }
#endif 

  term_PrintPrefix(clause_LiteralSignedAtom(Literal));
  fflush(stdout);
}


void clause_LiteralFPrint(FILE* File, LITERAL Lit)
/**************************************************************
  INPUT:   A file and a literal.
  RETURNS: none.
  MEMORY:  No effects.
************************************************************/
{
  term_FPrintPrefix(File, clause_LiteralSignedAtom(Lit));
}


/**************************************************************/
/* Clause functions                                           */
/**************************************************************/

void clause_Init(void)
/**************************************************************
  INPUT:   None.
  RETURNS: None, initializes the clausecounter.
  MEMORY:  No effects.
***************************************************************/
{
  int i;
  clause_CLAUSECOUNTER = 1;
  clause_STAMPID       = term_GetStampID();  
  for (i=0;i<=clause_MAXWEIGHT;i++)
    clause_SORT[i] = list_Nil();
}


CLAUSE clause_CreateBody(int ClauseLength)
/**************************************************************
  INPUT:   The number of literals as integer.
  RETURNS: A new generated clause node for 'ClauseLength'
  MEMORY:  Allocates a CLAUSE_NODE and the needed array for LITERALs.
*************************************************************/
{
  CLAUSE Result;

  Result                = (CLAUSE)memory_Malloc(sizeof(CLAUSE_NODE));

  Result->clausenumber  = clause_CLAUSECOUNTER++;
  Result->maxVar        = symbol_GetInitialStandardVarCounter();
  Result->flags         = 0;
  Result->depth         = 0;
  clause_InitSplitData(Result);
  Result->weight          = 0;
  Result->parentCls     = list_Nil();
  Result->parentLits    = list_Nil();

  Result->c             = 0;
  Result->a             = 0;
  Result->s             = 0;

  if (ClauseLength != 0)
    Result->literals =
      (LITERAL *)memory_Malloc((ClauseLength) * sizeof(LITERAL));

  clause_SetFromInput(Result);

  return Result;
}


CLAUSE clause_Create(LIST Constraint, LIST Antecedent, LIST Succedent)
/**************************************************************
  INPUT:   Three lists of pointers to atoms.
  RETURNS: The new generated clause.
  MEMORY:  Allocates a CLAUSE_NODE and the needed LITERAL_NODEs,
           uses the terms from the lists, additionally allocates
	   termnodes for the fol_Not() in Const. and Ante.
*************************************************************/
{
  CLAUSE Result;
  int i,c,a,s;
  
  Result                = (CLAUSE)memory_Malloc(sizeof(CLAUSE_NODE));
  
  Result->clausenumber  = clause_CLAUSECOUNTER++;
  Result->flags         = 0;
  Result->depth         = 0;
  clause_InitSplitData(Result);
  Result->parentCls     = list_Nil();
  Result->parentLits    = list_Nil();

  Result->c             = (c = list_Length(Constraint));
  Result->a             = (a = list_Length(Antecedent));
  Result->s             = (s = list_Length(Succedent));

  if (!clause_IsEmptyClause(Result))
    Result->literals =
      (LITERAL *) memory_Malloc((c+a+s) * sizeof(LITERAL));
  
  for(i = 0; i < c; i++) {
    Result->literals[i] =
      clause_LiteralCreate(term_Create(fol_Not(),
	list_List((TERM)list_Car(Constraint))),Result);
    Constraint = list_Cdr(Constraint);
  }
  
  a += c;
  for( ; i < a; i++) {
    Result->literals[i] =
      clause_LiteralCreate(term_Create(fol_Not(),
	list_List((TERM)list_Car(Antecedent))), Result);
    Antecedent = list_Cdr(Antecedent);
  }
  
  s += a;
  for( ; i < s; i++) {
    Result->literals[i] =
      clause_LiteralCreate((TERM) list_Car(Succedent), Result);
    Succedent = list_Cdr(Succedent);
  }
  
  clause_OrientAndReInit(Result);
  clause_SetFromInput(Result);
  
  /*puts("In clause_Create wurde folgende Clause erzeugt:");
    clause_PrintVerbose(Result);
    putchar('\n');*/
  
  return Result;
}


CLAUSE clause_CreateCrude(LIST Constraint, LIST Antecedent, LIST Succedent, BOOL Con)
/**************************************************************
  INPUT:   Three lists of pointers to literals (!) and a Flag indicating
           whether the clause comes from the conjecture part of of problem.
	   It is assumed that Constraint and Antecedent literals are negative,
	   whereas Succedent literals are positive.
  RETURNS: The new generated clause, where a clause_OrientAndReInit has still
           to be performed, i.e., variables are not normalized, maximal literal
	   flags not set, equations not oriented, the weight is not set.
  MEMORY:  Allocates a CLAUSE_NODE and the needed LITERAL_NODEs,
           uses the terms from the lists, additionally allocates
	   termnodes for the fol_Not() in Const. and Ante.
****************************************************************/
{
  CLAUSE Result;
  int i,c,a,s;
  
  Result                = (CLAUSE)memory_Malloc(sizeof(CLAUSE_NODE));
  
  Result->clausenumber  = clause_CLAUSECOUNTER++;
  Result->flags         = 0;
  if (Con)
    clause_SetFlag(Result, CONCLAUSE);

  Result->depth         = 0;
  clause_InitSplitData(Result);
  Result->parentCls     = list_Nil();
  Result->parentLits    = list_Nil();

  Result->c             = (c = list_Length(Constraint));
  Result->a             = (a = list_Length(Antecedent));
  Result->s             = (s = list_Length(Succedent));

  if (!clause_IsEmptyClause(Result))
    Result->literals = (LITERAL *)memory_Malloc((c+a+s) * sizeof(LITERAL));
  
  for(i = 0; i < c; i++) {
    Result->literals[i] =
      clause_LiteralCreate(list_Car(Constraint),Result);
    Constraint = list_Cdr(Constraint);
  }
  
  a += c;
  for( ; i < a; i++) {
    Result->literals[i] =
      clause_LiteralCreate(list_Car(Antecedent), Result);
    Antecedent = list_Cdr(Antecedent);
  }
  
  s += a;
  for( ; i < s; i++) {
    Result->literals[i] = clause_LiteralCreate(list_Car(Succedent), Result);
    Succedent = list_Cdr(Succedent);
  }
  
  clause_SetFromInput(Result);
  
  /*puts("In clause_CreateCrude the following clause was generated:");
    clause_PrintVerbose(Result);
    putchar('\n');*/
  
  return(Result);
}


CLAUSE clause_CreateUnnormalized(LIST Constraint, LIST Antecedent,
				 LIST Succedent)
/**************************************************************
  INPUT:   Three lists of pointers to atoms.
  RETURNS: The new generated clause.
  MEMORY:  Allocates a CLAUSE_NODE and the needed LITERAL_NODEs,
           uses the terms from the lists, additionally allocates
	   termnodes for the fol_Not() in Const. and Ante.
****************************************************************/
{
  CLAUSE Result;
  int i,c,a,s;

  Result                = (CLAUSE)memory_Malloc(sizeof(CLAUSE_NODE));

  Result->clausenumber  = clause_CLAUSECOUNTER++;
  Result->flags         = 0;
  Result->depth         = 0;
  clause_InitSplitData(Result);
  Result->parentCls     = list_Nil();
  Result->parentLits    = list_Nil();

  Result->c             = (c = list_Length(Constraint));
  Result->a             = (a = list_Length(Antecedent));
  Result->s             = (s = list_Length(Succedent));

  if (!clause_IsEmptyClause(Result))
    Result->literals = (LITERAL *)memory_Malloc((c+a+s) * sizeof(LITERAL));
  
    for(i = 0; i < c; i++) {
      Result->literals[i] =
	clause_LiteralCreate(term_Create(fol_Not(),
			list_List((TERM)list_Car(Constraint))),Result);
      Constraint = list_Cdr(Constraint);
    }

    a += c;
    for( ; i < a; i++) {
      Result->literals[i]  =
	clause_LiteralCreate(term_Create(fol_Not(),
			list_List((TERM)list_Car(Antecedent))), Result);
      Antecedent = list_Cdr(Antecedent);
    }

    s += a;
    for( ; i < s; i++) {
      Result->literals[i] =
	  clause_LiteralCreate((TERM)list_Car(Succedent), Result);
      Succedent = list_Cdr(Succedent);
    }
  clause_UpdateMaxVar(Result);

  /*puts("In clause_CreateUnnormalized wurde folgende Clause erzeugt:");
  clause_PrintVerbose(Result);
  putchar('\n');*/

  return Result;
}


CLAUSE clause_CreateFromLiterals(LIST LitList, BOOL Sorts, BOOL Conclause, BOOL Ordering)
/**************************************************************
  INPUT:   A list of literals and three flags indicating whether
           sort constraint literals should be generated whether
	   the clause is a conjecture clause and whether the ordering
	   information should be established.
  RETURNS: The new generated clause.
  MEMORY:  Allocates a CLAUSE_NODE and the needed LITERAL_NODEs,
           uses the terms from the lists.
****************************************************************/
{
  CLAUSE Result;
  LIST   Antecedent,Succedent,Constraint;
  TERM   Atom;

  Antecedent = list_Nil();
  Succedent  = list_Nil();
  Constraint = list_Nil();

  while (!list_Empty(LitList)) {
    if (term_TopSymbol(list_Car(LitList)) == fol_Not()) {
      Atom = term_FirstArgument(list_Car(LitList));
      if (Sorts && symbol_IsBaseSort(term_TopSymbol(Atom)) && term_IsVariable(term_FirstArgument(Atom)))
	Constraint = list_Cons(list_Car(LitList),Constraint);
      else
	Antecedent = list_Cons(list_Car(LitList),Antecedent);
    }
    else
      Succedent = list_Cons(list_Car(LitList),Succedent);
    LitList = list_Cdr(LitList);
  }
  
  Result = clause_CreateCrude(Constraint, Antecedent, Succedent, Conclause);

  list_Delete(Constraint);
  list_Delete(Antecedent);
  list_Delete(Succedent);

  if (Ordering)
    clause_OrientAndReInit(Result);
  else {
      clause_Normalize(Result);
      clause_SetWeight(Result, clause_ComputeWeight(Result));
      clause_UpdateMaxVar(Result);
  }
  
  return Result;
}

void clause_SetSortConstraint(CLAUSE Clause, BOOL Strong)
/**************************************************************
  INPUT:   A clause and a flag indicating whether also negative
           monadic literals with a real term argument should be
	   put in the sort constraint.
  RETURNS: void.
  EFFECT:  Negative monadic literals are put in the sort constraint.
****************************************************************/
{
  LITERAL ActLit,Help;
  TERM    ActAtom;
  int     i,k,NewConLits;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_SetSortConstraint\n", stderr);
    misc_DumpCore();
  }
#endif 

  i          = clause_LastConstraintLitIndex(Clause);
  NewConLits = 0;

  for (k=clause_FirstAntecedentLitIndex(Clause);k<=clause_LastAntecedentLitIndex(Clause);k++) {
    ActLit  = clause_GetLiteral(Clause,k);
    ActAtom = clause_LiteralAtom(ActLit);
    if (symbol_IsBaseSort(term_TopSymbol(ActAtom)) &&
	(Strong || term_IsVariable(term_FirstArgument(ActAtom)))) {
      if (++i!= k) {
	Help = clause_GetLiteral(Clause,i);
	clause_SetLiteral(Clause,i,ActLit);
	clause_SetLiteral(Clause,k,Help);
      }
      NewConLits++;
    }
  }

  clause_SetNumOfConsLits(Clause, clause_NumOfConsLits(Clause) + NewConLits);
  clause_SetNumOfAnteLits(Clause, clause_NumOfAnteLits(Clause) - NewConLits);
  clause_ReInit(Clause);

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal output in clause_SetSortConstraint\n", stderr);
    misc_DumpCore();
  }
#endif 

}



void clause_Delete(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: None.
  MEMORY:  Frees the memory of the .
***************************************************************/
{
  int i, n;

#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause)) { /* Clause may be a byproduct of some hyper rule */
    fputs("\n\nIllegal input in clause_Delete\n", stderr);
    misc_DumpCore();
  }
#endif 

  /*  fputs("\nClause deleted :", stdout); clause_Print(Clause); */

  n = clause_Length(Clause);
  
  for(i = 0; i < n; i++)
    clause_LiteralDelete(clause_GetLiteral(Clause,i));
  
  if (n != 0)
    memory_Free(Clause->literals, n * sizeof(LITERAL));
  
  list_Delete(clause_ParentClauses(Clause)); 
  list_Delete(clause_ParentLiterals(Clause));
#ifdef DEBUG
  if ((Clause->splitfield != NULL) && (Clause->splitfield_length == 0))
    puts("Illegal splitfield_length");
  if ((Clause->splitfield == NULL) && (Clause->splitfield_length != 0))
    puts("Illegal splitfield");
#endif
  if (Clause->splitfield != NULL) {
    
    memory_Free(Clause->splitfield, 
		sizeof(SPLITFIELDENTRY) * Clause->splitfield_length);
  }
  clause_Free(Clause);
}

/**************************************************************/
/* Functions to use the sharing for clauses.                  */
/**************************************************************/

void clause_InsertIntoSharing(CLAUSE Clause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:	A Clause.
  RETURNS: none.
  SUMMARY: Inserts the unsigned atoms of 'Clause' into the sharing. 
  CAUTION: None.
  MEMORY:  No effects.
***************************************************************/
{
  int i, litnum;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_InsertIntoSharing\n", stderr);
    misc_DumpCore();
  }

  clause_Check(Clause);
#endif 

  litnum = clause_Length(Clause);

  for(i = 0; i < litnum; i++){
    /* printf("\nDas %d -te Literal ist :\n",i);
       clause_LiteralPrint(clause_GetLiteral(Clause,i));
       */
    clause_LiteralInsertIntoSharing(clause_GetLiteral(Clause,i), ShIndex);
  }
}


void clause_DeleteFromSharing(CLAUSE Clause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A Clause and an Index.
  RETURNS: none.
  SUMMARY: Deletes 'Clause' and all its literals from the sharing,
           frees the memory of 'Clause'.
  CAUTION: None.
  MEMORY:  No effects.
***************************************************************/
{
  int i, length;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_DeleteFromSharing\n", stderr);
    misc_DumpCore();
  }
#endif 

  /* fputs("\nClause deleted from Sharing: ", stdout); clause_Print(Clause); */
  
  length = clause_Length(Clause);
  
  for(i = 0; i < length; i++)
    clause_LiteralDeleteFromSharing(clause_GetLiteral(Clause,i),ShIndex);
  
  if (length != 0)
    memory_Free(Clause->literals, length * sizeof(LITERAL));
  
  list_Delete(clause_ParentClauses(Clause)); 
  list_Delete(clause_ParentLiterals(Clause));
#ifdef DEBUG
  if ((Clause->splitfield != NULL) && (Clause->splitfield_length == 0))
    puts("Illegal splitfield_length");
  if ((Clause->splitfield == NULL) && (Clause->splitfield_length != 0))
    puts("Illegal splitfield");
#endif
  if (Clause->splitfield != NULL) {
    memory_Free(Clause->splitfield, 
		sizeof(SPLITFIELDENTRY) * Clause->splitfield_length);
  }
  clause_Free(Clause);
}


void clause_MakeUnshared(CLAUSE Clause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:	A Clause and an Index.
  RETURNS: none.
  SUMMARY: Deletes the clauses literals from the sharing and
           replaces them by unshared copies.
  CAUTION: None.
  MEMORY:  No effects.
***************************************************************/
{
  LITERAL ActLit;
  TERM SharedAtom, AtomCopy;
  int i,LastAnte,length;

#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause)) {
    fputs("\n\nIllegal input in clause_MakeUnshared\n", stderr);
    misc_DumpCore();
  }
#endif 

  /* fputs("\nClause to make unshared: ", stdout); clause_Print(Clause); */

  LastAnte = clause_LastAntecedentLitIndex(Clause);
  length   = clause_Length(Clause);

  for(i = clause_FirstLitIndex(); i <= LastAnte; i++) {
    ActLit     = clause_GetLiteral(Clause, i);
    SharedAtom = clause_LiteralAtom(ActLit);
    AtomCopy   = term_Copy(SharedAtom);
    sharing_Delete(ActLit, SharedAtom, ShIndex);
    clause_LiteralSetNegAtom(ActLit, AtomCopy);
  }

  for( ; i < length; i++) {
    ActLit     = clause_GetLiteral(Clause, i);
    SharedAtom = clause_LiteralSignedAtom(ActLit);
    AtomCopy   = term_Copy(SharedAtom);
    sharing_Delete(ActLit, SharedAtom, ShIndex);
    clause_LiteralSetPosAtom(ActLit, AtomCopy);
  }
  /* fputs("\nClause made unshared: ", stdout); clause_Print(Clause); */
}

void clause_MoveSharedClause(CLAUSE Clause, SHARED_INDEX Source, SHARED_INDEX Destination)
/**************************************************************
  INPUT:   A Clause and two indexes.
  RETURNS: None.
  EFFECT:  Deletes <Clause> from <Source> and inserts it into <Destination>.
***************************************************************/
{
  LITERAL Lit;
  TERM    Atom,Copy;
  int     i,length;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_MoveSharedClause\n", stderr);
    misc_DumpCore();
  }
#endif 

  length   = clause_Length(Clause);

  for(i = clause_FirstLitIndex(); i < length; i++) {
    Lit  = clause_GetLiteral(Clause, i);
    Atom = clause_LiteralAtom(Lit);
    Copy = term_Copy(Atom);        /* sharing_Insert works destructively on <Copy>'s superterms */
    clause_LiteralSetAtom(Lit, sharing_Insert(Lit, Copy, Destination));
    sharing_Delete(Lit, Atom, Source);
    term_Delete(Copy);
  }
}


void clause_DeleteSharedLiteral(CLAUSE Clause, int Indice,
				SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:	A Clause, a literals indice and an Index.
  RETURNS: none.
  SUMMARY: Deletes the shared literal from the clause.
  MEMORY:  Various.
***************************************************************/
{

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_DeleteSharedLiteral\n", stderr);
    misc_DumpCore();
  }
#endif 

  clause_MakeUnshared(Clause, ShIndex);
  clause_DeleteLiteral(Clause, Indice);
  clause_InsertIntoSharing(Clause, ShIndex);
}


void clause_DeleteClauseList(LIST ClauseList)
/**************************************************************
  INPUT:   A list of unshared clauses.
  RETURNS: none.
  SUMMARY: Deletes all clauses in the list and the list.
  MEMORY:  Frees the lists and the clauses' memory.
 ***************************************************************/
{
  LIST Scan;

  for (Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
    if (clause_Exists(list_Car(Scan)))
      clause_Delete(list_Car(Scan));

  list_Delete(ClauseList);
}


void clause_DeleteSharedClauseList(LIST ClauseList, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:	A list of clauses and an index.
  RETURNS: none.
  SUMMARY: Deletes all clauses in the list from the sharing.
  MEMORY:  Frees the lists and the clauses' memory.
***************************************************************/
{
  LIST Scan;

  for (Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
    clause_DeleteFromSharing(list_Car(Scan), ShIndex);

  list_Delete(ClauseList);
}


void clause_DeleteAllIndexedClauses(SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An Index.
  RETURNS: none.
  SUMMARY: Deletes all clauses' terms from the sharing, frees their
           memory.
  CAUTION: None.
  MEMORY:  Frees the memory of all clauses with terms in the index.
***************************************************************/
{
  LIST TermList,DelList,Scan;
  TERM NewVar;
  SYMBOL NewVarSymbol;

  NewVar = term_CreateStandardVariable();
  NewVarSymbol = term_TopSymbol(NewVar);

  TermList = st_GetInstance(cont_LeftContext(), sharing_Index(ShIndex), NewVar);
  /* This should yield a list of all terms in the index
     and thus the sharing. */

  while (!list_Empty(TermList)) {
    /* puts("\nWir haben aus dem Index folgende Liste "
	   "von Instanzen einer neuen Variable :"); 
    fflush(stdout);
    term_TermListPrint(TermList); */

    DelList = sharing_GetDataList(list_Car(TermList), ShIndex);

    for (Scan = DelList;
	 !list_Empty(Scan);
	 Scan = list_Cdr(Scan))
      list_Rplaca(Scan, clause_LiteralOwningClause(list_Car(Scan)));

    DelList = list_PointerDeleteDuplicates(DelList);

    for (Scan = DelList;
	 !list_Empty(Scan);
	 Scan = list_Cdr(Scan))
      clause_DeleteFromSharing(list_Car(Scan), ShIndex);

    list_Delete(TermList);

    TermList = st_GetInstance(cont_LeftContext(), sharing_Index(ShIndex), NewVar);

    list_Delete(DelList);
  }
  term_Delete(NewVar);
  symbol_Delete(NewVarSymbol);
}


void clause_PrintAllIndexedClauses(SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An Index.
  RETURNS: none.
  SUMMARY: Prints all indexed clauses to stdout.
  MEMORY:  None.
***************************************************************/
{
  LIST TermList,ClList,PrintList,Scan;
  TERM NewVar;
  SYMBOL NewVarSymbol;

  NewVar = term_CreateStandardVariable();
  NewVarSymbol = term_TopSymbol(NewVar);

  puts("\n Alle Klauseln im Index:");

  TermList = st_GetInstance(cont_LeftContext(), sharing_Index(ShIndex), NewVar);
  /* This should yield a list of all terms in the index
     and thus the sharing. */

  PrintList = list_Nil();

  while (!list_Empty(TermList)) {
    /* puts("\nWir haben aus dem Index folgende Liste "
       "von Instanzen einer neuen Variable :");
       fflush(stdout);
       term_TermListPrint(TermList);
       */
    ClList = sharing_GetDataList(list_Car(TermList), ShIndex);

    for (Scan = ClList;
	 !list_Empty(Scan);
	 Scan = list_Cdr(Scan))
      list_Rplaca(Scan, clause_LiteralOwningClause(list_Car(Scan)));

    PrintList = list_NPointerUnion(ClList, PrintList);

    Scan = TermList;
    TermList = list_Cdr(TermList);
    list_Free(Scan);
  }
  clause_ListPrint(PrintList);

  list_Delete(PrintList);

  term_Delete(NewVar);
  symbol_Delete(NewVarSymbol);
}


LIST clause_AllIndexedClauses(SHARED_INDEX ShIndex) {
/**************************************************************
  INPUT:   An index
  RETURNS: A list of all the clauses in the index
  MEMORY: Memory is allocated for the list nodes
***************************************************************/
    LIST clauselist, scan;
    clauselist = sharing_GetAllSuperTerms(ShIndex);
    for(scan = clauselist; scan != list_Nil(); scan = list_Cdr(scan))
	list_Rplaca(scan, clause_LiteralOwningClause(list_Car(scan)));
    clauselist = list_PointerDeleteDuplicates(clauselist);             
    return clauselist;
}


/**************************************************************/
/* Clause Access Functions                                    */
/**************************************************************/

void clause_DeleteLiteral(CLAUSE Clause, int Indice)
/**************************************************************
  INPUT:   An unshared clause and a literals number.
  RETURNS: None.
  MEMORY:  The memory of the literal with the 'Indice' and memory
           of its atom is freed.
***************************************************************/
{
  int i,lc,la,length,shift;
  LITERAL *Literals;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || (clause_Length(Clause) <= Indice)) {
    fputs("\n\nIllegal input in clause_DeleteLiteral\n", stderr);
    misc_DumpCore();
  }
#endif 

  clause_LiteralDelete(clause_GetLiteral(Clause, Indice));

  lc               = clause_LastConstraintLitIndex(Clause);
  la               = clause_LastAntecedentLitIndex(Clause);
  length           = clause_Length(Clause);
  Literals         = Clause->literals;

  if (length > 1)
    Clause->literals = (LITERAL *)memory_Malloc((length-1) * sizeof(LITERAL));
  else
    Clause->literals = (LITERAL *)NULL;

  if (Indice <= lc)
    clause_SetNumOfConsLits(Clause, (clause_NumOfConsLits(Clause) - 1));

  else if (Indice <= la)
    clause_SetNumOfAnteLits(Clause, (clause_NumOfAnteLits(Clause) - 1));

  else
    clause_SetNumOfSuccLits(Clause, (clause_NumOfSuccLits(Clause) - 1));

  shift  = 0;
  length = length - 1;

  for(i = 0; i < length; i++)
    if (i == Indice) {
      shift = 1;
      Clause->literals[i] = Literals[i+shift];
      
    } else
      Clause->literals[i] = Literals[i+shift];
  
  length = length + 1;
  memory_Free(Literals, length*sizeof(LITERAL));

  clause_ReInit(Clause);
}


void clause_DeleteLiteralNN(CLAUSE Clause, int Indice)
/**************************************************************
  INPUT:   An unshared clause and a literals number.
  RETURNS: None.
  MEMORY:  The memory of the literal with the 'Indice' and memory
           of its atom is freed.
***************************************************************/
{
  int i,lc,la,length,shift;
  LITERAL *Literals;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || (clause_Length(Clause) <= Indice)) {
    fputs("\n\nIllegal input in clause_DeleteLiteralNN\n", stdout);
    misc_DumpCore();
  }
#endif 

  clause_LiteralDelete(clause_GetLiteral(Clause, Indice));

  lc       = clause_LastConstraintLitIndex(Clause);
  la       = clause_LastAntecedentLitIndex(Clause);
  length   = clause_Length(Clause);
  Literals = Clause->literals;

  if (length > 1)
    Clause->literals = (LITERAL *)memory_Malloc((length-1) * sizeof(LITERAL));
  else
    Clause->literals = (LITERAL *)NULL;
  
  if (Indice <= lc)
    clause_SetNumOfConsLits(Clause, (clause_NumOfConsLits(Clause) - 1));
  else if (Indice <= la)
    clause_SetNumOfAnteLits(Clause, (clause_NumOfAnteLits(Clause) - 1));
  else
    clause_SetNumOfSuccLits(Clause, (clause_NumOfSuccLits(Clause) - 1));

  shift  = 0;
  length = length - 1;

  for(i = 0; i < length; i++)
    if (i == Indice) {
      shift = 1;
      Clause->literals[i] = Literals[i+shift];
    } else
      Clause->literals[i] = Literals[i+shift];
  
  length = length + 1;
  memory_Free(Literals, length*sizeof(LITERAL));
}


void clause_DeleteLiterals(CLAUSE Clause, LIST Indices)
/**************************************************************
  INPUT:   An unshared clause and a list of literal indices.
  RETURNS: None.
  EFFECT:  The literals given by <Indices> are deleted.
  MEMORY:  The memory of the literals with the 'Indice' and memory
           of its atom is freed.
***************************************************************/
{
  LITERAL *NewLits;
  int     i, j, nc, na, ns, lc, la, olength, nlength;

  nc = 0;
  na = 0;
  ns = 0;
  lc = clause_LastConstraintLitIndex(Clause);
  la = clause_LastAntecedentLitIndex(Clause);

  olength = clause_Length(Clause);
  nlength = olength - list_Length(Indices);

  if (nlength != 0)
    NewLits = (LITERAL *)memory_Malloc((nlength) * sizeof(LITERAL));
  else
    NewLits = (LITERAL *)NULL;

  for(i=clause_FirstLitIndex(),j=clause_FirstLitIndex(); i < olength; i++)
 
    if (list_PointerMember(Indices, (POINTER) i))
      clause_LiteralDelete(clause_GetLiteral(Clause,i));
    else {

      NewLits[j++] = clause_GetLiteral(Clause,i);

      if (i <= lc)
	nc++;
      else if (i <= la)
	na++;
      else
	ns++;
    }
  memory_Free(Clause->literals, olength * sizeof(LITERAL));

  Clause->literals = NewLits;

  clause_SetNumOfConsLits(Clause, nc);
  clause_SetNumOfAnteLits(Clause, na);
  clause_SetNumOfSuccLits(Clause, ns);

  clause_ReInit(Clause);
}


/**************************************************************/
/* Clause Comparisons                                         */
/**************************************************************/
/* In sort.c in ComputeSortIntern benutzt! */

BOOL clause_IsHornClause(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE if 'Clause' is a hornclause'
           FALSE else.
  MEMORY:  No effects.
  ***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_IsHornClause\n", stderr);
    misc_DumpCore();
  }
#endif 
  return (clause_NumOfSuccLits(Clause) == 1);
}

BOOL clause_HasTermSortConstraintLits(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: TRUE iff there is at least one sort constraint atom having
           a term as its argument
***************************************************************/
{
  int i,n;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_HasTermSortConstraintLits\n", stderr);
    misc_DumpCore();
  }
#endif
  
  n = clause_LastConstraintLitIndex(Clause);

  for (i = clause_FirstConstraintLitIndex(Clause);i <= n; i++)
    if (!term_AllArgsAreVar(clause_GetLiteralAtom(Clause,i)))
      return TRUE;

  return FALSE;
}


  
BOOL clause_HasSolvedConstraint(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE if 'Clause' has a solved 
           constraint, i.e. only variables as sort arguments,
	   FALSE else.
  MEMORY:  No effects.
***************************************************************/
{
  int  i,c;
  LIST CVars, LitVars;

#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause)) {
    fputs("\n\nIllegal input in clause_HasSolvedConstraint\n", stderr);
    misc_DumpCore();
  }
#endif

  CVars = list_Nil();
  c     = clause_NumOfConsLits(Clause);

  if (c == 0)
    return TRUE;

  if (clause_HasTermSortConstraintLits(Clause))
    return FALSE;

  for (i = 0;i < c; i++)
    CVars = list_NPointerUnion(term_VariableSymbols(clause_GetLiteralAtom(Clause, i)), CVars);

  if (i == c) {
    c       = clause_Length(Clause);
    LitVars = list_Nil();

    for (; i<c; i++)
      LitVars = list_NPointerUnion(LitVars, term_VariableSymbols(clause_GetLiteralAtom(Clause, i)));
	  
    if (list_Empty(CVars = list_NPointerDifference(CVars, LitVars))) {
      list_Delete(LitVars);
      return TRUE;
    }
    list_Delete(LitVars);
  }

  list_Delete(CVars);

  return FALSE;
}

BOOL clause_HasSelectedLiteral(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE iff <Clause> has a selected literal
  MEMORY:  No effects.
***************************************************************/
{
  int  i,negs;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_HasSelectedLiteral\n", stderr);
    misc_DumpCore();
  }
#endif

  negs = clause_LastAntecedentLitIndex(Clause);

  for(i=clause_FirstAntecedentLitIndex(Clause);i<=negs;i++)
    if (clause_LiteralGetFlag(clause_GetLiteral(Clause,i),LITSELECT))
      return TRUE;

  return FALSE;
}

BOOL clause_IsDeclarationClause(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE, if 'Clause' has only variables
           as arguments in constraint literals.
  MEMORY:  No effects.
***************************************************************/
{
  int     i,length;
  LITERAL Lit;

#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause)) {
    fputs("\n\nIllegal input in clause_IsDeclarationClause\n", stderr);
    misc_DumpCore();
  }
#endif 
  
  if (!clause_HasSolvedConstraint(Clause))
    return FALSE;

  length = clause_Length(Clause);

  for(i=clause_FirstSuccedentLitIndex(Clause);i<length;i++) {
    Lit = clause_GetLiteral(Clause,i);
    if (clause_LiteralIsMaximal(Lit) && symbol_IsBaseSort(term_TopSymbol(clause_LiteralSignedAtom(Lit))))
      return TRUE;
  }    

  return FALSE;
}

BOOL clause_IsSortTheoryClause(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE, if 'Clause' has only variables
           as arguments in constraint literals, no antecedent literals
	   and exactly one monadic succedent literal.
  MEMORY:  No effects.
***************************************************************/
{
  LITERAL Lit;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_IsSortTheoryClause\n", stderr);
    misc_DumpCore();
  }
#endif 
  
  if (clause_NumOfAnteLits(Clause) > 0 ||
      clause_NumOfSuccLits(Clause) > 1 ||
      !clause_HasSolvedConstraint(Clause))
    return FALSE;

  Lit = clause_GetLiteral(Clause,clause_FirstSuccedentLitIndex(Clause));
  if (symbol_IsBaseSort(term_TopSymbol(clause_LiteralSignedAtom(Lit))))
    return TRUE;

  return FALSE;
}

BOOL clause_IsPotentialSortTheoryClause(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE, if 'Clause' has monadic literals
           only variables
           as arguments in antecedent/constraint literals, no other antecedent literals
	   and exactly one monadic succedent literal.
  MEMORY:  No effects.
***************************************************************/
{
  LITERAL Lit;
  int     i;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_IsPotentialSortTheoryClause\n", stderr);
    misc_DumpCore();
  }
#endif 
  
  if (clause_NumOfSuccLits(Clause) != 1)
    return FALSE;

  for (i=clause_FirstLitIndex();i<clause_FirstSuccedentLitIndex(Clause);i++) {
    Lit = clause_GetLiteral(Clause,i);
    if (!symbol_IsBaseSort(term_TopSymbol(clause_LiteralAtom(Lit))) ||
	!term_IsVariable(term_FirstArgument(clause_LiteralAtom(Lit))))
      return FALSE; 
  }

  Lit = clause_GetLiteral(Clause,clause_FirstSuccedentLitIndex(Clause));
  if (symbol_IsBaseSort(term_TopSymbol(clause_LiteralSignedAtom(Lit))))
    return TRUE;

  return FALSE;
}


BOOL clause_HasOnlyVarsInConstraint(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE, if 'Clause' has only variables
           as arguments in constraint literals.
  MEMORY:  No effects.
***************************************************************/
{
  int  i,nc;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_HasOnlyVarsInConstraint\n", stderr);
    misc_DumpCore();
  }
#endif 

  nc = clause_NumOfConsLits(Clause);

  for (i = 0;
       (i < nc) && term_AllArgsAreVar(clause_GetLiteralAtom(Clause, i));
       i++);

  return (i==nc);
}


BOOL clause_HasSortInSuccedent(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: The boolean value TRUE, if 'Clause' has a maximal succedent
           sort literal; FALSE, else.
  MEMORY:  No effects.
***************************************************************/
{
  LITERAL Lit;
  int     i,l;
  BOOL    result = FALSE;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_HasMaxSortInSuccedent\n", stderr);
    misc_DumpCore();
  }
#endif 
 
  l  = clause_Length(Clause);

  for (i = clause_FirstSuccedentLitIndex(Clause); i < l && !result ; i++) {
    Lit = clause_GetLiteral(Clause, i);
    result = (symbol_Arity(term_TopSymbol(clause_LiteralAtom(Lit))) == 1);
  }
  return result;
}


BOOL clause_LitsHaveCommonVar(LITERAL Lit1, LITERAL Lit2)
/**************************************************************
  INPUT:   Two literals.
  RETURNS: The boolean value TRUE, if 'Lit1' and 'Lit2' have
           common variables, FALSE, else.
  MEMORY:  No effects.
***************************************************************/
{
  LIST Vars1, Vars2;
  BOOL Result;

  Vars1  = term_VariableSymbols(clause_LiteralAtom(Lit1));
  Vars2  = term_VariableSymbols(clause_LiteralAtom(Lit2));
  Result = list_HasIntersection(Vars1, Vars2);
  list_Delete(Vars1); list_Delete(Vars2);

  return Result;
}


/**************************************************************/
/* Clause Input and Output Functions                          */
/**************************************************************/

void clause_Print(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: Nothing.
  MEMORY:  No effects.
***************************************************************/
{
  RULE Origin;
  LITERAL Lit;
  int i,c,a,s;

  /*  printf("\nClause at %ld :", (long)Clause); */

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_Print\n", stderr);
    misc_DumpCore();
  }
#endif 

  if (!clause_Exists(Clause))
    fputs("(CLAUSE)NULL", stdout);
  else {
    printf("%d",clause_Number(Clause));
    
    Origin = clause_Origin(Clause);
    printf("[%d:", clause_SplitLevel(Clause));

#ifdef DEBUG
    if (Clause->splitfield_length <= 1)
      fputs("0.", stdout);
    else
      for(i=Clause->splitfield_length-1; i > 0; i--)
	printf("%lu.", Clause->splitfield[i]);
    if (Clause->splitfield_length == 0)
      putchar('1');
    else
      printf("%lu", (Clause->splitfield[0] | 1));
    printf(":%c%c:%c:%d:%d:", clause_GetFlag(Clause, CONCLAUSE) ? 'C' : 'A',
	   clause_GetFlag(Clause, WORKEDOFF) ? 'W' : 'U', 
	   clause_GetFlag(Clause, NOPARAINTO) ? 'N' : 'P',
	   clause_Weight(Clause), 
	   clause_Depth(Clause));
#endif 
    
    clause_PrintOrigin(Clause);
    
    if (Origin == INPUT) {
      ;
    } else  {
      
      putchar(':');
      clause_PrintParentClauses(Clause);
    }
    putchar(']'); putchar(' ');

    c = clause_NumOfConsLits(Clause);
    a = clause_NumOfAnteLits(Clause);
    s = clause_NumOfSuccLits(Clause);

    for (i = 0; i < c; i++) {

      Lit = clause_GetLiteral(Clause, i);
      clause_LiteralPrintUnsigned(Lit);
      if (i+1<c)
	putchar(' ');
      
    }
    fputs(" || ", stdout);

    a += c;
    for ( ; i < a; i++) {

      Lit = clause_GetLiteral(Clause, i);
      clause_LiteralPrintUnsigned(Lit);
      if (clause_LiteralIsMaximal(Lit)) {
	putchar('*');
	if (clause_LiteralIsOrientedEquality(Lit))
	  putchar('*');
      }
      if (clause_LiteralGetFlag(Lit,LITSELECT))
	putchar('+');
      if (i+1 < a)
	putchar(' ');
    }
    fputs(" -> ",stdout);

    s += a;
    for( ; i < s; i++){

      Lit = clause_GetLiteral(Clause, i);
      clause_LiteralPrintUnsigned(Lit);
      if (clause_LiteralIsMaximal(Lit)) {
	putchar('*');
	if (clause_LiteralIsOrientedEquality(Lit))
	  putchar('*');
      }
      if (clause_LiteralGetFlag(Lit,LITSELECT))
	putchar('+');      /* Redundant, this case shouldn't happen ! */
      if (i+1<s)
	putchar(' ');
    }
    putchar('.');
  }
}


void clause_PrintMaxLitsOnly(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: none.
  MEMORY:  No effects.
***************************************************************/
{
  int i,c,a,s;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_PrintMaxLitsOnly\n", stderr);
    misc_DumpCore();
  }
#endif 

  c = clause_NumOfConsLits(Clause);
  a = clause_NumOfAnteLits(Clause);
  s = clause_NumOfSuccLits(Clause);

  for (i = 0; i < c; i++) {
    if (clause_LiteralIsMaximal(clause_GetLiteral(Clause, i)))
      clause_LiteralPrint(clause_GetLiteral(Clause, i));
    if (clause_LiteralGetFlag(clause_GetLiteral(Clause, i),STRICTMAXIMAL)) {
      clause_LiteralPrint(clause_GetLiteral(Clause, i));
      fputs("(strictly)", stdout);
    }
  }
  fputs(" || ", stdout);
  
  a += c;
  for ( ; i < a; i++) {
    if (clause_LiteralIsMaximal(clause_GetLiteral(Clause, i)))
      clause_LiteralPrint(clause_GetLiteral(Clause, i));
    if (clause_LiteralGetFlag(clause_GetLiteral(Clause, i),STRICTMAXIMAL)) {
      clause_LiteralPrint(clause_GetLiteral(Clause, i));
      fputs("(strictly)", stdout);
    }
  }
  fputs(" -> ", stdout);

  s += a;
  for ( ; i < s; i++) {
    if (clause_LiteralIsMaximal(clause_GetLiteral(Clause, i)))
      clause_LiteralPrint(clause_GetLiteral(Clause, i));
    if (clause_LiteralGetFlag(clause_GetLiteral(Clause, i),STRICTMAXIMAL)) {
      clause_LiteralPrint(clause_GetLiteral(Clause, i));
      fputs("(strictly)", stdout);
    }
  }
  puts(".");  /* with newline */
}


void clause_FPrint(FILE* File, CLAUSE Clause)
/**************************************************************
  INPUT:   A file and a clause.
  RETURNS: none.
  SUMMARY: Prints any clause to the file 'File'. 
  CAUTION: Uses the term_Output functions.
  MEMORY:  No effects.
***************************************************************/
{
  int i, c, a, s;

  c = clause_NumOfConsLits(Clause);
  a = clause_NumOfAnteLits(Clause);
  s = clause_NumOfSuccLits(Clause);

  for(i = 0; i < c; i++)
    term_FPrint(File, clause_GetLiteralAtom(Clause, i));

  fputs(" || ", stdout);

  a += c;
  for( ; i < a; i++)
    term_FPrint(File, clause_GetLiteralAtom(Clause, i));

  fputs(" -> ", stdout);

  s += a;
  for( ; i < s; i++)
    term_FPrint(File, clause_GetLiteralAtom(Clause, i));
  
  putc('.', File);
}


void clause_ListPrint(LIST ClauseList)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: none.
  SUMMARY: Prints the clauses  to stdout. 
  CAUTION: Uses the clause_Print function.
  MEMORY:  No effects.
***************************************************************/
{
  while(!(list_Empty(ClauseList))){
    clause_Print(list_First(ClauseList)); 
    ClauseList = list_Cdr(ClauseList);
    if (!list_Empty(ClauseList))
      putchar('\n');
  }
}


void clause_PrintParentClauses(CLAUSE Clause)
/**************************************************************
  INPUT:	A Clause.
  RETURNS: None, prints the clauses parentclauses and -literals to stdout.
  MEMORY:  No effects.
***************************************************************/
{
  LIST Scan1,Scan2;
  int  length;
  
  if (!list_Empty(clause_ParentClauses(Clause))) {
    
    Scan1 = clause_ParentClauses(Clause);
    Scan2 = clause_ParentLiterals(Clause);
    printf("%d.%d", (int)list_Car(Scan1), (int)list_Car(Scan2));
    
    if (!list_Empty(list_Cdr(Scan1))) {
      
      length = list_Length(Scan1) - 2;
      putchar(',');
      Scan1 = list_Cdr(Scan1);
      Scan2 = list_Cdr(Scan2);
      printf("%d.%d", (int)list_Car(Scan1), (int)list_Car(Scan2));
      
      for (Scan1 = list_Cdr(Scan1), Scan2 = list_Cdr(Scan2);
	   !list_Empty(Scan1);
	   Scan1 = list_Cdr(Scan1), Scan2 = list_Cdr(Scan2)) {
	
	length -= 2;
	
	/* if (length == 0)
	   printf("|%d.%d", (int)list_Car(Scan1), (int)list_Car(Scan2));
	   else PROVISORIUM!!! */
	printf(",%d.%d", (int)list_Car(Scan1), (int)list_Car(Scan2));
      }
    }
  }
}


RULE clause_GetOriginFromString(const char* RuleName)
/**************************************************************
  INPUT:   A string containing the abbreviated name of a rule.
  RETURNS: The RULE corresponding to the <RuleName>.
***************************************************************/
{
  if      (strcmp(RuleName, "App")==0) return CLAUSE_DELETION;
  else if (strcmp(RuleName, "EmS")==0) return EMPTY_SORT;
  else if (strcmp(RuleName, "SoR")==0) return SORT_RESOLUTION;
  else if (strcmp(RuleName, "EqR")==0) return EQUALITY_RESOLUTION;
  else if (strcmp(RuleName, "EqF")==0) return EQUALITY_FACTORING;
  else if (strcmp(RuleName, "MPm")==0) return MERGING_PARAMODULATION;
  else if (strcmp(RuleName, "SpR")==0) return SUPERPOSITION_RIGHT;
  else if (strcmp(RuleName, "SPm")==0) return PARAMODULATION;
  else if (strcmp(RuleName, "OPm")==0) return ORDERED_PARAMODULATION;
  else if (strcmp(RuleName, "SpL")==0) return SUPERPOSITION_LEFT;
  else if (strcmp(RuleName, "GeR")==0) return GENERAL_RESOLUTION;
  else if (strcmp(RuleName, "SHy")==0) return SIMPLE_HYPER;
  else if (strcmp(RuleName, "OHy")==0) return ORDERED_HYPER;
  else if (strcmp(RuleName, "GeF")==0) return GENERAL_FACTORING;
  else if (strcmp(RuleName, "Spt")==0) return SPLITTING;
  else if (strcmp(RuleName, "Inp")==0) return INPUT;
  else if (strcmp(RuleName, "Rew")==0) return REWRITING;
  else if (strcmp(RuleName, "Con")==0) return CONDENSING;
  else if (strcmp(RuleName, "Obv")==0) return OBVIOUS_REDUCTIONS;
  else if (strcmp(RuleName, "SSi")==0) return SORT_SIMPLIFICATION;
  else if (strcmp(RuleName, "ClR")==0) return CLAUSE_REDUCTION;
  else if (strcmp(RuleName, "UnC")==0) return UNIT_CONFLICT;
  else if (strcmp(RuleName, "Def")==0) return DEFAPPLICATION;
  else {
    fputs("\nError in clause_GetOriginFromString: Unknown clause origin", stderr);
    misc_DumpCore();
  }
  return(CLAUSE_DELETION);
}

void clause_FPrintOrigin(FILE* File, CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: None, prints the clause's origin to the file.
  MEMORY:  No effects.
***************************************************************/
{
  RULE Result;

  Result = clause_Origin(Clause);

  switch(Result) {
  case CLAUSE_DELETION:        fputs("App", File); break;
  case EMPTY_SORT:             fputs("EmS", File); break;
  case SORT_RESOLUTION:        fputs("SoR", File); break;
  case EQUALITY_RESOLUTION:    fputs("EqR", File); break;
  case EQUALITY_FACTORING:     fputs("EqF", File); break;
  case MERGING_PARAMODULATION: fputs("MPm", File); break;
  case SUPERPOSITION_RIGHT:    fputs("SpR", File); break;
  case PARAMODULATION:         fputs("SPm", File); break;
  case ORDERED_PARAMODULATION: fputs("OPm", File); break;  
  case SUPERPOSITION_LEFT:     fputs("SpL", File); break;
  case GENERAL_RESOLUTION:     fputs("GeR", File); break;
  case SIMPLE_HYPER:           fputs("SHy", File); break;
  case ORDERED_HYPER:          fputs("OHy", File); break;  
  case GENERAL_FACTORING:      fputs("GeF", File); break;
  case SPLITTING:              fputs("Spt", File); break;
  case INPUT:                  fputs("Inp", File); break;
  case REWRITING:              fputs("Rew", File); break;
  case CONDENSING:             fputs("Con", File); break;
  case OBVIOUS_REDUCTIONS:     fputs("Obv", File); break;
  case SORT_SIMPLIFICATION:    fputs("SSi", File); break;
  case CLAUSE_REDUCTION:       fputs("ClR", File); break;
  case UNIT_CONFLICT:          fputs("UnC", File); break;
  case DEFAPPLICATION:         fputs("Def", File); break;
  default:                     fputs("\nError: clause has no origin.", stdout);
  }
}


void clause_PrintOrigin(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: None, prints the clauses origin to stdout.
  MEMORY:  No effects.
***************************************************************/
{ clause_FPrintOrigin(stdout, Clause);  }


void clause_PrintVerbose(CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: None, prints almost all the information kept in the
           clause structure.
  MEMORY:  No effects.
***************************************************************/
{
  int c,a,s;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_PrintVerbose\n", stderr);
    misc_DumpCore();
  }
#endif 

  c = clause_NumOfConsLits(Clause);
  a = clause_NumOfAnteLits(Clause);
  s = clause_NumOfSuccLits(Clause);

  printf(" c = %d a = %d s = %d", c,a,s); 
  printf(" Weight : %d", clause_Weight(Clause));
  printf(" Depth  : %d", clause_Depth(Clause));
  printf(" %s %s ", 
	 (clause_GetFlag(Clause,WORKEDOFF) ? "WorkedOff" : "Usable"),
	 (clause_GetFlag(Clause,CLAUSESELECT) ? "Selected" : "NotSelected"));

  clause_Print(Clause);
}


CLAUSE clause_GetNumberedCl(int number, LIST ClList)
{
  while (!list_Empty(ClList) &&
	 (clause_Number((CLAUSE)list_Car(ClList)) != number))
    ClList = list_Cdr(ClList);
  
  if (list_Empty(ClList))
    return((CLAUSE)NULL);
  else
    return((CLAUSE)list_Car(ClList));
}


LIST clause_NumberSort(LIST List)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: The same list where the elements are sorted wrt their number.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1 = List; !list_Empty(Scan1); Scan1 = list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1); !list_Empty(Scan2); Scan2=list_Cdr(Scan2)) {
      if (clause_Number((CLAUSE)list_Car(Scan2))
	  < clause_Number((CLAUSE)list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
    }
  }
  return(List);
}


LIST clause_NumberDelete(LIST List, int Number)
/**************************************************************
  INPUT:   A list of clauses and an integer.
  RETURNS: The same list where the clauses with <Number> are deleted.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan1,Scan2;
  
  for (Scan1 = List; !list_Empty(Scan1); )
    if (clause_Number(list_Car(Scan1))==Number) {
      
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
      List  = list_PointerDeleteOneElement(List, list_Car(Scan2));
    } else
      Scan1 = list_Cdr(Scan1);

  return(List);
}


static NAT clause_NumberOfMaxLits(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: The number of maximal literals in a clause.
***************************************************************/
{
  NAT Result,i,n;

  Result = 0;
  n      = clause_Length(Clause);

  for (i = clause_FirstAntecedentLitIndex(Clause); i < n; i++)
    if (clause_LiteralIsMaximal(clause_GetLiteral(Clause,i)))
      Result++;

  return Result;
}

/* Unused ! */
NAT clause_NumberOfMaxAntecedentLits(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: The number of maximal antecedent literals in a clause.
***************************************************************/
{
  NAT Result,i,n;

  Result = 0;
  n      = clause_LastAntecedentLitIndex(Clause);

  for (i = clause_FirstAntecedentLitIndex(Clause); i <= n; i++)
    if (clause_LiteralIsMaximal(clause_GetLiteral(Clause,i)))
      Result++;

  return Result;
}


BOOL clause_NegativePropLiteral(CLAUSE Clause) 
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains a negative propositional
           literal, 
           FALSE, otherwise.
  MEMORY:  No Effects.
**********************************************************/
{
    NAT     i,n;
    BOOL    Prop;

    n   = clause_LastAntecedentLitIndex(Clause);
    i   = clause_FirstAntecedentLitIndex(Clause);
    Prop = FALSE;
    for (i = 0; i <= n; i++) 
      Prop = (Prop || (symbol_Arity(term_TopSymbol(clause_GetLiteralAtom(Clause,i))) == 0));
    return Prop;
}

BOOL clause_PropLiteral(CLAUSE Clause) 
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE, if the clause contains a propositional
           literal, 
           FALSE, otherwise.
  MEMORY:  No Effects.
**********************************************************/
{
    NAT     i,n;
    BOOL    Prop;

    n   = clause_Length(Clause);
    i   = clause_FirstLitIndex();
    Prop = FALSE;
    for (i = 0; i <= n; i++) 
      Prop = (Prop || (symbol_Arity(term_TopSymbol(clause_GetLiteralAtom(Clause,i))) == 0));
    return Prop;
}

int clause_LiteralArity(CLAUSE Clause) 
/*********************************************************
  INPUT:   A clause.
  RETURNS: The common arity of the literals in the clause, if 
           the arity of all literals in the clause is equal,
           -1 otherwise
	   For the empty clause the result will be 0.
  MEMORY:  No Effects.
**********************************************************/
{
  int i,Result,Arity;

  i      = clause_FirstLitIndex();
  Result = -1;
  Arity = 0;
  if (clause_Length(Clause) < i) {
    /* Clause is a empty clause */
    Result = 0;
  } else {
    while (i < clause_Length(Clause) &&
	   Arity >= 0) {
      Arity = symbol_Arity(term_TopSymbol(clause_GetLiteralAtom(Clause,i)));
      if (Result >= 0) {
	if (Arity != Result) {
	  Arity  = -1;
	  Result = -1;
	};
      } else 
	Result = Arity;
      i++;
    };
  };
  return Result;
}

BOOL clause_ContainsNDLiteral(CLAUSE Clause) 
/*********************************************************
  INPUT:   A clause.
  RETURNS: TRUE if the clause contains a ND literal,
           FALSE otherwise.
  MEMORY:  No Effects.
**********************************************************/
{
  int i;
  BOOL Result;

  i      = clause_FirstLitIndex();
  Result = FALSE;
  if (clause_Length(Clause) >= i) {
    /* Clause is not an empty clause */
    while (i < clause_Length(Clause) &&
	   !Result) {
      Result = (strncmp("nd",symbol_Name(term_TopSymbol(clause_GetLiteralAtom(Clause,i))),2) == 0);
      i++;
    };
  };
  return Result;
}


void clause_SelectLiteral(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: void.
  EFFECT:  If the clause contains more than 2 maximal literals,
           at least one antecedent literal, the literal with
	   the highest weight is selected.
***************************************************************/
{
  BOOL   Prop;
  
  Prop       = (clause_LiteralArity(Clause) == 0);
  if (clause_HasSolvedConstraint(Clause) &&
      !clause_GetFlag(Clause,CLAUSESELECT) &&
      clause_NumOfAnteLits(Clause) > 0 &&
      ((flag_GetFlagValue(flag_SELECT) >= 2) ||
       (flag_GetFlagValue(flag_SELECT) == 1 && clause_NumberOfMaxLits(Clause) > 1))) {
    NAT     i,n;
    LITERAL Lit;
    
    n   = clause_LastAntecedentLitIndex(Clause);
    i   = clause_FirstAntecedentLitIndex(Clause);
    Lit = clause_GetLiteral(Clause,i);
    i++;
    
    /*    for ( ; i <= n; i++)
      if (clause_LiteralWeight(Lit)
	  < clause_LiteralWeight(clause_GetLiteral(Clause,i)))
	Lit = clause_GetLiteral(Clause,i);
	*/ /* select literal with maximal weight */

    for ( ; i <= n; i++)
      if (symbol_PrecedenceGreater(term_TopSymbol(clause_GetLiteralAtom(Clause,i)),
				   term_TopSymbol(clause_LiteralAtom(Lit)))) 
	Lit = clause_GetLiteral(Clause,i);

    if (flag_GetFlagValue(flag_SELECT) == 3 &&
	!Prop && 
	clause_NegativePropLiteral(Clause) &&
	(symbol_Arity(term_TopSymbol(clause_LiteralAtom(Lit))) == 0)) {
      clause_LiteralSetFlag(Lit,LITSELECT);
      clause_SetFlag(Clause,CLAUSESELECT);
    } else {
      if (flag_GetFlagValue(flag_SELECT) == 4 &&
	  !Prop && 
	  (symbol_Weight(term_TopSymbol(clause_LiteralAtom(Lit))) > 1)) {
	clause_LiteralSetFlag(Lit,LITSELECT);
	clause_SetFlag(Clause,CLAUSESELECT);
      };
    }
  }
}

void clause_SetSpecialFlags(CLAUSE Clause, BOOL SortDecreasing)
/**************************************************************
  INPUT:   A clause and a flag indicating wether all equations are
           sort decreasing. 
  RETURNS: void.
  EFFECT:  If the clause is a sort theory clause and its declaration
           top symbol is a set delcaration sort,i.e., it occured in a 
	   declaration right from the beginning, the paramodulation/superposition
	   steps into the clause are folbidden by setting the
	   NOPARAINTO clause flag
***************************************************************/
{
  if (SortDecreasing && 
      clause_IsSortTheoryClause(Clause) &&
      symbol_HasProperty(term_TopSymbol(clause_GetLiteralTerm(Clause,clause_FirstSuccedentLitIndex(Clause))),
			 DECLSORT))
    clause_SetFlag(Clause,NOPARAINTO);
  
}
    
BOOL clause_ContainsPotPredDef(CLAUSE Clause, NAT* Index, LIST* Pair)
/**************************************************************
  INPUT:   A clause and a pointer to an index.
  RETURNS: TRUE iff a succedent literal of the clause is a predicate
           having only variables as arguments, the predicate occurs only
	   once in the clause and no other variables but the predicates'
	   occur.
	   In that case Index is set to the index of the predicate and
	   Pair contains two lists : the literals for which positive 
	   occurrences must be found and a list of literals for which negative
	   occurrences must be found for a complete definition.
***************************************************************/     
{
  NAT i;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_ContainsPotPredDef\n", stderr);
    misc_DumpCore();
  }
#endif   

  /* Iterate over all succedent literals */
  for(i=clause_FirstSuccedentLitIndex(Clause); i < clause_Length(Clause); i++) {
    LITERAL lit;
    TERM atom;
    LIST pair;

    lit = clause_GetLiteral(Clause, i);
    atom = clause_LiteralSignedAtom(lit);
    if(symbol_IsPredicate(term_TopSymbol(atom))) {
      LIST l;
      BOOL ok;
      ok = TRUE;
      pair = list_PairCreate(list_Nil(), list_Nil());
      
      /* Make sure all arguments of predicate are variables */
      for(l=term_ArgumentList(atom); !list_Empty(l); l=list_Cdr(l)) {
	if (!term_IsStandardVariable((TERM) list_Car(l))) {
	  ok = FALSE;
	  break;
	}
      }
      if(ok == TRUE) {
	/* Make sure predicate occurs only once in clause */
	NAT j, count;
	count = 0;
	for(j=0; (j < clause_Length(Clause)) && (count < 2); j++) {
	  TERM t;
	  t = clause_GetLiteralAtom(Clause, j);
	  if (symbol_Equal(term_TopSymbol(t), term_TopSymbol(atom)))
	    count++;
	}
	if(count > 1)
	  ok = FALSE;
      }
      if(ok == TRUE) {
	/* Build lists of positive and negative literals */
	/* At the same time check if other variables than those among
	   the predicates arguments are found */
	NAT j;
	LIST predvars, vars;
	predvars = fol_FreeVariables(atom);
	
	/* Build list of literals for which positive occurrences are required */
	for(j=0; (j < clause_FirstSuccedentLitIndex(Clause)) && (ok == TRUE); 
	    j++) {
	  list_Rplaca(pair, list_Cons(clause_GetLiteralAtom(Clause, j), list_PairFirst(pair)));
	  vars = fol_FreeVariables(clause_GetLiteralTerm(Clause, j));
	  for(l=vars; !list_Empty(l); l=list_Cdr(l)) {
	    if (!list_Member(predvars, (TERM) list_Car(l), term_Equal)) {
	      ok = FALSE;
	      break;
	    }
	  }
	  list_Delete(vars);
	}
	
	/* Build list of literals for which negative occurrences are required */
	for(j=clause_FirstSuccedentLitIndex(Clause);
	    (j < clause_Length(Clause)) && (ok == TRUE); j++) {
	  if (j != i) {
	    list_Rplacd(pair, list_Cons(clause_GetLiteralAtom(Clause, j), list_PairSecond(pair)));
	    vars = fol_FreeVariables(clause_GetLiteralAtom(Clause, j));
	    for(l=vars; !list_Empty(l); l=list_Cdr(l)) {
	      if (!list_Member(predvars, (TERM) list_Car(l), term_Equal)) {
		ok = FALSE;
		break;
	      }
	    }
	    list_Delete(vars);
	  }
	}
	list_Delete(predvars);
      }

      if(ok == TRUE) {
	*Index = i;
	*Pair = pair;
	return TRUE;
      }
      else {
	list_Delete(list_PairFirst(pair));
	list_Delete(list_PairSecond(pair));
	list_PairFree(pair);
      }
    }
  }
  return FALSE;
}

BOOL clause_IsPartOfDefinition(CLAUSE Clause, TERM Predicate, int *Index, LIST Pair)
/**************************************************************
  INPUT:   A clause, a term, a pointer to an int and a pair of term lists.
  RETURNS: TRUE iff the predicate occurs among the negative literals of
           the clause and the other negative and positive literals are found
	   in the pairs' lists.
	   In that case they are removed from the lists.
	   Index is set to the index of the defined predicate in Clause.
***************************************************************/
{
  NAT predindex, i;
  BOOL ok;

  ok = TRUE;

  /* Check whether Predicate is among antecedent or constraint literals */
  for(predindex=clause_FirstLitIndex(); 
      predindex < clause_FirstSuccedentLitIndex(Clause);
      predindex++)
    if (term_Equal(Predicate, clause_GetLiteralAtom(Clause, predindex))) 
      break;
  if (predindex == clause_FirstSuccedentLitIndex(Clause))
    return FALSE;
  *Index = predindex;

  /* Check if other negative literals are required for definition */
  for(i=clause_FirstLitIndex(); 
      (i < clause_FirstSuccedentLitIndex(Clause)) && (ok == TRUE); i++) {
    if (i != predindex) 
      if(!list_Member((LIST) list_PairSecond(Pair), 
		      clause_GetLiteralAtom(Clause, i), term_Equal))
	ok = FALSE;
  }

  /* Check if positive literals are required for definition */
  for(i=clause_FirstSuccedentLitIndex(Clause); 
      (i < clause_Length(Clause)) && (ok == TRUE);
      i++)
    if(!list_Member((LIST) list_PairFirst(Pair), 
		    clause_GetLiteralAtom(Clause, i), term_Equal))
      ok = FALSE;
  
  if (ok == FALSE)
    return FALSE;
  else {
    /* Complement for definition found, remove literals from pair lists */
    for(i=0; i < clause_FirstSuccedentLitIndex(Clause); i++)
      if (i != predindex)
	list_Rplacd(Pair, 
		    list_DeleteElement((LIST) list_PairSecond(Pair), 
				       clause_GetLiteralAtom(Clause, i), 
				       term_Equal));
    for(i=clause_FirstSuccedentLitIndex(Clause); i < clause_Length(Clause); i++)
      list_Rplaca(Pair, 
		  list_DeleteElement((LIST) list_PairFirst(Pair), 
				     clause_GetLiteralAtom(Clause, i),
				     term_Equal));
    return TRUE;
  }
}
    
void clause_FPrintRule(FILE* File, CLAUSE Clause)
/**************************************************************
  INPUT:   A file and a clause.
  RETURNS: none.
  SUMMARY: Prints any term of the clause to file in rule format.
  CAUTION: Uses the term_Output functions.
***************************************************************/
{
  int  i,n;
  TERM Literal;
  LIST scan,antecedent,succedent,constraints;

  n = clause_Length(Clause);

  constraints = list_Nil();
  antecedent  = list_Nil();
  succedent   = list_Nil();

  for(i = 0; i < n; i++) {
    Literal = clause_GetLiteralTerm(Clause,i);
    if (symbol_Equal(term_TopSymbol(Literal),fol_Not())) {
      if (symbol_Arity(term_TopSymbol(fol_Atom(Literal))) == 1 &&
	  symbol_IsVariable(term_TopSymbol(term_FirstArgument(fol_Atom(Literal)))))
	constraints = list_Cons(Literal,constraints);
      else
	antecedent = list_Cons(Literal,antecedent);
    }
    else
      succedent = list_Cons(Literal,succedent);
  }
  /*fputs("||", File);*/
  for(scan = constraints; !list_Empty(scan); scan = list_Cdr(scan)) {
    term_FPrint(File, fol_Atom(list_Car(scan)));
    putc(' ', File);
  }
  fputs("||", File);
  for(scan = antecedent; !list_Empty(scan); scan = list_Cdr(scan)) {
    putc(' ', File);
    term_FPrint(File,fol_Atom(list_Car(scan)));
    if (list_Empty(list_Cdr(scan)))
      putc(' ', File);
  }
  fputs("->", File);
  for(scan = succedent; !list_Empty(scan); scan = list_Cdr(scan)) {
    putc(' ', File);
    term_FPrint(File,list_Car(scan));
  }
  fputs(".\n", File);
  
  list_Delete(antecedent);
  list_Delete(succedent);
  list_Delete(constraints);
}


void clause_FPrintOtter(FILE* File, CLAUSE clause)
/**************************************************************
  INPUT:   A file and a clause.
  RETURNS: none.
  SUMMARY: Prints any clause to File. 
  CAUTION: Uses the other clause_Output functions.
***************************************************************/
{
  int     n,j;
  LITERAL Lit;
  TERM    Atom;
  
  n = clause_Length(clause);

  for (j = 0; j < n; j++) {
    Lit  = clause_GetLiteral(clause,j);
    Atom = clause_LiteralAtom(Lit);
    if (clause_LiteralIsNegative(Lit))
      putc('-', File);
    if (fol_IsEquality(Atom)) {
      if (clause_LiteralIsNegative(Lit))
	putc('(', File);
      term_FPrintOtterPrefix(File,term_FirstArgument(Atom));
      fputs(" = ", File);
      term_FPrintOtterPrefix(File,term_SecondArgument(Atom));
      if (clause_LiteralIsNegative(Lit))
	putc(')', File);
    }
    else
      term_FPrintOtterPrefix(File,Atom);
    if (j <= (n-2))
      fputs(" | ", File);
  }
  fputs(".\n", File);
}


void clause_FPrintCnfDFG(FILE* File, BOOL OnlyProductive, LIST Axioms, LIST Conjectures)
/**************************************************************
  INPUT:   A file, a list of axiom clauses and a list of conjecture clauses.
           A flag indicating whether only potentially productive clauses should
	   be printed.
  RETURNS: none.
  SUMMARY: Prints a  the respective clause lists to <File> dependant on <OnlyProductive>. 
***************************************************************/
{
  LIST   scan;
  CLAUSE Clause;

  if (Axioms) {
    fputs("list_of_clauses(axioms, cnf).\n", File);
    for (scan=Axioms;!list_Empty(scan);scan=list_Cdr(scan)) {
      Clause = (CLAUSE)list_Car(scan);
      if (!OnlyProductive || 
	  (clause_HasSolvedConstraint(Clause) && !clause_HasSelectedLiteral(Clause)))
	clause_FPrintDFG(File,Clause,FALSE);
    }
    fputs("end_of_list.\n\n", File);
  }

  if (Conjectures) {
    fputs("list_of_clauses(conjectures, cnf).\n", File);
    for (scan=Conjectures;!list_Empty(scan);scan=list_Cdr(scan)) {
      Clause = (CLAUSE)list_Car(scan);
      if (!OnlyProductive || 
	  (clause_HasSolvedConstraint(Clause) && !clause_HasSelectedLiteral(Clause)))
	clause_FPrintDFG(File,Clause,FALSE);
    }
    fputs("end_of_list.\n\n", File);
  }
}


void clause_FPrintCnfOtter(FILE* File, LIST Clauses)
/**************************************************************
  INPUT:   A file, a list of axiom clauses and a list of conjecture clauses.
  RETURNS: none.
  SUMMARY: Prints a  the respective clause lists to <File>. 
***************************************************************/
{
  LIST   scan;
  int    i;
  BOOL   Equality;
  CLAUSE Clause;

  Equality = FALSE;

  for (scan=Clauses;!list_Empty(scan) && !Equality; scan=list_Cdr(scan)) {
    Clause = (CLAUSE)list_Car(scan);
    for(i=clause_FirstAntecedentLitIndex(Clause);i<clause_Length(Clause);i++)
      if (fol_IsEquality(clause_GetLiteralAtom(Clause,i))) {
	Equality = TRUE;
	i       = clause_Length(Clause);
      }
  }

  if (flag_GetFlagValue(flag_TDFG2OTTEROPTIONS) == 1) {
    fputs("\nset(process_input).", File);
    fputs("\nset(binary_res).", File);
    fputs("\nset(factor).", File);
    /*fputs("\nassign(pick_given_ratio, 4).", File);*/
    fputs("\nclear(print_kept).", File);
    fputs("\nassign(max_seconds, 20).", File);

    if (Equality) {
      fputs("\nclear(print_new_demod).", File);
      fputs("\nclear(print_back_demod).", File);
      fputs("\nclear(print_back_sub).", File);
      /*fputs("\nset(knuth_bendix).", File);*/
      fputs("\nset(para_from).", File);
      fputs("\nset(para_into).", File);
      fputs("\nset(para_from_vars).", File);
      fputs("\nset(back_demod).", File);    
    }
    fputs("\nset(auto).", File);
  }

  fputs("\n\n",File);

  if (Clauses) {
    fputs("list(usable).\n", File);
    if (Equality)
      fputs("x=x.\n", File);
    for (scan=Clauses;!list_Empty(scan);scan=list_Cdr(scan))
      clause_FPrintOtter(File,list_Car(scan));
    fputs("end_of_list.\n\n", File);
  }
}


void clause_FPrintCnfDFGDerivables(FILE* File, LIST Clauses, BOOL Type)
/**************************************************************
  INPUT:   A file, a list of clauses and a bool flag Type.
  RETURNS: none.
  SUMMARY: If <Type> is true then all axiom clauses in <Clauses> are
	   written to <File>. Otherwise all conjecture clauses in
	   <Clauses> are written to <File>. 
***************************************************************/
{
  CLAUSE Clause;

  while (Clauses) {
    Clause = (CLAUSE)list_Car(Clauses);
    if ((Type && !clause_GetFlag(Clause,CONCLAUSE)) ||
	(!Type && clause_GetFlag(Clause,CONCLAUSE)))
      clause_FPrintDFG(File,Clause,FALSE);
    Clauses = list_Cdr(Clauses);
  }
}


void clause_FPrintDFG(FILE* File, CLAUSE Clause, BOOL Justif)
/**************************************************************
  INPUT:   A file, a clause and a boolean value.
  RETURNS: none.
  SUMMARY: Prints any clause together with a label (the clause number)
	   to File. If Justif is TRUE also the labels of the parent
	   clauses are printed. 
  CAUTION: Uses the other clause_Output functions.
***************************************************************/
{
  int     n,j;
  LITERAL Lit;
  TERM    Atom;
  LIST    Variables,Iter;
  
  n = clause_Length(Clause);

  fputs("  clause(", File);
  Variables = list_Nil();

  for (j = 0; j < n; j++)
    Variables =
      list_NPointerUnion(Variables,
	term_VariableSymbols(clause_GetLiteralAtom(Clause,j)));

  if (!list_Empty(Variables)) {
    symbol_FPrint(File, fol_All());
    fputs("([", File);
    for(Iter = Variables; !list_Empty(Iter); Iter = list_Cdr(Iter)) {
      symbol_FPrint(File, (SYMBOL) list_Car(Iter));
      if (!list_Empty(list_Cdr(Iter)))
	putc(',', File);
    }
    fputs("],", File);
  }
  
  symbol_FPrint(File, fol_Or());
  putc('(', File);

  for (j = 0; j < n; j++) {
    Lit  = clause_GetLiteral(Clause,j);
    Atom = clause_LiteralSignedAtom(Lit);
    term_FPrintPrefix(File,Atom);
    if (j+1 < n)
      putc(',', File);
  }
  if (n==0)
    symbol_FPrint(File,fol_False());

  if (!list_Empty(Variables)) {
    list_Delete(Variables);
    putc(')', File);
  }
  fprintf(File, "),%d", clause_Number(Clause));

  if (Justif) {
    putc(',', File);
    clause_FPrintOrigin(File, Clause);
    fputs(",[", File);
    for(Iter = clause_ParentClauses(Clause);
	!list_Empty(Iter);
	Iter = list_Cdr(Iter)) {
      fprintf(File, "%d", (int)list_Car(Iter));
      if (!list_Empty(list_Cdr(Iter)))
	putc(',', File);
    }
    putc(']', File);
    fprintf(File, ",%d", clause_SplitLevel(Clause)); 
  }
    
  fputs(").\n", File);
}


void clause_Check(CLAUSE Clause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: void.
  EFFECT:  Checks whether the clause is in a proper state. If
           not, a core is dumped.
***************************************************************/
{
  CLAUSE Copy;
  if (!clause_Exists(Clause))
    return;
    
  if (!clause_IsClause(Clause)) {
    fputs("\n Clause not consistent !\n", stderr);
    misc_DumpCore();
  }

  Copy = clause_Copy(Clause);
  clause_OrientAndReInit(Copy);
  if ((clause_Weight(Clause) != clause_Weight(Copy)) ||
      (clause_MaxVar(Clause) != clause_MaxVar(Copy))) {
    fputs("\nClause not properly initialized: ", stdout);
    clause_Print(Clause);
    printf("\nWeight is: %d    Should be: %d", clause_Weight(Clause),
	   clause_Weight(Copy));
    printf("\nMaxVar is: %d    Should be: %d", clause_MaxVar(Clause),
	   clause_MaxVar(Copy));
    printf("\n");
    misc_DumpCore();
  }
  clause_Delete(Copy);
}

/* The following are output procedures for clauses with parent pointers */


void clause_PParentsFPrintParentClauses(FILE* File, CLAUSE Clause, BOOL ParentPts)
/**************************************************************
  INPUT:   A Clause.
  RETURNS: None, prints the clauses parent clauses and -literals to stdout.
  EFFECTS: No effects.
***************************************************************/
{
  LIST Scan1,Scan2;
  int  length;
  int  ParentNum;
  
  if (!list_Empty(clause_ParentClauses(Clause))) {
    
    Scan1 = clause_ParentClauses(Clause);
    Scan2 = clause_ParentLiterals(Clause);
    
    if (ParentPts)
      ParentNum = clause_Number(list_Car(Scan1));
    else 
      ParentNum = (int)list_Car(Scan1);

    fprintf(File, "%d.%d", ParentNum, (int)list_Car(Scan2));
    
    if (!list_Empty(list_Cdr(Scan1))) {
      
      length = list_Length(Scan1) - 2;
      fprintf(File, ",");
      Scan1 = list_Cdr(Scan1);
      Scan2 = list_Cdr(Scan2);
      
      if (ParentPts)
	ParentNum = clause_Number(list_Car(Scan1));
      else 
	ParentNum = (int)list_Car(Scan1);

      fprintf(File, "%d.%d", ParentNum, (int)list_Car(Scan2));
      
      for (Scan1 = list_Cdr(Scan1), Scan2 = list_Cdr(Scan2);
	   !list_Empty(Scan1);
	   Scan1 = list_Cdr(Scan1), Scan2 = list_Cdr(Scan2)) {
	
	length -= 2;
	 
	if (ParentPts)
	  ParentNum = clause_Number(list_Car(Scan1));
	else 
	  ParentNum = (int)list_Car(Scan1);

	fprintf(File, ",%d.%d", ParentNum, (int)list_Car(Scan2));
      }
    }
  }
}

void clause_PParentsLiteralFPrintUnsigned(FILE* File, LITERAL Literal)
/**************************************************************
  INPUT:   A Literal.
  RETURNS: none.
  EFFECTS:  No effects.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_LiteralIsLiteral(Literal)) {
    fputs("\n\nIllegal input in clause_LiteralPrintUnsigned\n", stderr);
    misc_DumpCore();
  }
#endif 

  term_FPrintPrefix(File, clause_LiteralAtom(Literal));
  fflush(stdout);
}

void clause_PParentsFPrintGen(FILE* File, CLAUSE Clause, BOOL ParentPts)
/**************************************************************
  INPUT:   A file, a clause, a flag
  RETURNS: Nothing.
  EFFECTS: Prints the clause to file in SPASS format. If <ParentPts>
           is TRUE, the parents of <Clause> are interpreted as pointers.
***************************************************************/
{
  RULE Origin;
  LITERAL Lit;
  int i,c,a,s;

  /*  printf("\nClause at %ld :", (long)Clause); */

#ifdef DEBUG
  /* if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in clause_Print\n", stderr);
    misc_DumpCore(); 
  } */
#endif 

  if (!clause_Exists(Clause))
    fputs("(CLAUSE)NULL", File);
  else {
    fprintf(File, "%d",clause_Number(Clause));
    
    Origin = clause_Origin(Clause);
    fprintf(File, "[%d:", clause_SplitLevel(Clause));

#ifdef DEBUG
    if (Clause->splitfield_length <= 1)
      fputs("0.", File);
    else
      for(i=Clause->splitfield_length-1; i > 0; i--)
	fprintf(File, "%lu.", Clause->splitfield[i]);
    if (Clause->splitfield_length == 0)
      fprintf(File,"1");
    else
      fprintf(File, "%lu", (Clause->splitfield[0] | 1));
    fprintf(File,":%c%c:%c:%d:%d:", clause_GetFlag(Clause, CONCLAUSE) ? 'C' : 'A',
	   clause_GetFlag(Clause, WORKEDOFF) ? 'W' : 'U', 
	   clause_GetFlag(Clause, NOPARAINTO) ? 'N' : 'P',
	   clause_Weight(Clause), 
	   clause_Depth(Clause));
#endif 
    
    clause_FPrintOrigin(File, Clause);
    
    if (!list_Empty(clause_ParentClauses(Clause)))
      {
	fprintf(File, ":");
	clause_PParentsFPrintParentClauses(File, Clause, ParentPts);
      }
    fprintf(File, "]");

    c = clause_NumOfConsLits(Clause);
    a = clause_NumOfAnteLits(Clause);
    s = clause_NumOfSuccLits(Clause);

    for (i = 0; i < c; i++) {

      Lit = clause_GetLiteral(Clause, i);
      clause_PParentsLiteralFPrintUnsigned(File, Lit);
      if (i+1<c)
	fprintf(File," ");
      
    }
    fputs(" || ", File);

    a += c;
    for ( ; i < a; i++) {

      Lit = clause_GetLiteral(Clause, i);
      clause_PParentsLiteralFPrintUnsigned(File, Lit);
      if (clause_LiteralIsMaximal(Lit)) {
	fprintf(File, "*");
	if (clause_LiteralIsOrientedEquality(Lit))
	  fprintf(File, "*");
      }
      if (clause_LiteralGetFlag(Lit,LITSELECT))
	fprintf(File, "+"); 
      if (i+1 < a)
	fprintf(File, " ");
    }
    fputs(" -> ",File);

    s += a;
    for( ; i < s; i++){

      Lit = clause_GetLiteral(Clause, i);
      clause_PParentsLiteralFPrintUnsigned(File, Lit);
      if (clause_LiteralIsMaximal(Lit)) {
	fprintf(File, "*"); 
	if (clause_LiteralIsOrientedEquality(Lit))
	  fprintf(File, "*"); 
      }
      if (clause_LiteralGetFlag(Lit,LITSELECT))
	fprintf(File, "+");      /* Redundant, this case shouldn't happen ! */
      if (i+1<s)
	fprintf(File," ");
    }
    fprintf(File,".");
  }
}

void clause_PParentsFPrint(FILE* File, CLAUSE Clause)
/**************************************************************
 INPUT  : A file handle, a clause                                                     
 RETURNS: -                                                     
 EFFECTS: Prints out the clause to file in SPASS output format
***************************************************************/
{
  clause_PParentsFPrintGen(File, Clause, TRUE);
}

void clause_PParentsListFPrint(FILE* File, LIST L)
/**************************************************************
 INPUT  : A file handle, a list of clauses with parent pointers
 RETURNS: -                                                      
 EFFECTS: Print the list to <file>                                                   
***************************************************************/
{
  while (!list_Empty(L)) {
    clause_PParentsFPrint(File, list_Car(L));
    fprintf(File, "\n");
    L = list_Cdr(L);
  }
}

void clause_PParentsPrint(CLAUSE Clause)
/**************************************************************
 INPUT  : A clause with parent pointers
 RETURNS: -
 EFFECTS: The clause is printed                                                     
***************************************************************/
{
  clause_PParentsFPrint(stdout, Clause);
}

void clause_PParentsListPrint(LIST L)
/**************************************************************
 INPUT  : A file handle, a list of clauses with parent pointers
 RETURNS: -                                                      
 EFFECTS: Print the list                                                    
***************************************************************/
{
  clause_PParentsListFPrint(stdout, L);
}
 
