/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                  PARSER FOR DFG SYNTAX                 * */
/* *                                                        * */
/* *  $Module:   DFG                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 1.13 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:46:58 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfgparser.y,v $ */

%{

#include <ctype.h>
#include <math.h>
#include "flags.h"
#include "dfg.h"
#include "symbol.h"
#include "term.h"
#include "foldfg.h"
#include "eml.h"
#include "clause.h"

/* Used for the texts from description section. */
typedef struct {
  char*     name;
  char*     author;
  char*     version;
  char*     logic;
  DFG_STATE status;
  char*     description;
  char*     date;
} DFG_DESCRIPTIONTYPE;

static DFG_DESCRIPTIONTYPE dfg_DESC;
static LIST                dfg_AXIOMLIST;
static LIST                dfg_CONJECLIST;
static LIST                dfg_SORTDECLLIST;
static LIST                dfg_AXCLAUSES;
static LIST                dfg_CONCLAUSES;
static LIST                dfg_PROOFLIST;     /* list_of_proofs */
static LIST                dfg_TERMLIST;      /* list_of_terms  */

NAT                        dfg_LINENUMBER;  /* used also in the scanner */


void yyerror(const char*);
int  yylex(void);		/* Defined in dfgscanner.l */

static void   dfg_SymbolDecl(int, char*, int);
static SYMBOL dfg_Symbol(char*, NAT);
static TERM   dfg_CreateQuantifier(SYMBOL, LIST, TERM);
static void   dfg_SubSort(char*, char*);
static void   dfg_Settings(char*);

static __inline__ TERM dfg_TermCreate(char* Name, LIST Arguments)
/* Look up the symbol, check its arity and create the term */
{
  return term_Create(dfg_Symbol(Name,list_Length(Arguments)), Arguments);
}

static void dfg_TranslPairDecl(char*, char*);

/**************************************************************/
/* Functions that handle symbols with unspecified arity       */
/**************************************************************/

/* The symbol list holds all symbols whose arity wasn't       */
/* specified in the symbol declaration section.               */
/* If the arity of a symbol was not specified in this section */
/* it is first set to 0. If the symbol occurs with always the */
/* same arity 'A' the arity of this symbol is set to 'A'.     */
static LIST   dfg_SYMBOLLIST;

static void dfg_SymAdd(SYMBOL);
static void dfg_SymCheck(SYMBOL, NAT);
static void dfg_SymCleanUp(void);

/**************************************************************/
/* Functions that handle variable names                       */
/**************************************************************/

/* List of quantified variables in the current input formula. */
/* This list is used to find symbols that by mistake weren't  */
/* declared in the symbol declaration section                 */
/* --> free variables                                         */
/* This is a list of lists, since each time a quantifier is   */
/* reached, a new list is added to the global list.           */
static LIST dfg_VARLIST;
static BOOL dfg_VARDECL;

static void   dfg_VarStart(void);
static void   dfg_VarStop(void);
static void   dfg_VarBacktrack(void);
static void   dfg_VarCheck(void);
static SYMBOL dfg_VarLookup(char*);

#define YYERROR_VERBOSE

%}

%union {
  int       number;
  char*     string;
  SYMBOL    symbol;
  TERM      term;
  LIST      list;
  DFG_STATE state;
  BOOL      bool;
}

%token DFG_AND DFG_AUTHOR DFG_AXIOMS DFG_BEGPROB DFG_CLAUSE DFG_CNF DFG_CONJECS DFG_DATE
%token DFG_DESC DFG_DNF DFG_ENDLIST DFG_ENDPROB DFG_EQUAL DFG_EQUIV DFG_EXISTS DFG_FALSE
%token DFG_FORMULA DFG_FORALL DFG_FUNC DFG_HYPOTH DFG_IMPLIED DFG_IMPLIES DFG_KIV DFG_LEM
%token DFG_CLSLIST DFG_DECLLIST DFG_DESCLIST DFG_FORMLIST DFG_GENSET DFG_PRFLIST
%token DFG_SETTINGS DFG_SYMLIST DFG_LOGIC DFG_NAME DFG_NOT DFG_OPERAT DFG_OR DFG_OTTER
%token DFG_PRED DFG_PRDICAT DFG_PROTEIN DFG_QUANTIF DFG_RESOLU DFG_SATIS DFG_SATURATE
%token DFG_SETHEO DFG_SORTS DFG_SPASS DFG_STATUS DFG_SUBSORT DFG_TERMLIST DFG_TRUE
%token DFG_UNKNOWN DFG_UNSATIS DFG_VERSION DFG_3TAP

/* Extension for extended modal logics */
%token DFG_EML DFG_PROP_FORMULA DFG_REL_FORMULA DFG_SPECIALFORMLIST DFG_TRANSLPAIRS 
%token DFG_BOX DFG_DIA DFG_TEST DFG_DOMAIN DFG_RANGE
%token DFG_COMP DFG_SUM DFG_CONV DFG_IDENTITY DFG_DIVERSITY DFG_DOMRESTR DFG_RANRESTR

%token <number> DFG_NUM DFG_MINUS1
%token <string> DFG_ID DFG_TEXT
%type <number>	arity slabel
%type <string>	id labelopt
%type <symbol>  binsymbol nsymbol quantsymbol
%type <term>	aterm atom clause cnfclause cnfclauseopt cnfclausebody lit qterm
%type <term>    term termopt
%type <list>	arglist atermlist atomlist cnfclausesopt formulalistopt labellist
%type <list>    litlist qtermlist sortlist
%type <state>	log_state
%type <bool>    origin

/* Extension for extended modal logics */
%type <symbol>  relpropbinsymbol relnsymbol
%type <term>    propterm proptermopt relterm reltermopt
%type <list>	proparglist relarglist specialformulalistopt

%%

/****************/
/* TOP RULE     */
/****************/

problem:	DFG_BEGPROB '(' id ')' '.'
		description
		logicalpart
		settinglistsopt
		DFG_ENDPROB '.'		{ misc_StringFree($3);
					  YYACCEPT; /* Stop immediately */ }
;

/****************/
/* DESCRIPTION  */
/****************/

description:	DFG_DESCLIST '.'
		name
		author
		versionopt
		logicopt
		status
		desctext
		dateopt
		DFG_ENDLIST '.'
;

name: 		DFG_NAME '(' DFG_TEXT ')' '.'
					{ dfg_DESC.name = $3; }
;

author:		DFG_AUTHOR '(' DFG_TEXT ')' '.'
					{ dfg_DESC.author = $3; }
;

status:		DFG_STATUS '(' log_state ')' '.'
					{ dfg_DESC.status = $3; }
;

desctext:	DFG_DESC '(' DFG_TEXT ')' '.'
					{ dfg_DESC.description = $3; }
;

versionopt:	/* empty */
		| DFG_VERSION '(' DFG_TEXT ')' '.'
					{ dfg_DESC.version = $3; }
;

logicopt:	/* empty */
		| DFG_LOGIC '(' DFG_TEXT ')' '.'
					{ dfg_DESC.logic = $3; }
;

dateopt:	/* empty */
		| DFG_DATE '(' DFG_TEXT ')' '.'
					{ dfg_DESC.date = $3; }
;

log_state:	DFG_SATIS		{ $$ = DFG_SATISFIABLE; }
		| DFG_UNSATIS		{ $$ = DFG_UNSATISFIABLE; }
		| DFG_UNKNOWN		{ $$ = DFG_UNKNOWNSTATE; }
;

/****************/
/* LOGICAL PART */
/****************/

logicalpart:	symbollistopt
		declarationlistopt
		formulalistsopt
		clauselistsopt
		specialformulalistsopt
		prooflistsopt
		listOfTermsopt
;

/****************/
/* SYMBOLS      */
/****************/

symbollistopt:	/* empty */
		| DFG_SYMLIST '.'
		  functionsopt
		  predicatesopt
		  sortsopt
		  operatorsopt
		  quantifiersopt
		  translpairsopt
		  DFG_ENDLIST '.'
;

functionsopt:	/* empty */
		| DFG_FUNC '[' functionlist ']' '.'
;

functionlist:	func
		| functionlist ',' func
;

func:		id			{ dfg_SymbolDecl(DFG_FUNC, $1, -2); }
		| '(' id ',' arity ')'
					{ dfg_SymbolDecl(DFG_FUNC, $2, $4); }
;

predicatesopt:	/* empty */
		| DFG_PRDICAT '[' predicatelist ']' '.'
;

predicatelist:	pred
		| predicatelist ',' pred
;

pred:		id			{ dfg_SymbolDecl(DFG_PRDICAT, $1, -2); }
		| '(' id ',' arity ')'	{ dfg_SymbolDecl(DFG_PRDICAT, $2, $4); }
;

sortsopt:	/* empty */
		| DFG_SORTS '[' sortlist ']' '.'
;

sortlist:	id			{ dfg_SymbolDecl(DFG_PRDICAT, $1, 1); }
		| sortlist ',' id	{ dfg_SymbolDecl(DFG_PRDICAT, $3, 1); }
;

operatorsopt:	/* empty */
		| DFG_OPERAT '[' operatorlist ']' '.'
;

operatorlist:	op
		| operatorlist ',' op
;

op:		id			{ dfg_SymbolDecl(DFG_OPERAT, $1, -2); }
		| '(' id ',' arity ')'	{ dfg_SymbolDecl(DFG_OPERAT, $2, $4); }
;

quantifiersopt:	/* empty */
		| DFG_QUANTIF '[' quantifierlist ']' '.'
;

quantifierlist:	quant
		| quantifierlist ',' quant
;

quant:		id			{ dfg_SymbolDecl(DFG_QUANTIF, $1, -2); }
		| '(' id ',' arity ')'	{ dfg_SymbolDecl(DFG_QUANTIF, $2, $4); }
;

arity:		DFG_MINUS1		{ $$ = -1; }
		| DFG_NUM		{ $$ = $1; }
;

translpairsopt:	/* empty */
		| DFG_TRANSLPAIRS '[' translpairlist ']' '.'
;

translpairlist:	translpair
		| translpairlist ',' translpair
;

translpair:	| '(' id ',' id ')'		{ dfg_TranslPairDecl($2, $4); }
;

/****************/
/* DECLARATIONS */
/****************/

declarationlistopt:	/* empty */
		| DFG_DECLLIST '.' 
		  decllistopt
		  DFG_ENDLIST '.'
;

decllistopt:	/* empty */
		| decllistopt decl
;

decl:		DFG_SUBSORT '(' id ',' id ')' '.'
					{ dfg_SubSort($3,$5); }
		| atom '.'		{ dfg_SORTDECLLIST = list_Nconc(dfg_SORTDECLLIST,list_List(list_PairCreate(NULL,$1))); }
		| DFG_PRED '(' id ',' sortdecl ')' '.'	/* ignored */
					{ misc_StringFree($3); }
		| DFG_FORALL '(' '['	{ dfg_VarStart(); }
		  qtermlist 		{ dfg_VarStop();  }
		  ']' ',' atom ')' '.'	{ TERM term;
					  dfg_VarBacktrack();
					  dfg_VarCheck();
					  term = dfg_CreateQuantifier(fol_All(),$5,$9);
					  dfg_SORTDECLLIST = list_Nconc(dfg_SORTDECLLIST,list_List(list_PairCreate(NULL,term)));
					}
;

sortdecl:	id			{ misc_StringFree($1); }
		| sortdecl ',' id	{ misc_StringFree($3); }
;

/****************/
/* FORMULAS     */
/****************/

formulalistsopt:	/* empty */
		| formulalistsopt formulalist
;

formulalist:	DFG_FORMLIST '(' origin ')' '.'
		formulalistopt
		DFG_ENDLIST '.'	{ if ($3) /* Axioms */
				    dfg_AXIOMLIST = list_Nconc(dfg_AXIOMLIST, $6);
				  else
				    dfg_CONJECLIST = list_Nconc(dfg_CONJECLIST, $6);
				}
;

origin:		DFG_AXIOMS		{ $$ = TRUE;  }
		| DFG_CONJECS		{ $$ = FALSE; }
;

formulalistopt:	/* empty */		{ $$ = list_Nil(); }
		| formulalistopt DFG_FORMULA '(' termopt labelopt ')' '.'
					{ LIST pair;
					  if ($4 == NULL) { /* No term */
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					  dfg_VarCheck();
					}
;

labelopt:	/* empty */		{ $$ = NULL; }
		| ',' id		{ $$ = $2;   }
;

/****************/
/* CLAUSES      */
/****************/

clauselistsopt:	/* empty */
		| clauselistsopt clauselist
;

clauselist:	DFG_CLSLIST '(' origin ',' DFG_CNF ')' '.'
		cnfclausesopt
		DFG_ENDLIST '.'		{ if ($3) /* Axioms */
                                            dfg_AXCLAUSES = list_Nconc(dfg_AXCLAUSES,$8);
                                          else
                                            dfg_CONCLAUSES = list_Nconc(dfg_CONCLAUSES,$8);
					}
		| DFG_CLSLIST '(' origin ',' DFG_DNF ')' '.'
		  dnfclausesopt
		  DFG_ENDLIST '.'	/* ignored */
;

cnfclausesopt:	/* empty */		{ $$ = list_Nil(); }
		| cnfclausesopt DFG_CLAUSE '(' cnfclauseopt labelopt ')' '.'
					{ LIST pair;
					  if ($4 == NULL) { /* No clause */
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					  dfg_VarCheck();
					}
;

cnfclauseopt:	/* empty */		{ $$ = NULL; }
		| cnfclause		{ $$ = $1; }
;

cnfclause:	cnfclausebody		{ $$ = $1; }
		| DFG_FORALL '(' '['	{ dfg_VarStart(); }
		  qtermlist		{ dfg_VarStop();  }
		  ']' ',' cnfclausebody ')'
					{ dfg_VarBacktrack();
					  $$ = dfg_CreateQuantifier(fol_All(),$5,$9);
					}
;

cnfclausebody:	DFG_OR '(' litlist ')'	{ $$ = term_Create(fol_Or(), $3); }
;

litlist:	lit			{ $$ = list_List($1); }
		| litlist ',' lit	{ $$ = list_Nconc($1, list_List($3)); }
;

lit:		atom			{ $$ = $1; }
		| DFG_NOT '(' atom ')'	{ $$ = term_Create(fol_Not(),list_List($3)); }
;

atomlist:	atom			{ $$ = list_List($1); }
		| atomlist ',' atom	{ $$ = list_Nconc($1, list_List($3)); }
;

atom:		id			{ SYMBOL s = dfg_Symbol($1,0);
                                          if (symbol_IsVariable(s) || !symbol_IsPredicate(s)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,s);
					    fputs(" isn't a predicate\n",stderr);
					    misc_Error();
					  }
					  $$ = term_Create(s, list_Nil());
					}
		| DFG_TRUE		{ $$ = term_Create(fol_True(),list_Nil()); }
		| DFG_FALSE		{ $$ = term_Create(fol_False(),list_Nil()); }
		| DFG_EQUAL '(' aterm ',' aterm ')'
					{ $$ = term_Create(fol_Equality(), list_Cons($3,list_List($5))); }
		| id '(' atermlist ')'	{ SYMBOL s;
					  s = dfg_Symbol($1, list_Length($3));
					  if (symbol_IsVariable(s) || !symbol_IsPredicate(s)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,s);
					    fputs(" isn't a predicate\n",stderr);
					    misc_Error();
					  }
					  $$ = term_Create(s, $3);
					}
;

atermlist:	aterm			{ $$ = list_List($1); }
		| atermlist ',' aterm	{ $$ = list_Nconc($1,list_List($3)); }
;

aterm:		id			{ SYMBOL s = dfg_Symbol($1,0);
					  if (!symbol_IsVariable(s) && !symbol_IsConstant(s)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,s);
					    fputs(" isn't a constant\n",stderr);
					    misc_Error();
					  }
					  $$ = term_Create(s, list_Nil());
					}
		| id '(' atermlist ')'	{ SYMBOL s = dfg_Symbol($1,list_Length($3));
					  if (!symbol_IsVariable(s) && !symbol_IsFunction(s)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,s);
					    fputs(" isn't a function\n",stderr);
					    misc_Error();
					  }
					  $$ = term_Create(s, $3);
					}
;


/* Ignore DNF clauses, since SPASS can only handle CNF */

dnfclausesopt:	/* empty */
		| dnfclausesopt DFG_CLAUSE '(' dnfclauseopt labelopt ')' '.'
				{ if ($5 != NULL) /* Delete the label */
				    misc_StringFree($5);
				}
;

dnfclauseopt:	/* empty */
		| dnfclause
;

dnfclause:	dnfclausebody
		| DFG_EXISTS '(' '['	{ dfg_VarStart(); }
		  atomlist		{ dfg_VarStop();  }
		  ']' ',' dnfclausebody ')'
				{ dfg_VarBacktrack();
				  dfg_VarCheck();
				  list_DeleteWithElement($5, term_Delete);
				}
;

dnfclausebody:	DFG_AND '(' litlist ')'
				{ list_DeleteWithElement($3, term_Delete); }
;

/************************/
/* SPECIAL FORMULAE     */
/************************/

specialformulalistsopt:	/* empty */
		| specialformulalistsopt specialformulalist
;

specialformulalist:	DFG_SPECIALFORMLIST '(' origin ',' DFG_EML ')' '.'
			specialformulalistopt
			DFG_ENDLIST '.'	{ if ($3) /* Axioms */
				    	    dfg_AXIOMLIST = list_Nconc(dfg_AXIOMLIST, $8);
				  	  else
				            dfg_CONJECLIST = list_Nconc(dfg_CONJECLIST, $8);
				          flag_SetFlagValue(flag_EML, 1);
					}
;

specialformulalistopt:	/* empty */		{ $$ = list_Nil(); }
			| specialformulalistopt DFG_FORMULA '(' termopt labelopt ')' '.'
			/* First order formula for specifying additonal (frame) properties */
					{ LIST pair;
					  if ($4 == NULL) { /* No term */
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					  dfg_VarCheck();
					}
			| specialformulalistopt DFG_PROP_FORMULA '(' proptermopt labelopt ')' '.'
			/* Formula of extended (propsitional) modal logic */
					{ LIST pair;
					  if ($4 == NULL) { /* No term */
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					}
			| specialformulalistopt DFG_REL_FORMULA '(' reltermopt labelopt ')' '.'
			/* Formula of extended (propsitional) modal logic */
					{ LIST pair;
					  if ($4 == NULL) { /* No term */
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					}
;

/****************/
/* TERMS        */
/****************/

termopt:	/* empty */		{ $$ = NULL; }
		| term			{ $$ = $1; }

arglist:	term			{ $$ = list_List($1); }
		| arglist ',' term	{ $$ = list_Nconc($1, list_List($3)); }
;

term:		id		{ $$ = dfg_TermCreate($1, list_Nil()); }
		| DFG_TRUE	{ $$ = term_Create(fol_True(),list_Nil()); }
		| DFG_FALSE	{ $$ = term_Create(fol_False(),list_Nil()); }
		| DFG_NOT '(' term ')'
				{ $$ = term_Create(fol_Not(),list_List($3)); }
		| binsymbol '(' term ',' term ')'
				{ $$ = term_Create($1, list_Cons($3, list_List($5))); }
		| nsymbol '(' arglist ')'
					{ $$ = term_Create($1, $3); }
		| id '(' arglist ')'
				{ $$ = dfg_TermCreate($1, $3); }
		| quantsymbol '(' '[' 	{ dfg_VarStart(); }
		  qtermlist 		{ dfg_VarStop(); }
		  ']' ',' term ')'
			  	{ dfg_VarBacktrack();
				  $$ = dfg_CreateQuantifier($1,$5,$9);
				}
		| id '(' '['		{ dfg_VarStart(); }
		  qtermlist		{ dfg_VarStop(); }
		  ']' ',' term ')'
				{ fprintf(stderr, "\nLine %d: SPASS can't handle the quantifier %s\n", dfg_LINENUMBER, $1);
				  misc_Error();
				}
		| '(' term ')'  { $$ = $2; }
;

binsymbol:	DFG_EQUAL		{ $$ = fol_Equality(); }
		| DFG_EQUIV		{ $$ = fol_Equiv();    }
		| DFG_IMPLIED		{ $$ = fol_Implied();  }
		| DFG_IMPLIES		{ $$ = fol_Implies();  }
;

nsymbol:	DFG_AND			{ $$ = fol_And(); }
		| DFG_OR		{ $$ = fol_Or();  }
;

quantsymbol:	DFG_EXISTS		{ $$ = fol_Exist(); }
		| DFG_FORALL		{ $$ = fol_All(); }
;

id:		DFG_ID			{ $$ = $1; }
		| DFG_NUM		{ $$ = misc_IntToString($1); }
;

qtermlist:	qterm			{ $$ = list_List($1); }
		| qtermlist ',' qterm	{ $$ = list_Nconc($1, list_List($3)); }
;

qterm:		id			{ SYMBOL s = dfg_Symbol($1,0);
					  if (!symbol_IsVariable(s)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,s);
					    fputs(" isn't a variable\n",stderr);
					    misc_Error();
					  }
					  $$ = term_Create(s, list_Nil());
					}
		| id '(' id ')'		{ SYMBOL p, v;
					  p = dfg_Symbol($1, 1);
					  if (!symbol_IsPredicate(p)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,p);
					    fputs(" isn't a predicate\n",stderr);
					    misc_Error();
					  }
					  v = dfg_Symbol($3, 0);
					  if (!symbol_IsVariable(v)) {
					    fprintf(stderr, "\nLine %d: ",dfg_LINENUMBER);
					    symbol_FPrint(stderr,v);
					    fputs(" isn't a variable\n",stderr);
					    misc_Error();
					  }
					  $$ = term_Create(p, list_List(term_Create(v,list_Nil())));
					}
;

/* TERMS for extended (propsitional) modal logics */

proptermopt:	/* empty */	{ $$ = NULL; }
		| propterm	{ $$ = $1; }
;

reltermopt:	/* empty */	{ $$ = NULL; }
		| relterm	{ $$ = term_Create(eml_Box(), 
					list_Cons(term_Create(fol_Not(), list_List($1)), 
					  list_List(term_Create(fol_False(), list_Nil())))); 
					/* since r = 1 iff (r':1)' = 1 */
                                }
;

proparglist:	propterm			{ $$ = list_List($1); }
		| proparglist ',' propterm	{ $$ = list_Nconc($1, list_List($3)); }
;

relarglist:	relterm				{ $$ = list_List($1); }
		| relarglist ',' relterm	{ $$ = list_Nconc($1, list_List($3)); }
;

propterm:	id		{ $$ = dfg_TermCreate($1, list_Nil()); }
		| DFG_TRUE	{ $$ = term_Create(fol_True(),list_Nil()); }
		| DFG_FALSE	{ $$ = term_Create(fol_False(),list_Nil()); }
		| DFG_NOT '(' propterm ')'
				{ $$ = term_Create(fol_Not(),list_List($3)); }
		| DFG_DOMAIN '(' relterm ')'
				{ $$ = term_Create(eml_Domain(),list_List($3)); }
		| DFG_RANGE '(' relterm ')'
				{ $$ = term_Create(eml_Range(),list_List($3)); }
		| binsymbol '(' propterm ',' propterm ')'
				{ $$ = term_Create($1, list_Cons($3, list_List($5))); }
		| nsymbol '(' proparglist ')'
					{ $$ = term_Create($1, $3); }
		| relpropbinsymbol '(' relterm ',' propterm ')'
				{ $$ = term_Create($1, list_Cons($3, list_List($5))); }
;

relterm:	id		{ $$ = dfg_TermCreate($1, list_Nil()); }
		| DFG_TRUE	{ $$ = term_Create(fol_True(),list_Nil()); }
		| DFG_FALSE	{ $$ = term_Create(fol_False(),list_Nil()); }
		| DFG_NOT '(' relterm ')'
				{ $$ = term_Create(fol_Not(),list_List($3)); }
		| DFG_IDENTITY	{ $$ = term_Create(eml_Id(),list_Nil()); }
		| DFG_DIVERSITY	{ $$ = term_Create(eml_Div(),list_Nil()); }
		| DFG_CONV '(' relterm ')'
				{ $$ = term_Create(eml_Conv(),list_List($3)); }
		| DFG_TEST '(' propterm ')'
				{ $$ = term_Create(eml_Test(),list_List($3)); }
		| DFG_DOMRESTR '(' relterm ',' propterm ')'
				{ $$ = term_Create(eml_DomRestr(), list_Cons($3, list_List($5))); }
		| DFG_RANRESTR '(' relterm ',' propterm ')'
				{ $$ = term_Create(eml_RanRestr(), list_Cons($3, list_List($5))); }
		| binsymbol '(' relterm ',' relterm ')'
				{ $$ = term_Create($1, list_Cons($3, list_List($5))); }
		| relnsymbol '(' relarglist ')'
					{ $$ = term_Create($1, $3); }
;

relpropbinsymbol:	DFG_BOX		{ $$ = eml_Box();  }
			| DFG_DIA	{ $$ = eml_Dia();  }
;

relnsymbol:	DFG_AND			{ $$ = fol_And();  }
		| DFG_OR		{ $$ = fol_Or();   }
		| DFG_COMP		{ $$ = eml_Comp(); }
		| DFG_SUM		{ $$ = eml_Sum();  }
;

/****************/
/* PROOFS       */
/****************/

prooflistsopt:	/* empty */
		| prooflistsopt prooflist
;

prooflist:	DFG_PRFLIST '(' DFG_RESOLU ')' '.'
		prooflistopt
		DFG_ENDLIST '.'
;

prooflistopt:	/* empty */
		| prooflistopt DFG_CLAUSE '(' clause ',' id ',' id ',' '[' labellist ']' ',' DFG_NUM ')' '.'
				{ LIST tupel;
				  RULE Rule = clause_GetOriginFromString($8);
                                  misc_StringFree($8);
				  if ($4 != NULL) {
				    /* CNF clause */
				    /* Build a tupel (label,clause,labellist,split level,origin) */
				    tupel = list_Cons((POINTER)$14,list_List((POINTER)Rule));
                                    tupel = list_Cons($6,list_Cons($4,list_Cons($11,tupel)));
				    dfg_PROOFLIST = list_Cons(tupel, dfg_PROOFLIST);
				  } else {
				    /* ignore DNF clause */
				    misc_StringFree($6);
				    list_Apply(misc_StringFree, $11);
				  }
				  dfg_VarCheck();
				}		
;

clause:		cnfclause		{ $$ = $1;   }
		| dnfclause		{ $$ = NULL; }
;

labellist:	id			{ $$ = list_List($1); }
		| labellist ',' id	{ $$ = list_Nconc($1, list_List($3)); }
;

/****************/
/* TERM LIST    */
/****************/

listOfTermsopt: /* empty */
                | listOfTermsopt listOfTerms
;

listOfTerms:    DFG_TERMLIST '.'        { dfg_VarStart(); }
                termlist DFG_ENDLIST '.' {
                                          dfg_VarStop();
                                          dfg_VarBacktrack();
                                          dfg_VarCheck(); }
;

termlist:       /* empty */
                | termlist term '.'      { dfg_TERMLIST = list_Nconc(dfg_TERMLIST, list_List($2)); }
; 

/****************/
/* SETTINGS     */
/****************/

settinglistsopt:	/* empty */
		| settinglistsopt settinglist
;

settinglist:	DFG_GENSET '.' gsettings DFG_ENDLIST '.'
		| DFG_SETTINGS '(' slabel ')' '.' DFG_TEXT DFG_ENDLIST '.' 
				{ if ($3 == DFG_SPASS)
				    dfg_Settings($6);/* analyze the text */
				  misc_StringFree($6);
				}
;

gsettings:	gsetting
		| gsettings gsetting
;

gsetting:	DFG_HYPOTH '[' labellist ']' '.'
				{ list_DeleteWithElement($3, misc_StringFree); }
;

slabel:		DFG_KIV			{ $$ = DFG_KIV;      }
		| DFG_LEM		{ $$ = DFG_LEM;      }
		| DFG_OTTER		{ $$ = DFG_OTTER;    }
		| DFG_PROTEIN		{ $$ = DFG_PROTEIN;  }
		| DFG_SATURATE		{ $$ = DFG_SATURATE; }
		| DFG_3TAP		{ $$ = DFG_3TAP;     }
		| DFG_SETHEO		{ $$ = DFG_SETHEO;   }
		| DFG_SPASS		{ $$ = DFG_SPASS;    }
;

%%

void yyerror(const char *s)
{
  fprintf(stderr,"line %i: %s\n", dfg_LINENUMBER, s);
  misc_Error();

}

static void dfg_Init(FILE* Input)
/**************************************************************
  INPUT:   The input file stream for the parser.
  RETURNS: Nothing.
  EFFECT:  The parser and scanner are initialized.
***************************************************************/
{
  extern FILE* dfg_in;  /* declared in dfgscanner */

  dfg_in               = Input;
  dfg_LINENUMBER       = 1;
  dfg_AXIOMLIST        = list_Nil();
  dfg_CONJECLIST       = list_Nil();
  dfg_SORTDECLLIST     = list_Nil();
  dfg_AXCLAUSES        = list_Nil();
  dfg_CONCLAUSES       = list_Nil();
  dfg_PROOFLIST        = list_Nil();
  dfg_TERMLIST         = list_Nil();
  dfg_SYMBOLLIST       = list_Nil();
  dfg_VARLIST          = list_Nil();
  dfg_VARDECL          = FALSE;
  dfg_DESC.name        = (char*) NULL;
  dfg_DESC.author      = (char*) NULL;
  dfg_DESC.version     = (char*) NULL;
  dfg_DESC.logic       = (char*) NULL;
  dfg_DESC.status      = DFG_UNKNOWNSTATE;
  dfg_DESC.description = (char*) NULL;
  dfg_DESC.date        = (char*) NULL;
}


void dfg_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Frees memory used by the problem description.
***************************************************************/
{
  if (dfg_DESC.name != NULL)
    misc_StringFree(dfg_DESC.name);
  if (dfg_DESC.author != NULL)
    misc_StringFree(dfg_DESC.author);
  if (dfg_DESC.version != NULL)
    misc_StringFree(dfg_DESC.version);
  if (dfg_DESC.logic != NULL)
    misc_StringFree(dfg_DESC.logic);
  if (dfg_DESC.description != NULL)
    misc_StringFree(dfg_DESC.description);
  if(dfg_DESC.date != NULL)
    misc_StringFree(dfg_DESC.date);
}

const char* dfg_ProblemName(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's name from the description section.
***************************************************************/
{
  return dfg_DESC.name;
}

const char* dfg_ProblemAuthor(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's author from the description section.
***************************************************************/
{
  return dfg_DESC.author;
}

const char* dfg_ProblemVersion(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's version from the description section.
***************************************************************/
{
  return dfg_DESC.version;
}

const char* dfg_ProblemLogic(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's logic from the description section.
***************************************************************/
{
  return dfg_DESC.logic;
}

DFG_STATE dfg_ProblemStatus(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's status from the description section.
***************************************************************/
{
  return dfg_DESC.status;
}

const char* dfg_ProblemStatusString(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The string representation of the problem's status.
***************************************************************/
{
  char* result = "";

  switch (dfg_DESC.status) {
  case DFG_SATISFIABLE:
    result = "satisfiable"; break;
  case DFG_UNSATISFIABLE:
    result = "unsatisfiable"; break;
  case DFG_UNKNOWNSTATE:
    result = "unknown"; break;
  default:
    fprintf(stderr, "Invalid status in function dfg_StatusString\n\n");
    misc_Error();
  }
  return result;
}

const char* dfg_ProblemDescription(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's description from the decsription section.
***************************************************************/
{
  return dfg_DESC.description;
}

const char* dfg_ProblemDate(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's date from the description section.
***************************************************************/
{
  return dfg_DESC.date;
}

void dfg_FPrintDescription(FILE* File)
/**************************************************************
  INPUT:   A file stream.
  RETURNS: Nothing.
  EFFECT:  The description section from the input file
           is printed to 'File'. You must call the parser first
           before calling this function.
***************************************************************/
{
  fputs("list_of_descriptions.\n  name(", File);
  if (dfg_DESC.name != NULL)
    fputs(dfg_DESC.name, File);
  else
    fputs("{* *}", File);
  fputs(").\n  author(", File);
  if (dfg_DESC.author != NULL)
    fputs(dfg_DESC.author, File);
  else
    fputs("{* *}", File);
  fputs(").\n", File);
  if (dfg_DESC.version != NULL) {
    fputs("  version(", File);
    fputs(dfg_DESC.version, File);
    fputs(").\n", File);
  }
  if (dfg_DESC.logic != NULL) {
    fputs("  logic(", File);
    fputs(dfg_DESC.logic, File);
    fputs(").\n", File);
  }
  fputs("  status(", File);
  fputs(dfg_ProblemStatusString(), File);
  fputs(").\n  description(", File);
  if (dfg_DESC.description != NULL)
    fputs(dfg_DESC.description, File);
  else
    fputs("{* *}", File);
  fputs(").\n", File);
  if (dfg_DESC.date != NULL) {
    fputs("  date(", File);
    fputs(dfg_DESC.date, File);
    fputs(").\n", File);
  }
  fputs("end_of_list.", File);
}


LIST dfg_DFGParser(FILE* File, LIST* Axioms, LIST* Conjectures, LIST* SortDecl)
/**************************************************************
  INPUT:   The input file containing clauses or formulae in DFG syntax.
           Axioms, Conjectures and SortDecl are pointers to lists used
           as return values.
  RETURNS: The list of clauses from File.
  EFFECT:  Reads formulae and clauses from the input file.
           The axioms, conjectures and sort declarations are appended
	   to the respective lists, the lists are not deleted !
	   All lists except the clauselist contain pairs
	   (label, term), where <label> may be NULL, if no
	   label was specified for that term.
***************************************************************/
{
  LIST  scan, tupel;
  TERM  clauseTerm;

  dfg_Init(File);      /* Initialize the parser and scanner */
  dfg_parse();          /* Invoke the parser */

  dfg_SymCleanUp();

  if (!flag_GetFlagValue(flag_INTERACTIVE) &&
      list_Empty(dfg_AXIOMLIST) && list_Empty(dfg_CONJECLIST) &&
      list_Empty(dfg_AXCLAUSES) && list_Empty(dfg_CONCLAUSES)) {
    fputs("No formulae and clauses found in input file!\n\n", stderr);
    misc_Error();
  }

  /* Remove clause labels and create clauses from the terms */
  for (scan = dfg_AXCLAUSES; !list_Empty(scan); scan = list_Cdr(scan)) {
    tupel = list_Car(scan);
    clauseTerm = list_PairSecond(tupel);
    list_Rplaca(scan, dfg_CreateClauseFromTerm(clauseTerm,TRUE));
    if (list_PairFirst(tupel) != NULL)        /* Label is defined */
      misc_StringFree(list_PairFirst(tupel));  /* Delete the label */
    list_PairFree(tupel);
  }
  for (scan = dfg_CONCLAUSES; !list_Empty(scan); scan = list_Cdr(scan)) {
    tupel = list_Car(scan);
    clauseTerm = list_PairSecond(tupel);
    list_Rplaca(scan, dfg_CreateClauseFromTerm(clauseTerm,FALSE));
    if (list_PairFirst(tupel) != NULL)        /* Label is defined */
      misc_StringFree(list_PairFirst(tupel));  /* Delete the label */
    list_PairFree(tupel);
  }

  /* Delete the proof list */
  dfg_DeleteProofList(dfg_PROOFLIST);

  /* Delete the list_of_terms, since it'll be ignored */
  list_DeleteWithElement(dfg_TERMLIST, term_Delete);

  scan = list_Nconc(dfg_AXCLAUSES, dfg_CONCLAUSES);

  *Axioms      = list_Nconc(*Axioms, dfg_AXIOMLIST);
  *Conjectures = list_Nconc(*Conjectures, dfg_CONJECLIST);
  *SortDecl    = list_Nconc(*SortDecl, dfg_SORTDECLLIST);

  return scan;
}


LIST dfg_ProofParser(FILE* File)
/**************************************************************
  INPUT:   The input file containing clauses in DFG syntax.
  RETURNS: A list of tupels (label,clause,justificationlist,splitlevel,origin)
           representing a proof.
  EFFECT:  Reads inputs clauses with labels and the proof lists
           from the input file.
           The elements of the list are lists with five items.
	   1. the label (a string) of a clause,
	   2. the clause in TERM format,
           3. the list of justification labels,
           4. the split level of the clause,
           5. the origin of the clause (RULE struct from clause.h).
	   Note that the justification list is empty for input
	   clauses.
***************************************************************/
{
  LIST  scan, tupel;
  TERM  term;

  dfg_Init(File);       /* Initialize the parser and scanner */
  dfg_parse();          /* Invoke the parser */

  dfg_SymCleanUp();

  /* Build the union of axiom and conjecture clauses */
  dfg_AXCLAUSES  = list_Nconc(dfg_AXCLAUSES, dfg_CONCLAUSES);
  dfg_CONCLAUSES = list_Nil();
  for (scan = dfg_AXCLAUSES; !list_Empty(scan); scan = list_Cdr(scan)) {
    tupel = list_Car(scan);
    term  = list_PairSecond(tupel);
    if (list_PairFirst(tupel) == NULL) {
      /* Ignore input clauses without label */
      term_Delete(term);
      list_PairFree(tupel);
      list_Rplaca(scan, NULL);
    } else
      /* Expand the pair to a tupel                            */
      /* (label,clause,justificationlist, split level, origin) */
      /* For input clauses the justificationlist is empty.     */
      /* Input clauses have split level 0.                     */
      list_Rplacd(tupel, list_Cons(term,list_Cons(list_Nil(),list_Cons(0, list_List((POINTER)INPUT)))));
  }
  /* Now delete the list items without labels */
  dfg_AXCLAUSES = list_PointerDeleteElement(dfg_AXCLAUSES, NULL);

  /* Delete the formula lists */
  dfg_DeleteFormulaPairList(dfg_AXIOMLIST);
  dfg_DeleteFormulaPairList(dfg_CONJECLIST);
  /* Delete the list of sort declarations */
  dfg_DeleteFormulaPairList(dfg_SORTDECLLIST);
  /* Delete the list_of_terms, since it'll be ignored */
  list_DeleteWithElement(dfg_TERMLIST, term_Delete);

  /* Finally append the proof list to the list of input clauses with labels */
  dfg_PROOFLIST = list_NReverse(dfg_PROOFLIST);
  dfg_AXCLAUSES = list_Nconc(dfg_AXCLAUSES, dfg_PROOFLIST);

  return dfg_AXCLAUSES;
}


LIST dfg_TermParser(FILE* File)
/**************************************************************
  INPUT:   The input file containing a list of terms in DFG syntax.
  RETURNS: The list of terms from <File>.
  EFFECT:  Reads terms from the list_of_terms from the input file.
***************************************************************/
{
  dfg_Init(File);       /* Initialize the parser and scanner */
  dfg_parse();          /* Invoke the parser */

  dfg_SymCleanUp();

  /* Delete the clause lists */
  dfg_DeleteFormulaPairList(dfg_AXCLAUSES);
  dfg_DeleteFormulaPairList(dfg_CONCLAUSES);
  /* Delete the formula lists */
  dfg_DeleteFormulaPairList(dfg_AXIOMLIST);
  dfg_DeleteFormulaPairList(dfg_CONJECLIST);
  /* Delete the proof list */
  dfg_DeleteProofList(dfg_PROOFLIST);
  /* Delete the list of sort declarations */
  dfg_DeleteFormulaPairList(dfg_SORTDECLLIST);

  return dfg_TERMLIST;
}


void dfg_DeleteFormulaPairList(LIST FormulaPairs)
/**************************************************************
  INPUT:   A list of pairs (label, formula).
  RETURNS: Nothing.
  EFFECT:  The list and the pairs with their strings and terms
           are completely deleted.
***************************************************************/
{
  LIST pair;

  for ( ; !list_Empty(FormulaPairs); FormulaPairs = list_Pop(FormulaPairs)) {
    pair = list_Car(FormulaPairs);  /* (label, term) */
    term_Delete(list_PairSecond(pair));
    if (list_PairFirst(pair) != NULL)
      misc_StringFree(list_PairFirst(pair));  /* Free the label */
    list_PairFree(pair);
  }
}

void dfg_StripLabelsFromList(LIST FormulaPairs)
/**************************************************************
  INPUT:   A list of pairs (label, formula).
  RETURNS: Nothing.
  EFFECT:  The pairs are replaced by the respective formula
           and the pairs with their label strings are deleted.
***************************************************************/
{
  LIST pair, scan;

  for (scan = FormulaPairs; !list_Empty(scan); scan = list_Cdr(scan)) {
    pair = list_Car(scan);  /* (label, term) */
    list_Rplaca(scan, list_PairSecond(pair));
    if (list_PairFirst(pair) != NULL)
      misc_StringFree(list_PairFirst(pair));  /* Free the label */
    list_PairFree(pair);
  }
}

void dfg_DeleteProofList(LIST Proof)
/**************************************************************
  INPUT:   A list of tupels (label, term, justificationlist, split level).
  RETURNS: Nothing.
  EFFECT:  All memory used by the proof list is freed.
           The labels must NOT be NULL entries!
***************************************************************/
{
  /* Delete the proof list */
  for ( ; !list_Empty(Proof); Proof = list_Pop(Proof)) {
    LIST tupel = list_Car(Proof);
    misc_StringFree(list_First(tupel));
    term_Delete(list_Second(tupel));
    list_Apply(misc_StringFree, list_Third(tupel));
    list_Delete(list_Third(tupel));
    list_Delete(tupel);
  }
}

/**************************************************************/
/* Static Functions                                           */
/**************************************************************/

static void dfg_SymbolDecl(int SymbolType, char* Name, int Arity)
/**************************************************************
  INPUT:   The type of a symbol, the name, and the arity.
  RETURNS: Nothing.
  EFFECT:  This function handles the declaration of symbols.
           If <Arity> is -2, it means that the arity of the symbol
	   was not specified, if it is -1 the symbol is declared
	   with arbitrary arity.
	   The <Name> is deleted.
***************************************************************/
{
  NAT    arity, length;
  SYMBOL symbol;

  if (Arity == -2)  /* not specified */
    arity = 0;
  else
    arity = Arity;

  /* Pay attention to the maximum symbol name length */
  length = strlen(Name);
  if (length >= symbol__SYMBOLMAXLEN)
    Name[symbol__SYMBOLMAXLEN-1] = '\0';

  /* Check if this symbol was declared earlier */
  symbol = symbol_Lookup(Name);
  if (symbol != 0) {
    /* Symbol was declared before */
    /* Check if the old and new symbol type are equal */
    if ((SymbolType == DFG_FUNC && !symbol_IsFunction(symbol)) ||
	(SymbolType == DFG_PRDICAT && !symbol_IsPredicate(symbol)) ||
	((SymbolType == DFG_OPERAT || SymbolType == DFG_QUANTIF) && 
	 !symbol_IsJunctor(symbol))) {
      fprintf(stderr, "Line %u: symbol %s was already declared as ",
	      dfg_LINENUMBER, Name);
      switch (symbol_Type(symbol)) {
      case symbol_CONSTANT:
      case symbol_FUNCTION:
	fputs("function\n", stderr); break;
      case symbol_PREDICATE:
	fputs("predicate\n", stderr); break;
      case symbol_JUNCTOR:
	fputs("junctor", stderr); break;
      default:
	fputs("unknown type\n", stderr);
      }
      misc_Error();
    }
    /* Now check the old and new arity if specified */
    if (Arity != -2 && Arity != symbol_Arity(symbol)) {
      fprintf(stderr, "Line %u: symbol %s was already declared with arity %d\n",
	      dfg_LINENUMBER, Name, symbol_Arity(symbol));
      misc_Error();
    }
  } else {
    /* Symbol was not declared before */
    switch (SymbolType) {
    case DFG_FUNC:
      symbol = symbol_CreateFunction(Name, arity, symbol_STATLEX);
      break;
    case DFG_PRDICAT:
      symbol = symbol_CreatePredicate(Name, arity, symbol_STATLEX);
      break;
    default:
      symbol = symbol_CreateJunctor(Name, arity, symbol_STATLEX);
    }
    if (Arity == -2)
      /* Arity wasn't specified so check the arity for each occurrence */
      dfg_SymAdd(symbol);
  }

  if (length >= symbol__SYMBOLMAXLEN) {
    /* To avoid a memory error restore the old string length */
    Name[symbol__SYMBOLMAXLEN-1] = ' ';  /* Something != '\0' */
  }
  misc_StringFree(Name);  /* Name was copied */
}


static SYMBOL dfg_Symbol(char* Name, NAT Arity)
/**************************************************************
  INPUT:   The name of a symbol and the actual arity of the symbol.
  RETURNS: The corresponding SYMBOL.
  EFFECT:  This function checks if the <Name> was declared as
           symbol or variable. If not, an error message is printed
	   to stderr.
	   The <Name> is deleted.
***************************************************************/
{
  SYMBOL symbol;
  char   old;
  NAT    length;

  old = ' ';  /* Just to avoid a compiler warning */
  /* Pay attention to the maximum symbol name length */
  length = strlen(Name);
  if (length >= symbol__SYMBOLMAXLEN) {
    old = Name[symbol__SYMBOLMAXLEN-1];
    Name[symbol__SYMBOLMAXLEN-1] = '\0';
  }

  symbol = symbol_Lookup(Name);
  if (length >= symbol__SYMBOLMAXLEN) {
    /* To avoid a memory error restore the old string */
    Name[symbol__SYMBOLMAXLEN-1] = old;
  }
  if (symbol != 0) {
    misc_StringFree(Name);
    dfg_SymCheck(symbol, Arity); /* Check the arity */
  } else {
    /* Variable */
    if (Arity > 0) {
      fprintf(stderr, "\nLine %d: Undefined symbol %s\n",dfg_LINENUMBER,Name);
      misc_Error();
    }
    symbol = dfg_VarLookup(Name);
  }
  return symbol;
}


static TERM dfg_CreateQuantifier(SYMBOL Symbol, LIST VarTermList, TERM Term)
/**************************************************************
  INPUT:   A quantifier symbol, a list possibly containing sorts,
           and a term.
  RETURNS: The created quantifier term.
***************************************************************/
{
  LIST varlist, sortlist, scan;
  TERM helpterm;

  /* First collect the variable symbols in varlist and the sorts in sortlist */
  varlist = sortlist = list_Nil();
  for ( ; !list_Empty(VarTermList); VarTermList = list_Pop(VarTermList)) {
    helpterm = list_Car(VarTermList);
    if (term_IsVariable(helpterm)) {
      varlist = list_Nconc(varlist, list_List((POINTER)term_TopSymbol(helpterm)));
      term_Delete(helpterm);
    } else {
      SYMBOL var = term_TopSymbol(term_FirstArgument(helpterm));
      varlist  = list_Nconc(varlist, list_List((POINTER)var));
      sortlist = list_Nconc(sortlist, list_List(helpterm));
    }
  }

  varlist = list_PointerDeleteDuplicates(varlist);
  /* Now create terms from the variables */
  for (scan = varlist; !list_Empty(scan); scan = list_Cdr(scan))
    list_Rplaca(scan, term_Create((SYMBOL)list_Car(scan), list_Nil()));

  if (!list_Empty(sortlist)) {
    if (symbol_Equal(fol_All(), Symbol)) {
      /* The conjunction of all sortterms implies the Term */
      if (symbol_Equal(fol_Or(), term_TopSymbol(Term))) {
	/* Special treatment if <Term> is a term with "or" like */
	/* in clauses: add all sort terms negated to the args    */
	/* of the "or" */
	for (scan = sortlist; !list_Empty(scan); scan = list_Cdr(scan))
	  /* Negate the sort terms */
	  list_Rplaca(scan, term_Create(fol_Not(), list_List(list_Car(scan))));
	sortlist = list_Nconc(sortlist, term_ArgumentList(Term));
	term_RplacArgumentList(Term, sortlist);
      } else {
	/* No "or" term, so build the implication term */
	if (list_Empty(list_Cdr(sortlist))) {
	  /* Only one sort term */
	  list_Rplacd(sortlist, list_List(Term));
	  Term = term_Create(fol_Implies(), sortlist);
	} else {
	  /* More than one sort term */
	  helpterm = term_Create(fol_And(), sortlist);
	  Term = term_Create(fol_Implies(), list_Cons(helpterm, list_List(Term)));
	}
      }
    } else if (symbol_Equal(fol_Exist(), Symbol)) {
      /* Quantify the conjunction of all sort terms and <Term> */
      if (symbol_Equal(fol_And(), term_TopSymbol(Term))) {
	/* Special treatment if <Term> has an "and" as top symbol: */
	/* just add the sort terms to the args of the "and".       */
	sortlist = list_Nconc(sortlist, term_ArgumentList(Term));
	term_RplacArgumentList(Term, sortlist);
      } else {
	sortlist = list_Nconc(sortlist, list_List(Term));
	Term = term_Create(fol_And(), sortlist);
      }
    }
  }
  helpterm = fol_CreateQuantifier(Symbol, varlist, list_List(Term));
  return helpterm;
}


CLAUSE dfg_CreateClauseFromTerm(TERM Clause, BOOL IsAxiom)
/**************************************************************
  INPUT:   A clause term and a boolean value.
  RETURNS: The clause term converted to a CLAUSE.
  EFFECT:  This function converts a clause stored as term
           into an EARL clause structure.
	   If 'IsAxiom' is TRUE the clause is treated as axiom
	   clause else as conjecture clause.
  MEMORY:  The clause term is deleted.
***************************************************************/
{
  LIST   literals, scan;
  TERM   literal;
  CLAUSE result;
  
  if (term_TopSymbol(Clause) == fol_All()) {
    /* Remove and free the quantifier and the OR term */
    literals = term_ArgumentList(term_SecondArgument(Clause));
    term_RplacArgumentList(term_SecondArgument(Clause), list_Nil());
  } else {
    /* Remove and free the OR term */
    literals = term_ArgumentList(Clause);
    term_RplacArgumentList(Clause, list_Nil());
  }
  term_Delete(Clause);

  for (scan = literals; !list_Empty(scan); scan = list_Cdr(scan)) {
    literal = (TERM) list_Car(scan);
    if (symbol_IsPredicate(term_TopSymbol(literal))) {  /* Positive literal */
      if (fol_IsTrue(literal)) {
	/* Clause is a tautology */
	list_PointerDeleteElement(literals, NULL);
	/* Remove possible NULL elements to avoid a crash in term_Delete */
	list_DeleteWithElement(literals, term_Delete);
	return (CLAUSE) NULL;
      } else if (fol_IsFalse(literal)) {
	/* Ignore this literal */
	term_Delete(literal);
	list_Rplaca(scan, NULL); /* Mark the actual list element */
      }
    } else {
      /* Found a negative literal */
      TERM atom = term_FirstArgument(literal);
      if (fol_IsFalse(atom)) {
	/* Clause is a tautology */
	list_PointerDeleteElement(literals, NULL);
	/* Remove possible NULL elements to avoid a crash in term_Delete */
	list_DeleteWithElement(literals, term_Delete);
	return (CLAUSE) NULL;
      } else if (fol_IsTrue(atom)) {
	/* Ignore this literal */
	term_Delete(literal);
	list_Rplaca(literals, NULL); /* Mark the actual list element */
      }
    }
  }

  literals = list_PointerDeleteElement(literals, NULL);
  /* Remove the special literals treated above from the list */
  result = clause_CreateFromLiterals(literals, FALSE, !IsAxiom, FALSE);
  /* Don't create sorts! */
  list_Delete(literals);

  return result;
}


static void dfg_SubSort(char* Name1, char* Name2)
/**************************************************************
  INPUT:   Two sort symbol names.
  RETURNS: Nothing.
  EFFECT:  This functions adds the formula
           forall([U], implies(Name1(U), Name2(U)))
	   to the list of axiom formulas. Both <Name1> and <Name2>
	   are deleted.
***************************************************************/
{
  SYMBOL s1, s2;
  TERM   varterm, t1, t2, term;

  s1 = dfg_Symbol(Name1, 1);   /* Should be unary predicates */
  s2 = dfg_Symbol(Name2, 1);
  if (!symbol_IsPredicate(s1)) {
    fprintf(stderr, "\nLine %d: ", dfg_LINENUMBER);
    symbol_FPrint(stderr, s1);
    fputs(" isn't a sort predicate\n", stderr);
    misc_Error();
  }
  if (!symbol_IsPredicate(s2)) {
    fprintf(stderr, "\nLine %d: ", dfg_LINENUMBER);
    symbol_FPrint(stderr, s2);
    fputs(" isn't a sort predicate\n", stderr);
    misc_Error();
  }

  varterm = term_Create(symbol_CreateStandardVariable(), list_Nil());
  symbol_ResetStandardVarCounter();
  
  t1   = term_Create(s1, list_List(varterm));
  t2   = term_Create(s2, list_List(term_Copy(varterm)));
  term = term_Create(fol_Implies(), list_Cons(t1, list_List(t2)));
  term = fol_CreateQuantifier(fol_All(), list_List(term_Copy(varterm)),
			      list_List(term));
  dfg_SORTDECLLIST = list_Nconc(dfg_SORTDECLLIST, list_List(list_PairCreate(NULL,term)));
}


static char* dfg_SettingsScanner(char* Text, int* Token, char** Id)
/**************************************************************
  INPUT:   A string, a pointer to a token, and a pointer to a string.
  RETURNS: A pointer to a character.
  EFFECT:  This function scans <Text> from the beginning. If it finds
           the first token, its sets <Token> to the value of this
	   token and returns the text position just after the end
	   of the token. Token may be set to the ASCII value of
	   a single character or to DFG_ID, if a string containing
	   alphanumerical characters was found. If <Token> is set
	   to DFG_ID, <Id> is set to the string representation of
	   the identifier, else <Id> is set to NULL.
	   Note that positive numbers are also treated as identifiers.
	   If the string "-1" is found <Token> is set to -1.
***************************************************************/
{
  char* scan;
  BOOL  loop;

  *Id  = NULL;
  scan = Text;
  loop = TRUE;
  do {
    while (isspace((int)*scan))
      scan++;
    if (*scan == '%')  /* one-line comment */
      while (*scan != '\n')
	scan++;
    else
      loop = FALSE;
  } while (loop);

  if (*scan == '-' && *(scan+1) == '1') {
    *Token = -1;
    scan += 2;
  } else if (!isalnum((int)*scan) && *scan != '_')  /* A single character */
    *Token = *scan++;
  else {
    /* Found identifier */
    char help, *idstart;

    idstart = scan;
    while (isalnum((int)*scan) || *scan == '_')
      scan++;

    help = *scan;
    *scan = '\0'; /* Mark the end of the identifier */
    *Id = memory_Malloc(sizeof(char)*(scan-idstart+1));
    strcpy(*Id, idstart);
    *scan = help; /* Restore old string */
    *Token = DFG_ID;
  }
  return scan;
}


static void dfg_Match(int RequiredToken, int ActToken)
/**************************************************************
  INPUT:   Two tokens.
  RETURNS: Nothing.
  EFFECT:  This function checks if <RequiredToken> and <ActToken>
           match. If they don't match an error message is printed
	   to stderr and the program exits.
***************************************************************/
{
  if (RequiredToken != ActToken) {
    fputs("\nParse error in settings: ", stderr);
    if (RequiredToken == DFG_ID)
      fputs("Identifier", stderr);
    else
      putc(RequiredToken, stderr);
    fputs(" expected, ", stderr);
    if (ActToken == DFG_ID)
      fputs("Identifier", stderr);
    else if (ActToken == '\0')
      fputs("unexpected end", stderr);
    else
      putc(ActToken, stderr);
    fputs(" found\n", stderr);
    misc_Error();
  }
}


static void dfg_Settings(char* Text)
/**************************************************************
  INPUT:   A string starting with "{*" and ending with "*}".
  RETURNS: Nothing.
  EFFECT:  The string must contain settings for SPASS. This
           function parses the string and sets the precedence
	   of symbols or some flags. If a syntax error occurs
	   am error message is written to stderr and the program
	   exits.
***************************************************************/
{
  int    token;
  char*  id;
  int    number;
  SYMBOL symbol;

  Text += 2;                   /* Skip the "{*" at the beginning */
  Text[strlen(Text)-2] = '\0'; /* Ignore the "*}" at the end of <Text> */

  Text = dfg_SettingsScanner(Text, &token, &id);
  while (token != '\0') {
    dfg_Match(DFG_ID, token);   /* Error, if token is not an identifier */
    if (strcmp(id, "set_precedence") == 0) {
      misc_StringFree(id);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('(', token);
      do {
        Text = dfg_SettingsScanner(Text, &token, &id);
        if (token == '(') { /* Specified with weight */
          Text = dfg_SettingsScanner(Text, &token, &id);
          dfg_Match(DFG_ID, token);
	  symbol = symbol_Lookup(id);
	  if (symbol == 0) {
	    fprintf(stderr, "\nUndefined symbol %s in precedence list\n", id);
	    misc_Error();
	  }
	  misc_StringFree(id);
          Text = dfg_SettingsScanner(Text, &token, &id);
          dfg_Match(',', token);
          Text = dfg_SettingsScanner(Text, &token, &id);
          dfg_Match(DFG_ID, token);  /* A number */
	  number = misc_StringToInt(id);
	  symbol_SetWeight(symbol, number);
	  misc_StringFree(id);
          Text = dfg_SettingsScanner(Text, &token, &id);
          if (token == ',') { /* Specified with symbol status */
            Text = dfg_SettingsScanner(Text, &token, &id);
            dfg_Match(DFG_ID, token);
	    if (strcmp(id, "r") == 0)
	      /* Symbol status ORDRIGHT */
	      symbol_AddProperty(symbol, ORDRIGHT);
	    else if (strcmp(id, "m") == 0)
	      /* Symbol status ORDMUL */
	      symbol_AddProperty(symbol, ORDMUL);
	    else if (strcmp(id, "l") != 0) {
	      fprintf(stderr, "\nUnknown symbol status \"%s\" in precedence list\n\n", id);
	      misc_Error();
	    }
	    misc_StringFree(id);
	    Text = dfg_SettingsScanner(Text, &token, &id);
          }
	  dfg_Match(')', token);
	} else {
	  dfg_Match(DFG_ID, token);
	  symbol = symbol_Lookup(id);
	  if (symbol == 0) {
	    fprintf(stderr, "\nUndefined symbol %s in precedence list\n", id);
	    misc_Error();
	  }
	  misc_StringFree(id);
	}
	symbol_SetOrdering(symbol, symbol_GetIncreasedOrderingCounter());
        Text = dfg_SettingsScanner(Text, &token, &id);
      } while (token == ',');
      dfg_Match(')', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('.', token);
    } else if (strcmp(id, "set_flag") == 0) {
      int  flag;
      misc_StringFree(id);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('(', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match(DFG_ID, token);
      flag = flag_Id(id);
      if (flag == -1) {
	fputs("\nFound unkown flag ", stderr);
	fputs(id, stderr); putc('\n', stderr);
	misc_Error();
      }
      misc_StringFree(id);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match(',', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      if (token == -1)
	number = -1;
      else {
	dfg_Match(DFG_ID, token);  /* Must be a number */
	number = misc_StringToInt(id);
	misc_StringFree(id);
      }
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match(')', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('.', token);
      flag_SetFlagValue(flag, number);
    } else if (strcmp(id, "set_DomPred") == 0) {
      misc_StringFree(id); 
      Text = dfg_SettingsScanner(Text, &token, &id); 
      dfg_Match('(', token); 
      do {
	Text = dfg_SettingsScanner(Text, &token, &id);  
	dfg_Match(DFG_ID, token);
	symbol = symbol_Lookup(id); 
	if (symbol == 0) { 
	  fprintf(stderr, "\nUndefined symbol %s in precedence list\n", id); 
	  misc_Error(); 
	} 
	misc_StringFree(id); 
	symbol_AddProperty(symbol, DOMPRED);
	Text = dfg_SettingsScanner(Text, &token, &id);  
      } while (token == ',');
      dfg_Match(')', token);
      Text = dfg_SettingsScanner(Text, &token, &id); 
      dfg_Match('.', token); 
    } else {
      fputs("\nUnknown identifier in settings: ", stderr);
      fputs(id, stderr); putc('\n', stderr);
      misc_Error();
    }
    Text = dfg_SettingsScanner(Text, &token, &id);
  }

  *(Text-1) = '*'; /* Restore the contents at the end of <Text> */
}    

static void dfg_TranslPairDecl(char* PropName, char* FoName)
/**************************************************************
  INPUT:   A propositional name/first order name pair
  RETURNS: Nothing.
  EFFECT:  Updates the appropriate list.
***************************************************************/
{
	SYMBOL PropSymbol, FoSymbol;

	PropSymbol = symbol_Lookup(PropName);
	FoSymbol = symbol_Lookup(FoName);

	if (PropSymbol == 0) {
		fprintf(stderr, "Line %u: Undefined symbol %s\n", 
										dfg_LINENUMBER, PropName); 
	  misc_Error(); 
	}
	else if (FoName == 0) {
		fprintf(stderr, "Line %u: Undefined symbol %s\n", 
										dfg_LINENUMBER, FoName); 
	  misc_Error(); 
	}
	else 
	  eml_SetPropFoSymbolAssocList(PropSymbol,list_List((POINTER)FoSymbol));
}

/**************************************************************/
/* Functions for the Symbol Table                             */
/**************************************************************/

typedef struct {
  SYMBOL symbol;
  BOOL   valid;
  int    arity;
} DFG_SYMENTRY, *DFG_SYM;

static __inline__ DFG_SYM dfg_SymCreate(void)
{
  return (DFG_SYM) memory_Malloc(sizeof(DFG_SYMENTRY));
}

static __inline__ void dfg_SymFree(DFG_SYM Entry)
{
  memory_Free(Entry, sizeof(DFG_SYMENTRY));
}


static void dfg_SymAdd(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: Nothing.
  EFFECT:  This function adds 'Symbol' to the symbol list.
           The arity of these symbols will be checked every time
	   the symbol occurs.
***************************************************************/
{
  DFG_SYM newEntry = dfg_SymCreate();
  newEntry->symbol = Symbol;
  newEntry->valid  = FALSE;
  newEntry->arity  = 0;
  dfg_SYMBOLLIST = list_Cons(newEntry, dfg_SYMBOLLIST);
}


static void dfg_SymCheck(SYMBOL Symbol, NAT Arity)
/**************************************************************
  INPUT:   A symbol and the current arity of this symbol.
  RETURNS: Nothing.
  EFFECT:  This function compares the previous arity of 'Symbol'
           with the actual 'Arity'. If these values differ
	   the symbol's arity is set to arbitrary.
	   The arity of symbols whose arity was specified in
	   the symbol declaration section is checked and a warning
	   is printed to stderr in case of differences.
***************************************************************/
{
  LIST scan = dfg_SYMBOLLIST;
  while (!list_Empty(scan)) {
    DFG_SYM actEntry = (DFG_SYM) list_Car(scan);
    if (actEntry->symbol == Symbol) {
      if (actEntry->valid) {
	if (actEntry->arity != Arity)
	  actEntry->arity = -1;    /* Arbitrary Arity */
      } else {
	/* Not valid => first time */
	actEntry->arity = Arity;
	actEntry->valid = TRUE;
      }
      return;
    }
    scan = list_Cdr(scan);
  }

  /* Symbol isn't in SymbolList, so its arity was specified.        */
  /* Check if the specified arity corresponds with the actual arity */
  if (symbol_Arity(Symbol) != symbol_ArbitraryArity() &&
      symbol_Arity(Symbol) != Arity) {
    fprintf(stderr, "Line %u: Symbol ", dfg_LINENUMBER);
    symbol_FPrint(stderr, Symbol);
    fprintf(stderr, " was declared with arity %u\n", symbol_Arity(Symbol));
    misc_Error();
  }
}


static void dfg_SymCleanUp(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function corrects all symbols whose arity wasn't
           specified in the symbol declaration section but seem
	   to occur with always the same arity.
	   The memory for the symbol list is freed.
***************************************************************/
{
  while (!list_Empty(dfg_SYMBOLLIST)) {
    DFG_SYM actEntry  = (DFG_SYM) list_Car(dfg_SYMBOLLIST);
    SYMBOL  actSymbol = actEntry->symbol;

    if (actEntry->arity != symbol_Arity(actSymbol))
      symbol_SetArity(actSymbol, actEntry->arity);

    dfg_SymFree(actEntry);
    dfg_SYMBOLLIST = list_Pop(dfg_SYMBOLLIST);
  }
}


/**************************************************************/
/* Functions for the Variable Table                           */
/**************************************************************/
  
typedef struct {
  char*  name;
  SYMBOL symbol;
} DFG_VARENTRY, *DFG_VAR;

static __inline__ char* dfg_VarName(DFG_VAR Entry)
{
  return Entry->name;
}

static __inline__ SYMBOL dfg_VarSymbol(DFG_VAR Entry)
{
  return Entry->symbol;
}

static __inline__ DFG_VAR dfg_VarCreate(void)
{
  return (DFG_VAR) memory_Malloc(sizeof(DFG_VARENTRY));
}

static void dfg_VarFree(DFG_VAR Entry)
{
  misc_StringFree(Entry->name);
  memory_Free(Entry, sizeof(DFG_VARENTRY));
}

static void dfg_VarStart(void)
{
  dfg_VARLIST = list_Push(list_Nil(), dfg_VARLIST);
  dfg_VARDECL = TRUE;
}

static void dfg_VarStop(void)
{
  dfg_VARDECL = FALSE;
}

static void dfg_VarBacktrack(void)
{
  list_DeleteWithElement(list_Top(dfg_VARLIST), dfg_VarFree);
  dfg_VARLIST = list_Pop(dfg_VARLIST);
}

static void dfg_VarCheck(void)
/* Should be called after a complete clause or formula was parsed */
{
  if (!list_Empty(dfg_VARLIST)) {
    fputs("\ndfgparser: List of variables should be empty!\n", stderr);
    misc_DumpCore();
  }
  symbol_ResetStandardVarCounter();
}

static SYMBOL dfg_VarLookup(char* Name)
/**************************************************************
  INPUT:   A variable name.
  RETURNS: The corresponding variable symbol.
  EFFECT:  If the variable name was quantified before, the
           corresponding symbol is returned and the <Name> is freed.
	   If the variable name was not quantified, and <dfg_VARDECL>
	   is TRUE, a new variable is created, else an error
	   message is printed and the program exits.
***************************************************************/
{
  LIST   scan, scan2;
  SYMBOL symbol = symbol_Null();

  /*   puts("\n----------------"); */
  /*   for (scan = dfg_VARLIST; !list_Empty(scan); scan = list_Cdr(scan)) { */
  /*     for (scan2 = list_Car(scan); !list_Empty(scan2); scan2=list_Cdr(scan2)) */
  /*       printf("%d:%s ", dfg_VarSymbol(list_Car(scan2)), dfg_VarName(list_Car(scan2))); */
  /*     putchar('\n'); */
  /*   }  DBG */

  scan  = dfg_VARLIST;
  scan2 = list_Nil();
  while (!list_Empty(scan) && list_Empty(scan2)) {
    scan2 = list_Car(scan);
    while (!list_Empty(scan2) &&
	   (strcmp(dfg_VarName(list_Car(scan2)), Name) != 0))
      scan2 = list_Cdr(scan2);
    scan = list_Cdr(scan);
  }

  if (!list_Empty(scan2)) {
    /* Found variable */
    misc_StringFree(Name);
    symbol = dfg_VarSymbol(list_Car(scan2));
  } else {
    /* Variable not found */
    if (dfg_VARDECL) {
      DFG_VAR newEntry = dfg_VarCreate();
      newEntry->name   = Name;
      newEntry->symbol = symbol_CreateStandardVariable();
      /* Add <newentry> to the first list in dfg_VARLIST */
      list_Rplaca(dfg_VARLIST, list_Cons(newEntry,list_Car(dfg_VARLIST)));
      symbol = dfg_VarSymbol(newEntry);
    } else {
      fprintf(stderr, "\nLine %u: Free Variable %s\n\n", dfg_LINENUMBER, Name);
      misc_Error();
    }
  }
  return symbol;
}
