/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              FIRST ORDER LOGIC SYMBOLS                 * */
/* *                                                        * */
/* *  $Module:   FOL                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.50 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:37 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: folspass.c,v $ */

#include "folspass.h"

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

SYMBOL  fol_ALL;
SYMBOL  fol_EXIST;
SYMBOL  fol_AND;
SYMBOL  fol_OR;
SYMBOL  fol_QUERY;
SYMBOL  fol_NOT;
SYMBOL  fol_IMPLIES;
SYMBOL  fol_IMPLIED;
SYMBOL  fol_EQUIV;
SYMBOL  fol_VARLIST;
SYMBOL  fol_EQUALITY;

static LIST fol_SYMBOLS;


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  FUNCTIONS                                             * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void fol_Init(BOOL All)
/**************************************************************
  INPUT:   A boolean value detrmining whether only 'equality' or
           all fol symbols are created.
  RETURNS: None.
  USAGE:   extern public.
  SUMMARY: Initializes the Fol Module.
  EFFECTS: If <All> then all fol-symbols are created,
           only 'equality' otherwise.
  CAUTION: MUST BE CALLED BEFORE ANY OTHER fol-FUNCTION.
***************************************************************/
{
  if (All) {
    fol_ALL      = symbol_CreateJunctor("all", 2, symbol_STATLEX);
    fol_EXIST    = symbol_CreateJunctor("exist", 2, symbol_STATLEX);
    fol_AND      = symbol_CreateJunctor("and", -1, symbol_STATLEX);
    fol_OR       = symbol_CreateJunctor("or", -1, symbol_STATLEX);
    fol_QUERY    = symbol_CreateJunctor("query", -1, symbol_STATLEX);
    fol_NOT      = symbol_CreateJunctor("not", 1, symbol_STATLEX);
    fol_IMPLIES  = symbol_CreateJunctor("impl", 2, symbol_STATLEX);
    fol_IMPLIED  = symbol_CreateJunctor("impd", 2, symbol_STATLEX);
    fol_EQUIV    = symbol_CreateJunctor("equiv", 2, symbol_STATLEX);
    fol_VARLIST  = symbol_CreateJunctor("", -1, symbol_STATLEX);
    fol_EQUALITY = symbol_CreatePredicate("=", 2, symbol_STATLEX);

    fol_SYMBOLS =
      list_Cons((POINTER)fol_ALL, list_Cons((POINTER)fol_EXIST, 
	list_Cons((POINTER)fol_AND, list_Cons((POINTER)fol_OR,
	  list_Cons((POINTER)fol_QUERY, list_Cons((POINTER)fol_NOT,
            list_Cons((POINTER)fol_IMPLIES, list_Cons((POINTER)fol_EQUIV,
              list_Cons((POINTER)fol_VARLIST, list_Cons((POINTER)fol_IMPLIED,
                list_List((POINTER)fol_EQUALITY)))))))))));
  } else {
    fol_EQUALITY = symbol_CreatePredicate("=", 2, symbol_STATLEX);
    fol_NOT      = symbol_CreateJunctor("spassnot", 1, symbol_STATLEX);
    fol_SYMBOLS  = list_Cons((POINTER)fol_NOT,
			     list_List((POINTER)fol_EQUALITY));
  }
}


SYMBOL fol_IsStringPredefined(const char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: The symbol iff String is a predefined fol string,
           symbol NULL otherwise
  EFFECT:  None
***************************************************************/
{
  LIST Scan;
  for(Scan = fol_SYMBOLS; !list_Empty(Scan); Scan = list_Cdr(Scan))
    if (strcmp(String, symbol_Name((SYMBOL)list_Car(Scan))) == 0)
      return ((SYMBOL)list_Car(Scan));
  return symbol_Null();
}


TERM fol_CreateQuantifier(SYMBOL Quantifier, LIST VarList, LIST Arguments)
/*******************************************************
  INPUT:   A symbol (which MUST be a fol quantifier),
           a list of variables that will be bound, and
           a list of arguments.
  RETURNS: A quantified term.
********************************************************/
{                                                                  
#ifdef DEBUG                                                          
  if (!(fol_IsQuantifier(Quantifier))) {
    printf("\n\nIllegal input in fol_CreateQuantifier\n");
    misc_Error();
  }
#endif                                                             
                                                                   
  return term_Create(Quantifier,
		     list_Cons(term_Create(fol_Varlist(), VarList),   
			       Arguments));  
}


TERM fol_ComplementaryTerm(TERM Term)
/*********************************************************
  INPUT:   A formula.
  RETURNS: The (copied) complementary (in sign) formula of <Term>
**********************************************************/
{
  if (symbol_Equal(term_TopSymbol(Term), fol_Not()))
    return term_Copy((TERM)list_First(term_ArgumentList(Term)));
  else
    return term_Create(fol_Not(), list_List(term_Copy(Term)));
}


void fol_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  The fol memory used by the fol modul is freed.
***************************************************************/
{
  list_Delete(fol_SYMBOLS);
}


void fol_PrintDFGSignature(FILE* File)
/**************************************************************
  INPUT:   A file stream.
  RETURNS: Nothing.
  SUMMARY: Prints all signature symbols in DFG format to the
           file stream.
***************************************************************/

{
  NAT    i;
  SYMBOL symbol;
  LIST   functions, predicates;

  functions = predicates = list_Nil();

  for (i = 1; i < symbol_ActIndex(); i++) 
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i)) {
      symbol = symbol_Signature(i)->info;
      switch (symbol_Type(symbol)) {

      case symbol_CONSTANT:
      case symbol_FUNCTION:
	functions = list_Push((POINTER)symbol, functions);
	break;
      case symbol_PREDICATE:
	predicates = list_Push((POINTER)symbol, predicates);
	break;
      }
    }

  if (!list_Empty(functions)) {
    fprintf(File, "  functions[");
    i = 0;
    do {
      symbol = (SYMBOL) list_Top(functions);
      fprintf(File, "(%s, %d)", symbol_Name(symbol), symbol_Arity(symbol));
      functions = list_Pop(functions);
      if (!list_Empty(functions))
	fprintf(File, ", ");
      if (i<15)
	i++;
      else {
	i= 0;
	fprintf(File, "\n\t");
      }
      
    } while (!list_Empty(functions));
    fprintf(File, "].\n");
  }

  predicates = list_DeleteElement(predicates, (POINTER)fol_VARLIST, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_EQUALITY, symbol_Equal);
  if (!list_Empty(predicates)) {
    i = 0;
    fprintf(File, "  predicates[");
    do {
      symbol = (SYMBOL) list_Top(predicates);
      fprintf(File, "(%s, %d)", symbol_Name(symbol), symbol_Arity(symbol));
      predicates = list_Pop(predicates);
      if (!list_Empty(predicates))
	fprintf(File, ", ");
      if (i<15)
	i++;
      else {
	i= 0;
	fprintf(File, "\n\t");
      }
    } while (!list_Empty(predicates));
    fprintf(File, "].\n");
  }
  list_Delete(predicates);
  list_Delete(functions);
}
