/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *        DYNAMIC MEMORY MANAGEMENT MODULE                * */
/* *                                                        * */
/* *  $Module:   MEMORY                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.48 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:47:01 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: memory.c,v $ */

#include "memory.h"

int        memory_PAGESIZE;
int        memory_MAXMEM;
static int memory__EOF = EOF;
unsigned   memory_NEWBYTES;
unsigned   memory_FREEDBYTES;

/* In debug--mode, one int for magic number and one int for size */
#ifdef DEBUG
const unsigned int memory_MARKSIZE = sizeof(int) + sizeof(int);
#else
const unsigned int memory_MARKSIZE = 0;
#endif

const unsigned int memory_MAGICMALLOC = 4711;
const unsigned int memory_MAGICFREE = 1147;

const unsigned int memory_ALIGN = sizeof(POINTER);

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  INITIALIZATION                                        * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void memory_Init(int Maxmem)
/*************************************************************/
/* INPUT  : The maximal amount of memory available in KB for */
/*          the memory module; if <=0 the module allocates   */
/*          as much memory as  possible by calloc()          */
/* SUMMARY: Initializes the memory management.               */
/*************************************************************/
{
  int i,extra;
  
  memory_FREEDBYTES = 0;
  memory_NEWBYTES   = 0;
  memory_PAGESIZE   = 8192; /* Standard value, may be changed */

  extra = 2*sizeof(POINTER) + memory_MARKSIZE; /* Extra Space per page */

  while (memory_PAGESIZE < (2*(memory__DYNMAXSIZE + memory_MARKSIZE) + extra))
    memory_PAGESIZE += 4096;    /* Minimum two objects per allocated page */

  if (Maxmem <= 0)
    memory_MAXMEM = -1;
  else
    memory_MAXMEM = Maxmem/(memory_PAGESIZE/1024);

  for (i=1; i<memory__DYNMAXSIZE; i++) {
    memory_ARRAY[i].free    = &memory__EOF;
    memory_ARRAY[i].new     = &memory__EOF;
    memory_ARRAY[i].page    = &memory__EOF;
    memory_ARRAY[i].size    = 
      memory_MARKSIZE + i + ((i % memory_ALIGN) ?
			      (memory_ALIGN - (i % memory_ALIGN)) : 0);
    memory_ARRAY[i].offset  = 
      ((memory_PAGESIZE-extra)/memory_ARRAY[i].size)*memory_ARRAY[i].size 
      + sizeof(POINTER) + memory_MARKSIZE;
  }
}


void memory_Restrict(int Maxmem)
/*************************************************************/
/* INPUT  : The maximal amount of memory available in KB for */
/*          the memory module; if <=0 the module allocates   */
/*          as much memory as  possible by calloc()          */
/*************************************************************/
{
  if (Maxmem <= 0)
    memory_MAXMEM = -1;
  else
    memory_MAXMEM = Maxmem/(memory_PAGESIZE/1024);
}


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  MALLOC                                                * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

#ifdef PURIFY

POINTER memory_Malloc(unsigned int Bytes)
{
  char *mem;

  mem = (char *)calloc(1,Bytes);
  
  if (mem == NULL)
    abort();
  
  return mem;
}

#else

POINTER memory_Malloc(unsigned int Bytes)
/*********************************************************/
/* INPUT:   A number which corresponds to the size of a  */
/*          memory.                                      */
/* RETURNS: A pointer to a block of "Bytes" bytes.       */
/*********************************************************/
{
  char *NewMemory;
  
#ifdef DEBUG
  if (Bytes == 0) {
    printf("memory: Tried to allocate 0 Bytes!");
    misc_DumpCore();
  }
#endif

  if (Bytes >= memory__DYNMAXSIZE) {
    NewMemory = (char *)calloc(1,Bytes + memory_MARKSIZE);
#ifdef DEBUG
    NewMemory = (char *)((int *)(NewMemory) + 2);
    *((int *)(NewMemory) - 2) = memory_MAGICMALLOC;
    *((int *)(NewMemory) - 1) = Bytes;
#endif
    memory_NEWBYTES    += Bytes + memory_MARKSIZE;
    return NewMemory;
  }

  if (*((int *)memory_ARRAY[Bytes].free) != EOF) {

    /* Use already freed block */
    NewMemory                = (char *)(memory_ARRAY[Bytes].free);
    memory_ARRAY[Bytes].free = *((POINTER *)(memory_ARRAY[Bytes].free));
    memory_FREEDBYTES       -= memory_ARRAY[Bytes].size;
#ifdef DEBUG
    *((int *)(NewMemory) - 2) = memory_MAGICMALLOC;
    *((int *)(NewMemory) - 1) = Bytes;  
#endif
    return NewMemory;

  } 
  else 
      if (*((int *)memory_ARRAY[Bytes].new) != EOF) {

	  /* Use new block from current page */
	  NewMemory               = (char *)(memory_ARRAY[Bytes].new);
	  memory_ARRAY[Bytes].new = NewMemory + memory_ARRAY[Bytes].size;
	  memory_NEWBYTES        += memory_ARRAY[Bytes].size;
#ifdef DEBUG
	  if ((int)NewMemory >= (int)memory_ARRAY[Bytes].page +
	      (int)memory_ARRAY[Bytes].offset) {
	      printf("memory: Address Overflow %d \n",Bytes);
	      misc_Error();
	  }
	  *((int *)(NewMemory) - 2) = memory_MAGICMALLOC;	  
	  *((int *)(NewMemory) - 1) = Bytes;  
#endif
	  return NewMemory;
    
      } 
      else {    
	  /* Allocate a new page of data.    */
	  /* First Block used to chain pages */
	  if (memory_MAXMEM >=0) {
	    if (memory_MAXMEM == 0) {
	      printf("\nSPASS terminated by Memory Flag.\n");
	      misc_Error();
	    } 
	    else
	      memory_MAXMEM--;
	  }
	  if ( (NewMemory=(char *)calloc(1,memory_PAGESIZE)) == NULL) {
	      printf("\nSPASS terminated, ran out of system memory.\n");
	      misc_Error();
	  }
	  *((POINTER *)NewMemory)     = memory_ARRAY[Bytes].page;
	  memory_ARRAY[Bytes].page    = NewMemory;
	  memory_NEWBYTES            += memory_ARRAY[Bytes].size;
	  memory_ARRAY[Bytes].new     = NewMemory + memory_ARRAY[Bytes].size +
	      sizeof(POINTER) + memory_MARKSIZE;
	  *((int *)(NewMemory + memory_ARRAY[Bytes].offset)) = EOF;
#ifdef DEBUG
	  *(int *)(NewMemory + sizeof(POINTER)) = memory_MAGICMALLOC;
	  *((int *)(NewMemory + sizeof(POINTER)) + 1) = Bytes;
#endif
	  return (NewMemory + sizeof(POINTER) + memory_MARKSIZE);
      }
}

#endif


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  DEBUGGING INFORMATION                                 * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void memory_Print(void)
/**************************************************************
  INPUT:   None.
  RETURNS: None.
  SUMMARY: Prints the number of used bytes to the screen.
***************************************************************/
{
  int Pages;
  int i;
  POINTER ActPage;

  Pages = 0;
  for(i = 1; i < memory__DYNMAXSIZE; i++) {
    ActPage = memory_ARRAY[i].page;
    while (*((int *)ActPage) != EOF) {
      Pages++;
      ActPage = *((POINTER *)ActPage);
    }
  }
    
  printf("\n###\n");
  printf("### Pagesize: %d\n",memory_PAGESIZE);
  printf("### Marksize: %d\n",(int)memory_MARKSIZE);
  printf("### Memory demanded:  %d KBytes\n", memory_NEWBYTES/1024);
  printf("### Memory freed:     %d KBytes\n", memory_FREEDBYTES/1024);
  printf("### Memory remaining: %d Bytes\n", memory_NEWBYTES-memory_FREEDBYTES);
  printf("### Pages allocated:  %d Pages\n", Pages);
  printf("###\n");
}

void memory_FPrint(FILE* File)
/**************************************************************
  INPUT:   A file.
  RETURNS: None.
  SUMMARY: Prints the number of used bytes to the file.
***************************************************************/
{
  int Pages;
  int i;
  POINTER ActPage;

  Pages = 0;
  for(i = 1; i < memory__DYNMAXSIZE; i++) {
    ActPage = memory_ARRAY[i].page;
    while (*((int *)ActPage) != EOF) {
      Pages++;
      ActPage = *((POINTER *)ActPage);
    }
  }
    
  fprintf(File,"\n###\n");
  fprintf(File,"### Pagesize: %d\n",memory_PAGESIZE);
  fprintf(File,"### Marksize: %d\n",(int)memory_MARKSIZE);
  fprintf(File,"### Memory demanded:  %d KBytes\n", memory_NEWBYTES/1024);
  fprintf(File,"### Memory freed:     %d KBytes\n", memory_FREEDBYTES/1024);
  fprintf(File,"### Memory remaining: %d KBytes\n", memory_NEWBYTES-memory_FREEDBYTES);
  fprintf(File,"### Pages allocated:  %d Pages\n", Pages);
  fprintf(File,"###\n");
}


void memory_PrintDetailed(void)
/**************************************************************
  INPUT:   None.
  RETURNS: None.
  SUMMARY: Prints the number of used bytes to the screen.
***************************************************************/
{
  int i;
  POINTER ActPage, ActData, ActFree;

  printf("\nEOF Pointer: %d\n", (int)(&memory__EOF));

  for (i=1; i<memory__DYNMAXSIZE; i++) {
    printf("\n\n Entry: %d Size: %d\n", i , memory_ARRAY[i].size);
    ActPage  = memory_ARRAY[i].page;
    ActData  = memory_ARRAY[i].new;
    ActFree  = memory_ARRAY[i].free;

    if (*((int *)ActPage) == EOF) {
      if (*((int *)ActData) == EOF)
	printf("   No request so far\n");
      else {
	printf("No Page entry but New entry");
	misc_Error();
      }
    }
    else {
      printf("   Current Page: %d Current New: %d\n",
	     (int)ActPage, (int)ActData);
      while(*((int *)ActPage) != EOF) {
	printf("\n   Page: %d Next Page: %d\n",
	       (int)ActPage, *((int *)ActPage));

	ActData = ((char *)ActPage + sizeof(POINTER) + memory_MARKSIZE);

	printf("   Data: ");
	while (*((int *)ActData) != EOF) {
	  printf("%d ", (int)ActData);
	  ActData = (char *)ActData + memory_ARRAY[i].size;
	}

#ifdef DEBUG
	ActData = ((char *)ActPage + sizeof(POINTER) + memory_MARKSIZE);

	while ((*((int *)ActData) != EOF) &&
	       (*((int*)ActData-2) != memory_MAGICMALLOC))
	  ActData = (char *)ActData + memory_ARRAY[i].size;

	if (*((int *)ActData) == EOF)
	  printf("\n   No memory allocated from this page\n");
	else {
	  printf("\n   Allocated: ");
	  do  {
	    if (*((int*)ActData-2) == memory_MAGICMALLOC)
	      printf("%d ", (int)ActData);
	    ActData = (char *)ActData + memory_ARRAY[i].size;
	  } while (*((int *)ActData) != EOF);
	}
#endif

	ActPage = *((POINTER *)ActPage);
      }
      if (*((int *)ActFree) == EOF)
	printf("\n   No freed memory\n");
      else {
	printf("\n   Free: ");
	while(*((int *)ActFree) != EOF) {
	  printf("%d ", (int)ActFree);
	  ActFree = *((POINTER *)ActFree);
	}
      }  
    }
  }
}
