/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *             PROOF CHECKING                             * */
/* *                                                        * */
/* $Revision: 0.44 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:47:02 $                             * */
/* $Author: schmidt $                                         * */
/* *                                                        * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */

/* $RCSfile: proofcheck.c,v $ */

/**************************************************************

   Given a dfg proof file, this program generates
   a set of proof files corresponding to each proof step.

	Options:

	-d    Prefix of generated files 
	-t    time limit for proof attempt
	      (to set the TimeLimit option in the  
	       file passed to SPASS)
	-s    file suffix of generated proof files 
	-q    quiet mode

**************************************************************/

#include "proofcheck.h"

/* forwarding */

static void   PrintFormulaList(FILE* File, LIST Formulas, char* Name);
static void   pcheck_SaveAsDFG2(FILE* File, LIST Axioms, LIST Conjectures);
static string pcheck_StrConc(char* s1, char* s2);
static string pcheck_StrEmpty();
static BOOL   pcheck_IsRightSplitHalf(CLAUSE C);

static BOOL    pcheck_StrToInt(char*, int*);
void           pcheck_IntToStr(int i, char* s);
static BOOL    pcheck_ClauseIsGround(CLAUSE C);

/* options */

int    pcheck_Timelimit;
char*  pcheck_ProofFileSuffix;
BOOL   pcheck_Quiet;

/* options for graph generation */
 
#ifdef USE_CG
BOOL   pcheck_ClauseCg;
BOOL   pcheck_GenNamedCg, pcheck_GenRedCg;
char   *pcheck_CgName, *pcheck_RedCgName;
GRAPHFORMAT pcheck_GraphFormat;
#endif


static void dfg_StringFree(char* String) 
/**************************************************************
 INPUT:   A dfg string                                  
 RETURNS: -                                             
 EFFECTS: frees memory                                  
 **************************************************************/
{
  memory_Free(String, sizeof(char)*(strlen(String)+1));
} 

static int pcheck_MaxParentSplitLevel(CLAUSE Clause) 
/**************************************************************
 INPUT  : A clause                                                      
 RETURNS: The max split level of the parent clauses                                                
 EFFECTS: -
***************************************************************/
{
  LIST Parents;
  int Max;
  int Act;

  Max = 0;
  Parents = clause_ParentClauses(Clause);

  while (!list_Empty(Parents)) {
    Act = clause_SplitLevel(list_Car(Parents));
    if (Act > Max) 
      Max = Act;
    Parents = list_Cdr(Parents); }
  return(Max);
}

static BOOL pcheck_ClauseIsFromLeftSplit(CLAUSE Clause)
/**************************************************************
 INPUT  : A Clause                                                      
 RETURNS: TRUE iff the clause if the left half of a split                                                     
 EFFECTS: -                                                     
 CAUTION: This works also for clauses without parents, since
          pcheck_MaxParentSplitLevel returns 0 in that case.
***************************************************************/
{
  if (clause_SplitLevel(Clause) > pcheck_MaxParentSplitLevel(Clause)) {
    return(TRUE); }
  return(FALSE);
}

static BOOL pcheck_ClauseIsFromRightSplit(CLAUSE Clause)
/**************************************************************
 INPUT  : A clause
 RETURNS: TRUE iff the clause is from the right half of a split
 EFFECTS: -
***************************************************************/
{
  if (list_Empty(clause_ParentClauses(Clause)))
    return(FALSE);

  return(clause_IsEmptyClause(list_Car(clause_ParentClauses(Clause))));
}

BOOL pcheck_ClauseIsFromSplit(CLAUSE Clause)
/**************************************************************
 INPUT  : A clause                                                      
 RETURNS: TRUE iff the clause is from a split
 EFFECTS: -
***************************************************************/
{
  return (pcheck_ClauseIsFromRightSplit(Clause) ||
      pcheck_ClauseIsFromLeftSplit(Clause));
}


static int pcheck_LabelToNumber(char* Label)
/**************************************************************
 INPUT  : A clause label                                                      
 RETURNS: This label as a number
 EFFECTS: -                                                     
***************************************************************/
{
  int Number;

  if (!pcheck_StrToInt(Label, &Number)) {
    fprintf(stderr, "could not convert clause label %s to number.\n", Label);
    misc_Error(); }
  return(Number);
}


static BOOL pcheck_BinSearchClauseInVector(CLAUSE* Clauses, int Position, int n, CLAUSE* C, int(*Weight)())
/**************************************************************
 INPUT  : A vector of clauses, a vector index,
	  number <n>, a clause reference.
 RETURNS: Searches <Clauses>[0..<Position>] for clause
          with number <n> with binary search. Returns
	  TRUE iff clause has been found, pointer to clause
	  in reference <C>
 EFFECTS: -                                                     
***************************************************************/
{
  int Upper;
  int Lower;
  int Mid;

  Upper = Position;
  Lower = 0;

  while (Upper != Lower) { 

    Mid = (Upper-Lower)/2+Lower;
    if (Weight(Clauses[Mid]) < n) /* so clause must be searched above */
      Lower = Mid+1;
    else
      Upper = Mid;  }

  (*C) = Clauses[Upper];

  return(Weight(*C) == n);
}

static void pcheck_Nop(POINTER P)
/**************************************************************
 INPUT  : A pointer
 RETURNS: -                                                     
 EFFECTS: -                                                     
***************************************************************/
{
  return;
}

void static pcheck_ParentNumbersToPointersInVector(CLAUSE* ClauseVector, int Size)
/**************************************************************
 INPUT  : A clause vector without duplicate clauses                                                     
 RETURNS: -                                                     
 EFFECTS: All clause parent numbers are replaced by pointers
          to the parents
***************************************************************/
{
  int     Position;
  LIST    NewParents, OldParents;
  LIST    ScanParents;
  int     ParentNum;
  CLAUSE  Parent;
  
  for (Position = 0; Position < Size; Position++)  {
     
    OldParents = clause_ParentClauses(ClauseVector[Position]);
    NewParents = list_Copy(OldParents);
    
    for (ScanParents = NewParents; !list_Empty(ScanParents); ScanParents = list_Cdr(ScanParents)) {
      
      ParentNum = (int)list_Car(ScanParents);
      if (pcheck_BinSearchClauseInVector(ClauseVector, Position, ParentNum, &Parent, clause_Number))
	list_Rplaca(ScanParents, Parent);  
      else {
	fprintf(stderr, "error: failed to convert parent numbers to pointers.\n");
	fprintf(stderr, "\nin clause:\n");
	clause_PParentsFPrintGen(stderr, ClauseVector[Position], FALSE);
	fprintf(stderr, "\nparent clause with number: %d not yet read.\n", ParentNum); 
	misc_Error();}
    }
    
    clause_SetParentClauses(ClauseVector[Position], NewParents);
    list_Delete(OldParents); }

}

LIST static pcheck_ForceParentNumbersToPointersInVector(CLAUSE* ClauseVector, int Size)
/**************************************************************
 INPUT  : A clause vector, possibly with duplicates
 RETURNS: TRUE if all clauses could be converted.
 EFFECTS: All MARKED flags are changed. All parent numbers
          are converted to pointers, if the parents exist,
	  otherwise the parent number is deleted from the list.
***************************************************************/
{
  int     Position;
  LIST    NewParents, OldParents;
  LIST    ScanParents, ScanPLits;
  LIST    OldPLits, NewPLits;
  int     ParentNum, PLitNum;
  CLAUSE  Parent, Clause;
  BOOL    Ok;
  LIST    Missing;

  Ok = TRUE;
  Missing = list_Nil();

  for (Position = 0; Position < Size; Position++) {
    clause_RemoveFlag(ClauseVector[Position], MARKED); 
    clause_RemoveFlag(ClauseVector[Position], HIDDEN); }

  for (Position = 0; Position < Size; Position++)  {
     
    Clause = ClauseVector[Position];
    if (!clause_GetFlag(Clause, MARKED)) { 
      clause_SetFlag(Clause, MARKED);
      OldParents = clause_ParentClauses(Clause);
      OldPLits   = clause_ParentLiterals(Clause);

      NewParents  = list_Nil();
      NewPLits    = list_Nil();
      ScanParents = OldParents;
      ScanPLits   = OldPLits;
      while  (!list_Empty(ScanParents)) {
	
	ParentNum = (int)list_Car(ScanParents);
	PLitNum   = (int)list_Car(ScanPLits);
	if (!pcheck_BinSearchClauseInVector(ClauseVector, Size-1, ParentNum, &Parent, clause_Number)) {
	  Missing = list_Cons((POINTER)ParentNum, Missing); 
	  clause_SetFlag(Clause, HIDDEN); }
	else {
	  if (clause_GetFlag(Parent, HIDDEN))
	    clause_SetFlag(Clause, HIDDEN);
	  NewParents  = list_Cons((POINTER)Parent, NewParents);
	  NewPLits    = list_Cons((POINTER)PLitNum,NewPLits); }
	ScanParents = list_Cdr(ScanParents); 
	ScanPLits   = list_Cdr(ScanPLits); }
      
      NewParents = list_NReverse(NewParents);
      NewPLits   = list_NReverse(NewPLits);
      clause_SetParentClauses(Clause, NewParents);
      clause_SetParentLiterals(Clause, NewPLits);
      list_Delete(OldPLits);
      list_Delete(OldParents);  
    } /* if clause is not marked */
  } /* for all clauses */
 
  return(Missing);
}

static int pcheck_PasteListInVector(LIST L, CLAUSE* Vector, int Position)
/**************************************************************
 INPUT  : A clause list, a clause vector
 RETURNS: <Position>+list_Length(L)                                                      
 EFFECTS: Write all clauses into the vector, starting
          from <Position>
***************************************************************/
{
  for (; !list_Empty(L); L = list_Cdr(L)) 
    { Vector[Position] = list_Car(L);
    Position++; }
  return(Position);
}



LIST pcheck_ConvertParentsInList(LIST List)
/**************************************************************
 INPUT  : A list of clauses
 RETURNS: The same list                                                      
 EFFECTS: Parent numbers are converted to pointers
***************************************************************/
{
  int     Size;
  CLAUSE* ClauseVector;
  LIST    Missing;

  Size = list_Length(List);
  if (Size == 0)
    return(list_Nil());

  /* convert list into vector for binary search */
  ClauseVector = (CLAUSE*)memory_Malloc(Size*sizeof(CLAUSE*));
  pcheck_PasteListInVector(List, ClauseVector, 0);

  /* convert parent lists */
  Missing = pcheck_ForceParentNumbersToPointersInVector(ClauseVector, Size);
  
  memory_Free(ClauseVector, Size*sizeof(CLAUSE*));
  return(Missing);
}

LIST pcheck_ConvertParentsInSPASSProof(PROOFSEARCH Search, LIST EmptyClauses) 
/**************************************************************
 INPUT  : A proofsearch object with clauses sorted by weight
          and an unsorted list <EmptyClauses>
 RETURNS: The lists, where the clauses in <EmptyClauses> are
          now sorted by weight, and parent numbers
	  in the clauses are replaced by parent pointers
 EFFECTS:                                                      
***************************************************************/
{
  LIST    AllLists; 
  LIST    Missing;

  AllLists = list_Nconc(list_Copy(prfs_WorkedOffClauses(Search)), 
			list_Nconc(list_Copy(prfs_UsableClauses(Search)),
				   list_Nconc(list_Copy(prfs_DocProofClauses(Search)), 
					      list_Nconc(list_Copy(EmptyClauses), 
							 list_Nil()))));
  
  AllLists = pcheck_ClauseNumberMergeSort(AllLists);
  Missing  = pcheck_ConvertParentsInList(AllLists);
  list_Delete(AllLists);

  return(Missing);
}

LIST pcheck_ParentNumbersToParents(LIST Proof)
/**************************************************************
 INPUT  : A list of clauses, representing a proof.
 RETURNS: The list, where parent numbers in the
          parent list of clauses are replaced
          by the parents (pointers).
 EFFECTS: -                                                       
 CAUTION: For finding the clause corresponding to a
          a clause number, the <Proof> list is searched
	  with binary search. This is correct only if
	  the clause numbers in <Proof> are increasing.
************************************************************/
{
  LIST    ScanClauses;
  int     ProofLength, Position;
  CLAUSE* ClauseVector;

  if (list_Empty(Proof))
    return(list_Nil());
  
  /* convert list into vector for binary search */
  ProofLength  = list_Length(Proof);
  ClauseVector = (CLAUSE*)memory_Malloc(ProofLength*sizeof(CLAUSE*));

  Position = 0;
  for (ScanClauses = Proof, Position = 0; 
       !list_Empty(ScanClauses);
       ScanClauses = list_Cdr(ScanClauses), Position++)
    ClauseVector[Position] = list_Car(ScanClauses);

  /* convert parent lists */
  pcheck_ParentNumbersToPointersInVector(ClauseVector, ProofLength);
 
  memory_Free(ClauseVector, ProofLength*sizeof(CLAUSE*));

  return(Proof);
}



LIST pcheck_ParentPointersToParentNumbers(LIST Clauses)
/**************************************************************
 INPUT  : A list of clauses                                                      
 RETURNS: The list with parent pointers replaced by
          parent numbers
 EFFECTS: Sets marks on all clauses.                            
***************************************************************/
{
  LIST ScanClauses;
  LIST ScanParents;

  pcheck_ClauseListRemoveFlag(Clauses, MARKED);

  for (ScanClauses = Clauses; !list_Empty(ScanClauses); ScanClauses = list_Cdr(ScanClauses)) {
    if (!clause_GetFlag(list_Car(ScanClauses), MARKED)){
      for (ScanParents = clause_ParentClauses(list_Car(ScanClauses)); !list_Empty(ScanParents);
	   ScanParents = list_Cdr(ScanParents)) 
	list_Rplaca(ScanParents, (POINTER)clause_Number(list_Car(ScanParents))); 
      clause_SetFlag(list_Car(ScanClauses), MARKED); }
  }
  return(Clauses);
}


LIST pcheck_ConvertTermListToClauseList(LIST ProofRest)
/**************************************************************
 INPUT  : A list of quintupels (lists) representing a proof
 RETURNS: The conversion of this list into the internal clause
          structure. For each clause, the numbers of parent clauses 
	  are replaced by pointers to the parents.
 EFFECTS: -                                                     
***************************************************************/
{
  LIST    Clauses;
  LIST    ProofLine;
  TERM    ClauseTerm;
  CLAUSE  Clause;
  int     Level;
  int     ClauseNumber;
  LIST    ParentLabels;
  LIST    ParentIds;     
  LIST    ParentLits;    /* this is a dummy list; parent lits are not yet specified in a SPASS proof */
  RULE    Origin;
  char*   ClauseLabel;

  Clauses = list_Nil();  /* result */

  while (!list_Empty(ProofRest)) {
    
    /* break proof line into components */

    ProofLine     = list_Car(ProofRest);
    
    ClauseLabel   = list_First(ProofLine);
    ClauseTerm    = list_Second(ProofLine);    
    /* replace by NULL clause, since dfg_CreateClauseFromTerm deletes clause ! */
    list_Rplaca(list_Cdr(ProofLine), clause_Null()); 
    ParentLabels  = (LIST)list_Third(ProofLine);
    Level         = (int)list_Fourth(ProofLine);
    Origin        = (RULE)list_Fifth(ProofLine);

    /* Conversion */

    Clause       = dfg_CreateClauseFromTerm(ClauseTerm, TRUE);
    ClauseNumber = pcheck_LabelToNumber(ClauseLabel);
    ParentIds    = list_Nil();
    ParentLits   = list_Nil();

    while (!list_Empty(ParentLabels)) {
      ParentIds  = list_Cons((POINTER)pcheck_LabelToNumber(list_Car(ParentLabels)), ParentIds); 
      ParentLits = list_Cons(0, ParentLits);
      ParentLabels   = list_Cdr(ParentLabels); }
    
    /* set all data */
   
    clause_SetNumber(Clause, ClauseNumber);
    ParentIds = list_NReverse(ParentIds);
    clause_SetParentClauses(Clause, ParentIds);
    clause_SetParentLiterals(Clause, ParentLits);
    Clause->origin = Origin;

    clause_SetSplitLevel(Clause, Level); 
    if (Level > 0) {
      clause_ClearSplitField(Clause);
      clause_SetSplitFieldBit(Clause, Level); }
    else 
      clause_SetSplitField(Clause, (SPLITFIELD)NULL,0);

    clause_RemoveFlag(Clause, MARKED);
    Clauses   = list_Cons(Clause, Clauses);
    ProofRest = list_Cdr(ProofRest);  }

  Clauses = list_NReverse(Clauses);
  
  /* convert parent numbers to pointers */
  Clauses = pcheck_ParentNumbersToParents(Clauses);
  
  return(Clauses);
}

static int pcheck_MaxClauseLevel(LIST Clauses)
/**************************************************************
 INPUT  : A list of clauses 
 RETURNS: The maximum split level of a clause
 EFFECTS: -                                                     
***************************************************************/
{
 int Max;
 int Act;
 LIST Parents;
 
 Max = 0;
 Parents = list_Nil();

 while (!list_Empty(Clauses)) {
   Act = clause_SplitLevel(list_Car(Clauses));
   if (Act > Max) 
     Max = Act;
   Clauses = list_Cdr(Clauses); }
 return(Max);
}


LIST pcheck_CollectMarkedClauses(LIST Clauses, CLAUSE_FLAGS Flag, BOOL Marked)
/**************************************************************
 INPUT  : A list of clauses, a clause flag, a boolean flag 
 RETURNS: A list with all clauses whose clause flag <Flag>
          has the value <MARKED>
 EFFECTS: -                                                     
***************************************************************/
{
  LIST Res;

  Res = list_Nil();
  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses))
    if (Marked == clause_GetFlag(list_Car(Clauses), Flag))
      Res = list_Cons(list_Car(Clauses), Res);
  
  return(Res);
}





static BOOL pcheck_ClauseIsUnmarked(CLAUSE C)
/**************************************************************
 INPUT  : A clause                                                      
 RETURNS: The value of the clauses MARKED flag                                                     
 EFFECTS: -                                                     
***************************************************************/
{
  return(!clause_GetFlag(C, MARKED));
}


static void pcheck_RemoveUnmarkedFromTableau(TABLEAU T)
/**************************************************************
 INPUT  : A tableau
 RETURNS: -
 EFFECTS: Delete all clauses that have the MARKED flag
          set from tableau.
***************************************************************/
{

  if (tab_IsEmpty(T))
    return;
  
  tab_SetClauses(T, list_DeleteElementIf(tab_Clauses(T), pcheck_ClauseIsUnmarked, pcheck_Nop));
  
  pcheck_RemoveUnmarkedFromTableau(tab_LeftBranch(T));
  pcheck_RemoveUnmarkedFromTableau(tab_RightBranch(T));

}

static void pcheck_CollectUnmarkedSplits(TABLEAU T, LIST* Splits)
/**************************************************************
 INPUT  : A tableau, a list of clauses by reference 
 RETURNS: -
 EFFECTS: Add all split clauses in the tableau that are not
          marked to the <Splits> list.
***************************************************************/
{
  LIST Scan;

  if (tab_IsEmpty(T))
    return;

  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (!clause_GetFlag(list_Car(Scan), MARKED) && clause_IsFromSplitting(list_Car(Scan))) 
      (*Splits) = list_Cons(list_Car(Scan), *Splits);  }
  
  pcheck_CollectUnmarkedSplits(tab_LeftBranch(T), Splits);
  pcheck_CollectUnmarkedSplits(tab_RightBranch(T), Splits);
}


void pcheck_TableauSplitsComplete(TABLEAU T)
/**************************************************************
 INPUT  : A tableau                                       
 RETURNS: Checks that every split has exactly two or no 
          successors. This condition must be true after tab_RemoveRedundantSplits
	  has been called.
 EFFECTS: -                                                     
***************************************************************/
{
  if (tab_IsEmpty(T))
    return;
  
  if (tab_RightBranchIsEmpty(T) && !tab_LeftBranchIsEmpty(T)) {
    
      fprintf(stderr, "\nerror: split of clause ");
      clause_PParentsFPrint(stderr, tab_SplitClause(T));
      fprintf(stderr, "\nhas no right branch.\n");
      misc_Error(); }
 
  if (!tab_RightBranchIsEmpty(T) && tab_LeftBranchIsEmpty(T)) {
    fprintf(stderr, "\nerror: split of clause ");
    clause_PParentsFPrint(stderr, tab_SplitClause(T));
    fprintf(stderr, "\nhas no left branch.\n");
    misc_Error(); }
  

  pcheck_TableauSplitsComplete(tab_LeftBranch(T));
  pcheck_TableauSplitsComplete(tab_RightBranch(T));
}

static void pcheck_RightSplitParents(CLAUSE SplitClause, CLAUSE RightSplitClause, CLAUSE LeftSplitClause)
/**************************************************************
 INPUT  : A split clause, and it left and right successors 
 RETURNS: -
 EFFECTS: Prints an error message if                                                      
***************************************************************/
{
  LIST Scan;
  BOOL HasEmpty, ContainsSplitClause;

  HasEmpty = ContainsSplitClause = FALSE;

  for (Scan = clause_ParentClauses(RightSplitClause);
  !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (clause_IsEmptyClause(list_Car(Scan)))
      HasEmpty = TRUE;
    if (clause_Number(list_Car(Scan)) == clause_Number(SplitClause)) 
      ContainsSplitClause = TRUE; 
  }
  
  if (!HasEmpty) {
    fprintf(stderr, "error: right split clause does not contain any empty clause as parent.\n");
    fprintf(stderr, "split clause: ");
    clause_PParentsFPrint(stderr, SplitClause);
    fprintf(stderr, "\nright clause: ");
    clause_PParentsFPrint(stderr, RightSplitClause);
    fprintf(stderr, "\n");
    misc_Error();  }

  if (!ContainsSplitClause) {
    fprintf(stderr, "error: right split clause does not have its split parent as a parent clause.\n");
    fprintf(stderr, "split clause: ");
    clause_PParentsFPrint(stderr, SplitClause);
    fprintf(stderr, "\nright clause: ");
    clause_PParentsFPrint(stderr, RightSplitClause);
    fprintf(stderr, "\n");
    misc_Error();  }
  
}


static void pcheck_SplitFormats(TABLEAU T)
/**************************************************************
 INPUT  : A tableau                                                      
 RETURNS: TRUE iff all splits:
          - have generated negations of the left split clause
	    if the left split clause is ground
	  - the conditions for right split clauses are
            as demanded in pcheck_RightSplitParents  
 EFFECTS: -                                                     
***************************************************************/
{
  LIST Scan;

  if (tab_IsEmpty(T))
    return;

  /* for right splits, check that parents have an empty clause and the split clause as a parent */

  for (Scan = tab_RightSplitClauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    pcheck_RightSplitParents(tab_SplitClause(T), list_Car(Scan), tab_LeftSplitClause(T)); }
    

  pcheck_SplitFormats(tab_RightBranch(T));
  pcheck_SplitFormats(tab_LeftBranch(T));
    
}

static void pcheck_SplitLevels(TABLEAU T)
/**************************************************************
 INPUT  : A Tableau                                                      
 RETURNS: TRUE iff all clauses in the tableau that
          are not splitting clauses have the max split
	  level of their parents.
 EFFECTS: -                                                     
 CAUTION: We assume that <T> has correct and complete split
          entries. See pcheck_SplitToProblems.
***************************************************************/
{
  LIST   Scan;
  CLAUSE Clause;
  int    CorrectLevel;

  if (tab_IsEmpty(T))
    return;

  /* check split levels */
  
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = list_Car(Scan);
    if (!list_Empty(clause_ParentClauses(Clause))
	&& !clause_IsFromSplitting(Clause)) {
 
     CorrectLevel = pcheck_MaxParentSplitLevel(Clause);
     if (clause_SplitLevel(Clause) != CorrectLevel) {
	fprintf(stderr, "error: split level of clause\n");
	clause_PParentsFPrint(stderr, Clause);
	fprintf(stderr, "\nshould be %d.\n", CorrectLevel); 
        misc_Error();}
    }
  }

  pcheck_SplitLevels(tab_RightBranch(T));
  pcheck_SplitLevels(tab_LeftBranch(T));
}

static void pcheck_SplitPrecheck(TABLEAU T)
/**************************************************************
 INPUT  : A tableau
 RETURNS: -
 EFFECTS: Stops and prints an error message if a left half
          of a split does not subsume its parents, or
	  if negations have been generated for a non-ground
	  left split clause.
***************************************************************/
{
  if (tab_IsEmpty(T))
    return;
  
  if (!subs_Subsumes(tab_LeftSplitClause(T), tab_SplitClause(T),-1,-1)) {
    fprintf(stderr, "error: incorrect split.\n");
    fprintf(stderr, "left half of split does not subsume splitted clause.\n");
    fprintf(stderr, "splitted clause: ");
    clause_PParentsFPrint(stderr, tab_SplitClause(T));
    fprintf(stderr, "\nleft split half: ");
    clause_PParentsFPrint(stderr, tab_LeftSplitClause(T));
    fprintf(stderr, "\n"); 
    misc_Error(); }

  if (list_Length(tab_RightSplitClauses(T)) > 1 &&
	!pcheck_ClauseIsGround(tab_LeftSplitClause(T))) {
    fprintf(stderr, "error: incorrect split.\n");
    fprintf(stderr, "non-ground split generated more than two clauses.\n");
    fprintf(stderr, "splitted clause: ");
    clause_PParentsFPrint(stderr, tab_SplitClause(T));
    fprintf(stderr, "\nleft split half: ");
    clause_PParentsFPrint(stderr, tab_LeftSplitClause(T));
    fprintf(stderr, "Right split clauses:\n"); 
    clause_PParentsListFPrint(stderr, tab_RightSplitClauses(T));
    fprintf(stderr, "\n");
    misc_Error(); }

  pcheck_SplitPrecheck(tab_LeftBranch(T));
  pcheck_SplitPrecheck(tab_RightBranch(T));

}

BOOL pcheck_BuildTableauFromProof(LIST Proof, TABLEAU* Tableau)
/**************************************************************
 INPUT  : A list of clauses representing a proof,
          a pointer to a tableau
 RETURNS: A tableau representing the proof in <Tableau>,
          TRUE iff no errors occurred.
 EFFECTS: Errors are commented when they occur.                                                     
***************************************************************/
{

  LIST     ProofRest;
  TABLEAU  SplitPos;
  TABPATH  Path;
  CLAUSE   Clause;
  int      ClauseLevel, SplitLevel;
  int      ProofDepth;

  if (list_Empty(Proof)) {
    *Tableau = tab_EmptyTableau();
    return(TRUE); }

  ProofDepth = pcheck_MaxClauseLevel(Proof);
  *Tableau   = tab_CreateNode();
  Path       = tab_PathCreate(ProofDepth, *Tableau);

  ProofRest = Proof;
  while (!list_Empty(ProofRest)) {

    SplitLevel  = tab_PathLength(Path);
    Clause      = list_Car(ProofRest);
    ClauseLevel = clause_SplitLevel(Clause); 

    if (pcheck_ClauseIsFromSplit(Clause)) {

      if (ClauseLevel == 0) {
	fprintf(stderr, "\nerror:level of split clause is 0.");
	fprintf(stderr, "\nclause: "); 
	clause_PParentsFPrint(stderr, Clause);
	fprintf(stderr, "\n");
	return(FALSE); }

      if (ClauseLevel > SplitLevel+1) {
	fprintf(stderr, "\nerror: split level of clause\n");
	clause_PParentsFPrint(stderr, Clause);
	fprintf(stderr, "\nis higher than current split level+1: %d\n", SplitLevel); 
	return(FALSE);}


      SplitPos = tab_PathNthNode(Path, ClauseLevel-1);

      if (pcheck_ClauseIsFromLeftSplit(Clause)) {
	if (!tab_LeftBranchIsEmpty(SplitPos)) {
	  fprintf(stderr, "\nerror: multiple left splits on level %d", ClauseLevel);
	  fprintf(stderr, "\nformer split clause: ");
	  clause_PParentsFPrint(stderr, tab_SplitClause(SplitPos));
	  fprintf(stderr, "\nformer left split clause: ");
	  clause_PParentsFPrint(stderr, tab_LeftSplitClause(SplitPos));
	  fprintf(stderr, "\ncurrent clause:");
	  clause_PParentsFPrint(stderr, Clause);
	  fprintf(stderr, "\n");
	  return(FALSE); }
	
	else {
	  Path = tab_PathPrefix(ClauseLevel-1, Path);
	  tab_SetSplitClause(SplitPos, list_Car(clause_ParentClauses(Clause)));
	  tab_SetLeftSplitClause(SplitPos, Clause);
	  tab_AddSplitAtCursor(Path, TRUE); }
      }
      
      /* now we have a right split */

      else {
	if (tab_RightBranchIsEmpty(SplitPos)) {
	  
	  if (tab_LeftBranchIsEmpty(SplitPos)) {
	    fprintf(stderr, "\nerror: right split without left split on level %d\n", ClauseLevel); 
	    fprintf(stderr, "clause: ");
	    clause_PParentsFPrint(stderr, Clause);
	    fprintf(stderr, "\n");
	    return(FALSE); } 
	  
	  else { 
	    Path = tab_PathPrefix(ClauseLevel-1, Path);
	    tab_AddSplitAtCursor(Path, FALSE); }
	}
	tab_AddRightSplitClause(SplitPos, Clause); 
      } /* clause from right split */
    } /* clause from split */
      
    if (ClauseLevel > tab_PathLength(Path)) {
      fprintf(stderr, "error: split level of clause\n");
      clause_PParentsFPrint(stderr, Clause);
      fprintf(stderr, "\nis greater than current split level %d\n", tab_PathLength(Path));
      return(FALSE); }
    tab_AddClauseOnItsLevel(Clause, Path); 
    
    ProofRest = list_Cdr(ProofRest);   
  } /* while proof is not empty */


  tab_PathDelete(Path);

  return(TRUE);
}


BOOL pcheck_TableauJustificationsRec(TABLEAU T, TABPATH Path)
/**************************************************************
 INPUT  : A Tableau <T>, a <Path> in the tableau                                                   
 RETURNS: TRUE iff all clauses in the last node of the
          path are correctly justified.
 EFFECTS: -                                                     
***************************************************************/
{
  CLAUSE  Clause;
  LIST    ScanClauses;
  LIST    ScanParents;
  LIST    Parents;
  CLAUSE  Parent;
  BOOL    Ok, RightSplit;

  if (tab_IsEmpty(T))
    return(TRUE);

  Ok = TRUE; 

  /* for each clause, check that its parents have been justified */

  for (ScanClauses = tab_Clauses(tab_PathTop(Path));
       !list_Empty(ScanClauses); ScanClauses = list_Cdr(ScanClauses)) {

    Clause  = list_Car(ScanClauses);
    Parents = clause_ParentClauses(Clause);
    
    RightSplit = pcheck_ClauseIsFromRightSplit(Clause); 
    
    /* check all parents */
    
    for (ScanParents = Parents; !list_Empty(ScanParents); 
	 ScanParents = list_Cdr(ScanParents)) {
      
      Parent = list_Car(ScanParents);
      
      if ((!(RightSplit && clause_IsEmptyClause(Parent)) &&
	  !(RightSplit && pcheck_ClauseIsFromLeftSplit(Parent))) ||
	  (clause_Number(Parent) > clause_Number(Clause)) ) {
	if (!tab_PathContainsClause(Path, Parent)) {
	  fprintf(stderr,"\nerror: on level %d:\n", tab_PathLength(Path));
#ifdef USE_CG
	  fprintf(stderr,"in node with label: %d", T->Label);
#endif
	  fprintf(stderr,"\nin clause:"); 
	  clause_PParentsFPrint(stderr, list_Car(ScanClauses)); 
	  fprintf(stderr,"\nparent clause with number %d not yet justified.\n", clause_Number(Parent));
	  Ok = FALSE; } } 
    }
  }  /* for all clauses in current node */
  
  
  /* Recursion */
  
  if (!tab_LeftBranchIsEmpty(T)) {
    Path = tab_PathPush(tab_LeftBranch(T), Path);
    Ok   = Ok && pcheck_TableauJustificationsRec(tab_LeftBranch(T), Path);
    Path = tab_PathPop(Path); }
  
  if (!tab_RightBranchIsEmpty(T)) {
    Path = tab_PathPush(tab_RightBranch(T), Path);
    Ok   = Ok && pcheck_TableauJustificationsRec(tab_RightBranch(T), Path);
    Path = tab_PathPop(Path); }

  return(Ok);
}

static BOOL pcheck_TableauJustifications(TABLEAU T)
/**************************************************************
 INPUT  : A Tableau                                                     
 RETURNS: TRUE iff for each clause in tableau, all its parent
          clauses have been derived in a split on the same level
	  or below.
 EFFECTS: -                                                     
***************************************************************/
{
  TABPATH Path;
  BOOL    Ok;

  Path = tab_PathCreate(tab_Depth(T),T);
  Ok   = pcheck_TableauJustificationsRec(T,Path); 
  tab_PathDelete(Path);
  
  return(Ok);
}

BOOL pcheck_TableauProof(TABLEAU* Tableau, LIST Proof)
/**************************************************************
 INPUT  : 
 RETURNS:                                                      
 EFFECTS:                                                      
***************************************************************/
{
  LIST RedundantClauses;
  LIST EmptyClauses;
  LIST UnmarkedSplits;

#ifdef USE_CG
  tab_LabelNodes(*Tableau);
  /* print out current tableau */
  if (pcheck_GenNamedCg) 
    tab_WriteTableau(*Tableau, pcheck_CgName, pcheck_GraphFormat);
#endif
  
  RedundantClauses = list_Nil();
  if (!pcheck_Quiet) {
    printf("pruning closed branches..."); fflush(stdout); }
  (*Tableau) = tab_PruneClosedBranches(*Tableau, &RedundantClauses);   /* delete descendants of already closed branches */
  if (!pcheck_Quiet) {
    printf("finished.\n"); }

  if (!pcheck_Quiet) {
    printf("removing incomplete splits..."); fflush(stdout); }
  (*Tableau) = tab_RemoveIncompleteSplits(*Tableau, &RedundantClauses);  /* reduce open node redundancies */
  if (!pcheck_Quiet) {
    printf("finished.\n"); } 
  
  list_Delete(RedundantClauses);

  /* remove all clauses that are not needed for the empty clauses of the proof or
     for the tableau structure (splits) */

  EmptyClauses = list_Nil();
  tab_GetEarliestEmptyClauses(*Tableau, &EmptyClauses);
  pcheck_ClauseListRemoveFlag(Proof, MARKED);
  pcheck_MarkRecursive(EmptyClauses);
  UnmarkedSplits = list_Nil();
  pcheck_CollectUnmarkedSplits(*Tableau, &UnmarkedSplits);
  pcheck_MarkRecursive(UnmarkedSplits);
  pcheck_RemoveUnmarkedFromTableau(*Tableau); 
  list_Delete(UnmarkedSplits);
  list_Delete(EmptyClauses);

#ifdef USE_CG
  /* print reduced graph */
  if (pcheck_GenRedCg) 
    tab_WriteTableau(*Tableau, pcheck_RedCgName, pcheck_GraphFormat);
#endif 

  tab_SetSplitLevels(*Tableau);
  pcheck_SplitLevels(*Tableau);
  tab_CheckEmpties(*Tableau);

 if (!tab_IsClosed(*Tableau)) {
    printf("\nerror: tableau is not closed.\n");  
    return(FALSE); }

  /* check justifications */
  if (!pcheck_Quiet) {
    printf("checking justifications..."); fflush(stdout); }
  if (!pcheck_TableauJustifications(*Tableau))
    return(FALSE);
  if (!pcheck_Quiet) {
    printf("finished.\n"); } 
 
  return(TRUE);
}




void pcheck_MarkRecursive(LIST Clauses)
/**************************************************************
 INPUT  : A list of clauses 
 RETURNS: -                                                      
 EFFECTS: Marks all <Clauses> and its ancestors with the
          MARKED clause flag.
***************************************************************/
{
  CLAUSE Clause;

  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) {
    Clause = list_Car(Clauses);
    if (!clause_GetFlag(Clause, MARKED)) 
      { pcheck_MarkRecursive(clause_ParentClauses(Clause)); 
      clause_SetFlag(Clause, MARKED);  }
  }
}


static void pcheck_CollectTermVariables(TERM Term, LIST* Vars)
/**************************************************************
 INPUT  :  
 RETURNS: The list with variable terms in term
 EFFECTS:                                                      
***************************************************************/
{
  LIST Scan;

  if (term_IsComplex(Term)) 
    for (Scan = term_ArgumentList(Term); !list_Empty(Scan);
	 Scan = list_Cdr(Scan))
      pcheck_CollectTermVariables(list_Car(Scan), Vars);
  else if (term_IsVariable(Term))   { 
    if (!list_Member(*Vars, Term, term_Equal))
      (*Vars) = list_Cons(term_Copy(Term), *Vars); }
  
}


static BOOL pcheck_IsRightSplitHalf(CLAUSE C)
/**************************************************************
 INPUT  : A clause                                                      
 RETURNS: TRUE iff the following conditions are fulfilled:
          - the first parent clause is an empty clause 
	  - the clause subsumes its third parent
 EFFECTS: -                                                 
***************************************************************/
{
  LIST Parents;
  BOOL Ok;

  Parents = list_Copy(clause_ParentClauses(C));
  Parents = list_PointerDeleteDuplicates(Parents);

  Ok = FALSE;
  if (list_Length(Parents) == 2) {
    if (clause_IsEmptyClause(list_First(Parents)))
      Ok = subs_Subsumes(C, list_Second(Parents),-1,-1); }
  list_Delete(Parents);

  return(Ok);
}



static BOOL pcheck_ClauseIsGround(CLAUSE C)
/**************************************************************
 INPUT  : A clause
 RETURNS: TRUE iff <C> is ground
 EFFECTS: -                                                      
***************************************************************/
{
  int i;
  LIST Vars;

  Vars = list_Nil();
  for (i = clause_FirstLitIndex(); i <= clause_LastLitIndex(C); i++) {
    pcheck_CollectTermVariables(clause_LiteralAtom(clause_GetLiteral(C,i)), &Vars);
    if (!list_Empty(Vars)) {
      term_ListDelete(Vars);
      return(FALSE);}
  }
    
  return(TRUE);
}

static TERM pcheck_UnivClosure(TERM T)
/**************************************************************
 INPUT  : A term, representing a formula
 RETURNS: The universal closure of the term.
 EFFECTS: <T> is part of the returned term!
***************************************************************/
{
  LIST FreeVars;

  FreeVars = list_Nil();
  pcheck_CollectTermVariables(T, &FreeVars);
  
  if (list_Empty(FreeVars))
    return(T);
  return(fol_CreateQuantifier(fol_All(), FreeVars, list_List(T)));
}

static TERM pcheck_ClauseToTerm(CLAUSE Clause)
/**************************************************************
 INPUT  : A clause 
 RETURNS: The clause as a term                                                      
 EFFECTS: -                                                     
***************************************************************/
{
  int  LitScan;
  LIST Args;
  TERM Lit;
  TERM ClauseTerm;

  Args = list_Nil();

  for (LitScan = clause_FirstLitIndex(); LitScan <= clause_LastLitIndex(Clause);
       LitScan++) {
    Lit = clause_LiteralSignedAtom(clause_GetLiteral(Clause, LitScan));
    Args = list_Cons(term_Copy(Lit), Args); }

  if (list_Empty(Args))
    Args = list_List(term_Create(fol_False(), list_Nil()));
  
  ClauseTerm = term_Create(fol_Or(), Args);
  ClauseTerm = pcheck_UnivClosure(ClauseTerm);

  return(ClauseTerm);
}

static LIST pcheck_ClauseListToTermList(LIST Clauses)
/**************************************************************
 INPUT  : A list of clauses                                                     
 RETURNS: The clauses as terms                                                     
 EFFECTS: -                                                     
***************************************************************/
{
  LIST Terms;

  Terms = list_Nil();
  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses))
    Terms = list_Cons(pcheck_ClauseToTerm(list_Car(Clauses)), Terms);
  
  return(Terms);
}

static void pcheck_SaveNumberedDFGProblem(int Number, LIST Axioms, LIST Conjectures, string ProofFileName, string DestPrefix)
/**************************************************************
 INPUT  : A (clause) number, a list of axioms and conjecturs,
          and a filename (of the proof file of the currently
	  checked proof)
 RETURNS: -                                                     
 EFFECTS: Saves a DFG file containing <Axioms> and <Conjectures> 
          under the name "<DestPrefix><Number>_<ProofFileName>" 
***************************************************************/
{
  char*     Filename;
  char*     Tmp;
  char      NumStr[100];
  FILE*     File;
  

  pcheck_IntToStr(Number, NumStr);
  Tmp      = pcheck_GenericFilename(ProofFileName, NumStr);
  Filename = pcheck_StrConc(DestPrefix, Tmp);
    
  File = misc_OpenFile(Filename, "w");
  pcheck_SaveAsDFG2(File, Axioms, Conjectures);
  misc_CloseFile(File, Filename);

  pcheck_StrDelete(Tmp);
  pcheck_StrDelete(Filename);
}

static void pcheck_SplitToProblems(TABLEAU T, string ProofFileName, string DestPrefix)
/**************************************************************
 INPUT  : 
 RETURNS:                                                       
 EFFECTS:                                                      
 CAUTION: We assume that we get non-null clauses when calling 
          tab_SplitClause and tab_LeftSplitClause.
***************************************************************/
{
  TERM SplitClauseTerm;
  TERM LeftClauseTerm, RightClauseTerm;
  TERM Equiv, Disj, Tmp;
  
  LIST   Conj, Args, Negations;

  SplitClauseTerm = pcheck_ClauseToTerm(tab_SplitClause(T));
  LeftClauseTerm  = pcheck_ClauseToTerm(tab_LeftSplitClause(T));
  
  /* by default, take all right split clauses as negations 
   * if the first clause is the second half of a split clause,
   * take only the rest of the right split clauses as negations. 
   */

  Negations = tab_RightSplitClauses(T);
  if (!list_Empty(Negations) && pcheck_IsRightSplitHalf(list_Car(Negations))) {
    Negations = list_Cdr(Negations);
    /* build C <=> C' v C''  */
    RightClauseTerm = pcheck_ClauseToTerm(list_Car(tab_RightSplitClauses(T)));
    Disj  = term_Create(fol_Or(), list_Cons(LeftClauseTerm, list_Cons(RightClauseTerm, list_Nil())));
    Equiv = term_Create(fol_Equiv(), list_Cons(SplitClauseTerm, list_Cons(Disj, list_Nil())));
    Conj  = list_List(Equiv);
    
    pcheck_SaveNumberedDFGProblem(clause_Number(tab_LeftSplitClause(T)), list_Nil(), 
				  Conj, ProofFileName, DestPrefix);
    term_ListDelete(Conj); }
  
  Args = list_Nil();

  /* build conjunction of negations, if there are any. */
  if (!list_Empty(Negations)) {

    LeftClauseTerm  = pcheck_ClauseToTerm(tab_LeftSplitClause(T));
    Args  = pcheck_ClauseListToTermList(Negations);
    Tmp   = term_Create(fol_And(), Args);
    Tmp   = term_Create(fol_Not(), list_List(Tmp));
    Equiv = term_Create(fol_Implies(), list_Cons(Tmp, list_Cons(LeftClauseTerm, list_Nil()))); 
    Conj  = list_List(Equiv);
    
    /* problem id is number of right part of split clause, if it exists, of of first negation */
    pcheck_SaveNumberedDFGProblem(clause_Number(list_Car(tab_RightSplitClauses(T))), 
				  list_Nil(), Conj, ProofFileName, DestPrefix);
    term_ListDelete(Conj);
  }

}

void pcheck_TableauToProofTask(TABLEAU T, string ProofFileName, string DestPrefix)
/**************************************************************
 INPUT  : A Tableau, two strings for filename generation  
 RETURNS: -                                                      
 EFFECTS: Generates a DFG problem files for each clause in the tableau.
          The problem asserts that the clause follows from
	  its parents. For splits, see pcheck_SplitToProblems
***************************************************************/
{
  LIST   Scan;
  LIST   Axioms, Conj;
  CLAUSE Clause;

  if (tab_IsEmpty(T))
    return;

  /* treat split clauses */
  if (!tab_IsLeaf(T)) {
    pcheck_SplitToProblems(T, ProofFileName, DestPrefix);  }
 
  /* treat other clauses */
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {

    Clause = list_Car(Scan);
    if (!clause_IsFromSplitting(Clause) && 
	!list_Empty(clause_ParentClauses(Clause))) {
      Axioms = pcheck_ClauseListToTermList(clause_ParentClauses(Clause));
      Conj   = list_List(pcheck_ClauseToTerm(Clause)); 
      pcheck_SaveNumberedDFGProblem(clause_Number(Clause), Axioms, Conj, 
				    ProofFileName, DestPrefix);  
      term_ListDelete(Axioms);
      term_ListDelete(Conj); }
  } 
  
  
  /* recursion */
  pcheck_TableauToProofTask(tab_RightBranch(T), ProofFileName, DestPrefix);
  pcheck_TableauToProofTask(tab_LeftBranch(T), ProofFileName, DestPrefix);
  
}


int pcheck_SeqProofDepth(LIST Proof)
/**************************************************************
 INPUT  : A sequential proof (list of clauses) 							       
 RETURNS: The maximum clause depth in the proof
 EFFECTS: -
***************************************************************/ 
{
  int  Max;

  Max = 0;
  
  for (; !list_Empty(Proof); Proof = list_Cdr(Proof)) 
    if (clause_Depth(list_Car(Proof)) > Max)
      Max = clause_Depth(list_Car(Proof));
  
  return(Max);
}


LIST pcheck_ReduceSPASSProof(LIST Proof)
/**************************************************************
 INPUT  : A list of clauses representing a SPASS proof.
          Parents are pointers.
 RETURNS: A list of clauses were incomplete splits
          and closed branches with descendants have been
	  removed.
 EFFECTS: -                                                     
***************************************************************/  
{
  LIST    EmptyClauses, RedundantClauses;
  LIST    ReducedProof;
  TABLEAU Tableau;
  LIST    UnmarkedSplits;

  if (!pcheck_BuildTableauFromProof(Proof, &Tableau))
    misc_Error();

  RedundantClauses = list_Nil();
  Tableau = tab_PruneClosedBranches(Tableau, &RedundantClauses);
  Tableau = tab_RemoveIncompleteSplits(Tableau, &RedundantClauses);
  list_Delete(RedundantClauses);
  
  tab_SetSplitLevels(Tableau);
  
  /* 
   *  get minimal proof: First find earliest derived empty clauses,
   *  then recursively mark ancestors of these clauses. Put
   *  only marked clauses in <ReducedProof>.
   */

  EmptyClauses = list_Nil();
  tab_GetEarliestEmptyClauses(Tableau, &EmptyClauses);
  pcheck_ClauseListRemoveFlag(Proof, MARKED);
  pcheck_MarkRecursive(EmptyClauses);
  UnmarkedSplits = list_Nil();
  pcheck_CollectUnmarkedSplits(Tableau, &UnmarkedSplits);
  pcheck_MarkRecursive(UnmarkedSplits);
  pcheck_RemoveUnmarkedFromTableau(Tableau); 
  list_Delete(UnmarkedSplits);

  ReducedProof = list_Nil(); 
  tab_ToClauseList(Tableau, &ReducedProof);
  ReducedProof = pcheck_ClauseNumberMergeSort(ReducedProof);

  tab_Delete(Tableau);
  list_Delete(EmptyClauses); 

  return(ReducedProof);
}



void pcheck_DeleteProof(LIST Proof)
/**************************************************************
 INPUT:   A Proof                                       
 RETURNS: -                                             
 EFFECTS: Frees memory associated with proof            
 **************************************************************/
{
  LIST Line, Scan2, Scan1;
  
  Scan1 = Proof;
  while (!list_Empty(Scan1)) {
    Line = list_Car(Scan1);
    
    dfg_StringFree(list_Car(Line));
    if (list_Second(Line) != clause_Null()) {                /* clause */
      term_Delete(list_Second(Line)); }

    /* delete labels in justification list and list itself */

    for(Scan2 = list_Third(Line); !list_Empty(Scan2); Scan2 = list_Cdr(Scan2)) 
      dfg_StringFree(list_Car(Scan2));
    list_Delete(list_Third(Line));     

    /* now contents of line are deleted. Delete line. */
    list_Delete(Line);     
    Scan1 = list_Cdr(Scan1);  }
 
 list_Delete(Proof);
}


string pcheck_GenericFilename(string Filename, string Id)
/**************************************************************
 INPUT  : Two strings                                   
 RETURNS: A string with Suffix as new extension to      
          of Filename                                   
          (Filename = name.ext -> name_<Id>.prf)      
 EFFECTS: -                                             
 **************************************************************/
{	
  string Help1, Help2, Help3, Res;
  int i;

  Help1 = pcheck_StrConc("_", Id); 
  Help2 = pcheck_StrConc(Help1, pcheck_ProofFileSuffix); 
  
  /* remove filename extension */
  for (i = 0; Filename[i] != '.' && i < pcheck_StrLength(Filename); i++) ;
  Help3 = pcheck_StrPrefix(Filename, i);

  Res = pcheck_StrConc(Help3, Help2);
  pcheck_StrDelete(Help1);
  pcheck_StrDelete(Help2); 
  pcheck_StrDelete(Help3); 

  return(Res);
}

static void pcheck_StrReverse(char *s)
/**************************************************************
 INPUT  : A string                                                      
 RETURNS: -                                                     
 EFFECTS: The string is reversed                                                     
***************************************************************/
{
    int i, j;
    char temp;

    for (i = 0, j = strlen(s)-1; i<j; i++, j--) {
	temp = s[i];
	s[i] = s[j];
	s[j] = temp;
	}
}  

void pcheck_IntToStr(int i, char* s)
/**************************************************************
 INPUT  : An integer, a string by reference                                                     
 RETURNS: -                                                     
 EFFECTS: Converts i to a number in <s>                                                     
***************************************************************/
{
  int j, sign;

  if ((sign = i) < 0)
    i = -i;

  j = 0;
  if (i == 0)
    s[j++] = '0';
  else {
    while (i > 0) {
      s[j++] = i % 10 + '0';
      i = i / 10;
    }
  }
  if (sign < 0)
    s[j++] = '-';
  s[j] = '\0';
  pcheck_StrReverse(s);
} 


static string pcheck_StrConc(char* s1, char* s2)
/**************************************************************
 INPUT: Two strings                                     
 RETURNS: A new string s1.s2                            
 EFFECTS: Malloc for new string                         
 **************************************************************/
{ 
  string dst;

  dst = memory_Malloc(pcheck_StrLength(s1)+pcheck_StrLength(s2)+1);
  strcpy(dst, s1);
  return(strcat(dst,s2)); 

}

static string pcheck_StrEmpty()
/**************************************************************
 INPUT:   -                                             
 RETURNS: a string of length 0                          
 EFFECTS: Malloc for string                             
 **************************************************************/
{ string s ;

  s = memory_Malloc(1);
  s[0] = '\0';
  return(s);
}

string pcheck_StrPrefix(string s, int i)
/**************************************************************
 INPUT: A string and a string length                    
 RETURNS: The prefix of s of length i                   
 EFFECTS: -                                             
 **************************************************************/
{
  string dst;

  dst = memory_Malloc(i+1);
  strncpy(dst, s, i);
  dst[i] = '\0';
  return(dst);
}


string pcheck_StrSuffix(string s, int i)
/**************************************************************
 INPUT: A string and a string length                    
 RETURNS: The string that results from cutting the first
          i characters from string s.                   
          Returns empty string if i>= length(s)         
 EFFECTS: -                                             
 **************************************************************/
{
  int l;
  string dst;
  
  l =  pcheck_StrLength(s);
  if (i >= l)
    return(pcheck_StrEmpty());

  dst = memory_Malloc(l-i+1);
  strcpy(dst, s+i);
  return(dst);
}

void pcheck_StrDelete(string s)
/**************************************************************
 INPUT: A string                                        
 EFFECTS: Frees memory occupied by string               
 **************************************************************/
{ 
  memory_Free(s, strlen(s)+1); 
} 

int pcheck_StrLength(string s)
/**************************************************************
 INPUT: A string                                        
 RETURNS: Length of string                              
 EFFECTS: -                                             
 **************************************************************/
{ 
return(strlen(s));
}

void pcheck_ClauseListSetFlag(LIST Clauses, CLAUSE_FLAGS Flag)
/**************************************************************
 INPUT:   A list of clauses, a clause flag
 RETURNS: -                              
 EFFECTS: Sets the <Flag> in all clauses in the list                                             
 **************************************************************/
{
  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) 
    clause_SetFlag(list_Car(Clauses), Flag);

}

void pcheck_ClauseListRemoveFlag(LIST Clauses, CLAUSE_FLAGS Flag)
/**************************************************************
 INPUT:   A list of clauses, a clause flag
 RETURNS: -                              
 EFFECTS: Removes the <Flag> in all clauses in the list                                             
 **************************************************************/
{
  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) 
    clause_RemoveFlag(list_Car(Clauses), Flag);

}

NAT pcheck_ClauseNumber(CLAUSE C)
/**************************************************************
 INPUT  : A clause
 RETURNS: Its number                                                     
 EFFECTS: -                                                     
***************************************************************/
{
  return(clause_Number(C));
}

LIST pcheck_ClauseNumberMergeSort(LIST L)
/**************************************************************
  INPUT:   A list of clauses  
  RETURNS: The sorted list: clause_Number(L[i]) < clause_Number(L[i+1])
  EFFECTS: Destructive
***************************************************************/
{
  LIST SingleStep, DoubleStep, Prev;
  LIST L1, L2;

  if (list_Length(L) < 2)
    return(L);

  
  /* divide list in two halves */
  Prev = L;
  SingleStep = list_Cdr(L);
  DoubleStep = list_Cdr(SingleStep);
  
  while (!list_Empty(DoubleStep) && !list_Empty(list_Cdr(DoubleStep))) {
    Prev       = SingleStep;
    SingleStep = list_Cdr(SingleStep);
    DoubleStep = list_Cdr(list_Cdr(DoubleStep)); }
  
  L1 = L;
  L2 = SingleStep;
  list_Rplacd(Prev, list_Nil());

  L = list_NNumberMerge(pcheck_ClauseNumberMergeSort(L1), pcheck_ClauseNumberMergeSort(L2), pcheck_ClauseNumber);

  return(L);
}

static void PrintFormulaList(FILE* File, LIST Formulas, char* Name)
/**************************************************************
 INPUT:   A file, a list of formulas, a name              
 EFFECTS: Print a list formulas in DFG format,          
          with given list name.                         
 **************************************************************/
{
  LIST scan;

  fprintf(File, "list_of_formulae(%s).\n", Name);
  for (scan = Formulas; !list_Empty(scan); scan= list_Cdr(scan)) {
    fprintf(File, "  formula(");
    fol_FPrintDFG(File, list_Car(scan));
    fprintf(File, ").\n");
  }
  fprintf(File, "end_of_list.\n\n"); 
} 

static void pcheck_SaveAsDFG2(FILE* File, LIST Axioms, LIST Conjectures)
/**************************************************************
 INPUT: A file, two lists of formulas                   
 EFFECTS: Prints DFG file containing these lists        
          to file.                                      
 **************************************************************/
{
  fprintf(File, "begin_problem(Unknown).\n\n"); 

  fprintf(File, "list_of_descriptions.\n");
  fprintf(File, "name({*Sub Proof*}).\n");
  fprintf(File, "author({* Proof Checker *}).\n");
  fprintf(File, "status(unsatisfiable).\n");
  fprintf(File, "description({* Checks correctness of proof line *}).\n");
  fprintf(File, "end_of_list.\n\n");
  
  fprintf(File, "list_of_symbols.\n");
  fol_FPrintDFGSignature(File);
  fprintf(File, "end_of_list.\n\n");

  PrintFormulaList(File, Axioms, "axioms");
  PrintFormulaList(File, Conjectures, "conjectures");

  fprintf(File, "end_problem.\n");
}


static BOOL pcheck_StrToInt(char* s, int* np)
/**************************************************************
  INPUT: a string and an integer by reference            
  RETURNS: TRUE iff string represents some integer       
  EFFECTS: np takes value of represented integer         
**************************************************************/
{
    int i, sign, n;

    i = 0;
    sign = 1;
    if (s[0] == '+' || s[0] == '-') {
	if (s[0] == '-')
	    sign = -1;
	i = 1;
	}
    if (s[i] == '\0')
	return(0);
    else {
	n = 0;
	for( ; s[i] >= '0' && s[i] <= '9'; i++)
	    n = n * 10 + s[i] - '0';
	*np = n * sign;
	return(s[i] == '\0');
	}
}


