/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                  PARSER FOR PROLOG SYNTAX              * */
/* *                                                        * */
/* *  $Module:   pro                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.42 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 18:04:21 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: proparser.y,v $ */

%{

#include "dfg.h"
#include "symbol.h"
#include "term.h"
#include "foldfg.h"
#include "clause.h"


void yyerror(const char*);
int  yylex(void);		/* Defined in proscanner.l */

#define YYERROR_VERBOSE


NAT pro_LINENUMBER;  /* used also in the scanner */

static LIST pro_RULESLIST;
static TERM pro_QUESTIONTERM;
static LIST pro_SYMBOLLIST;
static LIST pro_VARLIST;

static void   pro_AppendRule(TERM, LIST);
static void   pro_StoreQuestion(LIST);
static TERM   pro_TermCreate(char*, LIST);
static TERM   pro_VarTermCreate(char*);
static TERM   pro_CheckAtom(TERM Atom);
static void   pro_CheckArguments(TERM);
static void   pro_RestartVarList(void);
static SYMBOL pro_VarLookup(char*);

static __inline__ void pro_AppendFact(TERM Term)
{
  Term = term_Create(fol_Or(), list_List(Term));
  pro_RULESLIST = list_Nconc(pro_RULESLIST, list_List(Term));
  pro_RestartVarList();
}

%}

%union {
  char*     string;
  TERM      term;
  LIST      list;
}

%token PRO_IF PRO_QUESTION
%token <string> PRO_ID PRO_VAR PRO_NUM 
%type  <string>  id
%type  <list>    atomlist termlist
%type  <term>    term atom

%%

/****************/
/* TOP RULE     */
/****************/

prologprogram:	list_of_rules
                | list_of_rules question
;

question:    PRO_QUESTION atomlist '.' { pro_StoreQuestion($2); }
;

list_of_rules:	/* empty */
		| list_of_rules rule_or_fact
;

rule_or_fact:	fact
		| rule
;

fact:		atom '.'		{ pro_AppendFact($1); }
;

rule:		atom PRO_IF atomlist '.' { pro_AppendRule($1,$3); }
;

atomlist:	atom			{ $$ = list_List($1); }
		| atomlist ',' atom	{ $$ = list_Nconc($1,list_List($3)); }
;

atom:		term			{ $$ = pro_CheckAtom($1); }
;

term:		id			{ $$ = pro_TermCreate($1,list_Nil()); }
		| PRO_VAR		{ $$ = pro_VarTermCreate($1); } 
		| PRO_ID '(' termlist ')' { $$ = pro_TermCreate($1, $3); }
;

termlist:	term			{ $$ = list_List($1); }
		| termlist ',' term	{ $$ = list_Nconc($1, list_List($3)); }
;

id:		PRO_ID			{ $$ = $1; }
		| PRO_NUM		{ $$ = $1; }
;


%%

void yyerror(const char *s)
{
  fprintf(stderr,"line %i: %s\n", pro_LINENUMBER, s);
  misc_Error();

}

static __inline__ CLAUSE pro_CreateAxiomClause(TERM Term)
{
  return dfg_CreateClauseFromTerm(Term, TRUE);
}

LIST pro_PrologParser(FILE* InputFile)
{
  extern FILE* pro_in;  /* declared in proscanner */

  pro_in           = InputFile;
  pro_LINENUMBER   = 1;
  pro_QUESTIONTERM = (TERM) NULL;
  pro_RULESLIST    = list_Nil();
  pro_SYMBOLLIST   = list_Nil();
  pro_VARLIST      = list_Nil();

  pro_parse();  /* Call the parser */

  /* Cleanup */
  list_Delete(pro_SYMBOLLIST);
  pro_RestartVarList();

  /* Create Clauses */
  list_NMapCar(pro_RULESLIST, (POINTER (*)())pro_CreateAxiomClause);
  if (pro_QUESTIONTERM != NULL) {
    pro_RULESLIST =
      list_Nconc(pro_RULESLIST,
		 list_List(dfg_CreateClauseFromTerm(pro_QUESTIONTERM, FALSE)));
  }
  return pro_RULESLIST;
}


static TERM pro_TermCreate(char* Name, LIST Arguments)
/* Look up the symbol, create the term and check arguments */
{
  SYMBOL s;
  TERM   result;
  NAT    length;

  /* Pay attention to the maximum symbol name length */
  length = strlen(Name);
  if (length >= symbol__SYMBOLMAXLEN)
    Name[symbol__SYMBOLMAXLEN-1] = '\0';

  s = symbol_Lookup(Name);
  if (s == 0)
    s = symbol_CreateFunction(Name, list_Length(Arguments), symbol_STATLEX);

  /* Check the symbol's arity */
  if (symbol_Arity(s) != list_Length(Arguments)) {
    fprintf(stderr, "line %i: Symbol '%s' was defined with arity %d\n",
	    pro_LINENUMBER, Name, symbol_Arity(s));
    misc_Error();
  }

  if (length >= symbol__SYMBOLMAXLEN) {
    /* To avoid a memory error restore the old string length */
    Name[symbol__SYMBOLMAXLEN-1] = ' ';  /* Something != '\0' */
  }
  misc_StringFree(Name);    /* was copied by symbol_CreateFunction */

  result = term_Create(s, Arguments);
  pro_CheckArguments(result);
      
  return result;
}

static TERM pro_VarTermCreate(char* Name)
{
  SYMBOL s;

  s = pro_VarLookup(Name);
  return term_Create(s, list_Nil());
}

static TERM pro_CheckAtom(TERM Atom)
{
  SYMBOL s;

  s = term_TopSymbol(Atom);
  if (!symbol_IsPredicate(s)) {
    if (list_PointerMember(pro_SYMBOLLIST, (POINTER)s)) {
      fprintf(stderr, "parse error line %i: '%s' is not a predicate\n",
	      pro_LINENUMBER, symbol_Name(term_TopSymbol(Atom)));
      misc_Error();
    } else {
      /* First occurence, so change it to a predicate */
      s = symbol_ChangeType(s, symbol_PREDICATE);
      pro_SYMBOLLIST = list_Cons((POINTER)s, pro_SYMBOLLIST);
      term_RplacTop(Atom, s);
    }
  }
  return Atom;
}

static void pro_CheckArguments(TERM Term)
{
  LIST scan;

  for (scan = term_ArgumentList(Term); !list_Empty(scan); scan = list_Cdr(scan)) {
    SYMBOL argsymbol = term_TopSymbol(list_Car(scan));
    if (symbol_IsPredicate(argsymbol)) {
      fprintf(stderr, "parse error line %i: Term has predicate term"
	      " as argument: ", pro_LINENUMBER);
      symbol_FPrint(stderr, term_TopSymbol(Term));
      fputs("(...,", stderr);
      symbol_FPrint(stderr, argsymbol);
      fputs("(),...)\n", stderr);
      misc_Error();
    }
    if (!symbol_IsVariable(argsymbol)) {
      /* Add <argsymbol> to the front of the symbol list */
      pro_SYMBOLLIST = list_PointerDeleteOneElement(pro_SYMBOLLIST,
						    (POINTER)argsymbol);
      pro_SYMBOLLIST = list_Cons((POINTER)argsymbol, pro_SYMBOLLIST);
    }
  }
}

static void pro_AppendRule(TERM Positive, LIST Negatives)
{
  LIST scan;

  for (scan = Negatives; !list_Empty(scan); scan = list_Cdr(scan))
    list_Rplaca(scan, term_Create(fol_Not(), list_List(list_Car(scan))));
  Negatives = list_Cons(Positive, Negatives);
  pro_RULESLIST = list_Cons(term_Create(fol_Or(),Negatives), pro_RULESLIST);
  pro_RestartVarList();
}

static void pro_StoreQuestion(LIST List)
{
  TERM term;
  LIST scan;
  for (scan = List; !list_Empty(scan); scan = list_Cdr(scan))
    list_Rplaca(scan, term_Create(fol_Not(), list_List(list_Car(scan))));
  term = term_Create(fol_Or(), List);
  pro_QUESTIONTERM = term;
  pro_RestartVarList();
}

/**************************************************************/
/* Functions for the Variable Table                           */
/**************************************************************/
  
typedef struct {
  char*  name;
  SYMBOL symbol;
} PRO_VARENTRY, *PRO_VARIABLE;

static __inline__ char* pro_VarName(PRO_VARIABLE Entry)
{
  return Entry->name;
}

static __inline__ SYMBOL pro_VarSymbol(PRO_VARIABLE Entry)
{
  return Entry->symbol;
}

static SYMBOL pro_VarCreate(char* Name)
{
  PRO_VARIABLE item = (PRO_VARIABLE) memory_Malloc(sizeof(PRO_VARENTRY));

  item->name   = Name;
  item->symbol = symbol_CreateStandardVariable();

  pro_VARLIST = list_Cons(item, pro_VARLIST);

  return item->symbol;
}

static void pro_VarFree(PRO_VARIABLE Entry)
{
  misc_StringFree(Entry->name);
  memory_Free(Entry, sizeof(PRO_VARENTRY));
}

static void pro_RestartVarList(void)
{
  list_DeleteWithElement(pro_VARLIST, pro_VarFree);
  pro_VARLIST = list_Nil();
  symbol_ResetStandardVarCounter();
}

static SYMBOL pro_VarLookup(char* Name)
/**************************************************************
  INPUT:   A variable name.
  RETURNS: The corresponding variable symbol.
  EFFECT:  If the variable name occured before, the corresponding
           symbol is returned and the <Name> is freed.
           If the variable name didn't occur
           a new variable is created.
	   Every occurence of the special variable "_" generates
           a new variable symbol.
***************************************************************/
{
  LIST scan;

  if (strcmp(Name, "_") != 0) {
    for (scan = pro_VARLIST; !list_Empty(scan); scan = list_Cdr(scan)) {
      if (strcmp(Name, pro_VarName(list_Car(scan))) == 0) {
	misc_StringFree(Name);
	return pro_VarSymbol(list_Car(scan));
      }
    }
  }
  /* Create a new variable */
  return pro_VarCreate(Name);
}
