/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              REGULAR CLAUSES                           * */
/* *                                                        * */
/* *  $Module:   regular                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 1.1 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 14:46:09 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

/* $RCSfile: regular.c,v $ */

#include "regular.h"

static BOOL reg_TermHasArgSymbol(TERM Term, SYMBOL Symbol)
/**************************************************************
  INPUT:   A terms and a symbol.
  EFFECT:  None.
  RETURNS: TRUE, if symbol (usually, a variable or constant symbol) is the top
           symbol of an argument of Term.
***************************************************************/
{
  LIST scan;

#ifdef DEBUG
  if (!term_IsTerm(Term) || !symbol_IsSymbol(Symbol)) {
    fputs("\nIllegal Input in function reg_TermHasArgSymbol\n", stderr);
    misc_DumpCore();
  }
#endif

  for (scan=term_ArgumentList(Term); !list_Empty(scan); scan=list_Cdr(scan))
    if (symbol_Equal(Symbol, term_TopSymbol(list_Car(scan))))
      return TRUE;

  return FALSE;
}

static BOOL reg_DominatesTermIntern(TERM Term1, TERM Term2)
/**************************************************************
  INPUT:   Two real terms.
  EFFECT:  None.
  RETURNS: TRUE, if Term1 strictly dominates Term2.
***************************************************************/
{
#ifdef DEBUG
  if (!term_IsTerm(Term1) || !term_IsTerm(Term2) ||
     fol_IsJunctor(term_TopSymbol(Term1)) || fol_IsLiteral(Term1) ||
     fol_IsJunctor(term_TopSymbol(Term2)) || fol_IsLiteral(Term2)) {
    fputs("\nIllegal Input in function reg_DominatesTermIntern\n", stderr);
    misc_DumpCore();
  }
#endif

  if (term_Size(Term1) < term_Size(Term2))
    return FALSE;

  if (term_IsVariable(Term2) && 
        reg_TermHasArgSymbol(Term1, term_TopSymbol(Term2)))
    return TRUE;
  
  if (!term_IsVariable(Term1) && symbol_IsConstant(term_TopSymbol(Term2))) 
    return TRUE;

  if (symbol_IsFunction(term_TopSymbol(Term1)) &&
      symbol_IsFunction(term_TopSymbol(Term2))) {
    LIST scan1, scan2;
    scan1=term_ArgumentList(Term1); 
    scan2=term_ArgumentList(Term2); 
    while (!list_Empty(scan2) && !list_Empty(scan1)) {
      if (!term_Equal(list_Car(scan1), list_Car(scan2)))
        return FALSE; 
      scan1 = list_Cdr(scan1);
      scan2 = list_Cdr(scan2);
    }
    if (list_Empty(scan2))
      return TRUE; 
  }

  return FALSE;
}

static BOOL reg_DominatesTerm(TERM Term1, TERM Term2)
/**************************************************************
  INPUT:   Two real terms.
  EFFECT:  None.
  RETURNS: TRUE, if Term1 dominates Term2.
***************************************************************/
{
#ifdef DEBUG
  if (!term_IsTerm(Term1) || !term_IsTerm(Term2) ||
     fol_IsJunctor(term_TopSymbol(Term1)) || fol_IsLiteral(Term1) ||
     fol_IsJunctor(term_TopSymbol(Term2)) || fol_IsLiteral(Term2)) {
    fputs("\nIllegal Input in function reg_DominatesTerm\n", stderr);
    misc_DumpCore();
  }
#endif

  if (term_Size(Term1) == term_Size(Term2)) {
    if  (term_Equal(Term1,Term2))
      return TRUE;
      /* The dominates relation is reflexive */
  }
  return reg_DominatesTermIntern(Term1, Term2);
}

static BOOL reg_DominatesLiteral(TERM Lit1, TERM Lit2)
/**************************************************************
  INPUT:   Two literals.
  EFFECT:  None.
  RETURNS: TRUE, if Lit1 dominates Lit2.
***************************************************************/
{
  LIST args1, args2, scan1, scan2;
  TERM t1, t2;
  BOOL result;

#ifdef DEBUG
  if (!term_IsTerm(Lit1) || !fol_IsLiteral(Lit2)) {
    fputs("\nIllegal Input in function reg_DominatesLiteral\n", stderr);
    misc_DumpCore();
  }
#endif

  result = FALSE;

  args1 = (symbol_IsPredicate(term_TopSymbol(Lit1))) ?
    term_ArgumentList(Lit1) : term_ArgumentList(term_FirstArgument(Lit1));
  args2 = (symbol_IsPredicate(term_TopSymbol(Lit2))) ?
    term_ArgumentList(Lit2) : term_ArgumentList(term_FirstArgument(Lit2));
  
  if (list_Empty(args2))
    return TRUE;

  /* for every non-constant argument t2 of Lit2 there is an non-constant
   * argument t1 of Lit1 such that t1 dominates t2 */
  for (scan2=args2; !list_Empty(scan2); scan2 = list_Cdr(scan2)) {
    result = FALSE;
    t2     = list_Car(scan2);
    if (!term_IsConstant(t2))
      for (scan1=args1; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
        t1     = list_Car(scan1);
  	if (!term_IsConstant(t1) && reg_DominatesTerm(t1, t2)) {
  	  result = TRUE;
  	  break;
  	}
      }
    if (!result) 
	break;
  }
  return result;
}

static BOOL reg_SimilarLiteral(TERM Lit1, TERM Lit2)
{
  return reg_DominatesLiteral(Lit1, Lit2) && 
         reg_DominatesLiteral(Lit2, Lit1);
}

static BOOL reg_Dominates(TERM Term1, TERM Term2)
/**************************************************************
  INPUT:   Two `terms'.  It is assumed both are either literals or real
           terms.
  EFFECT:  None.
  RETURNS: TRUE, if Term1 dominates Term2.
***************************************************************/
{

#ifdef DEBUG
  if (!term_IsTerm(Term1) || !term_IsTerm(Term2)) {
    fputs("\nIllegal Input in function reg_Dominates\n", stderr);
    misc_DumpCore();
  }
#endif

  	printf("\n  "); term_Print(Term1);
  	printf(" ? "); term_Print(Term2);
  	printf("  sizes = (%d,%d)", term_Size(Term1), term_Size(Term2));
  if (term_Size(Term1) == term_Size(Term2) && term_Equal(Term1,Term2))
    return TRUE;
      /* The dominates relation is reflexive */

  if (fol_IsLiteral(Term1) && fol_IsLiteral(Term2))
    return reg_DominatesLiteral(Term1, Term2);
  else
    return reg_DominatesTermIntern(Term1, Term2);

  return FALSE;
}

static ord_RESULT reg_DominatesCompare(TERM Term1, TERM Term2)
/**************************************************************
  INPUT:   Two `terms'.  It is assumed either both are literals or real
           terms.
  EFFECT:  None.
  RETURNS: ord_GREATER_THAN, if Term1 dominates Term2,
           ord_LESS_THAN, if Term2 strictly dominates Term1.
***************************************************************/
{
  if (reg_Dominates(Term1, Term2))
    return ord_GreaterThan();  /* actually, greater than or equal to */
                               /* possibly, the type ord_RESULT needs to
                                * be extended */
  else if (reg_Dominates(Term2, Term1))
    return ord_SmallerThan();  /* actually, smaller than or equal to */
  
  return ord_Uncomparable();
  /* e.g., different variables are incomparable */
}

static BOOL reg_IsRegularTerm(TERM Term)
/**************************************************************
  INPUT:   A real term.
  EFFECT:  None.
  RETURNS: TRUE, if Term is regular, that is,
           it dominates all of its arguments.
***************************************************************/
{
  LIST   scan;
  BOOL   result;

#ifdef DEBUG
  if (!term_IsTerm(Term) || fol_IsLiteral(Term)) {
    fputs("\nIllegal Input in function reg_IsRegularTerm\n", stderr);
    misc_DumpCore();
  }
#endif

  	/* printf("\n  In reg_IsRegularTerm "); term_Print(Term); */

  if (!term_IsComplex(Term))  /* variable or constant */
    return TRUE;

  result = TRUE;
  /* for every argument t of Term we have that Term dominates t */
  for (scan=term_ArgumentList(Term); !list_Empty(scan); scan = list_Cdr(scan)) {
  	/* printf("\n    "); term_Print(list_Car(scan)); */
    if (!reg_DominatesTerm(Term, list_Car(scan))) {
      result = FALSE;
      break;
    }
  }
  return result;
}

static BOOL reg_IsRegularLiteral(TERM Lit)
/**************************************************************
  INPUT:   A literal.
  EFFECT:  None.
  RETURNS: TRUE, if Lit is regular, that is,
           if it does not contain a compound argument, or
	   one of its arguments is a regular compound term which
	   dominates all arguments of Lit.
***************************************************************/
{
  LIST   scan, args;
  TERM   max, t;
  BOOL   succ, complex;

#ifdef DEBUG
  if (!term_IsTerm(Lit) || !fol_IsLiteral(Lit)) {
    fputs("\nIllegal Input in function reg_IsRegularLiteral\n", stderr);
    misc_DumpCore();
  }
#endif

  max = t = Lit;  /* dummy */
  	printf("\n"); term_Print(Lit);
  args = (symbol_IsPredicate(term_TopSymbol(Lit))) ?
	   term_ArgumentList(Lit) :
	   term_ArgumentList(term_FirstArgument(Lit));
  
  if (list_Empty(args))
    return TRUE;

  /* Get the first regular compound term */
  succ    = FALSE;
  complex = FALSE;
  for (scan=args; !list_Empty(scan) && !succ; scan = list_Cdr(scan)) {
    max = list_Car(scan);
    if (term_IsComplex(max)) {
       complex = TRUE;
       if (reg_IsRegularTerm(max)) {
  		/* printf("\n  max = "); term_Print(max); */
         succ = TRUE;
       }
    }
  }
  if (complex) {
   if (!succ)  /* found a compound term but not a regular one */
    return FALSE;
  }
  else   /* found no compound term */
    return TRUE;

  succ = FALSE;
  for (scan=args; !list_Empty(scan); scan = list_Cdr(scan)) {
    t = list_Car(scan);
  		/* printf("\n  t = "); term_Print(t); */
    switch (reg_DominatesCompare(max,t)) {
      /* only three cases, because dominates is a quasi-ordering */
      case ord_GREATER_THAN:
  		printf(" gt ");
  	succ = TRUE;
  	break;
      case ord_SMALLER_THAN:
  		printf(" lt ");
  	if (reg_IsRegularTerm(t)) {
  	  max  = t;
    		/* printf("\n  max = "); term_Print(max); */
  	  succ = TRUE;
  	}
  	else
  	  succ = FALSE;
  	break;
      default:
  		printf(" incomp ");
  	if (term_IsComplex(t))
  	  return FALSE;  /* !! */
  	else
  	  succ = FALSE;
  	break;
    }
  }
	if (succ) printf(" reg. "); 
	else  printf(" not reg. "); 
  return succ;
}

/*
static BOOL reg_ContainsOnlyRegularLiterals(CLAUSE Clause)
{
  int  i;
  TERM lit;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\nIllegal Input in function reg_ContainsOnlyRegularLiterals\n", stderr);
    misc_DumpCore();
  }
#endif

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
    lit = clause_GetLiteralTerm(Clause,i);
    term_InstallSize(lit);
    if (!reg_IsRegularLiteral(lit))
      return FALSE;
  }

  return TRUE;
}

static BOOL reg_ContainsDominatingLiteral(CLAUSE Clause)
{
  int    i;
  TERM   max, t;
  BOOL   succ;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\nIllegal Input in function reg_ContainsOnlyRegularLiterals\n", stderr);
    misc_DumpCore();
  }
#endif

  succ = TRUE;
  max = clause_GetLiteralTerm(Clause,clause_FirstLitIndex());
  for (i = clause_FirstLitIndex()+1; i < clause_Length(Clause); i++) {
    t = clause_GetLiteralTerm(Clause,i);
    switch (reg_DominatesCompare(max,t)) {
      case ord_GREATER_THAN:
  		printf(" gt ");
  	succ = TRUE;
  	break;
      case ord_SMALLER_THAN:
  		printf(" lt ");
  	if (reg_IsRegularLiteral(t)) {
  	  max  = t;
  	  succ = TRUE;
  	}
  	else
  	  succ = FALSE;
  	break;
      default:
  		printf(" incomp ");
  	  succ = FALSE;
  	break;
    }
  }
	if (succ) printf(" reg. "); 
	else  printf(" not reg. "); 
  return succ;
}
*/


static int reg_SmallerFunctionalArity(TERM Term, int k)
/**************************************************************
  INPUT:   A term (literal or real term).
  EFFECT:  None.
  RETURNS: The smallest arity of any non-nullary function occurring in
           the Clause.
***************************************************************/
{
  int    help, result;
  TERM   helpTerm;
  LIST   scan;
  SYMBOL top;

#ifdef DEBUG
  if (!term_IsTerm(Term)) {
    fputs("\nIllegal Input in function reg_SmallerFunctionalArity\n", stderr);
    misc_DumpCore();
  }
#endif

  result = k;
  top = term_TopSymbol(Term);

  if (symbol_IsFunction(top) && !symbol_IsConstant(top)) {
    help = symbol_Arity(top);
    if ((help != symbol_ArbitraryArity()) && help < result)
      result =help;
  }
  if ((symbol_IsFunction(top) && !symbol_IsConstant(top)) || 
      symbol_Equal(top,fol_Not()) || symbol_IsPredicate(top))
    for (scan=term_ArgumentList(Term); !list_Empty(scan); scan = list_Cdr(scan)) {
      helpTerm = list_Car(scan);
      help = reg_SmallerFunctionalArity(helpTerm, result);
      if (help < result)
	result = help;
    }

  	printf(" %d ", result);
  return result;
}

BOOL reg_IsKRegularClause(CLAUSE Clause, int k)
/**************************************************************
  INPUT:   A clause, and a positive integer
  EFFECT:  Term sizes are installed.  Not destructive.
  RETURNS: TRUE, if Clause is k-regular according to Hustadt (1998).
  NOTE:    Not at all optimal (too many iterations over the clause).
***************************************************************/
{
  int    i, arity;
  TERM   max, lit, nonsingular;
  BOOL   succ, first;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || k < 0) {
    fputs("\nIllegal Input in function reg_IsKRegularClause\n", stderr);
    misc_DumpCore();
  }
#endif

  /* Initialisation */
  succ  = TRUE;               /* for Cond 3 */
  first = TRUE;               /* for Cond 3 */
  max   = clause_GetLiteralTerm(Clause,clause_FirstLitIndex());  
                              /* for Cond 3 */
  		/* printf("\n  max = "); term_Print(max); */
  
  nonsingular = (TERM) NULL;  /* for Cond 4 */

  for (i = clause_FirstLitIndex(); i < clause_Length(Clause); i++) {
    lit = clause_GetLiteralTerm(Clause,i);
  		printf("\n  lit = "); term_Print(lit);

    /* Condition 2:
     * k is less than or equal to the minimal arity of
     * any non-nullary function symbol occurring in Clause */
    arity = reg_SmallerFunctionalArity(lit, k);
    if (arity < k)
      return FALSE;

    term_InstallSize(lit);  /* Important in connection with regularity
                               and the dominates relation */

    /* Condition 1:
     * Clause contains only regular literals */
    if (!reg_IsRegularLiteral(lit))
      return FALSE;

    /* Condition 3:
     * Clause contains a literal which dominates all other literals. */
    if (first) {  /* Skip the first iteration */
      first = FALSE;
    }
    else {
      switch (reg_DominatesCompare(max,lit)) {
        /* only three cases, because dominates is a quasi-ordering */
        case ord_GREATER_THAN:
    		printf(" gt ");
    	  succ = TRUE;
    	break;
        case ord_SMALLER_THAN:
    		printf(" lt ");
    	  max  = lit;
          succ = TRUE;
    	break;
        default:
    		printf(" incomp ");
    	  succ = FALSE;
    	break;
      }
    }
    
    /* Condition 4:
     * All non-singular literals of depth one are similar wrt dominates,
     * i.e. depth one literals with more than one variable are similar */
    if (reg_IsNonSingularDepthTwoAtom(lit)) {
      if (!term_IsTerm(nonsingular)) {
        nonsingular = lit;
      		printf("\n  nonsingular = "); term_Print(nonsingular);
      }
      else if (!reg_SimilarLiteral(nonsingular,lit)) {
        return FALSE;
      }
    }
  }

  return succ;
}
