/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   RENAMING                             * */
/* *                                                        * */
/* *  $Module:   RENAMING                                   * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.57 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:47:03 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: renaming.c,v $ */

#include "term.h"
#include "renaming.h"

NAT ren_STAMPID;

static BOOL ren_AFactorOk(TERM,TERM);
static BOOL ren_BFactorOk(TERM,TERM);
static BOOL ren_AExtraFactorOk(TERM,TERM);
static BOOL ren_BExtraFactorOk(TERM,TERM);
static TERM ren_FormulaRename(TERM, LIST, LIST *, BOOL);
static LIST ren_GetRenamings(TERM, int, TERM, int, BOOL (*)());
static BOOL ren_PFactorOk(TERM);
static BOOL ren_PExtraFactorOk(TERM);
static int  ren_Polarity(TERM);
static BOOL ren_NotPFactorOk(TERM);
static BOOL ren_NotPExtraFactorOk(TERM);
static void ren_ResetTermStamp(TERM);

void ren_Init()
/**********************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  Initializes the renaming module, in particular
           the stamp id used in this module.
***********************************************************/
{
  ren_STAMPID = term_GetStampID();
}

static void ren_ResetTermStamp(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: void.
  EFFECT:  The Term stamp of term as well as the stamps of 
           all its subterms (up to atom level) are reset.
***********************************************************/
{
  SYMBOL Top;
 
  term_ResetTermStamp(Term);
  Top = term_TopSymbol(Term);

  if (!symbol_IsPredicate(Top)) {
    if (fol_IsQuantifier(Top))
      ren_ResetTermStamp(term_SecondArgument(Term));
    else {
      LIST Scan;
      for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	ren_ResetTermStamp(list_Car(Scan)); 
    }
  }
}

static BOOL ren_HasNEquivFathers(TERM Term1, TERM Term2, NAT n)
/**********************************************************
  INPUT:   Two terms, where <Term2> is a proper subterm of <Term1>
           and a number.
  RETURNS: TRUE if <Term2> has a <n>-father that are equivalences
           and  below <Term1>
***********************************************************/
{
  Term2 = term_Superterm(Term2);

  while (Term1 != Term2) {
    if (symbol_Equal(term_TopSymbol(Term2),fol_Equiv())) {
      n--;
      if (n == 0)
	return TRUE;
    }
    Term2 = term_Superterm(Term2);
  }

  return FALSE;
}


static BOOL ren_PExtraFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in positive polarity context
           results in more than two clauses.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv())) {
    if (!fol_IsLiteral(term_FirstArgument(Term)) && !fol_IsLiteral(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_And())) {
    if (list_Length(term_ArgumentList(Term))>2 ||
	ren_PFactorOk(term_FirstArgument(Term)) ||
	ren_PFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Implies()))
    return (ren_NotPExtraFactorOk(term_FirstArgument(Term)) || 
	    ren_PExtraFactorOk(term_SecondArgument(Term)));

  if (symbol_Equal(Top,fol_Not()))
    return ren_NotPExtraFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_PExtraFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_PExtraFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial disjunction or an atom */
}

static BOOL ren_PFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in positive polarity context
           results in more than one clause.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()) || symbol_Equal(Top,fol_And()))
    return TRUE;

  if (symbol_Equal(Top,fol_Implies()))
    return (ren_NotPFactorOk(term_FirstArgument(Term)) || ren_PFactorOk(term_SecondArgument(Term)));

  if (symbol_Equal(Top,fol_Not()))
    return ren_NotPFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_PFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_PFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial disjunction or an atom */
}

static BOOL ren_NotPExtraFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in negative polarity context
           results in more than two clauses.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv())) {
    if (!fol_IsLiteral(term_FirstArgument(Term)) && !fol_IsLiteral(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Or())) {
    if (list_Length(term_ArgumentList(Term))>2 ||
	ren_NotPFactorOk(term_FirstArgument(Term)) ||
	ren_NotPFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Implies())) {
    if (ren_PFactorOk(term_FirstArgument(Term)) ||
	ren_NotPFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Not()))
    return ren_PExtraFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_NotPExtraFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_NotPExtraFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial conjunction or an atom */
}

static BOOL ren_NotPFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in negative polarity context
           results in more than one clause.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()) || symbol_Equal(Top,fol_Or()) || symbol_Equal(Top,fol_Implies()))
    return TRUE;

  if (symbol_Equal(Top,fol_Not()))
    return ren_PFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_NotPFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_NotPFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial conjunction or an atom */
}


static BOOL ren_AFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the A-Factor of Term2 in Term1 is larger than one.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_And()) || fol_IsQuantifier(Top))
    return ren_AFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    TERM Sub;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2 && ren_PFactorOk(Sub))
	return TRUE;
    }
    return ren_AFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return ren_BFactorOk(Term1, Super);
    else
      return (ren_NotPFactorOk(term_FirstArgument(Super)) || ren_AFactorOk(Term1, Super));
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Pol == 0)
      return TRUE;
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 1)
      return (ren_NotPFactorOk(Term2) || ren_AFactorOk(Term1,Super));
    else
      return (ren_PFactorOk(Term2) ||  ren_BFactorOk(Term1,Super));
  }
  
  misc_DumpCoreOut("Unexpected case in ren_AFactorOk");
  return FALSE; 
}

static BOOL ren_AExtraFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the A-Factor of Term2 in Term1 is larger than two.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_And()) || fol_IsQuantifier(Top))
    return ren_AExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    TERM Sub;
    BOOL Extra;
    Extra = FALSE;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2) {
	if (ren_PFactorOk(Sub)) {
	  if (Extra || ren_PExtraFactorOk(Sub))
	    return TRUE;
	  else
	    Extra = TRUE;
	}
      }
    }
    return ren_AExtraFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return ren_BExtraFactorOk(Term1, Super);
    else
      return (ren_NotPExtraFactorOk(term_FirstArgument(Super)) || ren_AExtraFactorOk(Term1, Super));
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 0)
      return (ren_PFactorOk(Term2) ||  ren_NotPFactorOk(Term2) ||
	      ren_AFactorOk(Term1,Super) ||  ren_BFactorOk(Term1,Super));

    if (Pol == 1)
      return (ren_NotPExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Super));
    else
      return (ren_PExtraFactorOk(Term2) ||  ren_BExtraFactorOk(Term1,Super));
  }
  
  
  misc_DumpCoreOut("Unexpected case in ren_AExtraFactorOk");
  return FALSE; 
}

static BOOL ren_BFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the B-Factor of Term2 in Term1 is larger than one.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
    return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_AFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    TERM Sub;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2 && ren_NotPFactorOk(Sub))
	return TRUE;
    }
    return ren_BFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return (ren_PFactorOk(term_SecondArgument(Super)) || ren_AFactorOk(Term1, Super));
    else
      return ren_BFactorOk(Term1, Super);
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Pol == 0)
      return TRUE;
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 1)
      return (ren_PFactorOk(Term2) || ren_AFactorOk(Term1,Super));
    else
      return (ren_NotPFactorOk(Term2) || ren_BFactorOk(Term1,Super));
  }
  
  misc_DumpCoreOut("Unexpected case in ren_BFactorOk");
  return FALSE; 
}

static BOOL ren_BExtraFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the B-Factor of Term2 in Term1 is larger than two.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
    return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_AExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    TERM Sub;
    BOOL Extra;
    Extra = FALSE;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2) {
	if (ren_NotPFactorOk(Sub)) {
	  if (Extra || ren_NotPExtraFactorOk(Sub))
	    return TRUE;
	  else
	    Extra = TRUE;
	}
      }
    }
    return ren_BExtraFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return (ren_PExtraFactorOk(term_SecondArgument(Super)) || ren_AExtraFactorOk(Term1, Super));
    else
      return ren_BExtraFactorOk(Term1, Super);
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 0)
      return (ren_PFactorOk(Term2) ||  ren_NotPFactorOk(Term2) ||
	      ren_AFactorOk(Term1,Super) ||  ren_BFactorOk(Term1,Super));

    if (Pol == 1)
      return (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Super));
    else
      return (ren_NotPExtraFactorOk(Term2) ||  ren_BExtraFactorOk(Term1,Super));
    }
  
  misc_DumpCoreOut("Unexpected case in ren_BExtraFactorOk");
  return FALSE; 
}

BOOL ren_HasBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2)
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: TRUE if renaming <Term1> in <Term2> results in a positive benefit.
  CAUTION: It is assumed that all superterms are set !!!!
***********************************************************/
{ 
  BOOL  PFacOk, NotPFacOk, AFacOk, BFacOk;
  
  switch (Pol1*Pol2) {
    
  case 0:
    PFacOk    = ren_PFactorOk(Term2);
    NotPFacOk = ren_NotPFactorOk(Term2);
    AFacOk    = ren_AFactorOk(Term1,Term2);
    BFacOk    = ren_BFactorOk(Term1,Term2);
    /*printf("\n");fol_PrintDFG(Term2);
    printf("\n P: %d NotP: %d A: %d B: %d AE: %d BE: %d PE: %d NotPE: %d ",PFacOk,NotPFacOk,AFacOk,BFacOk,
	   ren_AExtraFactorOk(Term1,Term2),ren_BExtraFactorOk(Term1,Term2),
	   ren_PExtraFactorOk(Term2), ren_NotPExtraFactorOk(Term2));*/

    return ((AFacOk && BFacOk && PFacOk && NotPFacOk) ||
	    (AFacOk && PFacOk && (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Term2))) ||
	    (BFacOk && NotPFacOk && (ren_NotPExtraFactorOk(Term2) || ren_BExtraFactorOk(Term1,Term2))));
    break;

  case 1:
    return (ren_PFactorOk(Term2) && ren_AFactorOk(Term1,Term2));
    break;

  case -1:
    return (ren_NotPFactorOk(Term2) && ren_BFactorOk(Term1,Term2));
    break;
  }
  
  misc_DumpCoreOut("Unexpected polarity in ren_HasBenefit");
  return FALSE;  
}

static BOOL ren_HasNonZeroBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2)
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: TRUE if renaming <Term1> in <Term2> results in non-zero  positive benefit.
  CAUTION: It is assumed that all superterms are set !!!!
***********************************************************/
{ 
  BOOL  PFacOk, NotPFacOk, AFacOk, BFacOk, PEFacOk, NotPEFacOk, AEFacOk, BEFacOk;
  
  switch (Pol1*Pol2) {
    
  case 0:
    PFacOk     = ren_PFactorOk(Term2);
    NotPFacOk  = ren_NotPFactorOk(Term2);
    AFacOk     = ren_AFactorOk(Term1,Term2);
    BFacOk     = ren_BFactorOk(Term1,Term2);
    PEFacOk    = ren_PExtraFactorOk(Term2);
    NotPEFacOk = ren_NotPExtraFactorOk(Term2);
    AEFacOk    = ren_AExtraFactorOk(Term1,Term2);
    BEFacOk    = ren_BExtraFactorOk(Term1,Term2);
    
    return ((AFacOk && BFacOk && PFacOk && NotPFacOk && (AEFacOk || BEFacOk || PEFacOk || NotPEFacOk)) || 
	    (AFacOk && PFacOk && (ren_PExtraFactorOk(Term2) && ren_AExtraFactorOk(Term1,Term2))) ||
	    (Pol1 == 0 && (symbol_Equal(term_TopSymbol(Term2),fol_Equiv()) ||
			   ren_HasNEquivFathers(Term1,Term2,1))) ||
	    ren_HasNEquivFathers(Term1,Term2,2) ||
	    (BFacOk && NotPFacOk && (ren_NotPExtraFactorOk(Term2) && ren_BExtraFactorOk(Term1,Term2))));
    break;

  case 1:
    return (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Term2) ||
	    ren_HasNEquivFathers(Term1,Term2,2));
    break;

  case -1:
    return (ren_NotPExtraFactorOk(Term2) || ren_BExtraFactorOk(Term1,Term2) ||
	    ren_HasNEquivFathers(Term1,Term2,2));
    break;
  }
  
  misc_DumpCoreOut("Unexpected polarity in ren_HasNonZeroBenefit");
  return FALSE;  
}

BOOL ren_Complex(TERM Term1, int Pol1, TERM Term2, int Pol2)
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: TRUE if renaming <Term2> in <Term1> is a quantified formula,
           a conjunction or a disjunctions.
	   Note, negations are ignored.
  CAUTION: It is assumed that all superterms are set !!!!
***********************************************************/
{ 
  SYMBOL symbol;

  symbol = term_TopSymbol(Term2);
  if (fol_IsQuantifier(symbol) || 
      symbol_Equal(symbol, fol_AND) || symbol_Equal(symbol, fol_OR) || 
      symbol_Equal(symbol, fol_IMPLIES) || symbol_Equal(symbol, fol_EQUIV))
    return TRUE;
  else 
    return FALSE;
}

BOOL ren_Quantified(TERM Term1, int Pol1, TERM Term2, int Pol2)
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: TRUE if renaming <Term2> in <Term1> is a quantified formula.
  CAUTION: It is assumed that all superterms are set !!!!
***********************************************************/
{ 
  if (fol_IsQuantifier(term_TopSymbol(Term2)))
    return TRUE;
  else 
    return FALSE;
}

static LIST ren_GetRenamings(TERM Term1, int Pol1, TERM Term2, int Pol2,
		BOOL (*RenamingFunction)()) 
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: The list of subterms below <Term2> that have a positive renaming
           benefit.
  EFFECT:  All renamed formulae are stamped.
  CAUTION: None.
***********************************************************/
{ 
  SYMBOL Top;
  LIST   Result,Scan;

  Result = list_Nil();

  /* printf("\n %d: ",Pol1); term_Print(Term1); printf("\n %d: ",Pol2); term_Print(Term2); */

  if (fol_IsLiteral(Term2))
    return Result;

  Top = term_TopSymbol(Term2);

  if (symbol_Equal(Top,fol_Not()))
    return ren_GetRenamings(Term1, Pol1, term_FirstArgument(Term2), Pol2*-1,RenamingFunction);
  
  if (term_Superterm(Term2) && !fol_IsQuantifier(term_TopSymbol(term_Superterm(Term2))) &&
      RenamingFunction(Term1, Pol1, Term2, Pol2)) {
    	/* printf("\n*"); */
    Result = list_Cons(Term2,Result);
    term_SetTermStamp(Term2);
    Term1 = Term2;
    Pol1  = Pol2;
  }

  if (fol_IsQuantifier(Top))
    return list_Nconc(Result,ren_GetRenamings(Term1, Pol1,
		term_SecondArgument(Term2), Pol2, RenamingFunction));

  if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()))
    for (Scan=term_ArgumentList(Term2);!list_Empty(Scan);Scan=list_Cdr(Scan))
      Result = list_Nconc(Result,ren_GetRenamings(Term1,Pol1,list_Car(Scan),Pol2,RenamingFunction));
  else
    if (symbol_Equal(Top,fol_Implies()))
      Result = list_Nconc(Result,
        list_Nconc(ren_GetRenamings(Term1,Pol1,term_FirstArgument(Term2),Pol2*-1,RenamingFunction),
		   ren_GetRenamings(Term1,Pol1,term_SecondArgument(Term2),Pol2,RenamingFunction)));
    else
      if (symbol_Equal(Top,fol_Equiv()))
	Result = list_Nconc(Result,
			    list_Nconc(ren_GetRenamings(Term1,Pol1,term_FirstArgument(Term2),0,RenamingFunction),
				       ren_GetRenamings(Term1,Pol1,term_SecondArgument(Term2),0,RenamingFunction)));
      else
	misc_DumpCoreOut("Unexpected case in ren_GetRenamings");

  return Result;
}

static int ren_Polarity(TERM Term)
/**********************************************************
  INPUT:   A term where the existence of superterms is assumed!.
  RETURNS: The polarity of Term with respect to its superterms.
***********************************************************/
{
  TERM SuperTerm;

  SuperTerm = term_Superterm(Term);

  if (SuperTerm) {
    SYMBOL Top;
    Top = term_TopSymbol(SuperTerm);
    if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
      return ren_Polarity(SuperTerm);
    if (symbol_Equal(Top,fol_Not()))
      return (-1*ren_Polarity(SuperTerm));
    if (symbol_Equal(Top,fol_Equiv()))
      return 0;
    if (symbol_Equal(Top,fol_Implies())) {
      if (Term == term_FirstArgument(SuperTerm))
	return (-1*ren_Polarity(SuperTerm));
      else
	return ren_Polarity(SuperTerm);
    }
    misc_DumpCoreOut("Unexpected case in ren_Polarity");
  }

  return 1;
}


static TERM ren_FormulaRename(TERM Term, LIST Termlist, LIST *SkolemSymbols, BOOL NegNewTerm) 
/**********************************************************
  INPUT:   A term, a list of subterms of term, a list of symbols and a
           flag.
  RETURNS: The renamed formula with respect to termlist and all newly introduced
           Skolem symbols for renamings are added to <SkolemSymbols>
  CAUTION: The term is destructively changed.
***********************************************************/
{
  TERM   Result,ActTerm,DefTerm,Superterm,NewTerm;
  LIST   Scan,FreeVariables,Arg;
  SYMBOL ActSymbol;
 
  DefTerm = (TERM)NULL;

  if (!list_Empty(Termlist))
    Result = term_Create(fol_AND,list_List(Term));
  else
    return Term;

  /* To avoid warning */
  ActSymbol = 0;

  /* Iterate over all pairs */
  for(Scan=Termlist;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    BOOL done;
    int polarity;
    LIST scan2;

    done = FALSE;
    polarity = (int) list_PairSecond((LIST) list_Car(Scan)); 

    /* Iterate over all renamings in one list */
    for(scan2=(LIST) list_PairFirst((LIST)list_Car(Scan)); !list_Empty(scan2); scan2=list_Cdr(scan2)) {

      ActTerm        = (TERM)list_Car(scan2);
      FreeVariables  = fol_FreeVariables(ActTerm);
      Superterm      = term_Superterm(ActTerm);
      /* Make one new predicate symbol for all renamings in the list */
      if(done == FALSE) {
	int    Length;
	Length         = list_Length(FreeVariables);
	ActSymbol      = symbol_CreateSkolemPredicate(Length);
	*SkolemSymbols = list_Cons((POINTER)ActSymbol,*SkolemSymbols);      
      }

      /* Allways make new predicate term */
      NewTerm        = term_Create(ActSymbol, list_CopyWithElement(FreeVariables,(POINTER (*)())term_Copy));
      		/* printf("\n "); term_Print(NewTerm); */
      		/* printf(" [%d]:\t", ren_Polarity(ActTerm)); term_Print(ActTerm); */
      
      /* Add new definition only once */
      if(done == FALSE) {

	/* Build term for definition on top level*/
	switch (polarity) {
	case 0:
	  DefTerm = term_Create(fol_Equiv(),list_Cons(term_Copy(NewTerm),list_List(ActTerm)));
	  break;
	  
	case 1:
	  DefTerm = term_Create(fol_Implies(),list_Cons(term_Copy(NewTerm),list_List(ActTerm)));
	  break;
	  
	case -1:
	  if (NegNewTerm) {
	    /* The following ensures that the introduced literals are negative
	     * in the definitions */
	    DefTerm = term_Create(fol_Implies(),list_Cons(term_Copy(NewTerm),
                        list_List(term_Create(fol_Not(), list_List(ActTerm)))));
	    /* printf("\n  + ");term_Print(NewTerm); */
            term_RplacArgumentList(NewTerm,list_List(term_Copy(NewTerm)));
            term_RplacTop(NewTerm,fol_Not());
	    /* printf("\n  - ");term_Print(NewTerm); */
	  }
	  else
	    DefTerm = term_Create(fol_Implies(),list_Cons(ActTerm,list_List(term_Copy(NewTerm))));
	  break;
	}

	/* Build chain for arguments of new definition to superterm */
	term_RplacSuperterm(term_FirstArgument(DefTerm),DefTerm);
	term_RplacSuperterm(term_SecondArgument(DefTerm),DefTerm);
	/* If the new predicate has arguments they must be all quantified */
	if (!list_Empty(FreeVariables)) {
	  list_NMapCar(FreeVariables,(POINTER (*)())term_Copy);
	  DefTerm = fol_CreateQuantifier(fol_All(),FreeVariables, list_List(DefTerm));
	}
	/* Add new definition to top level */
	term_RplacArgumentList(Result,list_Nconc(term_ArgumentList(Result),list_List(DefTerm)));
      	done = TRUE;
      }
      else {
	/* If a new definition already exists for the renaming term because it matches
	   a previous renaming term, just delete it and later replace it by the 
	   predicate  */
	LIST ScanCdr;
	list_Delete(FreeVariables);
	/* Remove all subterms of ActTerm from renaming lists */
	/* Terms with lower depth, may be subterms */
	for(ScanCdr = list_Cdr(Scan);  !list_Empty(ScanCdr); ScanCdr=list_Cdr(ScanCdr)) {
	  LIST l;
	  for(l=(LIST) list_PairFirst((LIST) list_Car(ScanCdr)); !list_Empty(l); l=list_Cdr(l))
	    if(term_HasPointerSubterm(ActTerm, (TERM) list_Car(l)))
	      list_Rplaca(l, NULL);
	  list_Rplaca((LIST) list_Car(ScanCdr), 
		      list_PointerDeleteElement(list_PairFirst((LIST) list_Car(ScanCdr)), NULL));
	  if(list_Empty(list_PairFirst((LIST) list_Car(ScanCdr)))) {
	    list_PairFree(list_Car(ScanCdr));
	    list_Rplaca(ScanCdr, NULL);
	  }
	}
	list_PointerDeleteElement(list_Cdr(Scan), NULL);
	term_Delete(ActTerm);
      }
      /* Search Position in Argumentlist */
      for(Arg=term_ArgumentList(Superterm);
	  !list_Empty(Arg) && !((TERM)list_Car(Arg) == ActTerm);
	  Arg=list_Cdr(Arg));                                      
      /* Replace renaming term by predicate and add chain to Superterm */
      list_Rplaca(Arg,NewTerm);
      term_RplacSuperterm(NewTerm, Superterm);
    }
  }
  /* printf("\nRes:\t");term_Print(Result); */

  return Result;			   
}

static LIST ren_FreeRenaming(LIST Renamings)
/**********************************************************
  INPUT:   A list of pairs of renaming candidates.
           (list of matching terms, polarity for definition) 
  RETURNS: The list of candidates without renamings that have
           benefit zero.
  CAUTION: Destructive.
***********************************************************/
{
  LIST Scan;
  TERM Father,Term;

  for(Scan=Renamings;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    if(list_Length((LIST) list_PairFirst((LIST) list_Car(Scan))) == 1) {
      Term   = (TERM)list_Car((LIST) list_PairFirst((LIST) list_Car(Scan)));
      Father = term_Superterm(Term);
      while (!term_HasTermStamp(Father) && term_Superterm(Father)) {
	Father = term_Superterm(Father);
      }
	
      if (!ren_HasNonZeroBenefit( Father,  (term_Superterm(Father) ? ren_Polarity(Father) : 1),  
				  Term,  ren_Polarity(Term))) {
	list_Delete((LIST) list_PairFirst((LIST) list_Car(Scan)));
	list_Free((LIST) list_Car(Scan));
	list_Rplaca(Scan,NULL);
	term_ResetTermStamp(Term);
      }
    }
  }

  Renamings = list_PointerDeleteElement(Renamings,NULL);

  return Renamings;
}


TERM ren_Rename(TERM Term, LIST *SkolemSymbols, BOOL Document, BOOL Match,
		BOOL (*RenamingFunction)()) 
/**********************************************************
  INPUT:   A term and a pointer to a list of Skolem symbols,
           a flag indicating whether the renamings should be
	   documented and a flag indicating whether matching subterms should be renamed
	   using the same predicate
  RETURNS: The possibly changed Term where subformulae are renamed
           if this results in a smaller clause normal form, with
	   respect to the number of clauses. The newly introduced
	   Skolem predicates are added to <SkolemSymbols>.
  CAUTION: Destructive.
***********************************************************/
{
  LIST Renamings,Scan;
  TERM Father,Renaming;
  LIST allmatches;      /* All terms that are candidates for renaming */
                        /* Necessary, because two renamings may match */

  Renamings = list_Nil();

  if (term_StampOverflow(ren_STAMPID))
    ren_ResetTermStamp(Term);

#ifdef DEBUG
  fol_CheckFatherLinks(Term);
#endif

  term_StartStamp();

  /* Get list of terms */
  Renamings = ren_GetRenamings(Term,1,Term,1,RenamingFunction);

  /* Now sort this list wrt term depth so that deeper terms are handled
     first. This prevents subterms from being deleted before their 
     superterms are renamed */
  Renamings = list_GreaterNumberSort(Renamings, term_Depth);
    /* for(Scan=Renamings; !list_Empty(Scan); Scan=list_Cdr(Scan))
      printf("\n Depth %d", term_Depth((TERM) list_Car(Scan))); */

  allmatches = list_Nil();
  for(Scan=Renamings; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
    if(Match == TRUE) {
      int  polarity;
      LIST l, matchables;
      /* Get subterms that can be matched with the first renaming */
      /* Check if renaming was already used because it was matchable to a */
      /* renaming used before */

      if(!list_PointerMember(allmatches, (TERM) list_Car(Scan))) {
	matchables = fol_Matchables(Term, (TERM) list_Car(Scan));
	polarity   = 4711;

	/* Find polarity of definition */
	for(l=matchables; !list_Empty(l); l=list_Cdr(l)) {
	  int newpol;
	  if(list_PointerMember(allmatches, list_Car(l))) {
	    list_Rplaca(l, NULL);
	  }
	  else {
	    /* Determine polarity */
	    newpol = ren_Polarity((TERM) list_Car(l));
	    if(polarity == 4711)
	      polarity = newpol;
	    else {
	      if((newpol == 0) || (newpol != polarity)) {
		polarity = 0;
		break;
	      }
	    }
	  }
	}
	matchables = list_PointerDeleteElement(matchables, NULL);
	/* Replace list element by pair (list of matching terms, polarity) */
	list_Rplaca(Scan, list_PairCreate(matchables, (POINTER) polarity));
	allmatches = list_Nconc(list_Copy(matchables), allmatches);
      }
      else {
	/* Renaming was already applied */
	list_Rplaca(Scan, NULL);
      }
    }
    else {
      int polarity;
      polarity = ren_Polarity((TERM)list_Car(Scan));
      list_Rplaca(Scan, list_PairCreate(list_List(list_Car(Scan)), (LIST) polarity));
    }
  }
  list_Delete(allmatches);

  /* Delete entries for renamings that occurred several times */
  Renamings = list_PointerDeleteElement(Renamings, NULL);
  if (RenamingFunction == ren_HasBenefit)
    Renamings = ren_FreeRenaming(Renamings); /* Checks for non zero benefit */

  if (!list_Empty(Renamings) && Document) {
    printf("\n\nRenaming term:\n ");
    term_PrettyPrint(Term);
    for(Scan=Renamings;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      LIST l;
      fputs("\nMatching subterms:\n", stdout);
      for(l=(LIST) list_PairFirst((LIST) list_Car(Scan)); !list_Empty(l); l=list_Cdr(l)) {
	Renaming = (TERM)list_Car(l);
	fol_PrettyPrintDFG(Renaming);
	Father = Renaming;
	while (term_Superterm(Father) && 
	       (term_Superterm(term_Superterm(Father)) ||
		!symbol_Equal(term_TopSymbol(term_Superterm(Father)),fol_And())))
	  Father = term_Superterm(Father);
	printf("\n\t in\n ");
	fol_PrettyPrintDFG(Father);
	if (!list_Empty(list_Cdr(l)))
	  printf("\n\n ");
	else
	  printf("\n");
      }
    }
  }

  if (RenamingFunction == ren_Quantified)
    Term = ren_FormulaRename(Term,Renamings,SkolemSymbols,FALSE);
  else
    Term = ren_FormulaRename(Term,Renamings,SkolemSymbols,FALSE);
  
  if (!list_Empty(Renamings) && Document) {
    printf("\n\nRenamed term:\n");
    term_PrettyPrint(Term);
  }

  for(Scan=Renamings; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
    list_Delete((LIST) list_PairFirst((LIST) list_Car(Scan)));
    list_Free((LIST) list_Car(Scan));
  }
  list_Delete(Renamings);

  term_StopStamp();

  return Term;
}
