/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                                                        * */
/* *          Module: rescmp                                * */
/* *                                                        * */
/* *                                                        * */
/* *          Compare two SPASS results (clause sets )      * */
/* *                                                        * */
/* $Revision: 0.43 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:38 $                                      * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: rescmp.c,v $ */

#include "options.h"
#include "list.h"
#include "rules-inf.h"
#include "rules-sort.h"
#include "rules-split.h"
#include "rules-red.h"
#include "analyze.h"
#include "clock.h"

extern LIST dfg_DFGParser(FILE*, LIST*, LIST*, LIST*);
extern void dfg_Free();
extern void dfg_DeleteFormulaPairList(LIST);

static LIST  rescmp_ClausesFromFile(char * Filename);
static BOOL  rescmp_Lists(LIST* L1, LIST* L2, LIST *Unmapped1, LIST *Mapped2, char** File1, char** File2);
static BOOL  rescmp_FindImage(CLAUSE c1, LIST l, CLAUSE* c2, LIST* Pred);
static void  rescmp_MoveElement(LIST* L1, LIST Pred, LIST *L2) ;

int main(int argc, char* argv[]) 
/**************************************************************
  INPUT:   The command line args: possibly an option (-v),
           and two file names.
  RETURNS: Nothing
  EFFECTS:  Prints out whether each clause in file1 mutually subsumes
           with a clause in file2, and in verbose mode
	   information about which clauses could not be
	   mapped.
***************************************************************/
{
  LIST      Result1, Result2;
  LIST      Unmapped1, Mapped2;
  BOOL      Verbose;
  OPTID     Verbose_OptId;
  char      *File1, *File2;
  FLAGSTORE Flagstore;
  
  clock_Init(); 
  clock_StartCounter(clock_OVERALL); 
  memory_Init(-1);
  symbol_Init(TRUE);
  stack_Init();
  hash_Init();
  sort_Init();
  term_Init();
  fol_Init(TRUE);
  cont_Init();
  unify_Init();

  flag_Init();
  subs_Init();
  clause_Init();
  red_Init(); /* added */
  /*  split_Init(); */
  dp_Init();
  opts_Init();

  Flagstore   = flag_NewStore();
  flag_InitStoreByDefaults(Flagstore);
  flag_InstallStore(Flagstore);

  Unmapped1   = list_Nil();
  Mapped2     = list_Nil();

  Verbose_OptId = opts_Declare("v", opts_NOARGTYPE);
  if (!opts_Read(argc, argv)) {
     misc_Error();}
  if (opts_IsSet(Verbose_OptId)) 
    Verbose = TRUE;  
  else 
    Verbose = FALSE;

  if (opts_Indicator() != argc-2)    { 
      printf("\n\t       Usage: rescmp <input-file1> <input-file2> \n\n");
      return 0; }
  
  File1 = argv[opts_Indicator()];
  File2 = argv[opts_Indicator()+1];

  /* read files */

  if (Verbose) { 
    printf("Reading file '%s'...", File1); 
    fflush(stdout);}
  Result1 = rescmp_ClausesFromFile(File1);
  if (Verbose) { 
    printf("Ok.\n"); 
    fflush(stdout);}

  dfg_Free();
  if (Verbose) { 
    printf("Reading file '%s'...", File2); 
    fflush(stdout);}
  Result2 = rescmp_ClausesFromFile(File2);

  if (Verbose) { 
    printf("Ok.\n"); 
    fflush(stdout);}

  opts_Free();

  /* compute difference */
  
  if (rescmp_Lists(&Result1, &Result2, &Unmapped1, &Mapped2, &File1, &File2))
    printf("Results are equal.\n");
  else {
    printf("Results differ.\n");

    if (Verbose) {
      printf("Unmapped clauses in file '%s': ", File1);
      if (list_Empty(Unmapped1)) printf("None.");
      else {
	printf("\n");
	clause_ListPrint(Unmapped1); }

      printf("\n\nUnmapped clauses in file '%s': ", File2);
      if (list_Empty(Result2)) printf("None.");
      else {
	printf("\n");
	clause_ListPrint(Result2); }
      printf("\n");}  }


  /* clean up */
  clause_DeleteClauseList(Result1);
  list_Delete(Unmapped1); 
  clause_DeleteClauseList(Result2);
  clause_DeleteClauseList(Mapped2);

  flag_FreeStore(Flagstore);

  sort_Free(); 
  unify_Free();
  cont_Free();
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();
  flag_Free();
  
#ifdef DEBUG
  memory_Print();
  printf("\n");
#endif
  return 0;
}

static BOOL rescmp_Lists(LIST* L1, LIST* L2, LIST *Unmapped1, LIST *Mapped2, char** File1, char** File2)
/**************************************************************
  INPUT:   Four clause lists by reference, two file names by reference. 
  RETURNS: For each clause c1 in list L1, a corresponding clause  c2 
           in L2 is searched. If found, c2 is prepended to Mapped2
	   and removed from L2, if not, c1 is prepended to Unmapped1. 

	   Therefore: 
	   L2 contains on return all clauses that have
	   NOT been the image of some clause in L1,

	   Mapped2 contains all clauses that HAVE been the image
	   of some clause in L1,

	   L1 remains unchanged, and

	   Unmapped1 contains all clauses in L1 that have not been
	   mapped on some clause in L2.

	   TRUE is returned if all clauses in both lists were
	   mapped, that is, Empty(Unmapped1) && Empty(L2).

	   The lists may be swapped so that length(L1) > Length(L2),
	   in which case the filenames are also swapped.
	   
  EFFECTS:  L2 is destructivly changed as described above.
***************************************************************/

{
  LIST Tmp1;
  LIST Rest; /* contains rest of clause list to be mapped */
  LIST Pred; /* contains predecessor of an image for a mapped clause */ 
  CLAUSE c1,c2;

  char* Tmp2;

  /* order lists so that L2, the list that is always fully searched, is the smaller list */
 
  if (list_Length(*L2) > list_Length(*L1))     { 
    Tmp1 = *L1; *L1 = *L2; *L2 = Tmp1; 
    Tmp2 = *File1; *File1 = *File2; *File2 = Tmp2; }

  Rest = *L1;

  while (!list_Empty(Rest)) {
    
    c1 = list_Car(Rest);

    if (rescmp_FindImage(c1,*L2, &c2,&Pred)) {

      /* if an image is found, dequeue element from */
      /* image list, put it in 'Mapped2' */

      rescmp_MoveElement(L2, Pred, Mapped2); }
  
    else {

      /* otherwise, prepend mapped element to 'Mapped1' */ 
      *Unmapped1 = list_Cons(c1, *Unmapped1);}
   
    /* next clause */
    Rest = list_Cdr(Rest); 
  }
  
  if (list_Empty(*Unmapped1) && list_Empty(*L2))
    return(TRUE);
  else
    return(FALSE);
}


static BOOL rescmp_FindImage(CLAUSE c1, LIST l, CLAUSE* c2, LIST* Pred)
/**************************************************************
  INPUT:   A clause to be mapped, a list to be search for an image,
           by reference a clause (the image on return) and a list,
  RETURNS: Searches a clause c2 in list l such that
           c1 subs c2, c2 subs c1. Returns TRUE iff such
	   a clause is found. In this case, *c2 points to the
	   clause, and *Pred is the sublist of l that contains
	   *c2 as its second element. (So that the calling
	   function can efficiently dequeue c2.)
	   
  EFFECTS: None 
  MEMORY:  None
***************************************************************/
{    
  BOOL Mapped;
  int l1;

  Mapped = FALSE;
  l1 = clause_Length(c1);
  *Pred = list_Nil();

  while (!Mapped && !list_Empty(l)) {
    *c2 = list_Car(l);

    /* check subsumption on clauses with same length */

    if (clause_Length(*c2) == l1) {
      if (subs_Subsumes(c1,*c2, -1,-1)) {
	if (subs_Subsumes(*c2,c1, -1,-1)) 
	  Mapped = TRUE; }}
    
    if (!Mapped) {
      *Pred = l;
      l = list_Cdr(l); }
  }

  return(Mapped);
}
  
static void rescmp_MoveElement(LIST* L1, LIST Pred, LIST* L2) 
/*********************************************************************
  INPUT:   Three lists.
  RETURNS: Nix 
  EFFECTS: Move element c=list_Car(list_Cdr(Pred)) (Pred is a sublist
           of L1) with predecessor Pred from list L1 to list L2. 
	   Adjust L1 if c is first element (i.e., Pred is empty list). 
           A list node in L1 is freed, one is added to L2.
**********************************************************************/
{ 
  POINTER El;

  if (list_Empty(Pred))
    El = list_Car(*L1);
  else 
    El = list_Car(list_Cdr(Pred));

  *L2 = list_Cons(El, *L2);

  if (list_Empty(Pred)) 

    /* remove first element */
    { 
      Pred = *L1; /* use Pred as temporary */
      *L1 = list_Cdr(*L1);

      memory_Free(Pred, sizeof(LIST_NODE)); }

  else {

    /* remove other element */
    list_DequeueNext(Pred); }
  
}

static LIST rescmp_ClausesFromFile(char * Filename)
/*********************************************************************
  INPUT:   A Filename
  RETURNS: A list of the clauses contained in the file <Filename>
  EFFECTS: Allocates memory for result.
**********************************************************************/
{
  LIST Clauses;
  FILE* File;
  LIST Axioms, Conjectures, SortDecl;

  Axioms      = list_Nil();
  Conjectures = list_Nil();
  SortDecl    = list_Nil();
  
  File    = misc_OpenFile(Filename, "r");
  Clauses = dfg_DFGParser(File, &Axioms, &Conjectures, &SortDecl);
  misc_CloseFile(File, Filename);

  if (!list_Empty(Axioms) || !list_Empty(Conjectures)) {
    printf("\nerror: file %s contains formulas.\n", Filename);
    misc_Error();  }
    
  dfg_DeleteFormulaPairList(SortDecl);

  return(Clauses);
}

