/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                RESOLUTION                              * */
/* *                                                        * */
/* *  $Module:   RESOLUTION                                 * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.51 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:38 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: resolution.c,v $ */

#include "resolution.h"


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

BOOL res_IsEmptyClauseInList(LIST List)
/**************************************************************
  INPUT:   A list.
  RETURNS: TRUE if list contains the empty clause and FALSE 
           otherwise.
  EFFECTS: None.
***************************************************************/
{
  LIST   Scan;

  for (Scan=List;!list_Empty(Scan);Scan=list_Cdr(Scan)){
    if (clause_Length(list_Car(Scan)) ==0)
      return TRUE;
  }
  return FALSE;
}


CLAUSE res_CondenseClause(CLAUSE c, LIST condlist)
/**********************************************************
  INPUT:   A clause c.
  RETURNS: The condensed clause c.
  CAUTION: The clause c is destructively changed.
***********************************************************/
{
  if (!list_Empty(condlist)){
    LIST   scan,termlist;
    CLAUSE newclause;

    termlist = list_Nil();
    for (scan=condlist;!list_Empty(scan);scan=list_Cdr(scan))
      termlist =
	list_Cons(term_Copy(clause_GetLiteralTerm(c, (int) list_Car(scan))),
		  termlist);

    newclause = clause_CreateFromLiterals(termlist,FALSE,FALSE,TRUE);
    if (clause_GetFlag(c,WORKEDOFF))
      clause_SetFlag(newclause,WORKEDOFF);
    clause_Delete(c);
    list_Delete(termlist);
    return newclause;
  }

  return c;
}


SYMBOL res_GetMaxVarTerm(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: The maximal variable symbol of the term.
  CAUTION: None.
***********************************************************/
{  
  int          Stack;
  SYMBOL       Top,max;

#ifdef DEBUG
  if (!term_IsTerm(Term)) {
    printf("\n\nIllegal input in res_getMaxVarTerm\n");
    misc_DumpCore();
  }
#endif

  max = symbol_Null();
  Stack = stack_Bottom();
  do {
    if (term_IsComplex(Term)) {
      stack_Push(term_ArgumentList(Term));
    } else {
      Top = term_TopSymbol(Term);
      if (symbol_IsVariable(Top) && Top > max)
	max = Top;
    }

    while (!stack_Empty(Stack) && list_Empty(stack_Top()))
      stack_Pop();

    if (!stack_Empty(Stack)) {
      Term = (TERM)list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(Stack));
  
  return max;
    
}


SYMBOL res_GetMaxVarClause(CLAUSE Clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: The maximal variable symbol of the terms of the clause.
  CAUTION: None.
***********************************************************/
{  
  int    n,i;
  SYMBOL max,max_help;

  max = symbol_GetInitialStandardVarCounter();

  n = clause_Length(Clause);

  for (i = 0; i < n; i++) {
    max_help = res_GetMaxVarTerm(clause_GetLiteralTerm(Clause,i));
    if (max_help > max)
      max = max_help;
  }
  return max;
}


void res_InsertClauseIndex(CLAUSE clause, st_INDEX stindex)
/**********************************************************
  INPUT:   A st_INDEX and a clause.
  RETURNS: Inserts the clause in the st_INDEX stindex.
  CAUTION: None.
***********************************************************/
{
  int n,j;

  n = clause_Length(clause);
  for (j = 0; j < n; j++)
    st_EntryCreate(stindex,
		   clause_GetLiteral(clause,j),
		   clause_GetLiteralTerm(clause,j),
		   cont_LeftContext());
}


void res_DeleteClauseIndex(CLAUSE clause, st_INDEX stindex)
/**********************************************************
  INPUT:   A st_INDEX and a clause.
  RETURNS: Deletes the clause from the st_INDEX stindex.
  CAUTION: None.
***********************************************************/
{
  int n, j;

  n = clause_Length(clause);
  for (j = 0; j < n; j++)
    if (!st_EntryDelete(stindex,
			clause_GetLiteral(clause,j), 
			clause_GetLiteralTerm(clause,j),
			cont_LeftContext()))
      misc_DumpCore();
}


LIST res_BuildAllResolvents(CLAUSE clauseL, st_INDEX stindex)
/**********************************************************
  INPUT:   A clause and a st_Index.
  RETURNS: A list of all resolvents of clauseL with stindex.
  CAUTION: None.Not used.
***********************************************************/
{
  int    j,i,n; 
  LIST   scan,unifiables,resolvents,termlist,termlistLsubst,termlistRsubst;
  TERM   term,negterm,uniterm;
  SYMBOL maxvar;
  SUBST  Lsubst,Rsubst;
  CLAUSE resclause;
  LITERAL unilitres;
  
  resolvents = list_Nil();
  n          = clause_Length(clauseL);
  
  for (i = 0; i < n; i++) {
    term        = clause_GetLiteralTerm(clauseL,i);
    negterm     = fol_ComplementaryTerm(term);  
    unifiables  = st_GetUnifier(cont_LeftContext(), stindex, cont_RightContext(), negterm);
    term_Delete(negterm);
    
    maxvar = res_GetMaxVarClause(clauseL);
    
    for (scan=unifiables; !list_Empty(scan); scan=list_Cdr(scan)) {
      termlist       = list_Nil();
      termlistLsubst = list_Nil();
      termlistRsubst = list_Nil();
      unilitres      = (LITERAL)list_Car(scan);
      
      if (clause_LiteralArity(clause_LiteralOwningClause(unilitres)) >= 0) {

	if (res_GetMaxVarClause(clause_LiteralOwningClause(unilitres))
	    > maxvar)
	  maxvar = res_GetMaxVarClause(clause_LiteralOwningClause(unilitres));
      
	term_StartMaxRenaming(maxvar);
	for (j = 0; j < n; j++)
	  term_Rename(clause_GetLiteralTerm(clauseL,j));
      
	uniterm = clause_LiteralSignedAtom(unilitres);
      
	cont_Check();

	if (unify_UnifyNoOC(cont_LeftContext(),
			    fol_Atom(term),
			    cont_RightContext(),
			    fol_Atom(uniterm))){
	  subst_ExtractUnifier(cont_LeftContext(), &Lsubst, cont_RightContext(), &Rsubst);
	
	  for (j = 0; j < n; j++) {
	    if (!(j == i))
	      termlistLsubst =
		list_Cons(subst_Apply(Lsubst,
				      term_Copy(clause_GetLiteralTerm(clauseL,j))),
			  termlistLsubst);
	  }
	  
	  for (j = 0;
	       j < clause_Length(clause_LiteralOwningClause(unilitres));
	       j++){
	    if (!(term_Equal(uniterm, clause_GetLiteralTerm(
			     clause_LiteralOwningClause(unilitres),j))))
	      termlistRsubst =
		list_Cons(subst_Apply(Rsubst, term_Copy(clause_GetLiteralTerm(
			              clause_LiteralOwningClause(unilitres),j))),
			  termlistRsubst);
	  }
	
	  termlist  = list_Nconc(termlistLsubst,termlistRsubst);
	  resclause = clause_CreateFromLiterals(termlist,FALSE,FALSE,TRUE);
	
	  term_StartMinRenaming();
	  for (j = 0; j < clause_Length(resclause); j++)
	    term_Rename(clause_GetLiteralTerm(resclause,j));
	
	  term_StartMinRenaming();
	  for (j = 0; j < clause_Length(clauseL); j++)
	    term_Rename(clause_GetLiteralTerm(clauseL,j));
	  
	  resolvents = list_Cons(resclause,resolvents);
	
	} else{
	  printf("Error in res_BuildAllResolvents!\n");
	  misc_DumpCore();
	}

	cont_Reset();
      
	list_Delete(termlist);
	subst_Delete(Lsubst);
	subst_Delete(Rsubst);
      }
    }
    list_Delete(unifiables);
  }
  return resolvents;
}


void res_Res(LIST *usable, int resnumber)
/**********************************************************
  INPUT:   A list of clauses and an integer.
  RETURNS: Nothing.
  CAUTION: <clauselist> is consumed. Not used.
***********************************************************/
{
  LIST     workedoff,resolvents;
  CLAUSE   clause;
  st_INDEX stindex; 
 
  workedoff    = list_Nil();
  stindex      = st_IndexCreate();

  while (!list_Empty(*usable) && resnumber > 0) {
    clause     = (CLAUSE)list_Car(*usable);
    /* Resolution is only allowed on clauses where all 
       literals have the same arity; uh 12.01.99 */
    if (clause_LiteralArity(clause) >= 0) {
      resolvents = res_BuildAllResolvents(clause,stindex);
      if (res_IsEmptyClauseInList(resolvents))
	resnumber  = 0;
      else
	resnumber -= list_Length(resolvents);
      *usable      = list_Nconc(*usable,resolvents);
      
      res_InsertClauseIndex(clause,stindex);
      
      workedoff  = list_Cons(clause,workedoff);
      
      *usable = list_Pop(*usable);
    }
  }
  *usable = list_Nconc(workedoff,*usable);
  st_IndexDelete(stindex);
}


CLAUSE res_SelectLightestClause(LIST clauselist)
/**********************************************************
  INPUT:   A list of clauses.
  RETURNS: The lightest clause of the clauselist. 
  CAUTION: None.
***********************************************************/
{
  CLAUSE clause;
  LIST   scan;
  int    min;

  clause = list_Car(clauselist);
  min    = clause_Weight(clause);
  
  for (scan=list_Cdr(clauselist); !list_Empty(scan); scan=list_Cdr(scan)) {
    if (clause_Weight(list_Car(scan)) < min) {
      clause = list_Car(scan);
      min    = clause_Weight(clause);
    }
  }
  return clause;
}


BOOL res_HasTautology(CLAUSE clause)
/**********************************************************
  INPUT:   A clauses.
  RETURNS: TRUE if the clause contains a complementary 
           literal pair and FALSE otherwise. 
  CAUTION: None.
***********************************************************/
{
  BOOL found;
  TERM literal1;
  int  i, j, n;

  found = FALSE;
  n     = clause_Length(clause);
  
  for (i = 0; i < n && !found; i++) {
    literal1 = fol_ComplementaryTerm(clause_GetLiteralTerm(clause,i));
    for (j = 0; j < n && !found; j++)
      if (j != i && term_Equal(literal1, clause_GetLiteralTerm(clause,j)))
	found = TRUE;

    term_Delete(literal1);
  }
  return found;
}


LIST res_ResBuildAllResolvents(CLAUSE clauseL, st_INDEX stindex)
/**********************************************************
  INPUT:   A clause and a st_Index.
  RETURNS: A list of all resolvents of clauseL with the
           clauses in stindex which are worked off
  CAUTION: None.
***********************************************************/
{
  int    j,i,n; 
  LIST   scan,unifiables,unifiableshelp,resolvents,
         termlist,termlistLsubst,termlistRsubst;
  TERM   term,negterm,uniterm;
  SYMBOL maxvar;
  SUBST  Lsubst,Rsubst;
  CLAUSE resclause;
  LITERAL unilitres,litres;

  resolvents = list_Nil();
  n          = clause_Length(clauseL);
  
  for (i = 0; i < n; i++) {
    term            = clause_GetLiteralTerm(clauseL,i);
    negterm         = fol_ComplementaryTerm(term); 
    unifiableshelp  = st_GetUnifier(cont_LeftContext(), stindex, cont_RightContext(), negterm);
    unifiables      = list_Nil();

    for (scan=unifiableshelp; !list_Empty(scan); scan=list_Cdr(scan)) {
      litres = (LITERAL)list_Car(scan);
      if (clause_GetFlag(clause_LiteralOwningClause(litres),WORKEDOFF))
	unifiables = list_Cons(list_Car(scan),unifiables);
    }
    list_Delete(unifiableshelp);
    term_Delete(negterm);
    maxvar = res_GetMaxVarClause(clauseL);

    for (scan=unifiables; !list_Empty(scan); scan=list_Cdr(scan)) {
      termlist       = list_Nil();
      termlistLsubst = list_Nil();
      termlistRsubst = list_Nil();
      unilitres  = (LITERAL)list_Car(scan);
      if (res_GetMaxVarClause(clause_LiteralOwningClause(unilitres)) > maxvar)
	maxvar = res_GetMaxVarClause(clause_LiteralOwningClause(unilitres));
      term_StartMaxRenaming(maxvar);
      for (j = 0; j < n; j++)
	term_Rename(clause_GetLiteralTerm(clauseL,j));
      uniterm = clause_LiteralSignedAtom(unilitres);

      cont_Check();

      if (unify_UnifyNoOC(cont_LeftContext(),
			  fol_Atom(term),
			  cont_RightContext(),
			  fol_Atom(uniterm))) {
	subst_ExtractUnifier(cont_LeftContext(), &Lsubst, cont_RightContext(), &Rsubst);
	for (j = 0; j < n; j++) {
	  if (j != i)
	    termlistLsubst =
	      list_Cons(subst_Apply(Lsubst,
			  term_Copy(clause_GetLiteralTerm(clauseL,j))),
			termlistLsubst);
	}
	for (j = 0;
	     j < clause_Length(clause_LiteralOwningClause(unilitres));
	     j++) {
	  if (!term_Equal(uniterm,clause_GetLiteralTerm(
                 clause_LiteralOwningClause(unilitres),j)))
	    termlistRsubst =
	      list_Cons(subst_Apply(Rsubst, term_Copy(clause_GetLiteralTerm(
			  clause_LiteralOwningClause(unilitres),j))),
			termlistRsubst);
	}
	termlist      = list_Nconc(termlistLsubst,termlistRsubst);
	resclause     = clause_CreateFromLiterals(termlist,FALSE,FALSE,TRUE);
	term_StartMinRenaming();
	for (j=0;j<clause_Length(resclause);j++)
	  term_Rename(clause_GetLiteralTerm(resclause,j));

	term_StartMinRenaming();
	for (j = 0; j < clause_Length(clauseL); j++)
	  term_Rename(clause_GetLiteralTerm(clauseL,j));

	if (!res_HasTautology(resclause))
	  resolvents = list_Cons(resclause,resolvents);
	else
	  clause_Delete(resclause);

      } else{
	printf("Error in res_BuildAllResolvents!\n");
	misc_DumpCore();
      }
      
      cont_Reset();

      list_Delete(termlist);
      subst_Delete(Lsubst);
      subst_Delete(Rsubst);
    }
    list_Delete(unifiables);
  }
  return resolvents;
}


LIST res_ResBuildAllResolventsOpt(CLAUSE clauseL, st_INDEX stindex)
/**********************************************************
  INPUT:   A clause and a st_Index.
  RETURNS: A list of all resolvents of clauseL with the
           clauses in stindex which are worked off
  CAUTION: None.
***********************************************************/
{
  int    j,i,n; 
  LIST   scan,unifiables,unifiableshelp,resolvents,termlist,
         termlistLsubst,termlistRsubst;
  TERM   term,negterm,uniterm;
  SYMBOL maxvar;
  SUBST  Lsubst,Rsubst;
  CLAUSE resclause;
  LITERAL unilitres,litres;

  resolvents = list_Nil();
  n          = clause_Length(clauseL);
  
  for (i = 0; i < n; i++) {
    term            = clause_GetLiteralTerm(clauseL,i);
    negterm         = fol_ComplementaryTerm(term); 
    unifiableshelp  = st_GetUnifier(cont_LeftContext(), stindex, cont_RightContext(), negterm);
    unifiables      = list_Nil();

    for (scan=unifiableshelp; !list_Empty(scan); scan=list_Cdr(scan)) {
      litres = (LITERAL)list_Car(scan);
      if (clause_GetFlag(clause_LiteralOwningClause(litres),WORKEDOFF))
	unifiables = list_Cons(list_Car(scan),unifiables);
    }
    list_Delete(unifiableshelp);
    term_Delete(negterm);
    maxvar = res_GetMaxVarClause(clauseL);

    for (scan=unifiables; !list_Empty(scan); scan=list_Cdr(scan)) {
      termlist       = list_Nil();
      termlistLsubst = list_Nil();
      termlistRsubst = list_Nil();
      unilitres  = (LITERAL)list_Car(scan);
      if (res_GetMaxVarClause(clause_LiteralOwningClause(unilitres)) > maxvar)
	maxvar = res_GetMaxVarClause(clause_LiteralOwningClause(unilitres));
      term_StartMaxRenaming(maxvar);
      for (j = 0; j < n; j++)
	term_Rename(clause_GetLiteralTerm(clauseL,j));
      uniterm = clause_LiteralSignedAtom(unilitres);

      cont_Check();

      if (unify_UnifyNoOC(cont_LeftContext(),
			  fol_Atom(term),
			  cont_RightContext(),
			  fol_Atom(uniterm))){
	subst_ExtractUnifier(cont_LeftContext(), &Lsubst, cont_RightContext(), &Rsubst);
	for (j = 0; j < n; j++) {
	  if (j != i)
	    termlistLsubst =
	      list_Cons(subst_Apply(Lsubst,
			  term_Copy(clause_GetLiteralTerm(clauseL,j))),
			termlistLsubst);
	}
	for (j = 0;
	     j < clause_Length(clause_LiteralOwningClause(unilitres));
	     j++){
	  if (!(term_Equal(uniterm,clause_GetLiteralTerm(
			     clause_LiteralOwningClause(unilitres),j))))
	    termlistRsubst =
	      list_Cons(subst_Apply(Rsubst,term_Copy(clause_GetLiteralTerm(
			  clause_LiteralOwningClause(unilitres),j))),
			termlistRsubst);
	}
	termlist      = list_Nconc(termlistLsubst,termlistRsubst);
	resclause     = clause_CreateFromLiterals(termlist,FALSE,FALSE,TRUE);
	term_StartMinRenaming();
	for (j = 0; j < clause_Length(resclause); j++)
	  term_Rename(clause_GetLiteralTerm(resclause,j));

	term_StartMinRenaming();
	for (j = 0; j < clause_Length(clauseL); j++)
	  term_Rename(clause_GetLiteralTerm(clauseL,j));

	if (!res_HasTautology(resclause))
	  resolvents = list_Cons(resclause,resolvents);
	else
	  clause_Delete(resclause);
      } else{
	printf("Error in res_BuildAllResolvents!\n");
	misc_DumpCore();
      }

      cont_Reset();
      
      list_Delete(termlist);
      subst_Delete(Lsubst);
      subst_Delete(Rsubst);
    }
    list_Delete(unifiables);
  }
  return resolvents;
}


LIST res_GetSubsumendClauses(LIST subslist, LIST list1, LIST list2)
/**********************************************************
  INPUT:   Three  lists of clauses.
  RETURNS: The list of clauses of subslist which are subsumed
           by a clause from list1 or list2.
  CAUTION: None.
***********************************************************/
{
  LIST result,scan,scan1;

  result = list_Nil();

  for (scan = list1; !list_Empty(scan); scan = list_Cdr(scan))
    for (scan1  = subslist; !list_Empty(scan1); scan1 = list_Cdr(scan1))
      if (subs_Idc(list_Car(scan), list_Car(scan1)))
	if (!list_PointerMember(result, list_Car(scan1)))
	  result = list_Cons(list_Car(scan1), result);

  for (scan = list2; !list_Empty(scan); scan = list_Cdr(scan))
    for (scan1 = subslist; !list_Empty(scan1); scan1=list_Cdr(scan1))
      if (subs_Idc(list_Car(scan), list_Car(scan1)))
	if (!list_PointerMember(result, list_Car(scan1)))
	  result = list_Cons(list_Car(scan1), result);
  return result;
}


LIST res_ResSubs(LIST usable, LIST workedoff, int resnumber)
/**********************************************************
  INPUT:   A list of clauses and an integer.
  RETURNS: A list of all resolvents of clauseL with the
           clauses in stindex which are worked off
  CAUTION: <clauselist> is consumed. Not used.
***********************************************************/
{
  LIST     resolvents,subslist,scan,condlist;
  CLAUSE   clause;
  st_INDEX stindex; 
  
  stindex      = st_IndexCreate();

  for (scan = usable; !list_Empty(scan); scan = list_Cdr(scan))
    res_InsertClauseIndex(list_Car(scan),stindex);

  if (!list_Empty(workedoff))
    for (scan = workedoff; !list_Empty(scan); scan = list_Cdr(scan)){
      clause_SetFlag(list_Car(scan),WORKEDOFF);
      res_InsertClauseIndex(list_Car(scan),stindex);
    }

  while ((!list_Empty(usable)) && (resnumber != 0)) {
    clause     = res_SelectLightestClause(usable);
    usable     = list_PointerDeleteElement(usable,clause);
    resolvents = res_ResBuildAllResolvents(clause,stindex);
    workedoff  = list_Cons(clause,workedoff);
    clause_SetFlag(clause,WORKEDOFF);             

    if (res_IsEmptyClauseInList(resolvents)){
      resnumber  = 0;
      printf("\nJucheeeeeee !!!!!!!!!\n");
    } else {
      /*resnumber -= list_Length(resolvents);*/

      subslist = res_GetSubsumendClauses(resolvents,workedoff,usable);
      for (scan = subslist; !list_Empty(scan); scan = list_Cdr(scan))
	resolvents = list_PointerDeleteElement(resolvents,list_Car(scan));
      for (scan = resolvents; !list_Empty(scan); scan = list_Cdr(scan)){
	condlist = cond_CondFast(list_Car(scan));
	if (!list_Empty(condlist))
	  list_Rplaca(scan, res_CondenseClause(list_Car(scan),condlist));
	list_Delete(condlist);
      }
      list_DeleteWithElement(subslist,clause_Delete);

      subslist = res_GetSubsumendClauses(workedoff,resolvents,list_Nil());
      for (scan = subslist; !list_Empty(scan); scan = list_Cdr(scan)) {
	workedoff = list_PointerDeleteElement(workedoff,list_Car(scan));
	res_DeleteClauseIndex(list_Car(scan),stindex);
      }
      list_DeleteWithElement(subslist,clause_Delete);

      subslist = res_GetSubsumendClauses(usable,resolvents,list_Nil());
      for (scan = subslist; !list_Empty(scan); scan = list_Cdr(scan)) {
	usable = list_PointerDeleteElement(usable,list_Car(scan));
	res_DeleteClauseIndex(list_Car(scan),stindex);
      }
      list_DeleteWithElement(subslist,clause_Delete);
    }
    for (scan = resolvents; !list_Empty(scan); scan = list_Cdr(scan)){
      clause_RemoveFlag((CLAUSE)list_Car(scan),WORKEDOFF);
      res_InsertClauseIndex((CLAUSE)list_Car(scan),stindex);
    }
    usable = list_Nconc(usable,resolvents);
  }
  usable = list_Nconc(workedoff,usable);
  st_IndexDelete(stindex);
  return usable;
}


BOOL res_BackSub(CLAUSE clause, st_INDEX stindex)
/**********************************************************
  INPUT:   A clauses and an index.
  RETURNS: TRUE if a clause of the index subsumes the clause clause.
  CAUTION: None.
***********************************************************/
{
  int    n,i; 
  LIST   scan,generals;
  TERM   term;
  LITERAL litres;

  n = clause_Length(clause);
  for (i = 0; i < n; i++) {
    term      = clause_GetLiteralTerm(clause,i);
    generals  = st_GetGen(cont_LeftContext(), stindex, term);
	
    for (scan = generals; !list_Empty(scan); scan = list_Cdr(scan)){
      litres = (LITERAL)list_Car(scan);
      if (litres == clause_GetLiteral(clause_LiteralOwningClause(litres),0)) {
	if (subs_Idc(clause_LiteralOwningClause(litres),clause)){
	  list_Delete(generals);
	  return TRUE;
	}
      }
    }
    list_Delete(generals);
  }
  return FALSE;
}


BOOL res_BackSubWithLength(CLAUSE clause, st_INDEX stindex)
/**********************************************************
  INPUT:   A clauses and an index.
  RETURNS: TRUE if a clause of the index subsumes the clause clause
           and length(clause) >= length(clause of index).
  CAUTION: None.
***********************************************************/
{
  int    n,i; 
  LIST   scan,generals;
  TERM   term;
  LITERAL litres;

  n = clause_Length(clause);
  for (i = 0; i < n; i++) {
    term      = clause_GetLiteralTerm(clause,i);
    generals  = st_GetGen(cont_LeftContext(), stindex, term);
    for (scan = generals; !list_Empty(scan); scan = list_Cdr(scan)) {
      litres = (LITERAL) list_Car(scan);
      if (litres == clause_GetLiteral(clause_LiteralOwningClause(litres),0) &&
	  clause_Length(clause) >= clause_Length(clause_LiteralOwningClause(litres)) &&
	  clause_Weight(clause) >= clause_Weight(clause_LiteralOwningClause(litres)) &&
	  subs_Idc(clause_LiteralOwningClause(litres),clause)) {
	    list_Delete(generals);
	    return TRUE;
      }
    }
    list_Delete(generals);
  }
  return FALSE;
}


LIST res_ForwardSub(CLAUSE clause, st_INDEX stindex)
/**********************************************************
  INPUT:   A clause and an index.
  RETURNS: The list of clauses from the stindex which are 
           subsumed by clause.
  CAUTION: None.
***********************************************************/
{
  LIST   scan,instances,result;
  TERM   term;
  LITERAL litres;

  result = list_Nil();

  if (clause_Length(clause)>0){
    term      = clause_GetLiteralTerm(clause,0);
    instances = st_GetInstance(cont_LeftContext(), stindex, term);
	
    for (scan = instances; !list_Empty(scan); scan = list_Cdr(scan)){
      litres = (LITERAL)list_Car(scan);
      if ((subs_Idc(clause,clause_LiteralOwningClause(litres))) && 
	  (!list_PointerMember(result,clause_LiteralOwningClause(litres))))
	result = list_Cons(clause_LiteralOwningClause(litres),result);
    }
    list_Delete(instances);
  }
  return result;
}


void res_ClauseTermNormalize(CLAUSE clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: None.
  CAUTION: The terms of the clause are destrucively normalized.
***********************************************************/
{
  int n,j;

  n = clause_Length(clause);

  term_StartMinRenaming();
  for (j = 0; j < n; j++)
    term_Rename(clause_GetLiteralTerm(clause,j));
}


LIST res_BuildFactors(CLAUSE clause)
/**********************************************************
  INPUT:   A clause.
  RETURNS: a list of all one step factors that can be build from <clause>,
  CAUTION: None.
***********************************************************/
{
  int    j,i,n; 
  TERM   L1,L2;
  LIST   Factors;

  n       = clause_Length(clause);
  Factors = list_Nil();
  
  for (i = 0; i < n - 1; i++) {
    L1 = clause_GetLiteralTerm(clause,i);
    for (j = i + 1; j < n; j++) {
      L2   = clause_GetLiteralTerm(clause,j);

      cont_Check();

      if (unify_UnifyCom(cont_LeftContext(), L1, L2)) {
	int    k;
	SUBST  Subst;
	CLAUSE Factor;
	LIST   Literals;
	subst_ExtractUnifierCom(cont_LeftContext(), &Subst);
	Literals = list_Nil();
	for (k = 0; k < n; k++)
	  if (k != i)
	    Literals =
	      list_Cons(subst_Apply(Subst,
			  term_Copy(clause_GetLiteralTerm(clause,k))),
			Literals);
	Factor = clause_CreateFromLiterals(Literals,FALSE,FALSE,TRUE);
	term_StartMinRenaming();
	for (k = 0; k < n - 1; k++)
	  term_Rename(clause_GetLiteralTerm(Factor,k));
	Factors = list_Cons(Factor,Factors);
	list_Delete(Literals);
	subst_Delete(Subst);
      }

      cont_Reset();
    }
  }
  return Factors;
}


LIST res_ResSubsIndex(LIST usable, LIST workedoff,int resnumber,int weight,
		      BOOL* found)
/**********************************************************
  INPUT:   A list of clauses and an integer.
  RETURNS: A list of all resolvents of clauseL with the
           clauses in stindex which are worked off in the
           clause structure.
  CAUTION: <clauselist> is consumed
***********************************************************/
{
  LIST     resolvents,subsindexlist,scan,scan1,condlist;
  CLAUSE   clause,actres;
  st_INDEX stindex; 

  stindex = st_IndexCreate();

  for (scan = usable; !list_Empty(scan); scan = list_Cdr(scan))
    res_InsertClauseIndex(list_Car(scan),stindex);
  for (scan = workedoff; !list_Empty(scan); scan = list_Cdr(scan)) {
    clause_SetFlag(list_Car(scan),WORKEDOFF);
    res_InsertClauseIndex(list_Car(scan),stindex);
  }
  while (!list_Empty(usable) && (resnumber > 0)) {	
    clause = res_SelectLightestClause(usable);
    if (clause_Weight(clause) > weight)
      resnumber = 0;
    else {
      resolvents   = res_BuildFactors(clause);
      usable       = list_PointerDeleteElement(usable,clause);
      resolvents   = list_Nconc(resolvents,
				res_ResBuildAllResolvents(clause,stindex));

      /* res_ResBuildAllResolventsOpt makes all possible resolvents
	 with clauses in stindex, except the */
      /* clause clause itself, i.e., no self resolvents are build */
      workedoff    = list_Cons(clause,workedoff);
      clause_SetFlag(clause,WORKEDOFF);

      if (res_IsEmptyClauseInList(resolvents)){
	resnumber  = 0;
	usable = list_Nconc(resolvents,usable);
	*found = TRUE;
      } else {
	resnumber -= list_Length(resolvents);
	for (scan = resolvents; !list_Empty(scan); scan = list_Cdr(scan)) {
	  condlist = cond_CondFast(list_Car(scan));
	  if (!list_Empty(condlist)){
	    list_Rplaca(scan,res_CondenseClause(list_Car(scan),condlist));
	    res_ClauseTermNormalize(list_Car(scan));
	  }
	  list_Delete(condlist);
	}
	for (scan = resolvents; !list_Empty(scan); scan = list_Cdr(scan)){
	  actres = (CLAUSE)list_Car(scan);
	  if (res_BackSubWithLength(actres,stindex)){
	    clause_Delete(actres);
	  } else {
	    subsindexlist = res_ForwardSub(actres,stindex);  
	    for (scan1=subsindexlist;!list_Empty(scan1);scan1=list_Cdr(scan1)){
	      if (clause_GetFlag(list_Car(scan1),WORKEDOFF))
		workedoff = list_PointerDeleteElement(workedoff,
						      list_Car(scan1));
	      else
		usable    = list_PointerDeleteElement(usable,list_Car(scan1));
	      res_DeleteClauseIndex(list_Car(scan1), stindex);
	      clause_Delete(list_Car(scan1));
	    }

	    list_Delete(subsindexlist);
	    clause_RemoveFlag(actres,WORKEDOFF);
	    /*clause_Print(actres);*/ 
	    res_InsertClauseIndex(actres,stindex);
	    usable = list_Cons(actres,usable);
	  }
	}
	list_Delete(resolvents);
      }
    }
  }
  usable = list_Nconc(workedoff,usable);
  st_IndexDelete(stindex);
  return usable;
}
