/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *         RECURSIVE PATH ORDERING WITH STATUS            * */
/* *                                                        * */
/* *  $Module:   RPOS                                       * */
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.77 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/07/20 20:56:39 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

     
/* $RCSfile: rpos.c,v $ */

#include "rpos.h"

const NAT RPOS_NONGROUND = 0;
const NAT RPOS_DONTKNOW  = 1;

static RPOS_TERM rpos_TermCreate(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A new RPOS_TERM with no subterms.
***************************************************************/
{
  RPOS_TERM result   = memory_Malloc(sizeof(RPOS_TERMNODE));
  result->term       = NULL;
  result->parent     = NULL;
  result->children   = list_Nil();
  result->ord        = RPOS_DONTKNOW;
  result->size       = 1;   /* Every term has at least one symbol */
  result->count      = 0;
  result->newSubList = list_Nil();
  result->smaller    = list_Nil();
  result->equal      = list_Nil();
  result->visited    = FALSE;
  return result;
}

static void rpos_TermDelete(RPOS_TERM RTerm)
/**************************************************************
  INPUT:   A RPOS_TERM.
  EFFECT:  The RPOS_TERM structure with all its subterms
           is deleted.
***************************************************************/
{
  LIST scan;

  for (scan = RTerm->children; !list_Empty(scan); scan = list_Pop(scan))
    rpos_TermDelete(list_Car(scan));
  for (scan = RTerm->newSubList; !list_Empty(scan); scan = list_Pop(scan))
    rpos_TermDelete(list_Car(scan));
  list_Delete(RTerm->smaller);
  list_Delete(RTerm->equal);
  memory_Free(RTerm, sizeof(RPOS_TERMNODE));
}


static void rpos_CountVariables(TERM Term, BOOL *Ground, NAT Index)
/**************************************************************
  INPUT:   A term, a BOOL pointer and an index
  RETURNS: <Ground> is set to TRUE, if <Term> is a ground term,
           FALSE otherwise.
  EFFECT:  The global variable <ord_VARCOUNT> is used to count
           the variable occurrences in <Term>. Since <ord_VARCOUNT>
	   is a two-dimensional array <Index> gives the second index
	   for the array access. <Index> must be 0 or 1.
	   The elements of <ord_VARCOUNT> must be set to 0 before
	   calling this function, since this function just increases
	   those counters!
***************************************************************/
{
  int  stack;

#ifdef DEBUG
  if (!term_IsTerm(Term) || Index > 1) {
    fputs("\nIllegal Input in function rpos_CountVariables\n", stderr);
    misc_DumpCore();
  }
#endif

  stack   = stack_Bottom();
  *Ground = TRUE;             /* Assume term is ground */

  do {
    if (term_IsStandardVariable(Term)) {
      ord_VARCOUNT[term_TopSymbol(Term)][Index]++;
      *Ground = FALSE;
    } else {
      if (term_IsComplex(Term))
	stack_Push(term_ArgumentList(Term));
    }

    while (!stack_Empty(stack) && list_Empty(stack_Top()))
      stack_Pop();

    if (!stack_Empty(stack)) {
      Term = list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(stack));
}


static void rpos_VarCondition(TERM Term1, BOOL *VarCond1, BOOL *Ground1,
			      TERM Term2, BOOL *VarCond2, BOOL *Ground2)
/**************************************************************
  INPUT:   Two terms and four pointers to booleans.
  RETURNS: VarCond1 is set to FALSE, if a variable occurs less often
           in Term1 than in Term2.
	   VarCond2 is set to FALSE, if a variable occurs less often
	   in Term2 than in Term1.
	   Ground1 is set to TRUE, if Term1 is a ground term.
	   Ground2 is set to TRUE, if Term2 is a ground term.
***************************************************************/
{
  SYMBOL maxVar1, maxVar2;
  NAT    i;

  *VarCond1 = *VarCond2 = TRUE;
  maxVar1   = term_MaxVar(Term1);
  maxVar2   = term_MaxVar(Term2);

  if (symbol_GreaterVariable(maxVar2, maxVar1))
    maxVar1 = maxVar2;

  for (i = 0; i <= maxVar1; i++) {
    ord_VARCOUNT[i][0] = 0;
    ord_VARCOUNT[i][1] = 0;
  }

  rpos_CountVariables(Term1, Ground1, 0);
  rpos_CountVariables(Term2, Ground2, 1);

  for (i = 0; i <= maxVar1; i++) {
    if (ord_VARCOUNT[i][0] < ord_VARCOUNT[i][1]) {
      *VarCond1 = FALSE;
      if (!*VarCond2)
	return;
    }
    else if (ord_VARCOUNT[i][0] > ord_VARCOUNT[i][1]) {
      *VarCond2 = FALSE;
      if (!*VarCond1)
	return;
    }
  }
}


static RPOS_TERM rpos_InitComparison(TERM Term, PRIORITY_QUEUE PQueue)
/**************************************************************
  INPUT:   A term and a priority queue.
  RETURNS: A RPOS_TERM corresponding to <Term>.
  EFFECT:  All constants from <Term> are inserted into <PQueue>.
           For non-ground terms the "ord" entry is set to
	   RPOS_NONGROUND, for ground terms it's set to
	   RPOS_DONTKNOW.
	   For terms whose top symbol has ORD_RIGHT status
	   the argument list is reversed.
***************************************************************/
{
  RPOS_TERM rTerm;
  LIST      scan;

  rTerm = rpos_TermCreate();
  rTerm->term = Term;

  if (symbol_IsVariable(term_TopSymbol(Term)))
    rTerm->ord = RPOS_NONGROUND;
  else if (!term_IsComplex(Term))
    pq_Insert(PQueue, rTerm);   /* Insert contants and 0-ary predicates */
  else {  /* Term is complex */
    if (symbol_HasProperty(term_TopSymbol(Term), ORDRIGHT))
      /* For RIGHT symbols reverse the argument list of the term */
      rTerm->children = list_Reverse(term_ArgumentList(Term));
    else
      /* For LEFT and MUL symbols simply copy the argument list */
      rTerm->children = list_Copy(term_ArgumentList(Term));

    for (scan = rTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      RPOS_TERM child = rpos_InitComparison(list_Car(scan), PQueue);
      list_Rplaca(scan, child);
      child->parent = rTerm;
      if (child->ord == RPOS_NONGROUND)
	rTerm->ord = RPOS_NONGROUND;
      rTerm->size += child->size;
    }
  }
  return rTerm;
}


static RPOS_TERM rpos_ContInitComparison(CONTEXT Context, TERM Term, PRIORITY_QUEUE PQueue)
/**************************************************************
  INPUT:   A term and a priority queue.
  RETURNS: A RPOS_TERM corresponding to <Term>.
  EFFECT:  This function is the same as rpos_InitComparison except
           is considers variable bindings in <Context>.
	   If a variable is bound to a term, a new RPOS_TERM
	   is created for this term. In common this means the
	   resulting RPOS_TERM differs from <Term>!
***************************************************************/
{
  RPOS_TERM rTerm;
  LIST      scan;

  Term  = cont_Deref(&Context, Term);
  rTerm = rpos_TermCreate();
  rTerm->term = Term;

  if (symbol_IsVariable(term_TopSymbol(Term)))
    rTerm->ord = RPOS_NONGROUND;
  else if (!term_IsComplex(Term))
    pq_Insert(PQueue, rTerm);   /* Insert contants and 0-ary predicates */
  else {  /* Term is complex */
    if (symbol_HasProperty(term_TopSymbol(Term), ORDRIGHT))
      /* For RIGHT symbols reverse the argument list of the term */
      rTerm->children = list_Reverse(term_ArgumentList(Term));
    else
      /* For LEFT and MUL symbols simply copy the argument list */
      rTerm->children = list_Copy(term_ArgumentList(Term));

    for (scan = rTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      RPOS_TERM child = rpos_ContInitComparison(Context, list_Car(scan), PQueue);
      list_Rplaca(scan, child);
      child->parent = rTerm;
      if (child->ord == RPOS_NONGROUND)
	rTerm->ord = RPOS_NONGROUND;
      rTerm->size += child->size;
    }
  }
  return rTerm;
}


static ord_RESULT rpos_GroundTermCompare(RPOS_TERM RT1, RPOS_TERM RT2)
/**************************************************************
  INPUT:   Two RPOS_TERMs representing ground terms.
  RETURNS: ord_SMALLER_THAN, if RT1 is smaller than Rt2.
           ord_GREATER_THAN, if RT1 is greater than RT2.
	   ord_EQUAL, if RT1 is equal to RT2 wrt. the RPOS ordering.
  EFFECT:  This function is used for comparisons in the priority queue.
           See [Snyder93] for details..
	   This function only works correctly, if the arguments
           of both terms are already ordered, that means the <ord> field
	   of the arguments, must be set correctly.
***************************************************************/
{
  SYMBOL s1, s2;

  if (RT1->term == RT2->term)
    return ord_EQUAL;     /* A shortcut for shared terms ;-) */

  s1 = term_TopSymbol(RT1->term);
  s2 = term_TopSymbol(RT2->term);

  if (symbol_PrecedenceGreater(s1, s2))
    return ord_GREATER_THAN;
  else if (symbol_PrecedenceGreater(s2, s1))
    return ord_SMALLER_THAN;
  else {
    /* Symbols have same precedence => they must be equal */
    LIST scan1, scan2;
    
    for (scan1 = RT1->children, scan2 = RT2->children;
	 !list_Empty(scan1) && !list_Empty(scan2);
	 scan1 = list_Cdr(scan1), scan2 = list_Cdr(scan2)) {
      RT1 = list_Car(scan1);
      RT2 = list_Car(scan2);

      if (RT1->ord > RT2->ord)
	return ord_GREATER_THAN;
      else if (RT1->ord < RT2->ord)
	return ord_SMALLER_THAN;
    }

    if (list_Empty(scan1) && list_Empty(scan2))
      return ord_EQUAL;
    else if (list_Empty(scan1))
      /* Can't happen?  Terms with equal symbols should have same arity? */
      return ord_SMALLER_THAN;
    else
      /* Can't happen?  Terms with equal symbols should have same arity? */
      return ord_GREATER_THAN;
  }
}


static BOOL rpos_GroundTermEqual(RPOS_TERM RT1, RPOS_TERM RT2)
/**************************************************************
  INPUT:   Two RPOS_TERMs representing ground terms.
  RETURNS: TRUE, if RT1 is equal to RT2, FALSE otherwise.
  EFFECT:  This function is used by the ground term algorithm
           from [Snyder93].
***************************************************************/
{
  return rpos_GroundTermCompare(RT1, RT2) == ord_EQUAL;
}


static BOOL rpos_GroundTermSmaller(RPOS_TERM RT1, RPOS_TERM RT2)
/**************************************************************
  INPUT:   Two RPOS_TERMs representing ground terms.
  RETURNS: TRUE, if RT1 is smaller than RT2, FALSE otherwise.
  EFFECT:  This function is used by the ground term algorithm
           from [Snyder93].
***************************************************************/
{
  return rpos_GroundTermCompare(RT1, RT2) == ord_SMALLER_THAN;
}


static void rpos_UpdateMUL(RPOS_TERM RTerm)
{
  /* Update the argument list of MUL terms   */
  BOOL mul;
  LIST scan, nground;

  if (term_IsComplex(RTerm->term)) {
    nground = list_Nil();
    mul = symbol_HasProperty(term_TopSymbol(RTerm->term), ORDMUL);

    for (scan = RTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      rpos_UpdateMUL(list_Car(scan));
      if (mul && ((RPOS_TERM)list_Car(scan))->ord == RPOS_NONGROUND)
	nground = list_Cons(list_Car(scan), nground);
    }
    
    if (mul) {
      if (RTerm->ord == RPOS_NONGROUND) {
	/* Put all (ordered) ground args into <children> and */
	/* all non-ground args into <newSubList>             */
	list_Delete(RTerm->children);
	RTerm->children   = RTerm->newSubList;
	RTerm->newSubList = nground;
      } else {
	/* For ground MUL terms that where not completely ordered */
	/* (by the rpos_ShortGroundCompare function) delete the   */
	/* list of already ordered arguments. Otherwise we'd have */
	/* some trouble in the rpos_TermDelete function because   */
	/* some arguments might occur both in the children and    */
	/* newSubList lists.                                  */
	if (RTerm->ord < symbol_Arity(term_TopSymbol(RTerm->term)))
	  list_Delete(RTerm->newSubList);
      }
    }
  }
}

      
static LIST rpos_UpdateParent(RPOS_TERM RTerm)
{
  RPOS_TERM parent;
  LIST      result;

  parent = RTerm->parent;
  result = list_Nil();

  if (parent != NULL) {
    if (symbol_HasProperty(term_TopSymbol(parent->term), ORDMUL))
      parent->newSubList = list_Cons(RTerm, parent->newSubList);
    
    parent->count++;
    
    if (parent->count == symbol_Arity(term_TopSymbol(parent->term))) {
      if (symbol_HasProperty(term_TopSymbol(parent->term), ORDMUL)) {
	/* Now newSubList is the ordered list of parent's subterms.      */
	/* We want to use this list instead of the original subterm list */
	/* for the '<' of the priority queue.                            */ 
	list_Delete(parent->children);
	parent->children = parent->newSubList;
	parent->newSubList = list_Nil();
      }
      result = list_Cons(parent, result); /* parent is a new candidate */
    }
  }
  return result;
}


static ord_RESULT rpos_ShortCompareGround(CONTEXT Context1, TERM T1,
					  CONTEXT Context2, TERM T2)
{
  ord_RESULT     result;
  PRIORITY_QUEUE pqueue;  /* corresponds to candidates in [Snyder93] */
  RPOS_TERM      rTerm1, rTerm2;
  LIST           newCandidates;
  BOOL           T1finished, T2finished;
  NAT            order;

#ifdef DEBUG
  if ((Context1 == NULL && Context2 != NULL) ||
      (Context1 != NULL && Context2 == NULL)) {
    fputs("\nIllegal input in rpos_ShortCompareGround\n", stderr);
    misc_DumpCore();
  }
#endif

  newCandidates = list_Nil();
  T1finished    = T2finished = FALSE;
  order         = 1;  /* 0 is RPOS_NONGROUND and 1 is RPOS_DONTKNOW */

  /* Create the RPOS_TERM structures and init the priority queue */
  if (Context1 == NULL) {
    pqueue = pq_Create(term_ComputeSize(T1)+term_ComputeSize(T2), rpos_GroundTermSmaller);
    rTerm1 = rpos_InitComparison(T1, pqueue);
    rTerm2 = rpos_InitComparison(T2, pqueue);
  } else {
    /* Both contexts are != NULL */
    pqueue = pq_Create(cont_TermSize(Context1,T1)+cont_TermSize(Context2,T2),
		       rpos_GroundTermSmaller);
    rTerm1 = rpos_ContInitComparison(Context1, T1, pqueue);
    rTerm2 = rpos_ContInitComparison(Context2, T2, pqueue);
  }

  while (!pq_Empty(pqueue) && !T1finished && !T2finished) {
    RPOS_TERM s = pq_ExtractMin(pqueue);

    s->ord = ++order;
    newCandidates = rpos_UpdateParent(s);

    if (s == rTerm1)
      T1finished = TRUE;
    else if (s == rTerm2)
      T2finished = TRUE;

    while (!pq_Empty(pqueue) &&
	   rpos_GroundTermEqual(s, pq_GetMin(pqueue))) {
      /* The priority queue contains terms that are equal to s (wrt. RPOS) */
      s = pq_ExtractMin(pqueue);
      s->ord = order;
      newCandidates = list_Nconc(rpos_UpdateParent(s), newCandidates);

      if (s == rTerm1)
	T1finished = TRUE;
      else if (s == rTerm2)
	T2finished = TRUE;
    }

    while (!list_Empty(newCandidates)) {
      pq_Insert(pqueue, list_Car(newCandidates));
      newCandidates = list_Pop(newCandidates);
    }
  }

  result = ord_UNCOMPARABLE;
  if (T1finished) {
    if (T2finished)
      result = ord_EQUAL;
    else
      result = ord_SMALLER_THAN;
  }
  else if (T2finished)
    result = ord_GREATER_THAN;
  else {
    fputs("\nERROR: ground terms are uncomparable?\n", stderr);
    misc_DumpCore();
  }

  rpos_TermDelete(rTerm1);
  rpos_TermDelete(rTerm2);
  pq_Delete(pqueue);

  return result;
}

static void rpos_FullCompareGround(CONTEXT Context1, TERM T1,
				   CONTEXT Context2, TERM T2,
				   RPOS_TERM* rt1, RPOS_TERM* rt2)
{
  PRIORITY_QUEUE pqueue;  /* corresponds to candidates in [Snyder93] */
  RPOS_TERM      rTerm1, rTerm2;
  LIST           newCandidates;
  NAT            order;

#ifdef DEBUG
  if ((Context1 == NULL && Context2 != NULL) ||
      (Context1 != NULL && Context2 == NULL)) {
    fputs("\nIllegal input in rpos_ShortCompareGround\n", stderr);
    misc_DumpCore();
  }
#endif

  newCandidates = list_Nil();
  order         = 1;  /* 0 is RPOS_NONGROUND and 1 is RPOS_DONTKNOW */
 
  /* Create the RPOS_TERM structures and init the priority queue */
  if (Context1 == NULL) {
    pqueue = pq_Create(term_ComputeSize(T1)+term_ComputeSize(T2), rpos_GroundTermSmaller);
    rTerm1 = rpos_InitComparison(T1, pqueue);
    rTerm2 = rpos_InitComparison(T2, pqueue);
  } else {
    /* Both contexts are != NULL */
    pqueue = pq_Create(cont_TermSize(Context1,T1)+cont_TermSize(Context2,T2),
		       rpos_GroundTermSmaller);
    rTerm1 = rpos_ContInitComparison(Context1, T1, pqueue);
    rTerm2 = rpos_ContInitComparison(Context2, T2, pqueue);
  }

  while (!pq_Empty(pqueue)) {
    RPOS_TERM s = pq_ExtractMin(pqueue);

    s->ord = ++order;
    newCandidates = rpos_UpdateParent(s);

    while (!pq_Empty(pqueue) &&
	   rpos_GroundTermEqual(s, pq_GetMin(pqueue))) {
      /* The priority queue contains terms that are equal to s (wrt. RPOS) */
      s = pq_ExtractMin(pqueue);
      s->ord = order;
      newCandidates = list_Nconc(rpos_UpdateParent(s), newCandidates);
    }

    while (!list_Empty(newCandidates)) {
      pq_Insert(pqueue, list_Car(newCandidates));
      newCandidates = list_Pop(newCandidates);
    }
  }

  pq_Delete(pqueue);
  rpos_UpdateMUL(rTerm1);
  rpos_UpdateMUL(rTerm2);

  *rt1 = rTerm1;
  *rt2 = rTerm2;
}


static LIST rpos_SortByLength(RPOS_TERM RTerm)
{
  RPOS_TERM actTerm;
  LIST      *A, scan, result;        /* A is an array of lists */
  NAT       i, stack, size;

  A = (LIST*) memory_Malloc(sizeof(LIST) * RTerm->size);

  for (i = 0; i < RTerm->size; i++)
    A[i] = list_Nil();
  
  stack = stack_Bottom();
  A[RTerm->size - 1] = list_List(RTerm);

  if (!list_Empty(RTerm->children))    /* Ground arguments */
    stack_Push(RTerm->children);
  if (!list_Empty(RTerm->newSubList))  /* Non-ground arguments */
    stack_Push(RTerm->newSubList);

  while (!stack_Empty(stack)) {
    scan = stack_Top();
    actTerm = (RPOS_TERM) list_Car(scan);
    stack_RplacTop(list_Cdr(scan));

    size = actTerm->size;
    A[size - 1] = list_Cons(actTerm, A[size - 1]);

    if (!list_Empty(actTerm->children))    /* LEX all args, MUL ground args */
      stack_Push(actTerm->children);
    if (!list_Empty(actTerm->newSubList))  /* MUL only: non-ground args */
      stack_Push(actTerm->newSubList);
    
    while (!stack_Empty(stack) && list_Empty(stack_Top()))
      stack_Pop();
  }

  /*   for (i = 0; i < Length; i++) { */
  /*     printf("\nSize %d:", i+1); */
  /*     for (scan = A[i]; !list_Empty(scan); scan = list_Cdr(scan)) { */
  /*       putchar(' '); */
  /*       term_PrintPrefix(((RPOS_TERM) list_Car(scan))->term); */
  /*     } */
  /*   } */
  /*   fflush(stdout); DBG*/
  
  /* Now create the result list where the terms are sorted by their size */
  /* i is the index of the field for the next term */
  /* size is the index of the current list         */
  result = list_Nil();
  for (i = RTerm->size; i-- > 0; )
    result = list_Nconc(A[i], result);

  memory_Free(A, sizeof(LIST) * RTerm->size); 
  return result;
}


static ord_RESULT rpos_InternNGCompare(RPOS_TERM RT1, RPOS_TERM RT2)
/* For terms that were already compared */
{
  if (RT1->ord != RPOS_NONGROUND && RT2->ord != RPOS_NONGROUND) {
    /* Both terms are ground, so just compare the ord numbers */
    if (RT1->ord > RT2->ord)
      return ord_GREATER_THAN;
    else if (RT1->ord < RT2->ord)
      return ord_SMALLER_THAN;
    else
      return ord_EQUAL;    /* Ground terms are always comparable */
  }

  if (list_PointerMember(RT1->smaller, RT2))
    return ord_GREATER_THAN;
  if (list_PointerMember(RT2->smaller, RT1))
    return ord_SMALLER_THAN;
  if (list_PointerMember(RT1->equal, RT2))
    return ord_EQUAL;
  return ord_UNCOMPARABLE;
}


static ord_RESULT rpos_NGCompare(RPOS_TERM RT1, RPOS_TERM RT2)
{
  RPOS_TERM rTerm1, rTerm2, parent;
  LIST      list1, list2, scan1, scan2, argScan1, argScan2;
  BOOL      is_greater, is_smaller, equal;

  /*   puts("\n-----------------------"); */
  /*   rpos_TermPrint(RT1); fputs(" and ", stdout); */
  /*   rpos_TermPrint(RT2); putchar('\n'); DBG */

  list1 = rpos_SortByLength(RT1);
  list2 = rpos_SortByLength(RT2);
  /* We have two lists, one for each term, where the subterms are sorted */
  /* by their size. Now compare the terms from one list with all terms   */
  /* from the other list.                                                */

  /* First treat variables: if we find two equal variables add term1 to  */
  /* the equal list of term2 and vice versa. Now add term1 to the        */
  /* smaller lists of all parents of term2 and term2 to all parents of   */
  /* term1.                                                              */
  for (scan1 = list1; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
    rTerm1 = (RPOS_TERM) list_Car(scan1);
    if (rTerm1->size > 1)
      break;
    if (!term_IsStandardVariable(rTerm1->term))
      continue;

    for (scan2 = list2; !list_Empty(scan2); scan2 = list_Cdr(scan2)) {
      rTerm2 = (RPOS_TERM) list_Car(scan2);
      if (rTerm2->size > 1)
	break;
      if (term_EqualTopSymbols(rTerm1->term, rTerm2->term)) {
	/* Found the same variable */
	rTerm1->equal = list_Cons(rTerm2, rTerm1->equal);
	rTerm2->equal = list_Cons(rTerm1, rTerm2->equal);
	for (parent = rTerm1->parent; parent != NULL; parent = parent->parent)
	  parent->smaller = list_Cons(rTerm2, parent->smaller);
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent)
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
      }
    }
  }

  /* Now apply the ordering */
  for (scan1 = list1; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
    rTerm1 = (RPOS_TERM) list_Car(scan1);
    if (term_IsStandardVariable(rTerm1->term))
      continue;

    for (scan2 = list2; !list_Empty(scan2); scan2 = list_Cdr(scan2)) {
      LIST children1, children2, ngChildren1, ngChildren2;

      rTerm2 = (RPOS_TERM) list_Car(scan2);

      if (term_IsStandardVariable(rTerm2->term) ||
	  (rTerm1->ord != RPOS_NONGROUND && rTerm2->ord != RPOS_NONGROUND))
	/* Term2 is variable, already compared => continue */
	/* Both terms are ground => continue               */
	continue;

      if (rTerm2->visited) {
	rTerm2->visited = FALSE;
	continue;
      }

      /* RPOS case 2A and case 1B */
      /* Check if one argument of Term1 >= Term2 */
      /* Check if Term2 > all arguments of Term1 */
      /* and topSymbol(Term2) > topSymbol(Term1) */
      is_greater = FALSE;
      is_smaller = TRUE;
      for (argScan1 = rTerm1->children;
	   !list_Empty(argScan1) && !is_greater;
	   argScan1 = list_Cdr(argScan1)) {
	switch (rpos_InternNGCompare(list_Car(argScan1), rTerm2)) {
	case ord_EQUAL:
	case ord_GREATER_THAN:
	  is_greater = TRUE;
	case ord_UNCOMPARABLE:
	  is_smaller = FALSE;
	  break;
	default:
	  break;
	}
      }
      for (argScan1 = rTerm1->newSubList;
	   !list_Empty(argScan1) && !is_greater;
	   argScan1 = list_Cdr(argScan1)) {
	switch (rpos_InternNGCompare(list_Car(argScan1), rTerm2)) {
	case ord_EQUAL:
	case ord_GREATER_THAN:
	  is_greater = TRUE;
	case ord_UNCOMPARABLE:
	  is_smaller = FALSE;
	  break;
	default:
	  break;
	}
      }
      if (is_greater) {
	/* Argument of Term1 >= Term2 */
	rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	/* Don't use the parent "trick" from below */
	/* since Term1 is from the outer for-loop  */
	continue;
      }
      if (is_smaller &&
	  symbol_PrecedenceGreater(term_TopSymbol(rTerm2->term),
				   term_TopSymbol(rTerm1->term))) {
	/* Term2 > all arguments of Term1 and  */
	/* topsymbol(Term2) > topsymbol(Term1) */
	rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	/* Since Term1 is smaller than Term2, Term1 is also */
	/* smaller than all parents of Term2.               */
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
	continue;
      }
      
      /* RPOS case 2B and case 1A */
      /* Check if one argument of Term2 >= Term1 */
      /* Check if Term1 > all arguments of Term2 */
      /* and topSymbol(Term1) > topSymbol(Term2) */
      is_greater = TRUE;
      is_smaller = FALSE;
	
      for (argScan2 = rTerm2->children;
	   !list_Empty(argScan2) && !is_smaller;
	   argScan2 = list_Cdr(argScan2)) {
 	switch (rpos_InternNGCompare(rTerm1, list_Car(argScan2))) {
	case ord_EQUAL:
	case ord_SMALLER_THAN:
	  is_smaller = TRUE;
	case ord_UNCOMPARABLE:
	  is_greater = FALSE;
	  break;
 	default:
	  break;
	}
      }
      for (argScan2 = rTerm2->newSubList;
	   !list_Empty(argScan2) && !is_smaller;
	   argScan2 = list_Cdr(argScan2)) {
	switch (rpos_InternNGCompare(rTerm1, list_Car(argScan2))) {
	case ord_EQUAL:
	case ord_SMALLER_THAN:
	  is_smaller = TRUE;
	case ord_UNCOMPARABLE:
	  is_greater = FALSE;
	  break;
	default:
	  break;
	}
      }
      if (is_smaller) {
	/* Argument of Term2 >= Term1 */
	rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	/* Since Term1 is smaller than Term2, Term1 is also */
	/* smaller than all parents of Term2.               */
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
	continue; 
      }
      if (is_greater &&
	  symbol_PrecedenceGreater(term_TopSymbol(rTerm1->term),
		 		   term_TopSymbol(rTerm2->term))) {
	/* Term1 > all arguments of Term2 and  */
	/* topsymbol(Term1) > topsymbol(Term2) */
	rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	continue;
      }


      if (!term_EqualTopSymbols(rTerm1->term, rTerm2->term))
	/* Now we consider RPOS cases 3 and 4 where equal topsymbols are required.    */
	/* If they don't have the same topsymbols at this point they're uncomparable. */
	continue;

      
      /* RPOS Case 3: topsymbol(insertTerm) = topsymbol(iteratorTerm), LEX */
      if (!symbol_HasProperty(term_TopSymbol(rTerm1->term),ORDMUL)) {
	is_greater = FALSE;
	is_smaller = FALSE;
	equal      = TRUE;

	/* Compare the arguments of both terms as long as they're equal */
	for (argScan1 = rTerm1->children, argScan2 = rTerm2->children;
	     !list_Empty(argScan1) && equal;
	     argScan1 = list_Cdr(argScan1), argScan2 = list_Cdr(argScan2)) {
	  switch (rpos_InternNGCompare(list_Car(argScan1), list_Car(argScan2))) {
	  case ord_UNCOMPARABLE:
	    equal = FALSE;      /* Terms are uncomparable */
	    break;
	  case ord_SMALLER_THAN:
	    is_smaller = TRUE;
	    equal      = FALSE;
	    break;
	  case ord_GREATER_THAN:
	    is_greater = TRUE;
	    equal      = FALSE;
	    break;
	  default:
	    break;
	  }
	}
	  
	if (is_greater) {
	  /* Check if Term1 is greater than the remaining arguments of Term2 */
	  for ( ; !list_Empty(argScan2) && is_greater; argScan2 = list_Cdr(argScan2)) {
	    if (rpos_InternNGCompare(rTerm1, list_Car(argScan2)) != ord_GREATER_THAN)
	      is_greater = FALSE; 
	    /* ord_EQUAL and ord_SMALLER_THAN are impossible */
	    /* since they're handled by RPOS case 2B.        */
	  }
	}
	else if (is_smaller) {
	  /* Check if Term2 is greater than the remaining arguments of Term1 */
	  for ( ; !list_Empty(argScan1) && is_smaller; argScan1 = list_Cdr(argScan1)) {
	    if (rpos_InternNGCompare(list_Car(argScan1), rTerm2) != ord_SMALLER_THAN)
	      is_smaller = FALSE;
	    /* ord_EQUAL and ord_GREATER_THAN are impossible */
	    /* since they're handled by RPOS case 2A.        */
	  }
	}
	if (is_greater)
	  rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	else if (is_smaller)
	  rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	else if (equal) {
 	  rTerm1->equal = list_Cons(rTerm2, rTerm1->equal);
	  rTerm2->equal = list_Cons(rTerm1, rTerm2->equal);
	}
	if (is_smaller || equal) {
	  /* Since Term1 is smaller than or equal to Term2, Term1 is also */
	  /* smaller than all parents of Term2.                           */
	  for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	    parent->smaller = list_Cons(rTerm1, parent->smaller);
	    parent->visited = TRUE;
	  }
	}
	  
	continue;
      }


      /* RPOS Case 4: topsymbol(tail->t) = topsymbol(iterator->t), MUL */
      equal     = TRUE;
      children1 = children2 = list_Nil();
      /* First delete equal ground arguments. Since they're sorted in    */
      /* descending order, we later only consider the biggest non-equal. */ 
      for (argScan1 = rTerm1->children, argScan2 = rTerm2->children;
	   !list_Empty(argScan1) && equal;
	   argScan1 = list_Cdr(argScan1), argScan2 = list_Cdr(argScan2)) {
	if (((RPOS_TERM) list_Car(argScan1))->ord >
	    ((RPOS_TERM) list_Car(argScan2))->ord) {
	  equal     = FALSE;
	  children1 = list_Cons(list_Car(argScan1), children1);
	}
	else if (((RPOS_TERM) list_Car(argScan1))->ord <
		 ((RPOS_TERM) list_Car(argScan2))->ord) {
	  equal     = FALSE;
	  children2 = list_Cons(list_Car(argScan2), children2);
	}
      }
      /* Now delete equal non-ground subterms */
      ngChildren1 = list_Copy(rTerm1->newSubList);
      ngChildren2 = list_Copy(rTerm2->newSubList);
      for (argScan1 = ngChildren1; !list_Empty(argScan1); argScan1 = list_Cdr(argScan1))
	for (argScan2 = ngChildren2; !list_Empty(argScan2); argScan2 = list_Cdr(argScan2))
	  if (list_PointerMember(((RPOS_TERM)list_Car(argScan1))->equal,
				 list_Car(argScan2))) {
	    list_Rplaca(argScan1, NULL);
	    list_Rplaca(argScan2, NULL);
	    break;
	  }
      ngChildren1 = list_PointerDeleteElement(ngChildren1, NULL);
      ngChildren2 = list_PointerDeleteElement(ngChildren2, NULL);
      
      if (equal && list_Empty(ngChildren1)) {
	/* Terms are equal */
	rTerm1->equal = list_Cons(rTerm2, rTerm1->equal);
	rTerm2->equal = list_Cons(rTerm1, rTerm2->equal);
	/* Since Term1 is  equal to Term2, Term1 is */
	/* smaller than all parents of Term2.       */
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
	continue;
      }
	
      children1 = list_Nconc(children1, ngChildren1);
      children2 = list_Nconc(children2, ngChildren2);
      /* Now children1 contains args(Term1) \ args(Term2), */
      /* and children2 contains args(Term2) \ args(Term1)  */

      /* Case A: Check if Term1 > Term2 */
      ngChildren2 = list_Copy(children2);
      for (argScan1 = children1; !list_Empty(argScan1); argScan1 = list_Cdr(argScan1)) {
	for (argScan2 = ngChildren2; !list_Empty(argScan2);
	     argScan2 = list_Cdr(argScan2)) {
	  if (rpos_InternNGCompare(list_Car(argScan1), list_Car(argScan2))
	      == ord_GREATER_THAN) {
	    list_Rplaca(argScan2, NULL);
	  }
	}
	ngChildren2 = list_PointerDeleteElement(ngChildren2, NULL);
	if (list_Empty(ngChildren2))
	  break;
      }
      if (list_Empty(ngChildren2)) {
	/* Term1 > Term2 */
	rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	list_Delete(children1);
	list_Delete(children2);
	continue;
      }
      list_Delete(ngChildren2);   /* Delete the copy */

      /* Case B: Check if Term1 > Term2 */
      ngChildren1 = list_Copy(children1);
      for (argScan2 = children2; !list_Empty(argScan2); argScan2 = list_Cdr(argScan2)) {
	for (argScan1 = ngChildren1; !list_Empty(argScan1);
	     argScan1 = list_Cdr(argScan1)) {
	  if (rpos_InternNGCompare(list_Car(argScan2), list_Car(argScan1))
	      == ord_GREATER_THAN) {
	    list_Rplaca(argScan1, NULL);
	  }
	}
	ngChildren1 = list_PointerDeleteElement(ngChildren1, NULL);
	if (list_Empty(ngChildren1))
	  break;
      }
      if (list_Empty(ngChildren1)) {
	/* Term2 > Term1*/
	rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	/* Since Term1 is smaller than Term2, Term1 is also */
	/* smaller than all parents of Term2.               */
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
      } else
	list_Delete(ngChildren1);
      list_Delete(children1);
      list_Delete(children2);
    }
  }

  /*   rpos_ListPrint(list1); putchar('\n'); rpos_ListPrint(list2); DBG */

  list_Delete(list1);
  list_Delete(list2);

  return rpos_InternNGCompare(RT1, RT2);
}


ord_RESULT rpos_BorromUpCompare(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: ord _GREATER_THAN if <T1> is greater than <T2> wrt
           the RPOS ordering.
	   ord_SMALLER_THAN if <T2> is greater than <T1>.
	   ord_EQUAL if both terms are equal wrt the RPOS ordering.
	   ord_UNCOMPARABLE otherwise.
***************************************************************/
{
  BOOL       varCond1, varCond2, ground1, ground2;
  ord_RESULT result;

  /*   puts("\n--------------------"); */
  /*   term_PrintPrefix(T1); fputs(" and ", stdout); */
  /*   term_PrintPrefix(T2); fflush(stdout); DBG */

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (term_ContainsSymbol(T2, term_TopSymbol(T1)))
      result = ord_SMALLER_THAN;
    else
      result = ord_UNCOMPARABLE;

    /*     fputs(", result: ", stdout); term_PrintPrefix(T1); */
    /*     putchar(' '); ord_Print(result); */
    /*     putchar(' '); term_PrintPrefix(T2);  fflush(stdout); DBG */
    return result;
  }

  if (term_IsVariable(T2)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (term_ContainsSymbol(T1, term_TopSymbol(T2)))
      result = ord_GREATER_THAN;
    else
      result = ord_UNCOMPARABLE;
    
    /*     fputs(", result: ", stdout); term_PrintPrefix(T1); */
    /*     putchar(' '); ord_Print(result); */
    /*     putchar(' '); term_PrintPrefix(T2); fflush(stdout); DBG */
    return result;
  }

  rpos_VarCondition(T1, &varCond1, &ground1, T2, &varCond2, &ground2);

  if (ground1 && ground2)
    /* Both terms are ground, so use the ground algorithm */
    result = rpos_ShortCompareGround(NULL, T1, NULL, T2); /* No contexts */
  else {
    /* At least one term is non-ground */    
    if (varCond1 || varCond2) {
      RPOS_TERM  rt1, rt2;

      rpos_FullCompareGround(NULL, T1, NULL, T2, &rt1, &rt2); /* No contexts */
      result = rpos_NGCompare(rt1, rt2);
      
      rpos_TermDelete(rt1);
      rpos_TermDelete(rt2);
    } else
      result = ord_UNCOMPARABLE;
    /* If in T1 a variable x occurs more often than in T2, and */
    /* in T2 a variable y occurs more often than in T1, the    */
    /* terms are uncomparable.                                 */
  }
  
  /*   fputs(", result: ", stdout); term_PrintPrefix(T1); */
  /*   putchar(' '); ord_Print(result); */
  /*   putchar(' '); term_PrintPrefix(T2); fflush(stdout); DBG */
  
  return result;
}


static void rpos_ContCountVariables(CONTEXT Context, TERM Term, BOOL *Ground, NAT Index)
/**************************************************************
  INPUT:   A context, a term, a BOOL pointer and an index
  RETURNS: <Ground> is set to FALSE, if <Term> is not a ground term,
           Attention: <Ground> must be set to TRUE before calling
	   this function!
  EFFECT:  The global variable <ord_VARCOUNT> is used to count
           the variable occurrences in <Term>. Since <ord_VARCOUNT>
	   is a two-dimensional array <Index> gives the second index
	   for the array access. <Index> must be 0 or 1.
	   The elements of <ord_VARCOUNT> must be set to 0 before
	   calling this function, since this function just increases
	   those counters!
	   Unlike rpos_CountVariables variable bindings in <Context>
	   are considered. This function has to be recursive
	   because it uses the cont_Deref function.
***************************************************************/
{
  LIST scan;

#ifdef DEBUG
  if (!term_IsTerm(Term) || Index > 1) {
    fputs("\nIllegal Input in function rpos_CountVariables\n", stderr);
    misc_DumpCore();
  }
#endif

  Term = cont_Deref(&Context, Term);

  if (term_IsStandardVariable(Term)) {
    ord_VARCOUNT[term_TopSymbol(Term)][Index]++;
    *Ground = FALSE;
  }
  else if (term_IsComplex(Term)) {
    for (scan = term_ArgumentList(Term); !list_Empty(scan); scan = list_Cdr(scan))
      rpos_ContCountVariables(Context, list_Car(scan), Ground, Index);
  }
}

static void rpos_ContVarCondition(CONTEXT Context1, TERM Term1,
				  BOOL *VarCond1, BOOL *Ground1,
				  CONTEXT Context2, TERM Term2,
				  BOOL *VarCond2, BOOL *Ground2)
/**************************************************************
  INPUT:   Two contexts, two terms and four pointers to booleans.
  RETURNS: Variables in Term1 are interpreted wrt Context1,
           variables in Term2 wrt Context2.
	   VarCond1 is set to FALSE, if a variable occurs less often
           in Term1 than in Term2.
	   VarCond2 is set to FALSE, if a variable occurs less often
	   in Term2 than in Term1.
	   Ground1 is set to TRUE, if Term1 is a ground term.
	   Ground2 is set to TRUE, if Term2 is a ground term.
	   Unlike rpos_VarCondition variable bindings in
	   the respective contexts are considered.
***************************************************************/
{
  SYMBOL maxVar1, maxVar2;
  NAT    i;

  *VarCond1 = *VarCond2 = TRUE;
  maxVar1   = cont_TermMaxVar(Context1, Term1);
  maxVar2   = cont_TermMaxVar(Context2, Term2);

  if (symbol_GreaterVariable(maxVar2, maxVar1))
    maxVar1 = maxVar2;

  for (i = 0; i <= maxVar1; i++) {
    ord_VARCOUNT[i][0] = 0;
    ord_VARCOUNT[i][1] = 0;
  }

  *Ground1 = *Ground2 = TRUE;
  rpos_ContCountVariables(Context1, Term1, Ground1, 0);
  rpos_ContCountVariables(Context2, Term2, Ground2, 1);

  for (i = 0; i <= maxVar1; i++) {
    if (ord_VARCOUNT[i][0] < ord_VARCOUNT[i][1]) {
      *VarCond1 = FALSE;
      if (!*VarCond2)  /* If both conditions are violated => uncomparable */
	return;
    }
    else if (ord_VARCOUNT[i][0] > ord_VARCOUNT[i][1]) {
      *VarCond2 = FALSE;
      if (!*VarCond1)
	return;
    }
  }
}


ord_RESULT rpos_BottomUpContCompare(CONTEXT Context1, TERM T1,
			    CONTEXT Context2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: ord _GREATER_THAN if <T1> is greater than <T2> wrt
           the RPOS ordering.
	   ord_SMALLER_THAN if <T2> is greater than <T1>.
	   ord_EQUAL if both terms are equal wrt the RPOS ordering.
	   ord_UNCOMPARABLE otherwise.
	   Unlike rpos_Compare variable bindings in <Context>
	   are considered.
***************************************************************/
{
  BOOL       varCond1, varCond2, ground1, ground2;
  ord_RESULT result;

  /* fputs("\n--------------------\nCONT ", stdout);
     cont_TermPrintPrefix(Context1, T1); fputs(" and ", stdout);
     cont_TermPrintPrefix(Context2, T2); fflush(stdout); DBG */

  T1 = cont_Deref(&Context1, T1);
  T2 = cont_Deref(&Context2, T2);

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (cont_TermContainsSymbol(Context2, T2, term_TopSymbol(T1)))
      result = ord_SMALLER_THAN;
    else
      result = ord_UNCOMPARABLE;
  
    /*     fputs(", result: ", stdout); term_PrintPrefix(T1); */
    /*     putchar(' '); ord_Print(result); */
    /*     putchar(' '); term_PrintPrefix(T2); DBG */
    return result;
  }
  
  if (term_IsVariable(T2)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (cont_TermContainsSymbol(Context1, T1, term_TopSymbol(T2)))
      result = ord_GREATER_THAN;
    else
      result = ord_UNCOMPARABLE;
    
    /*     fputs(", result: ", stdout); term_PrintPrefix(T1); */
    /*     putchar(' '); ord_Print(result); */
    /*     putchar(' '); term_PrintPrefix(T2); DBG */
    return result;
  }

  rpos_ContVarCondition(Context1, T1, &varCond1, &ground1,
			Context2, T2, &varCond2, &ground2);

  if (ground1 && ground2)
    /* Both terms are ground, so use the ground algorithm */
    result = rpos_ShortCompareGround(Context1, T1, Context2, T2);
  else {
    /* At least one term is non-ground */    
    if (varCond1 || varCond2) {
      RPOS_TERM rt1, rt2;

      rpos_FullCompareGround(Context1, T1, Context2, T2, &rt1, &rt2);
      result = rpos_NGCompare(rt1, rt2);

      rpos_TermDelete(rt1);
      rpos_TermDelete(rt2);
    } else {
      result = ord_UNCOMPARABLE;
      /* If in T1 a variable x occurs more often than in T2, and */
      /* in T2 a variable y occurs more often than in T1, the    */
      /* terms are uncomparable.                                 */
    }
  }

  /*   fputs(", result: ", stdout); term_PrintPrefix(T1); */
  /*   putchar(' '); ord_Print(result); */
  /*   putchar(' '); term_PrintPrefix(T2); DBG */
  
  return result;
}
  
BOOL rpos_BottomUpContGreater(CONTEXT Context1, TERM T1, CONTEXT Context2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: TRUE, if <Term1> is greater than <Term2> with respect
           to the RPOS ordering, FALSE otherwise.
	   Variable bindings in the respective contexts are
	   considered.
***************************************************************/
{
  BOOL       varCond1, varCond2, ground1, ground2;
  ord_RESULT result;

  /* fputs("\n--------------------\nCONTGreater ", stdout);
     cont_TermPrintPrefix(Context1, T1); fputs(" and ", stdout);
     cont_TermPrintPrefix(Context2, T2); fflush(stdout); DBG */

  T1 = cont_Deref(&Context1, T1);
  T2 = cont_Deref(&Context2, T2);

  if (term_IsVariable(T1)) {
    /* A variable can't be greater than any other term! */

    /*fputs(", result: F", stdout); DBG */
    return FALSE;
  }
  
  if (term_IsVariable(T2)) {
    if (cont_TermContainsSymbol(Context1, T1, term_TopSymbol(T2)))
      result = TRUE;
    else
      result = FALSE;  /* uncomparable */
    
    /* fputs(", result: ", stdout); putchar(result ? 'T' : 'F'); DBG */
    return result;
  }

  rpos_ContVarCondition(Context1, T1, &varCond1, &ground1,
			Context2, T2, &varCond2, &ground2);

  if (!varCond1) {
    /* A variable occurs more often in T2 than in T1, */
    /* so T1 can't be greater than T2.                */

    /* fputs(", result: ", stdout); putchar('F'); */
    return FALSE;
  }

  result = FALSE;
  if (ground1 && ground2) {
    /* Both terms are ground, so use the ground algorithm */
    if (ord_IsGreaterThan(rpos_ShortCompareGround(Context1, T1, Context2, T2)))
      result = TRUE;
  } else {
    /* At least one term is non-ground */
    RPOS_TERM rt1, rt2;
    
    rpos_FullCompareGround(Context1, T1, Context2, T2, &rt1, &rt2);
    if (ord_IsGreaterThan(rpos_NGCompare(rt1, rt2)))
      result = TRUE;
    
    rpos_TermDelete(rt1);
    rpos_TermDelete(rt2);
  }

  /* fputs(", result: ", stdout); putchar(result ? 'T' : 'F'); DBG */
  
  return result;
}

void rpos_TermPrint(RPOS_TERM RTerm)
{
  LIST scan;

  symbol_Print(term_TopSymbol(RTerm->term));
  if (!list_Empty(RTerm->children) || !list_Empty(RTerm->newSubList)) {
    putchar('(');
    for (scan = RTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      rpos_TermPrint(list_Car(scan));
      if (!list_Empty(list_Cdr(scan)))
	putchar(',');
    }
    if (!list_Empty(RTerm->newSubList)) {
      putchar(':');
      for (scan = RTerm->newSubList; !list_Empty(scan); scan = list_Cdr(scan)) {
	rpos_TermPrint(list_Car(scan));
	if (!list_Empty(list_Cdr(scan)))
	  putchar(',');
      }
    }
    putchar(')');
  }
}


void rpos_ListPrint(LIST NGList)
{
  LIST scan;

  for ( ; !list_Empty(NGList); NGList = list_Cdr(NGList)) {
    RPOS_TERM rTerm;
    rTerm = (RPOS_TERM)list_Car(NGList);
    putchar('\n');
    rpos_TermPrint(rTerm);
    if (!list_Empty(rTerm->equal)) {
      fputs(" == ", stdout);
      for (scan = rTerm->equal; !list_Empty(scan); scan = list_Cdr(scan)) {
	rpos_TermPrint(list_Car(scan));
	putchar(' ');
      }
    }
    fputs(" > ", stdout);
    for (scan = rTerm->smaller; !list_Empty(scan); scan = list_Cdr(scan)) {
      rpos_TermPrint(list_Car(scan));
      putchar(' ');
    }
  }
  fflush(stdout);
}

/**************************************************************/
/* Top Down Version                                           */
/**************************************************************/

static LIST rpos_MultisetDifference(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: The multiset difference between the arguments
           of both terms with respect to rpos_Equal.
***************************************************************/
{
  LIST result;

  result = list_Copy(term_ArgumentList(T1));
  result = list_NMultisetDifference(result, term_ArgumentList(T2), rpos_Equal);
  return result;
}
  
  
static ord_RESULT rpos_MulGreaterEqual(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms with equal top symbols and multiset status.
  RETURNS: ord_GREATER_THAN if <T1> is greater than <T2>,
	   ord_EQUAL        if both terms are equal and
	   ord_UNCOMPARABLE otherwise.
***************************************************************/
{
  LIST l1, l2;

  l1 = rpos_MultisetDifference(T1, T2);
  if (list_Empty(l1))
    /* If |M| = |N| and M-N = {} then N-M = {} */ 
    return ord_Equal();   /* Terms are equal */
  else {
    LIST scan;
    BOOL greater;

    l2 = rpos_MultisetDifference(T2, T1);

    for (greater = TRUE; !list_Empty(l2) && greater; l2 = list_Pop(l2)) {
      for (scan = l1, greater = FALSE; !list_Empty(scan) && !greater; scan = list_Cdr(scan))
	greater = rpos_Greater(list_Car(scan), list_Car(l2));
    }
    list_Delete(l1); /* l2 was freed in the outer for loop */
    if (greater)
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  }
}

static ord_RESULT rpos_LexGreaterEqual(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms with equal top symbols and lexicographic status.
  RETURNS: ord_GREATER_THAN if <T1> is greater than <T2>,
	   ord_EQUAL        if both terms are equal and 
	   ord_UNCOMPARABLE otherwise.
***************************************************************/
{
  ord_RESULT result;
  LIST       l1, l2, scan1, scan2;

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    l1 = list_Reverse(term_ArgumentList(T1)); /* Create new lists */
    l2 = list_Reverse(term_ArgumentList(T2));
  } else {
    l1 = term_ArgumentList(T1);
    l2 = term_ArgumentList(T2);
  }
  /* First ignore equal arguments */
  result = ord_Equal();
  for (scan1 = l1, scan2 = l2; !list_Empty(scan1);
       scan1 = list_Cdr(scan1), scan2 = list_Cdr(scan2)) {
    result = rpos_GreaterEqual(list_Car(scan1), list_Car(scan2));
    if (!ord_IsEqual(result))
      break;
  }

  if (ord_IsEqual(result))  /* All arguments are equal, so the terms */
    /* empty */;            /* are equal with respect to RPOS */
  else if (ord_IsGreaterThan(result)) {
    /* Check if T1 > each remaining argument of T2 */
    for (scan2 = list_Cdr(scan2); !list_Empty(scan2) && rpos_Greater(T1, list_Car(scan2));
	 scan2 = list_Cdr(scan2)); /* Empty body */
    if (list_Empty(scan2))
      result = ord_GreaterThan();
    else
      result = ord_Uncomparable();
  }
  else {
    /* Argument of T1 was not >= argument of T2. */

    /* Try to find an argument of T1 that is >= T2 */
    for (scan1 = list_Cdr(scan1), result = ord_Uncomparable();
	 !list_Empty(scan1) && !ord_IsGreaterThan(result);
	 scan1 = list_Cdr(scan1)) {
      if (!ord_IsUncomparable(rpos_GreaterEqual(list_Car(scan1), T2)))
	result = ord_GreaterThan();
    }
  }

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    list_Delete(l1);  /* Delete the lists create above */
    list_Delete(l2);
  }
  return result;
}


BOOL rpos_Equal(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: TRUE, if <T1> is equal to <T2> and
           FALSE otherwise.
***************************************************************/
{
  LIST l1, l2;

  if (!term_EqualTopSymbols(T1, T2))
    return FALSE;
  else if (!term_IsComplex(T1))  /* Equal variable or constant */
    return TRUE;
  else {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL)) {  /* MUL case */
      l1 = rpos_MultisetDifference(T1, T2);
      if (list_Empty(l1))
	return TRUE;
      else {
	list_Delete(l1);
	return FALSE;
      }
    } else {   /* LEX case */
      for (l1 = term_ArgumentList(T1), l2 = term_ArgumentList(T2);
	   !list_Empty(l1) &&  rpos_Equal(list_Car(l1), list_Car(l2));
	   l1 = list_Cdr(l1), l2 = list_Cdr(l2)); /* empty body */
      return list_Empty(l1);  /* All arguments were equal */
    }
  }
}


ord_RESULT rpos_GreaterEqual(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: ord_GREATER_THAN if <T1> is greater than <T2>
	   ord_EQUAL        if both terms are equal
	   ord_UNCOMPARABLE otherwise.
***************************************************************/
{
  LIST scan;

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      return ord_Equal();   /* T2 is the same variable */
    else
      /* A variable can't be greater than another term */
      return ord_Uncomparable();
  } else if (term_IsVariable(T2)) {   /* T1 isn't a variable */
    if (term_ContainsSymbol(T1, term_TopSymbol(T2)))
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  } else if (term_EqualTopSymbols(T1, T2)) {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL))
      return rpos_MulGreaterEqual(T1, T2);
    else
      return rpos_LexGreaterEqual(T1, T2);
  } else {
    if (symbol_PrecedenceGreater(term_TopSymbol(T1), term_TopSymbol(T2))) {
      /* Different top symbols, symbol of T1 > symbol of T2. */
      /* Try if T1 > each argument of T2.                    */
      for (scan = term_ArgumentList(T2); !list_Empty(scan); scan = list_Cdr(scan))
	if (!rpos_Greater(T1,  list_Car(scan)))
	  return ord_Uncomparable();
      return ord_GreaterThan();
    } else {
      /* Try to find an argument of T1 that is >= T2 */
      for (scan = term_ArgumentList(T1); !list_Empty(scan); scan = list_Cdr(scan))
	if (!ord_IsUncomparable(rpos_GreaterEqual(list_Car(scan), T2)))
	  return ord_GreaterThan();    /* Argument of T1 >= T2 */
      return ord_Uncomparable();
    }
  }
}

ord_RESULT rpos_Compare(TERM T1, TERM T2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: The relation between the two terms with respect to the
           RPOS ordering:
           ord_GREATER_THAN if <T1> is greater than <T2>,
	   ord_EQUAL        if both terms are equal,
           ord_SMALLER_THAN if <T2> is greater than <T1> and
	   ord_UNCOMPARABLE otherwise.
***************************************************************/
{
  ord_RESULT result;

  result = rpos_GreaterEqual(T1, T2);
  if (!ord_IsUncomparable(result))
    return result;
  else if (rpos_Greater(T2, T1))
    return ord_SmallerThan();
  else
    return ord_UNCOMPARABLE;
}

/**************************************************************/
/* Term comparison with respect to bindings                   */
/**************************************************************/

static LIST rpos_ContMultisetDifference(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: The multiset difference between the arguments
           of both terms with respect to rpos_ContEqual.
  EFFECT:  Variable bindings are considered.
***************************************************************/
{
  LIST result, scan1, scan2;

  /* Don't apply bindings at top level, since that happened */
  /* in rpos_ContGreaterEqual */

  /* We can't use list_NMultisetDifference, since that function   */
  /* expects an equality functions for terms that takes two terms */
  /* as arguments. We also need the two contexts resolve variable */
  /* bindings. */
  result = list_Copy(term_ArgumentList(T1));
  for (scan2 = term_ArgumentList(T2); !list_Empty(scan2);
       scan2 = list_Cdr(scan2)) {
    /* Delete at most one occurence of the */
    /* current element of list2 from list1 */
    for (scan1 = result; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
      if (list_Car(scan1) != NULL &&
	  rpos_ContEqual(C1, list_Car(scan1), C2, list_Car(scan2))) {
	/* arg of list1 wasn't deleted earlier and terms are equal */
	list_Rplaca(scan1, NULL);  /* Mark argument of T1 as deleted */
	break;
      }
    }
  }
  return list_PointerDeleteElement(result, NULL); /* Delete all marked terms */
}
  
  
static ord_RESULT rpos_ContMulGreaterEqual(CONTEXT C1, TERM T1,
					   CONTEXT C2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms with equal top symbols
           and multiset status.
  RETURNS: ord_GREATER_THAN if <T1> is greater than <T2>,
	   ord_EQUAL        if both terms are equal and
	   ord_UNCOMPARABLE otherwise.
  EFFECT:  Variable bindings are considered.
***************************************************************/
{
  LIST l1, l2;

  /* Don't apply bindings at top level, since that happened */
  /* in rpos_ContGreaterEqual. */

  l1 = rpos_ContMultisetDifference(C1, T1, C2, T2);
  if (list_Empty(l1))
    /* If |M| = |N| and M-N = {} then N-M = {} */ 
    return ord_Equal();   /* Terms are equal */
  else {
    LIST scan;
    BOOL greater;

    l2 = rpos_ContMultisetDifference(C2, T2, C1, T1);

    for (greater = TRUE; !list_Empty(l2) && greater; l2 = list_Pop(l2)) {
      for (scan = l1, greater = FALSE; !list_Empty(scan) && !greater;
	   scan = list_Cdr(scan))
	greater = rpos_ContGreater(C1, list_Car(scan), C2, list_Car(l2));
    }
    list_Delete(l1); /* l2 was freed in the outer for loop */
    if (greater)
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  }
}

static ord_RESULT rpos_ContLexGreaterEqual(CONTEXT C1, TERM T1,
					   CONTEXT C2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms with equal top symbols
           and lexicographic status.
  RETURNS: ord_GREATER_THAN if <T1> is greater than <T2>,
	   ord_EQUAL        if both terms are equal and 
	   ord_UNCOMPARABLE otherwise.
  EFFECT:  Variable bindings are considered.
***************************************************************/
{
  ord_RESULT result;
  LIST       l1, l2, scan1, scan2;

  /* Don't apply bindings at top level, since that happened */
  /* in rpos_ContGreaterEqual */

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    l1 = list_Reverse(term_ArgumentList(T1)); /* Create new lists */
    l2 = list_Reverse(term_ArgumentList(T2));
  } else {
    l1 = term_ArgumentList(T1);
    l2 = term_ArgumentList(T2);
  }
  /* First ignore equal arguments */
  result = ord_Equal();
  for (scan1 = l1, scan2 = l2; !list_Empty(scan1);
       scan1 = list_Cdr(scan1), scan2 = list_Cdr(scan2)) {
    result = rpos_ContGreaterEqual(C1, list_Car(scan1), C2, list_Car(scan2));
    if (!ord_IsEqual(result))
      break;
  }

  if (ord_IsEqual(result))  /* All arguments are equal, so the terms */
    /* empty */;            /* are equal with respect to RPOS */
  else if (ord_IsGreaterThan(result)) {
    /* Check if T1 > each remaining argument of T2 */
    for (scan2 = list_Cdr(scan2);
	 !list_Empty(scan2) && rpos_ContGreater(C1, T1, C2, list_Car(scan2));
	 scan2 = list_Cdr(scan2)); /* Empty body */
    if (list_Empty(scan2))
      result = ord_GreaterThan();
    else
      result = ord_Uncomparable();
  }
  else {
    /* Argument of T1 was not >= argument of T2. */
    /* Try to find an argument of T1 that is >= T2 */
    for (scan1 = list_Cdr(scan1), result = ord_Uncomparable();
	 !list_Empty(scan1) && !ord_IsGreaterThan(result);
	 scan1 = list_Cdr(scan1)) {
      if (!ord_IsUncomparable(rpos_ContGreaterEqual(C1,list_Car(scan1),C2,T2)))
	result = ord_GreaterThan();
    }
  }

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    list_Delete(l1);  /* Delete the lists create above */
    list_Delete(l2);
  }
  return result;
}


BOOL rpos_ContEqual(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: TRUE, if <T1> is equal to <T2> and
           FALSE otherwise.
  EFFECT:  Variable bindings are considered.
***************************************************************/
{
  LIST l1, l2;

  T1 = cont_Deref(&C1, T1);
  T2 = cont_Deref(&C2, T2);

  if (!term_EqualTopSymbols(T1, T2))
    return FALSE;
  else if (!term_IsComplex(T1))
    return TRUE;
  else {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL)) {
      l1 = rpos_ContMultisetDifference(C1, T1, C2, T2);
      if (list_Empty(l1))
	return TRUE;
      else {
	list_Delete(l1);
	return FALSE;
      }
    } else {   /* LEX case */
      for (l1 = term_ArgumentList(T1), l2 = term_ArgumentList(T2);
	   !list_Empty(l1) &&  rpos_ContEqual(C1,list_Car(l1),C2,list_Car(l2));
	   l1 = list_Cdr(l1), l2 = list_Cdr(l2)); /* empty body */
      return list_Empty(l1);  /* All arguments were equal */
    }
  }
}


ord_RESULT rpos_ContGreaterEqual(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: ord_GREATER_THAN if <T1> is greater than <T2>
	   ord_EQUAL        if both terms are equal
	   ord_UNCOMPARABLE otherwise.
  EFFECT:  Variable bindings are considered.
***************************************************************/
{
  LIST scan;

  T1 = cont_Deref(&C1, T1);
  T2 = cont_Deref(&C2, T2);

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      return ord_Equal();   /* T2 is the same variable */
    else
      /* A variable can't be greater than another term */
      return ord_Uncomparable();
  } else if (term_IsVariable(T2)) {   /* T1 isn't a variable */
    if (cont_TermContainsSymbol(C1, T1, term_TopSymbol(T2)))
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  } else if (term_EqualTopSymbols(T1, T2)) {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL))
      return rpos_ContMulGreaterEqual(C1, T1, C2, T2);
    else
      return rpos_ContLexGreaterEqual(C1, T1, C2, T2);
  } else {
    if (symbol_PrecedenceGreater(term_TopSymbol(T1), term_TopSymbol(T2))) {
      /* Different top symbols, symbol of T1 > symbol of T2. */
      /* Try if T1 > each argument of T2.                    */
      for (scan = term_ArgumentList(T2); !list_Empty(scan); scan = list_Cdr(scan))
	if (!rpos_ContGreater(C1, T1, C2, list_Car(scan)))
	  return ord_Uncomparable();
      return ord_GreaterThan();
    } else {
      /* Try to find an argument of T1 that is >= T2 */
      for (scan = term_ArgumentList(T1); !list_Empty(scan); scan = list_Cdr(scan))
	if (!ord_IsUncomparable(rpos_ContGreaterEqual(C1,list_Car(scan),C2,T2)))
	  return ord_GreaterThan();    /* Argument of T1 >= T2 */
      return ord_Uncomparable();
    }
  }
}

ord_RESULT rpos_ContCompare(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: The relation between the two terms with respect to the
           RPOS ordering:
           ord_GREATER_THAN if <T1> is greater than <T2>,
	   ord_EQUAL        if both terms are equal,
           ord_SMALLER_THAN if <T2> is greater than <T1> and
	   ord_UNCOMPARABLE otherwise.
  EFFECT:  Variable bindings are considered.
***************************************************************/
{
  ord_RESULT result;

  T1 = cont_Deref(&C1, T1);
  T2 = cont_Deref(&C2, T2);

  result = rpos_ContGreaterEqual(C1, T1, C2, T2);
  if (!ord_IsUncomparable(result))
    return result;
  else if (rpos_ContGreater(C2, T2, C1, T1))
    return ord_SmallerThan();
  else
    return ord_UNCOMPARABLE;
}
  
