/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *          REPRESENTATION OF PROOF SEARCH                * */
/* *                                                        * */
/* *  $Module:   PROOF SEARCH                               * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.61 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 17:47:06 $                             * */
/* $Author: schmidt $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: search.c,v $ */


#include "search.h"
#include "defs.h"
#include "stat.h"



/**************************************************************/
/* Functions                                                  */
/**************************************************************/

static SPLIT prfs_SplitCreate(PROOFSEARCH PS)
/**************************************************************
  INPUT:   A proofsearch object
  RETURNS: A new split object, which is initialized.
  MEMORY:  Allocates memory for the SPLIT_NODE.
  EFFECT:  Increases the split level of the proof search object.
***************************************************************/
{
  SPLIT Result;

  prfs_IncValidLevel(PS);

  Result                 = (SPLIT)memory_Malloc(sizeof(SPLIT_NODE));
  Result->splitlevel     = prfs_ValidLevel(PS);
  Result->used           = FALSE;
  Result->blockedClauses = list_Nil();
  Result->deletedClauses = list_Nil();
  Result->father         = (CLAUSE) NULL;
  return Result;
}

static void prfs_SplitDelete(SPLIT S)
/**************************************************************
  INPUT:   A split 
  RETURNS: Nothing.
  MEMORY:  Deletes blocked and deleted clauses. Frees the split.
***************************************************************/
{
  clause_DeleteClauseList(S->blockedClauses);
  clause_DeleteClauseList(S->deletedClauses);
  if (S->father != (CLAUSE)NULL)
    clause_Delete(S->father);
  prfs_SplitFree(S);
}


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *		    DEBUGGING FUNCTIONS	         	    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

BOOL prfs_Check(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  None.
  RETURNS: TRUE if all invariants about <Search> are valid.
***************************************************************/
{
  LIST   Scan,Clauses;
  SPLIT  Split;
  CLAUSE Clause;

  for (Scan=prfs_UsableClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (!clause_IsClause(Clause) ||
	clause_GetFlag(Clause,WORKEDOFF) ||
	!prfs_IsClauseValid(Clause, prfs_ValidLevel(Search)))
      return FALSE;
  }

  for (Scan=prfs_WorkedOffClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (!clause_IsClause(Clause) ||
	!clause_GetFlag(Clause,WORKEDOFF) || 
	!prfs_IsClauseValid(Clause, prfs_ValidLevel(Search)))
      return FALSE;
  }

  for (Scan=prfs_SplitStack(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Split = (SPLIT)list_Car(Scan);
    if (Split->used) {
      if (!list_Empty(Split->blockedClauses) || !list_Empty(Split->deletedClauses)) {
	/*printf("\n");prfs_PrintSplit(Split); printf("\n");*/
	return FALSE;
      } else {      
	for (Clauses=prfs_UsableClauses(Search);!list_Empty(Clauses);Clauses=list_Cdr(Clauses))
	  if (clause_SplitLevel(list_Car(Clauses)) == Split->splitlevel) {
	    /*printf("\n\n");prfs_PrintSplit(Split); 
	      printf("\n Clause must not exist: ");clause_Print(list_Car(Clauses)); printf("\n");*/
	    return FALSE;
	  }
	for (Clauses=prfs_WorkedOffClauses(Search);!list_Empty(Clauses);Clauses=list_Cdr(Clauses))
	  if (clause_SplitLevel(list_Car(Clauses)) == Split->splitlevel) {
	    /*printf("\n\n");prfs_PrintSplit(Split);
	      printf("\n Clause must not exist: ");clause_Print(list_Car(Clauses)); printf("\n");*/
	    return FALSE;
	  }
      }
    } 
  }
  
  if (prfs_ValidLevel(Search) == 0) {
    if (!prfs_SplitStackEmpty(Search))
      return FALSE;
  } else {
    if (prfs_ValidLevel(Search) != prfs_SplitSplitLevel(prfs_SplitStackTop(Search)))
      return FALSE;
  }

  if (prfs_ValidLevel(Search) < prfs_LastBacktrackLevel(Search))
    return FALSE;

  return TRUE;
}

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *			HIGH LEVEL FUNCTIONS		    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

static void prfs_InsertInSortTheories(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  If the clause is a declaration clause it is inserted
           into the dynamic and approximated dynamic sort theory.
  RETURNS: void.
***************************************************************/
{
  if ((prfs_DynamicSortTheory(Search) != (SORTTHEORY)NULL
       || prfs_ApproximatedDynamicSortTheory(Search) != (SORTTHEORY)NULL)
      && clause_IsDeclarationClause(Clause)) {
    int     i,l;
    LITERAL lit;
    CLAUSE  copy;
    LIST    approx;
    l = clause_Length(Clause);
    for (i = clause_FirstSuccedentLitIndex(Clause); i < l; i++) {
      lit = clause_GetLiteral(Clause,i);
      if (clause_LiteralIsMaximal(lit) && 
	  symbol_IsBaseSort(term_TopSymbol(clause_LiteralSignedAtom(lit)))) {
	if (prfs_DynamicSortTheory(Search) != (SORTTHEORY)NULL) {
	  copy = clause_Copy(Clause);
	  list_Delete(clause_ParentClauses(copy));
	  clause_SetParentClauses(copy,list_Nil());
	  list_Delete(clause_ParentLiterals(copy));
	  clause_SetParentLiterals(copy,list_Nil());
	  clause_SetNumber(copy,clause_Number(Clause));
	  sort_TheoryInsertClause(prfs_DynamicSortTheory(Search),Clause,
				  copy,clause_GetLiteral(copy,i));
	}
	if (prfs_ApproximatedDynamicSortTheory(Search) != (SORTTHEORY)NULL) {
	  approx = sort_ApproxMaxDeclClauses(Clause);
	  for ( ; !list_Empty(approx); approx = list_Pop(approx)) {
	    copy = (CLAUSE)list_Car(approx);
	    sort_TheoryInsertClause(prfs_ApproximatedDynamicSortTheory(Search),
				    Clause, copy, 
				    clause_GetLiteral(copy,clause_FirstSuccedentLitIndex(copy)));
	  }
	}
      }
    }
  }
}

static void prfs_DeleteFromSortTheories(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  If the clause is a declaration clause it is deleted
           from the dynamic and approximated dynamic sort theory.
  RETURNS: void.
***************************************************************/
{
  if (clause_IsDeclarationClause(Clause)) {
    if (prfs_DynamicSortTheory(Search) != (SORTTHEORY)NULL)
      sort_TheoryDeleteClause(prfs_DynamicSortTheory(Search), Clause);
    if (prfs_ApproximatedDynamicSortTheory(Search) != (SORTTHEORY)NULL)
      sort_TheoryDeleteClause(prfs_ApproximatedDynamicSortTheory(Search), Clause);
  }
}

void prfs_DeleteDocProof(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  RETURNS: void.
  EFFECT:  The docproof structures are deleted.
***************************************************************/
{
  clause_DeleteSharedClauseList(prfs_DocProofClauses(Search),prfs_DocProofSharingIndex(Search));
  if (prfs_DocProofSharingIndex(Search))
    sharing_IndexDelete(prfs_DocProofSharingIndex(Search));
  Search->dpindex = NULL;
  Search->dplist  = NULL;
}

void prfs_Delete(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  RETURNS: void.
  EFFECT:  The whole structure including all its substructures 
           is deleted.
***************************************************************/
{
  LIST Scan;

  clause_DeleteClauseList(prfs_EmptyClauses(Search));
  list_DeleteWithElement(prfs_Definitions(Search), def_Delete);
  list_Delete(prfs_UsedEmptyClauses(Search));
  sort_TheoryDelete(prfs_StaticSortTheory(Search));
  sort_TheoryDelete(prfs_DynamicSortTheory(Search));
  sort_TheoryDelete(prfs_ApproximatedDynamicSortTheory(Search));
  if (prfs_Precedence(Search) != (PRECEDENCE)NULL)
    symbol_DeletePrecedence(prfs_Precedence(Search));
  clause_DeleteSharedClauseList(prfs_WorkedOffClauses(Search),prfs_WorkedOffSharingIndex(Search));
  clause_DeleteSharedClauseList(prfs_UsableClauses(Search),prfs_UsableSharingIndex(Search));
  clause_DeleteSharedClauseList(prfs_DocProofClauses(Search),prfs_DocProofSharingIndex(Search));
  sharing_IndexDelete(prfs_WorkedOffSharingIndex(Search));
  sharing_IndexDelete(prfs_UsableSharingIndex(Search));
  if (prfs_DocProofSharingIndex(Search))
    sharing_IndexDelete(prfs_DocProofSharingIndex(Search));
  for (Scan=prfs_SplitStack(Search);!list_Empty(Scan);Scan=list_Cdr(Scan))
    prfs_SplitDelete(list_Car(Scan));
  list_Delete(prfs_SplitStack(Search));
  flag_FreeStore(prfs_Store(Search));
  memory_Free(Search,sizeof(PROOFSEARCH_NODE));
}


void prfs_Clean(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  RETURNS: void.
  EFFECT:  All clauses are deleted. The structure is cleaned
           and initialized.
***************************************************************/
{
  LIST Scan;

  clause_DeleteClauseList(prfs_EmptyClauses(Search));
  list_DeleteWithElement(prfs_Definitions(Search), def_Delete);
  list_Delete(prfs_UsedEmptyClauses(Search));
  sort_TheoryDelete(prfs_StaticSortTheory(Search));
  sort_TheoryDelete(prfs_DynamicSortTheory(Search));
  sort_TheoryDelete(prfs_ApproximatedDynamicSortTheory(Search));
  clause_DeleteSharedClauseList(prfs_WorkedOffClauses(Search),prfs_WorkedOffSharingIndex(Search));
  clause_DeleteSharedClauseList(prfs_UsableClauses(Search),prfs_UsableSharingIndex(Search));
  clause_DeleteSharedClauseList(prfs_DocProofClauses(Search),prfs_DocProofSharingIndex(Search));
  for (Scan=prfs_SplitStack(Search);!list_Empty(Scan);Scan=list_Cdr(Scan))
    prfs_SplitDelete(list_Car(Scan));
  list_Delete(prfs_SplitStack(Search));

  Search->emptyclauses     = list_Nil();
  Search->definitions      = list_Nil();
  Search->usedemptyclauses = list_Nil();
  Search->wolist           = list_Nil();
  Search->uslist           = list_Nil();
  Search->astatic          = (SORTTHEORY)NULL;
  Search->adynamic         = (SORTTHEORY)NULL;
  Search->dynamic          = (SORTTHEORY)NULL;
  Search->dplist           = list_Nil();
  
  Search->stack          = list_StackBottom();
  Search->validlevel     = 0;
  Search->lastbacktrack  = 0;
  Search->splitcounter   = 0;
  Search->keptclauses    = 0;
  Search->derivedclauses = 0;
  Search->loops          = 0;
  Search->backtracked    = 0;  
}

PROOFSEARCH prfs_Create()
/**************************************************************
  INPUT:   None.
  RETURNS: A new proof search object. The worked off and usable
           indexes are created whilst the docproof index and the
	   sort theories are not created, since they are not
	   needed in general.
***************************************************************/
{
  PROOFSEARCH Result;

  Result = memory_Malloc(sizeof(PROOFSEARCH_NODE));

  Result->emptyclauses = list_Nil();
  Result->definitions  = list_Nil();
  Result->usedemptyclauses = list_Nil();
  Result->woindex      = sharing_IndexCreate();
  Result->wolist       = list_Nil();
  Result->usindex      = sharing_IndexCreate();
  Result->uslist       = list_Nil();

  Result->astatic      = (SORTTHEORY)NULL;
  Result->adynamic     = (SORTTHEORY)NULL;
  Result->dynamic      = (SORTTHEORY)NULL;

  Result->precedence   = (PRECEDENCE)NULL;

  Result->store        = flag_NewStore();
  flag_InitStoreByDefaults(Result->store);
  
  Result->dpindex      = (SHARED_INDEX)NULL;
  Result->dplist       = list_Nil();
  
  Result->stack          = list_StackBottom();
  Result->validlevel     = 0;
  Result->lastbacktrack  = 0;
  Result->splitcounter   = 0;
  Result->keptclauses    = 0;
  Result->derivedclauses = 0;
  Result->loops          = 0;
  Result->backtracked    = 0;
    
  return Result;  
}

void prfs_CopyIndices(PROOFSEARCH Search, PROOFSEARCH SearchCopy) 
/**************************************************************
  INPUT:   A proof search object and a clean proof search object
  RETURNS: NONE.
  EFFECT : Copies the indices from Search to SearchCopy
  WARNING: Splitstack and theories are not copied !
***************************************************************/
{
  LIST Scan;

  /* If a DocProof index is required but not yet allocated in SearchCopy,
     do it now */
  if((prfs_DocProofSharingIndex(Search) != NULL)
     && (prfs_DocProofSharingIndex(SearchCopy) == NULL))
    prfs_AddDocProofSharingIndex(SearchCopy);

  /* Copy usable, worked-off and docproof index */
  for(Scan = prfs_UsableClauses(Search); !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_InsertUsableClause(SearchCopy, clause_Copy((CLAUSE) list_Car(Scan)));

  for(Scan = prfs_WorkedOffClauses(Search); !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_InsertWorkedOffClause(SearchCopy, clause_Copy((CLAUSE) list_Car(Scan)));

  for(Scan = prfs_DocProofClauses(Search); !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_InsertDocProofClause(SearchCopy, clause_Copy((CLAUSE) list_Car(Scan)));
}


void prfs_InsertWorkedOffClause(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  MEMORY:  The clause is assumed to be unshared.
  EFFECT:  The clause is inserted into the worked off sharing index
           and list of <Search>. The unshared literals are deleted.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_InsertWorkedOffClause\n", stderr);
    misc_DumpCore();
  }
#endif 

  clause_SetFlag(Clause,WORKEDOFF);
  prfs_SetWorkedOffClauses(Search,list_Cons(Clause, prfs_WorkedOffClauses(Search)));
  clause_InsertIntoSharing(Clause, prfs_WorkedOffSharingIndex(Search));
  prfs_InsertInSortTheories(Search, Clause);
}


void prfs_InsertUsableClause(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  MEMORY:  The clause is assumed to be unshared.
  EFFECT:  The clause is inserted into the usable sharing index
           and list of <Search> sorted with respect to their weight. 
	   The unshared literals are deleted.
  RETURNS: void.
***************************************************************/
{

#ifdef DEBUG
  if (!clause_IsClause(Clause) || clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_InsertUsableClause\n", stderr);
    misc_DumpCore();
  }
  else {
    LIST scan;
    for (scan = prfs_UsableClauses(Search); !list_Empty(scan); scan = list_Cdr(scan)) {
      if (clause_Number(Clause) == clause_Number(list_Car(scan))) {
	fputs("\nDetected two clauses with same clause number in the usable clauses!\n", stderr);
	misc_DumpCore();
      }
    }
  }
#endif 
    
  prfs_SetUsableClauses(Search,clause_InsertWeighed(Clause, prfs_UsableClauses(Search)));
  clause_InsertIntoSharing(Clause, prfs_UsableSharingIndex(Search));
}

void prfs_InsertDocProofClause(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  MEMORY:  The clause is assumed to be unshared.
  EFFECT:  The clause is inserted into the proof documentation sharing index.
	   The unshared literals are deleted.
  RETURNS: void.
***************************************************************/
{ 
#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_InsertDocProofClause\n", stderr);
    misc_DumpCore();
  }
#endif   
  
  if (prfs_DocProofSharingIndex(Search) == (SHARED_INDEX)NULL)
    clause_Delete(Clause);
  else {
    prfs_SetDocProofClauses(Search,list_Cons(Clause, prfs_DocProofClauses(Search)));
    clause_InsertIntoSharing(Clause, prfs_DocProofSharingIndex(Search));
  }
}

void prfs_MoveUsableWorkedOff(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is inserted into the worked off sharing index
           and list and it is deleted from the usable index and list.
	   In particular, the WorkedOff flag is set and if <Clause> is a 
	   declaration clause, it is inserted into the respective sort theories.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause) || clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_MoveUsableWorkedOff\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_SetUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));
  clause_SetFlag(Clause,WORKEDOFF);
  clause_MoveSharedClause(Clause,prfs_UsableSharingIndex(Search),prfs_WorkedOffSharingIndex(Search));
  prfs_SetWorkedOffClauses(Search,list_Cons(Clause, prfs_WorkedOffClauses(Search)));
  prfs_InsertInSortTheories(Search, Clause);
}

void prfs_MoveWorkedOffDocProof(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is inserted into the doc proof sharing index
           and list of <Search> and it is deleted from the worked off
	   index and list.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause) || !clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_MoveWorkedOffDocProof\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_DeleteFromSortTheories(Search, Clause);
  prfs_SetWorkedOffClauses(Search,list_PointerDeleteElement(prfs_WorkedOffClauses(Search),Clause));
  clause_RemoveFlag(Clause,WORKEDOFF);

  if (prfs_DocProofSharingIndex(Search) == (SHARED_INDEX)NULL)
    clause_DeleteFromSharing(Clause,prfs_WorkedOffSharingIndex(Search));
  else {
    clause_MoveSharedClause(Clause,prfs_WorkedOffSharingIndex(Search),prfs_DocProofSharingIndex(Search));
    prfs_SetDocProofClauses(Search,list_Cons(Clause, prfs_DocProofClauses(Search)));
  }
}

void prfs_MoveUsableDocProof(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is inserted into the doc proof sharing index
           and list of <Search> and it is deleted from the usable
	   index and list.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause) || clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in  prfs_MoveUsableDocProof\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_SetUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));

  if (prfs_DocProofSharingIndex(Search) == (SHARED_INDEX)NULL)
    clause_DeleteFromSharing(Clause,prfs_UsableSharingIndex(Search));
  else {
    clause_MoveSharedClause(Clause,prfs_UsableSharingIndex(Search),prfs_DocProofSharingIndex(Search));
    prfs_SetDocProofClauses(Search,list_Cons(Clause, prfs_DocProofClauses(Search)));
  }
}

void prfs_MoveInvalidClausesDocProof(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  All clauses that have a split level higher than the
           current split level of <Search> are moved to the
	   proof documentation index. If it does not exist, i.e.,
	   no proof documentation required, the clauses are
	   deleted.
  RETURNS: void.
***************************************************************/
{
  LIST   scan, invalid;
  CLAUSE clause;

  invalid = list_Nil();
  for (scan = prfs_WorkedOffClauses(Search); !list_Empty(scan);
       scan = list_Cdr(scan)) {
    clause = (CLAUSE)list_Car(scan);
    if (!prfs_IsClauseValid(clause, prfs_ValidLevel(Search)))
      invalid = list_Cons(clause,invalid);
  }
  /* WARNING: The following move operation changes the worked off */
  /* set of the proof search object destructively.                */
  /* So it's impossible to move those function calls into the     */
  /* loop above.                                                  */
  for ( ; !list_Empty(invalid); invalid = list_Pop(invalid))
    prfs_MoveWorkedOffDocProof(Search,list_Car(invalid));

  invalid = list_Nil();
  for (scan = prfs_UsableClauses(Search); !list_Empty(scan);
       scan = list_Cdr(scan)) {
    clause = (CLAUSE)list_Car(scan);
    if (!prfs_IsClauseValid(clause, prfs_ValidLevel(Search)))
      invalid = list_Cons(clause,invalid);
  }
  /* WARNING: The following move operation changes the usable     */
  /* set of the proof search object destructively.                */
  /* So it's impossible to move those function calls into the     */
  /* loop above.                                                  */
  for ( ; !list_Empty(invalid); invalid = list_Pop(invalid))
    prfs_MoveUsableDocProof(Search,list_Car(invalid));
}

void prfs_ExtractWorkedOff(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is removed from the worked off index and
           list and returned as an unshared clause.
	   Sort theories are updated accordingly.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause) || !clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_ExtractWorkedOff\n", stderr);
    misc_DumpCore();
  }
#endif

  prfs_DeleteFromSortTheories(Search, Clause);
  clause_RemoveFlag(Clause,WORKEDOFF);
  prfs_SetWorkedOffClauses(Search,list_PointerDeleteElement(prfs_WorkedOffClauses(Search),Clause));
  clause_MakeUnshared(Clause,prfs_WorkedOffSharingIndex(Search));
}

void prfs_ExtractUsable(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is removed from the usable off index and
           list and returned as an unshared clause.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause) || clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_ExtractUsable\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_SetUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));
  clause_MakeUnshared(Clause,prfs_UsableSharingIndex(Search));
}

void prfs_ExtractDocProof(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is removed from the docproof off index and
           list and returned as an unshared clause.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsUnorderedClause(Clause)) {
    fputs("\n\nIllegal input in prfs_ExtractDocProof\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_SetDocProofClauses(Search,list_PointerDeleteElement(prfs_DocProofClauses(Search),Clause));
  clause_MakeUnshared(Clause,prfs_DocProofSharingIndex(Search));
}

void prfs_DeleteWorkedOff(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is deleted from the worked off index and list.
	   Sort theories are updated accordingly.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause) || !clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_DeleteWorkedOff\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_DeleteFromSortTheories(Search, Clause);
  prfs_SetWorkedOffClauses(Search,list_PointerDeleteElement(prfs_WorkedOffClauses(Search),Clause));
  clause_DeleteFromSharing(Clause,prfs_WorkedOffSharingIndex(Search));
}


void prfs_DeleteUsable(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is deleted from the usable index and list.
  RETURNS: void.
***************************************************************/
{
#ifdef DEBUG
  if (!clause_IsClause(Clause) || clause_GetFlag(Clause, WORKEDOFF)) {
    fputs("\n\nIllegal input in prfs_DeleteUsable\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_SetUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));
  clause_DeleteFromSharing(Clause,prfs_UsableSharingIndex(Search));
}

void prfs_PrintSplit(SPLIT Split)
/**************************************************************
  INPUT:   A split.
  RETURNS: None, prints the information kept in the split structure.
  MEMORY:  No effects.
***************************************************************/
{
  LIST Scan;

    printf("\n Split: %d %ld", prfs_SplitSplitLevel(Split), (long)Split);
    printf("\n Father: ");
    if (prfs_GetFatherClause(Split) != (CLAUSE)NULL)
      clause_Print(prfs_GetFatherClause(Split));
    else
      printf("No father, unnecessary split.");

    fputs("\n Split is ", stdout);
    if (prfs_SplitIsUnused(Split))
      puts("unused.");
    else
      puts("used.");
    fputs(" Blocked clauses:", stdout);
    for(Scan=prfs_SplitBlockedClauses(Split);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      printf("\n ");
      clause_Print(list_Car(Scan));
    }
    printf("\n Deleted clauses:");
    for(Scan=prfs_SplitDeletedClauses(Split);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      printf("\n ");
      clause_Print(list_Car(Scan));
    }
}


static void prfs_PrintSplitStack(PROOFSEARCH PS)
/**************************************************************
  INPUT:   None.
  RETURNS: None, prints almost all the information kept in the
           split structure.
  MEMORY:  No effects.
***************************************************************/
{
  LIST Scan;
  int  Level;

  printf("\n Splitstack:");
  Level = prfs_ValidLevel(PS);

  for (Scan = prfs_SplitStack(PS); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    printf("\n Level: %d",Level--);
    prfs_PrintSplit(list_Car(Scan));
  }
}

void prfs_Print(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  The proof search object is printed to stdout.
  RETURNS: void.
***************************************************************/
{
  LIST Scan;

#ifdef DEBUG
  if (!prfs_Check(Search)) {
    fputs("\n\nIllegal input in prfs_Print\n", stderr);
    misc_DumpCore();
  }
#endif 

  printf("\n\n Proofsearch: Current Level: %d Last Backtrack Level: %d Splits: %d Loops: %d Backtracked: %d",
	 prfs_ValidLevel(Search),prfs_LastBacktrackLevel(Search),prfs_SplitCounter(Search),
	 prfs_Loops(Search),prfs_BacktrackedClauses(Search));
  printf("\n Empty Clauses:");
  for(Scan=prfs_EmptyClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    printf("\n ");
    clause_Print(list_Car(Scan));
  }
  printf("\n Definitions:");
  for(Scan=prfs_Definitions(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    printf("\n ");
    term_Print(list_Car(Scan));
  }
  printf("\n Worked Off Clauses:");
  for(Scan=prfs_WorkedOffClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    printf("\n ");
    clause_Print(list_Car(Scan));
  }
  printf("\n Usable Clauses:");
  for(Scan=prfs_UsableClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    printf("\n ");
    clause_Print(list_Car(Scan));
  }
  prfs_PrintSplitStack(Search);
  printf("\n Static Sort Theory:");
  sort_TheoryPrint(prfs_StaticSortTheory(Search));
  printf("\n Dynamic Sort Theory:");
  sort_TheoryPrint(prfs_DynamicSortTheory(Search));
  printf("\n Approximated Dynamic Sort Theory:");
  sort_TheoryPrint(prfs_ApproximatedDynamicSortTheory(Search));
  printf("\n");
}

static CLAUSE prfs_DoSplitting(PROOFSEARCH PS, CLAUSE SplitClause, LIST Literals)
/**************************************************************
  INPUT:   An unshared clause to be splitted where 'Literals' is the list
           of literals to keep (in their order in the SplitClause) and 
	   an index where the splitted clause is inserted,
	   and a proofsearch object
  RETURNS: A pointer to the (stack-, not sharing-) inserted splitted clause.
  MEMORY:  The blocked parts and the actparts literals are created
           unshared, memory for the two (more for HornSplits) new 
	   clausenodes is allocated.
***************************************************************/
{

  SPLIT   NewSplit;
  CLAUSE  NewClause, BlockedClause;
  LITERAL NextLit,NewLit;
  int     i,j,lengthBlocked,lengthNew,lc,la,ls,nc,na,ns;
#ifdef DEBUG_RULES_SPLIT
  if (!(clause_IsClause(SplitClause))) {
    printf("\n\nIllegal input in prfs_DoSplitting\n");
    misc_Error();
  }
#endif

  prfs_DecSplitCounter(PS);
  NewSplit = prfs_SplitCreate(PS);

  prfs_SetFatherClause(NewSplit, SplitClause);

  lengthNew = list_Length(Literals);
  lengthBlocked = clause_Length(SplitClause) - lengthNew;

  NewClause = clause_CreateBody(lengthNew);
  BlockedClause = clause_CreateBody(lengthBlocked);
  clause_DecreaseCounter(); /* reset internally increased counter! */
  clause_SetNumber(BlockedClause, 0);
  /* To detect forgotten setting at insertion! */

  lc = clause_LastConstraintLitIndex(SplitClause);
  la = clause_LastAntecedentLitIndex(SplitClause);
  ls = clause_LastSuccedentLitIndex(SplitClause);

  nc = 0; na = 0; ns = 0;

  j = clause_FirstLitIndex();

  for (i = clause_FirstLitIndex(); i <= ls; i++) {
    NextLit = clause_GetLiteral(SplitClause, i);

    NewLit  = clause_LiteralCopy(NextLit);

    if ((lengthNew > 0) && /* To avoid access of Nirvana. */
	list_PointerMember(Literals,NextLit)) {
      /* NewLit is literal for the NewClause. */

      lengthNew--;
      clause_SetLiteral(NewClause, j++, NewLit);
      clause_LiteralSetOwningClause(NewLit, NewClause);
      clause_AddParentClause(NewClause, clause_Number(SplitClause));
      clause_AddParentLiteral(NewClause, i);
      if (i <= lc)
	nc++;
      else if (i <= la)
	na++;
      else 
	ns++;

    } else { /* NewLit is literal for the BlockedClause. */

      clause_SetLiteral(BlockedClause, (i-j), NewLit);
      clause_LiteralSetOwningClause(NewLit, BlockedClause);
      clause_AddParentClause(BlockedClause, clause_Number(SplitClause));
      clause_AddParentLiteral(BlockedClause, i);
    }
  } /* end of 'for all literals'. */

  clause_SetNumOfConsLits(NewClause, nc);
  clause_SetNumOfConsLits(BlockedClause,
			  (clause_NumOfConsLits(SplitClause) - nc));
  clause_SetNumOfAnteLits(NewClause, na);
  clause_SetNumOfAnteLits(BlockedClause,
			  (clause_NumOfAnteLits(SplitClause) - na));
  clause_SetNumOfSuccLits(NewClause, ns);
  clause_SetNumOfSuccLits(BlockedClause,
			  (clause_NumOfSuccLits(SplitClause) - ns));

  clause_ReInit(BlockedClause);
  clause_UpdateSplitDataFromNewSplitting(BlockedClause,
					 SplitClause, prfs_SplitSplitLevel(NewSplit));
  clause_SetFromSplitting(BlockedClause);
  clause_SetParentLiterals(BlockedClause,
    list_NReverse(clause_ParentLiterals(BlockedClause)));

  clause_SetDepth(BlockedClause, clause_Depth(SplitClause)+1);

  prfs_SplitAddBlockedClause(NewSplit, BlockedClause);
  prfs_SplitSetDeletedClauses(NewSplit, list_Nil());

  
  prfs_SplitStackPush(PS, NewSplit);

  clause_ReInit(NewClause);
  clause_UpdateSplitDataFromNewSplitting(NewClause, SplitClause,
					 prfs_SplitSplitLevel(NewSplit));
  clause_SetFromSplitting(NewClause);

  clause_SetParentLiterals(NewClause,
			   list_NReverse(clause_ParentLiterals(NewClause)));

  clause_SetDepth(NewClause, clause_Depth(SplitClause)+1);
  clause_RemoveFlag(NewClause, WORKEDOFF);

  if (clause_IsGround(NewClause)) {
    /* Keep Clauses made from NewClause for refutation case! */
    CLAUSE UnitClause;
    LIST   AtomList;

    la = clause_LastAntecedentLitIndex(NewClause);
    ls = clause_LastSuccedentLitIndex(NewClause);

    Literals = clause_ParentLiterals(NewClause);

    for (i = clause_FirstLitIndex(); i <= ls; i++) {

      NextLit    = clause_GetLiteral(NewClause, i);
      AtomList   = list_List(term_Copy(clause_LiteralAtom(NextLit)));

      if (i <= la)
	UnitClause = clause_Create(list_Nil(), list_Nil(), AtomList);
      else
	UnitClause = clause_Create(list_Nil(), AtomList, list_Nil());

      clause_SetNumber(UnitClause, -1);
      /* To detect forgotten setting at reinsertion! */
      clause_DecreaseCounter();
      /* Reset internally increased counter! */

      list_Delete(AtomList);

      clause_SetFromSplitting(UnitClause);
      clause_UpdateSplitDataFromNewSplitting(UnitClause, SplitClause,
					     prfs_SplitSplitLevel(NewSplit));
      clause_AddParentClause(UnitClause, clause_Number(NewClause));
      clause_AddParentLiteral(UnitClause, i);
      clause_AddParentClause(UnitClause, clause_Number(SplitClause));
      clause_AddParentLiteral(UnitClause, (int)list_Car(Literals));
      Literals = list_Cdr(Literals);
      prfs_SplitAddBlockedClause(NewSplit, UnitClause);
    }
  }
  /* printf("\n\nSPLITTING DONE!");
     printf("\nAus           : "); clause_Print(SplitClause); fflush(stdout);
     printf("\nDer erste Teil: "); clause_Print(NewClause); fflush(stdout);
     printf("\nDer zweite Teil: ");
     clause_Print(BlockedClause); fflush(stdout);
     printf("\nDaher als BlockedClauses:\n");
     clause_ListPrint(prfs_SplitBlockedClauses(NewSplit)); fflush(stdout);
     */
  return NewClause;
}


static LIST prfs_GetSplitLiterals(PROOFSEARCH PS, CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause and a proofsearch object
  RETURNS: A list of literals building the bigger part of a 
           variable-disjunct literal partition if one exists,
	   an empty list, else.
  MEMORY:  Allocates memory for the literal list.
***************************************************************/
{
  LITERAL NextLit;
  int     i,length,OldLength;
  LIST    LitList, VarOcc,NextOcc;
  LITSTAT *LitStatP;
  BOOL    Change;

#ifdef DEBUG_RULES_SPLIT
  if (!clause_IsClause(Clause)) {
    printf("\n\nIllegal input in prfs_GetSplitLiterals\n");
    misc_DumpCore();
  }
#endif 

  /* We aren't prepared to cope with empty clauses !!! */
  LitList = list_Nil();

  if (prfs_SplitCounter(PS) != 0) {

    if (clause_Length(Clause) == 1) {
      return LitList;
    } else {
      if (!list_Empty(stat_STORE)) {
	LitStatP = list_Car(stat_STORE);
#ifdef DDEBUG 
	printf("\nprfs_GSL: Current clause ");
	clause_Print(Clause);
	printf("\nprfs_GSL: LitStat clause ");
	clause_Print(LitStatP->owningClause);
#endif
	if (LitStatP->owningClause == Clause) {
	  if (clause_LiteralAtom(clause_GetLiteral(Clause,LitStatP->litindex)) == LitStatP->literal) {
#ifdef DDEBUG
	    printf("\nUsing pre-selected literal.\n");
#endif
	    LitList = list_Cons(clause_GetLiteral(Clause,LitStatP->litindex),LitList);
	    return LitList;
	  } else {
#ifdef DDEBUG
	    printf("\nFailure to detect pre-selected literal at %d: \n",LitStatP->litindex);  
	    clause_LiteralPrint(clause_GetLiteral(Clause,LitStatP->litindex));
#endif
	  }
	}
      }
    }

    if (clause_HasSuccLits(Clause)) {
      /* The clause contains more than one positive literal */
      if (clause_HasGroundSuccLit(Clause)) {

	NextLit = clause_GetGroundSuccLit(Clause);
	LitList = list_Cons(NextLit, LitList);

	for (i = clause_LastAntecedentLitIndex(Clause);i >= clause_FirstLitIndex();i--) {
	  NextLit = clause_GetLiteral(Clause, i);
	  if (term_IsGround(clause_LiteralAtom(NextLit)))
	    LitList = list_Cons(NextLit, LitList);
	}
	return LitList;
      }

      NextLit = clause_GetLiteral(Clause, clause_LastLitIndex(Clause)); /* Is a Succedent Literal */
      VarOcc  = term_VariableSymbols(clause_LiteralAtom(NextLit));
      LitList = list_List(NextLit);
      length  = clause_Length(Clause);
      Change  = TRUE;

      while(Change) {
	Change = FALSE;

	for (i=(clause_LastLitIndex(Clause)-1) ; i>=clause_FirstLitIndex(); i--) {

	  NextLit = clause_GetLiteral(Clause, i);
	
	  if (!list_PointerMember(LitList,NextLit)) {
	    NextOcc = term_VariableSymbols(clause_LiteralAtom(NextLit));
	    if (list_HasIntersection(VarOcc, NextOcc)) { 
	      OldLength = list_Length(VarOcc);
	      VarOcc    = list_NPointerUnion(VarOcc, NextOcc);
	      LitList   = list_Cons(NextLit, LitList);
	      if (OldLength < list_Length(VarOcc))
		Change = TRUE;
	    }
	    else
	      list_Delete(NextOcc);
	  }
	}
      }
      if (list_Length(LitList) == length) {
	list_Delete(LitList);
	LitList = list_Nil();
      }
      Change = TRUE;                  /* Check whether not all succedent literls are used */
      for(i=clause_FirstSuccedentLitIndex(Clause) ; i<length && Change;i++)
	if (!list_PointerMember(LitList,clause_GetLiteral(Clause, i)))
	  Change = FALSE;
      if (Change) {
	list_Delete(LitList);
	LitList = list_Nil();
      }
      list_Delete(VarOcc);
    }
  }
  return LitList;
}



CLAUSE prfs_PerformSplitting(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a usable clause.
  EFFECT:  If <Clause> can be split it is splitted, the first
           part of the split is returned  and the
	   splitted clause is removed from usable and kept in the split stack.
	   Otherwise <Clause> remains unchanged and NULL is returned.
  RETURNS: NULL if <Clause> is not splittable, the first split part otherwise.
***************************************************************/
{
  CLAUSE Result;

  Result = (CLAUSE)NULL;
  
  if (clause_IsEmptyClause(Clause)) {
    return Result;
  };

  if (clause_HasSolvedConstraint(Clause)) {
    LIST LitList;

    LitList = prfs_GetSplitLiterals(Search, Clause);

    if (!list_Empty(LitList)) {
      prfs_ExtractUsable(Search,Clause);
      Result = prfs_DoSplitting(Search, Clause, LitList);
      list_Delete(LitList);
    }
  }

  return Result;
}

