/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *             TABLEAU PROOF TREES                       * */
/* *                                                        * */
/* $Revision: 0.77 $                                          * */
/* $State: Exp $                                            * */
/* $Date: 1998/12/02 18:01:43 $                                      * */
/* $Author: schmidt $                                     * */
/* *                                                        * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */

/* $RCSfile: tableau.c,v $ */

#include "tableau.h" 

/* for graph output */
extern BOOL        pcheck_ClauseCg;
extern GRAPHFORMAT pcheck_GraphFormat;

TABPATH tab_PathCreate(int MaxLevel, TABLEAU Tab)
/**************************************************************
 INPUT  : A tableau, its maximum expected depth                                                      
 RETURNS: A path consiting of the root  node of the tableau
          which can have a max length of <MaxLevel>
 EFFECTS: - 
***************************************************************/
{
 TABPATH TabPath;

 TabPath            = (TABPATH)memory_Malloc(sizeof(TABPATH_NODE));
 TabPath->Path      = (TABLEAU*)memory_Malloc(sizeof(TABLEAU)*(MaxLevel+1));
 TabPath->Path[0]   = Tab;
 TabPath->MaxLength = MaxLevel;
 TabPath->Length    = 0;

 return(TabPath); 
}

void tab_PathDelete(TABPATH TabPath)
/**************************************************************
 INPUT  : A tableau path  
 RETURNS: -
 EFFECTS: The path is deleted                                                
***************************************************************/
{
  memory_Free(TabPath->Path, (TabPath->MaxLength+1)*sizeof(TABLEAU));
  memory_Free(TabPath, sizeof(TABPATH_NODE));
}


BOOL tab_PathContainsClause(TABPATH Path, CLAUSE Clause)
/**************************************************************
 INPUT  : A tableau path, a clause
 RETURNS: TRUE iff the clause exists on its level (wrt. the
          path) in the tableau
 EFFECTS: -
***************************************************************/
{
  LIST Scan;

  if (clause_SplitLevel(Clause) > tab_PathLength(Path))
    return(FALSE);

  for(Scan = tab_Clauses(tab_PathNthNode(Path, clause_SplitLevel(Clause))); 
	!list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (list_Car(Scan) == Clause)
      return(TRUE);  }
  return(FALSE);
}

BOOL tab_PathContainsClauseSoft(TABPATH Path, CLAUSE Clause)
/**************************************************************
 INPUT  : A tableau path, a clause
 RETURNS: TRUE iff the clause is on one of the levels
          in the tableau traversed by the path. Different
	  from tab_PathContainsClauseSoft, since it does
          not expect the clause on its split level.
 EFFECTS: -
***************************************************************/
{
  LIST Scan;
  int Level;

  if (clause_SplitLevel(Clause) > tab_PathLength(Path))
    return(FALSE);
  
  for (Level = 0; Level <= tab_PathLength(Path); Level++) {
    for(Scan = tab_Clauses(tab_PathNthNode(Path, Level));  
	!list_Empty(Scan); Scan = list_Cdr(Scan)) {
      if (list_Car(Scan) == Clause)
	return(TRUE);  } } 

  return(FALSE);
}


BOOL tab_PathContainsClauseRobust(TABPATH P, CLAUSE C)
/**************************************************************
 INPUT  : A tableau path, a clause
 RETURNS: TRUE if clause can be find on the path (not
          necessarily on its level)
 EFFECTS: Prints a note if clause cannot be found on
          its level. Intended for debugging.
***************************************************************/
{
  if (tab_PathContainsClause(P,C)) 
    return(TRUE);

  if (tab_PathContainsClauseSoft(P,C)) {
    fprintf(stderr, "NOTE: Clause is found on path, but not indexed by level.\n");
    clause_PParentsFPrint(stderr,C); fflush(stderr);
    return(TRUE);
  }
  return(FALSE);
}

void tab_AddSplitAtCursor(TABPATH Path, BOOL LeftSide)
/**************************************************************
 INPUT  : A tableau path, a flag
 RETURNS: -
 EFFECTS: Extends the tableau containing the path <Path> to
          the left if <LeftSide> is TRUE, to the right 
	  otherwise
***************************************************************/
{
  TABLEAU Tab, NewBranch;

  Tab = tab_PathTop(Path);
  NewBranch = tab_CreateNode();
  if (LeftSide) {

#ifdef DEBUG
    if (!tab_LeftBranchIsEmpty(Tab)) {
      fprintf(stderr, "\nrecreating existing left branch in tableau.\n");
      misc_Error(); }
#endif
      tab_SetLeftBranch(Tab,NewBranch); }

  else {

#ifdef DEBUG    
    if (!tab_RightBranchIsEmpty(Tab)) {
      fprintf(stderr, "\nrecreating existing right branch in tableau.\n");
      misc_Error(); }
#endif
    tab_SetRightBranch(Tab, NewBranch); }

  tab_PathPush(NewBranch, Path);
}

void tab_AddClauseOnItsLevel(CLAUSE C, TABPATH Path)
/**************************************************************
 INPUT  : A clause, a tableau path 
 RETURNS: -
 EFFECTS: Adds the clause on its split level which
          must belong to <Path>
***************************************************************/
{
  int Level; 

  Level = clause_SplitLevel(C);
  if (Level > tab_PathLength(Path)) {
    fprintf(stderr, "error: split level %d of clause: ", Level);
    clause_FPrint(stderr, C);
    fprintf(stderr, "\nis higher than existing level %d\n", tab_PathLength(Path)); 
    misc_Error();  }
  
  tab_AddClause(C, tab_PathNthNode(Path, clause_SplitLevel(C)));
}


int tab_Depth(TABLEAU T)
/**************************************************************
 INPUT  : A tableau
 RETURNS: The depth of the tableau
 EFFECTS: -
***************************************************************/
{
  if (tab_IsEmpty(T))
    return(0);
  if (tab_IsLeaf(T))
    return(0);
  else 
    return(misc_Max(tab_Depth(tab_RightBranch(T))+1, tab_Depth(tab_LeftBranch(T)))+1);
}

BOOL tab_HasEmptyClause(TABLEAU T)
/**************************************************************
 INPUT  : A tableau                                                     
 RETURNS: TRUE iff an empty clause is among the clauses
          on this level
 EFFECTS: -                                                     
***************************************************************/
{
  LIST Scan;
  
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) 
    if (clause_IsEmptyClause(list_Car(Scan)))
      return(TRUE);

  return(FALSE);
}


BOOL tab_IsClosed(TABLEAU T)
/**************************************************************
 INPUT  : A Tableau   
 RETURNS: TRUE iff the tableau is closed. (NOTE: FALSE
          if the tableau is empty)
 EFFECTS: -                                                     
***************************************************************/
{
  if (tab_IsEmpty(T))
    return(FALSE);

  if (tab_HasEmptyClause(T))
      return(TRUE);
  /* 
   *  now tableau can only be closed
   *  if there has been a split, and
   *  both subtableaus are closed 
   */
  
  if (tab_RightBranchIsEmpty(T) || tab_LeftBranchIsEmpty(T)) {

#ifdef USE_CG    
    printf("\nopen node label: %d", T->Label);
    fflush(stdout);
#endif

    return(FALSE);
    
 }
  return(tab_IsClosed(tab_RightBranch(T)) && tab_IsClosed(tab_LeftBranch(T)));
}

LIST tab_DeleteFlat(TABLEAU T)
/**************************************************************
 INPUT  : A tableau
 RETURNS: The list of its clauses on the first level of 
          the tableau
 EFFECTS: Frees the root tableau node.                                                     
***************************************************************/
{
  LIST Clauses;

  Clauses = tab_Clauses(T);
  memory_Free(T, sizeof(TABLEAU_NODE));
  
  return(Clauses);
}


void tab_DeleteGen(TABLEAU T, LIST* Clauses, BOOL DeleteClauses)
/**************************************************************
 INPUT  : A tableau, a list of clauses by reference, a flag 
 RETURNS: -
 EFFECTS: Depending on <DeleteClauses>, all clauses in the
          tableau are added to <Clauses> or just deleted.
	  The memory for the tableau and its clause lists is
	  freed.
***************************************************************/
{
  if (tab_IsEmpty(T))
    return;
  
  tab_DeleteGen(tab_RightBranch(T), Clauses, DeleteClauses);
  tab_DeleteGen(tab_LeftBranch(T),  Clauses, DeleteClauses);

  list_Delete(tab_RightSplitClauses(T));
  if (DeleteClauses) 
    list_Delete(tab_Clauses(T)); 
  else
    *Clauses = list_Nconc(tab_Clauses(T), *Clauses);
  
  tab_DeleteFlat(T);
  
}

void tab_DeleteCollectClauses(TABLEAU T, LIST* Clauses)
/**************************************************************
 INPUT  : A tableau, a list of clauses by reference 
 RETURNS: -                                                     
 EFFECTS: Frees the memory of the tableau, but collects
          its clauses
***************************************************************/
{
  tab_DeleteGen(T, Clauses, FALSE);
}

void tab_Delete(TABLEAU T)
/**************************************************************
 INPUT  : A tableau                                                     
 RETURNS: -                                                     
 EFFECTS: Frees the memory of the tableau                                                     
***************************************************************/
{
  LIST Redundant;
  
  Redundant = list_Nil();
  tab_DeleteGen(T, &Redundant, TRUE);
}

static void tab_SetSplitLevelsRec(TABLEAU T, int Level)
/**************************************************************
 INPUT  : A tableau                                                     
 RETURNS: -                                                     
 EFFECTS: The split levels of the clauses in the
          tableau are set to the level of the
	  tableau level they are contained in.
***************************************************************/
{
  LIST Scan;

  if (tab_IsEmpty(T))
    return;

  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    clause_SetSplitLevel(list_Car(Scan), Level); 
    if (Level >0) {
      clause_ClearSplitField(list_Car(Scan));
      clause_SetSplitFieldBit(list_Car(Scan), Level); }
    else 
      clause_SetSplitField(list_Car(Scan), (SPLITFIELD)NULL,0);

}
  
  tab_SetSplitLevelsRec(tab_RightBranch(T), Level+1);
  tab_SetSplitLevelsRec(tab_LeftBranch(T), Level+1);
  
}

void tab_SetSplitLevels(TABLEAU T)
/**************************************************************
 INPUT  : A tableau                                                     
 RETURNS: -                                                     
 EFFECTS: The split levels of the clauses in the
          tableau are set to the level of the
	  tableau they belong to.
***************************************************************/
{
  tab_SetSplitLevelsRec(T,0);
}

TABLEAU tab_PruneClosedBranches(TABLEAU T, LIST* Clauses) 
/**************************************************************
 INPUT  : A tableau, a list of clauses by reference. 
 RETURNS: The (destructively) reduced tableau: Descendants of 
          nodes that have an empty clause are deleted.
 EFFECTS: The tableau is modified.                                                      
***************************************************************/
{
  if (tab_IsEmpty(T))
    return(T);

  /* if there is an empty clause on this level, delete subtrees */

  if (tab_HasEmptyClause(T)) {

    tab_DeleteCollectClauses(tab_RightBranch(T), Clauses); 
    tab_DeleteCollectClauses(tab_LeftBranch(T), Clauses); 
    tab_SetRightBranch(T, tab_EmptyTableau());
    tab_SetLeftBranch(T, tab_EmptyTableau()); 
    list_Delete(tab_RightSplitClauses(T));
    tab_SetRightSplitClauses(T, list_Nil());
    tab_SetSplitClause(T,clause_Null());
    tab_SetLeftSplitClause(T, clause_Null()); }

  /* else recursively prune subtrees */
  
  else {
    tab_SetRightBranch(T, tab_PruneClosedBranches(tab_RightBranch(T), Clauses));
    tab_SetLeftBranch(T, tab_PruneClosedBranches(tab_LeftBranch(T), Clauses));  }
  
  return(T);
}


TABLEAU tab_RemoveIncompleteSplits(TABLEAU T, LIST* Clauses)
/**************************************************************
 INPUT  : A Tableau, a list of clauses by reference
 RETURNS: The reduced tableau: If a node has exactly one
          successor (that is, the corresponding split was
          not completed), delete the successor and move
	  its subtrees to <T>.
 EFFECTS: The successor node is deleted, and its clauses added
          to <Clauses>
***************************************************************/
{
  LIST    NewClauses;
  TABLEAU Child;

  if (tab_IsEmpty(T))
    return(T);

  if (tab_IsLeaf(T))
    return(T);

  if (!tab_IsEmpty(tab_RightBranch(T)) &&
      !tab_IsEmpty(tab_LeftBranch(T))) {	   
    tab_SetRightBranch(T, tab_RemoveIncompleteSplits(tab_RightBranch(T), Clauses));
    tab_SetLeftBranch(T, tab_RemoveIncompleteSplits(tab_LeftBranch(T), Clauses));
    return(T);  }
  
  if (tab_IsEmpty(tab_RightBranch(T))) 
    Child = tab_LeftBranch(T);
  else 
    Child = tab_RightBranch(T);

  Child = tab_RemoveIncompleteSplits(Child, Clauses);

  tab_SetLeftBranch(T, tab_LeftBranch(Child));   
  tab_SetRightBranch(T, tab_RightBranch(Child));

  /* copy split data */

  tab_SetSplitClause(T, tab_SplitClause(Child));
  tab_SetLeftSplitClause(T, tab_LeftSplitClause(Child));
  tab_SetRightSplitClauses(T, tab_RightSplitClauses(Child));

  /* delete ancestors of deleted clauses and remember */

  NewClauses = tab_DeleteFlat(Child);
  (*Clauses) = list_Nconc(NewClauses, *Clauses);

  return(T);

}

void tab_CheckEmpties(TABLEAU T)
/**************************************************************
 INPUT  : A tableau 
 RETURNS: -
 EFFECTS: Prints warnings if non-leaf nodes contain
          empty clauses (which should not be the case
          after pruning any more), of if leaf nodes
	  contain more than one empty clause
***************************************************************/
{
  LIST Scan, Empties;
  BOOL Printem;


  if (tab_IsEmpty(T))
    return;

  /* get all empty clauses in this node */ 
  Empties = list_Nil();
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (clause_IsEmptyClause(list_Car(Scan)))
      Empties = list_Cons(list_Car(Scan), Empties); }

  Printem = FALSE;
  if (!list_Empty(Empties) && !tab_IsLeaf(T)) {
    printf("\nNOTE: non-leaf node contains empty clauses.\n");
   Printem = TRUE;
  }
  
  if (tab_IsLeaf(T) && list_Length(Empties) > 1) {
    printf("\nNOTE: Leaf contains more than one empty clauses.\n");
    Printem =TRUE; }

  if (Printem) {
    printf("Clauses:\n");
    clause_PParentsListPrint(tab_Clauses(T)); }

  list_Delete(Empties);
  tab_CheckEmpties(tab_LeftBranch(T));
  tab_CheckEmpties(tab_RightBranch(T));
}

void tab_GetAllEmptyClauses(TABLEAU T, LIST* L)
/**************************************************************
 INPUT  : A tableau, a list by reference
 RETURNS: All empty clauses in the tableau prepended to <L>.
 EFFECTS: -                                                      
***************************************************************/
{

  if (tab_IsEmpty(T))
    return;
  
  tab_GetAllEmptyClauses(tab_LeftBranch(T), L);
  tab_GetAllEmptyClauses(tab_RightBranch(T), L);
}

void tab_GetEarliestEmptyClauses(TABLEAU T, LIST* L)
/**************************************************************
 INPUT  : A tableau, a list of clauses by reference
 RETURNS: - 
 EFFECTS: For each leaf node, adds empty clauses in 
          leaf nodes to <L>. If the leaf node contains only one
	  empty clause, it is added to <L> anyway.
          If the leaf node contains more than one empty clause,
	  the earliest derived empty clause is added to <L>.
***************************************************************/
{
  CLAUSE FirstEmpty;
  LIST   Scan;

  if (tab_IsEmpty(T))
    return;

  if (tab_IsLeaf(T)) {  
    FirstEmpty = clause_Null();
  
    for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      if (clause_IsEmptyClause(list_Car(Scan))) {
	if (FirstEmpty == clause_Null()) 
	  FirstEmpty = list_Car(Scan);
	else if (clause_Number(FirstEmpty) > clause_Number(list_Car(Scan)))
	  FirstEmpty = list_Car(Scan); }
    }
    
    if (FirstEmpty != clause_Null())
      (*L) = list_Cons(FirstEmpty, *L); }
  
  tab_GetEarliestEmptyClauses(tab_LeftBranch(T), L);
  tab_GetEarliestEmptyClauses(tab_RightBranch(T), L);

}

void tab_ToClauseList(TABLEAU T, LIST* Proof)
/**************************************************************
 INPUT  : A tableau <T>, a list of clauses 
 RETURNS: -
 EFFECTS: All clauses in T are added to <Proof>
***************************************************************/
{
  if (tab_IsEmpty(T))
    return;
  
  (*Proof) = list_Nconc(list_Copy(tab_Clauses(T)), *Proof);

  tab_ToClauseList(tab_LeftBranch(T),Proof);
  tab_ToClauseList(tab_RightBranch(T),Proof);
}


void tab_ToSeqProofOrdered(TABLEAU T, LIST* Proof)
/**************************************************************
 INPUT  : A tableau <T>, a list of clauses <Proof> representing a 
          proof by reference
 RETURNS: The sequential proof corresponding to the tableau.
 EFFECTS:                                                      
***************************************************************/
{
  LIST Scan;
  BOOL RightSplitRead, LeftSplitRead;
  
  if (tab_IsEmpty(T))
    return;

  Scan = tab_Clauses(T); 
  RightSplitRead = LeftSplitRead = FALSE;

  while (!list_Empty(Scan)) {

    /* insert left and right splits and descendants controlled by clause number */

    if (!RightSplitRead && !tab_RightBranchIsEmpty(T) &&
	clause_Number(list_Car(Scan)) <
	clause_Number(list_Car(tab_RightSplitClauses(T)))) {
      tab_ToSeqProofOrdered(tab_RightBranch(T), Proof);
      RightSplitRead = TRUE; }
    
    if (!LeftSplitRead && !tab_LeftBranchIsEmpty(T) &&
	clause_Number(list_Car(Scan)) < 
	clause_Number(tab_LeftSplitClause(T))) {
      tab_ToSeqProofOrdered(tab_LeftBranch(T), Proof);
      LeftSplitRead  = TRUE; }
    
    (*Proof) = list_Cons(list_Car(Scan), *Proof); 
    Scan  = list_Cdr(Scan);  }

  /* if a split clause with descendants has not been inserted yet, 
     it been generated after all other clauses */

  if (!RightSplitRead)
    tab_ToSeqProofOrdered(tab_RightBranch(T), Proof);
  if (!LeftSplitRead)
    tab_ToSeqProofOrdered(tab_LeftBranch(T), Proof);

}




/****************************************************************/
/*  SPECIALS FOR GRAPHS                                         */
/****************************************************************/

#ifdef USE_CG

void tab_LabelNodesRec(TABLEAU T, int* Num)
/**************************************************************
 INPUT  : A Tableau, a number by reference
 RETURNS: -
 EFFECTS: Labels the tableau nodes dflr, starting
          with <Num>
***************************************************************/
{
  if (tab_IsEmpty(T))
    return;
  T->Label = *Num;
  (*Num)++;
  tab_LabelNodesRec(tab_LeftBranch(T), Num);
  tab_LabelNodesRec(tab_RightBranch(T), Num);
}

void tab_LabelNodes(TABLEAU T)
/**************************************************************
 INPUT  : A Tableau, a number by reference
 RETURNS: -
 EFFECTS: Labels the tableau nodes dflr, starting with 0
***************************************************************/
{
  int Num;

  Num = 0;
  tab_LabelNodesRec(T, &Num);
}


void tab_FPrintNodeLabel(FILE *File, TABLEAU T)
/**************************************************************
 INPUT  : A file handle, a tableau 
 RETURNS: -
 EFFECTS: Prints the root node information to <File>:
          clauses, split information
***************************************************************/
{
  LIST Scan;

  /* start printing of node label string */
  fprintf(File, "\""); 
  
#ifdef USE_CG
  fprintf(File, "label: %d\\n", T->Label);
#endif
  
  /* print left and right parts of split */
  fprintf(File, "SplitClause : "); 
  clause_PParentsFPrint(File, tab_SplitClause(T));
  fprintf(File, "\\nLeft Clause : ");
  clause_PParentsFPrint(File, tab_LeftSplitClause(T));
  fprintf(File, "\\n"); 
  fprintf(File, "RightClauses: ");
  if (list_Empty(tab_RightSplitClauses(T)))
    fprintf(File,"[]\\n");
  else {
    clause_PParentsFPrint(File, list_Car(tab_RightSplitClauses(T)));
    fprintf(File, "\\n");
    for (Scan = list_Cdr(tab_RightSplitClauses(T)); !list_Empty(Scan); Scan = list_Cdr(Scan)){
      fprintf(File,"              ");
      clause_PParentsFPrint(File, list_Car(Scan)); 
      fprintf(File, "\\n"); }
 }
  
  /* print clause at this level */
  
  if (pcheck_ClauseCg) {
    if (list_Empty(tab_Clauses(T)))
      fprintf(File, "[]");
    else {
      for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)){
	clause_PParentsFPrint(File, list_Car(Scan)); 
	fprintf(File, "\\");
	fprintf(File, "n"); }
    }
  }
  fprintf(File, "\""); /* close string */
}

void tab_FPrintEdgeCgFormat(FILE* File, int Source, int Target, BOOL Left)
/**************************************************************
 INPUT  : A file handle, two node labels, a flag
 RETURNS: -
 EFFECTS: Prints the edge denoted by <Source> and <Target>
          to <File>, with an edge label (0/1) according to <Left>.
	  The edge label is added since xvcg cannot handle
          ordered edges.
***************************************************************/
{
  fprintf(File, "\nedge: {");
  fprintf(File, "\nsourcename: \"%d\"", Source);
  fprintf(File, "\ntargetname: \"%d\"\n", Target);
  fprintf(File, "\nlabel: \"");
  if (Left) 
    fprintf(File, "0");
  else
    fprintf(File, "1");
  fprintf(File, "\"");

  fprintf(File, "  }\n");

} 

void tab_FPrintEdgesCgFormat(FILE* File, TABLEAU T)
/**************************************************************
 INPUT  : A file handle, a tableau
 RETURNS: -
 EFFECTS: Prints egde information of <T> in xvcg graph format
          to <File>.
***************************************************************/
{
  if (tab_IsEmpty(T))
    return;

  if (!tab_LeftBranchIsEmpty(T))
    tab_FPrintEdgeCgFormat(File, T->Label, tab_LeftBranch(T)->Label, TRUE);
  if (!tab_RightBranchIsEmpty(T))
    tab_FPrintEdgeCgFormat(File, T->Label, tab_RightBranch(T)->Label, FALSE);
  
  tab_FPrintEdgesCgFormat(File, tab_LeftBranch(T));
  tab_FPrintEdgesCgFormat(File, tab_RightBranch(T));
}

void tab_FPrintNodesCgFormat(FILE* File, TABLEAU T)
/**************************************************************
 INPUT  : A file handle, a tableau
 RETURNS: -
 EFFECTS: Prints egde information of <T> in xvcg graph format
          to <File>.                                            
***************************************************************/
{

  if (tab_IsEmpty(T))
    return;

  fprintf(File, "\nnode: {\n");
  fprintf(File, "\nlabel: ");
  tab_FPrintNodeLabel(File, T);
  fprintf(File, "\n"); /* end label section */
  
  fprintf(File, "title: \"%d\"\n", T->Label);
  fprintf(File, "  }\n");
  
  /* recursion */
  tab_FPrintNodesCgFormat(File, tab_LeftBranch(T));
  tab_FPrintNodesCgFormat(File, tab_RightBranch(T));
  
} 

void tab_FPrintCgFormat(FILE* File, TABLEAU T)
/**************************************************************
 INPUT  : A file handle, a tableau
 RETURNS: -
 EFFECTS: Prints tableau as a graph in xvcg format to <File>
***************************************************************/
{
  fprintf(File, "graph: \n");
  fprintf(File,"{\n");
  fprintf(File, "display_edge_labels: yes\n");
  
  tab_FPrintNodesCgFormat(File, T);
  tab_FPrintEdgesCgFormat(File, T); 
  fprintf(File, "}\n");
}

void tab_PrintCgFormat(TABLEAU T)
/**************************************************************
 INPUT  : A tableau 
 RETURNS: -                                                     
 EFFECTS: Print tableau in xvcg format to stdout
***************************************************************/
{
  tab_FPrintCgFormat(stdout, T);
}

/**************************************************************/
/* procedures for printing graph in da Vinci format           */
/**************************************************************/

void tab_FPrintDaVinciEdge(FILE* File, int L1, int L2)
/**************************************************************
 INPUT  : A file handle, two numbers 
 RETURNS: - 
 EFFECTS: Print an edge in daVinci format
***************************************************************/
{
  fprintf(File, "l(\"%d->%d\","   ,L1,L2);
  fprintf(File, "e(\"\",[],\n");
  /* print child node as reference */
  fprintf(File, "r(\"%d\")))\n", L2);
}

void tab_FPrintDaVinciFormatRec(FILE* File, TABLEAU T)
/**************************************************************
 INPUT  : A file handle, a tableau
 RETURNS: -
 EFFECTS: Prints tableau to <File> in daVinci format
***************************************************************/
{
  /* print node label */
  fprintf(File, "l(\"%d\",", T->Label);
  /* print node attributes */
  fprintf(File, "n(\"\", [a(\"OBJECT\","); 
  tab_FPrintNodeLabel(File, T);
  fprintf(File, ")],\n");
  
  /* print egde list  */
  fprintf(File, "[");
  if (!tab_LeftBranchIsEmpty(T)) 
    tab_FPrintDaVinciEdge(File, T->Label, tab_LeftBranch(T)->Label); 

  if (!tab_RightBranchIsEmpty(T)) {
    if (!tab_LeftBranchIsEmpty(T))
      fprintf(File, ",");
    tab_FPrintDaVinciEdge(File, T->Label, tab_RightBranch(T)->Label); }
  fprintf(File, "]");

  fprintf(File, "))"); /* this ends the node description */

  if (!tab_LeftBranchIsEmpty(T)) {
    fprintf(File, ",");
    tab_FPrintDaVinciFormatRec(File, tab_LeftBranch(T)); }
  
  if (!tab_RightBranchIsEmpty(T)) {
    fprintf(File, ",");
    tab_FPrintDaVinciFormatRec(File, tab_RightBranch(T)); }
}

void tab_FPrintDaVinciFormat(FILE* File, TABLEAU T)
/**************************************************************
 INPUT  : A file handle <File>, a tableau
 RETURNS: -
 EFFECTS: Print tableau in daVinci format to <File>
***************************************************************/
{
  fprintf(File, "[\n");
  tab_FPrintDaVinciFormatRec(File,T);
  fprintf(File, "]\n");
}


void tab_WriteTableau(TABLEAU T, char* Filename, GRAPHFORMAT Format)
/**************************************************************
 INPUT  : A tableau, a filename 
 RETURNS: -
 EFFECTS: -
***************************************************************/
{
  FILE*  File;
  
  File =  misc_OpenFile(Filename, "w");

  if (Format == DAVINCI)
    tab_FPrintDaVinciFormat(File, T);
  else if (Format  == XVCG)
    tab_FPrintCgFormat(File, T);
  else {
    fprintf(stderr, "error: unknown output format for tableau.\n");
     misc_CloseFile(File, Filename);
    misc_Error(); }

  misc_CloseFile(File, Filename);
}

#endif
